package mark.core.io;
import java.io.*;
import java.util.*;

/**
 * Static utility methods.
 *
 * @author Mark Chavira
 */

public abstract class Util {

  /**
   * The newline separator.
   */
  
  public static final String NEWLINE = System.getProperty ("line.separator");

  /**
   * A byte[] representing the smallest integer in ascii.
   */

  private static final byte[] MIN_INT =
    Integer.toString (Integer.MIN_VALUE).getBytes ();
  
  /**
   * Adds an ascci representation of the given integer to the given byte[]
   * starting at the given index.
   * 
   * @param tgt the given byte[].
   * @param tgtIndex the given index. 
   * @param n the given integer.
   * @return the the index after the last byte written.
   */
  
  public static int writeInt (byte[] tgt, int tgtIndex, int n) {
    
    // Handle the special case where the integer is 0.  This is a special
    // case because the loop for the general case won't execute at all if
    // the integer is 0.
    
    if (n == 0) {
      tgt[tgtIndex] = '0';
      return tgtIndex + 1;
    }
    
    // Handle the special case where the integer is the minimum possible
    // value.  This is a special case, because the general case converts to
    // positive, which causes problems.
      
    if (n == Integer.MIN_VALUE) {
      for (byte b : MIN_INT) {tgt[tgtIndex++] = b;}
      return tgtIndex;
    }
      
    // Handle the general case.  First, if the number is negative, then write
    // '-'.  Next, compute the number of digits in the number.  Next, write
    // the number into the byte[].  I thought about different approaches
    // here.  One approach is to allocate scratch, write in reverse to
    // scratch, and then copy backwards into tgt.  This is bad, because
    // allocating would incur extra cost and memory management burden.
    // Allocating statically is not better, because then this code can't
    // be called by multiple threads.  Another approach is to write in
    // reverse to end of array and then copy to beginning.  This would have
    // the side effect of overwriting some bytes that need not be
    // overwritten.
  
    if (n < 0) {n *= -1; tgt[tgtIndex++] = '-';}
    int numDigits = 0;
    for (int temp = n; temp != 0; temp /= 10) {++numDigits;}
    int scratchIndex = tgtIndex + numDigits;
    for (; n!= 0; n /= 10) {tgt[--scratchIndex] = (byte)('0' + n % 10);}
    return tgtIndex + numDigits;
    
  }
  
  /**
   * Inputs all nonwhite tokens from the given reade, skipping comments,
   * which are introduced by a pound character and terminated by newline.
   * 
   * @param r the given reader.
   * @return the tokens.
   */
  
  public static LinkedList<String> tokens (BufferedReader r)
   throws Exception {
    LinkedList<String> ans = new LinkedList<String> ();
    for (;;) {
      String line = r.readLine ();
      if (line == null) {break;}
      line = line.replaceAll ("#.*", "");
      line = line.replaceAll (";", " ; ");
      if (line.matches ("^\\s*$")) {continue;}
      String[] toks = line.trim ().split ("\\s+");
      for (String t : toks) {ans.add (t);}
    }
    return ans;
  }
  
  /**
   * Inputs all characters in a reader into a string and closes the reader.
   *
   * @param r the reader.
   * @return the string.
   */

  public static String stringFromReader (Reader r) throws Exception {
    char[] buf = new char[1024];
    StringBuilder ans = new StringBuilder ();
    for (int num; (num = r.read (buf)) != -1; ans.append (buf, 0, num)) ;
    r.close ();
    return ans.toString ();
  }

  /**
   * Reads lines from the given reader.
   *
   * @param r the given reader.
   * @return the names.
   */

  public static String[] nonemptyLinesFromReader (Reader r) throws Exception {
    String a = stringFromReader (r);
    if (a.length () == 0) {
      return new String[0]; // split returns one empty string in this case 
    }
    return a.split ("[\\n\\r]+");
  }

  /**
   * Outputs the given strings to the file with the given name, along with
   * a newline after each string.
   *
   * @param s the given strings.
   * @param f the given name.
   */
  
  public static void linesToFile (CharSequence[] s, String f)
   throws Exception {
    BufferedWriter o =
      new BufferedWriter (new FileWriter (f));
    for (int i = 0; i < s.length; i++) {
      o.append (s[i]);
      o.newLine ();
    }
    o.close ();
  }

  /**
   * Outputs the given string to the file with the given name.
   *
   * @param s the given string.
   * @param f the given name.
   */

  public static void stringToFile (CharSequence s, String f)
   throws Exception {
    BufferedWriter o =
      new BufferedWriter (new FileWriter (f));
    o.append (s);
    o.close ();
  }

  /**
   * Processes files underneath the given file recursively using the given
   * file filter.
   * 
   * @param ff a file filter which processes each file and returns whether
   *   the traversal should continue underneath this file.
   * @param f the given file.
   */

  public static void processFilesRecursively (FileFilter ff, File f) {
    if (ff.accept (f) && f.isDirectory ()) {
      for (File child : f.listFiles ()) {
        processFilesRecursively (ff, child);
      }
    }
  }

  /**
   * Deletes the given file, recursively if it is a directory.
   * 
   * @param f the directory.
   */
  
  public static boolean delete (File f) {
    if (f.isDirectory()) {
      for (File ch : f.listFiles ()) {
        if (!delete (ch)) {
          return false;
        }
      }
    }
    return f.delete ();
  }
  
  /**
   * Deletes the file with the given name, and throws an exception if
   * unsucccessful.
   *
   * @param fn the given name.
   */
  
  public static void deleteIfExists (String fn) throws Exception {
    File f = new File (fn);
    if (f.exists () && !delete (f)) {
      throw new Exception ("Could not delete file: " + fn);
    }
  }

  /**
   * Deletes the file with the given name, prints a warning to standard error
   * if unsuccessful, and returns whether successful.
   *
   * @param fn the given name.
   * @return whether successful.
   */
  
  public static boolean deleteWarningIfExists (String fn) {
    File f = new File (fn);
    if (f.exists () && !delete (f)) {
      System.err.println ("Could not delete file: " + fn);
      return false;
    }
    return true;
  }

  /**
   * Renames the file with the first given name to have the second given
   * name.
   *
   * @param fn1 the first given name.
   * @param fn2 the second given name.
   */
  
  public static void rename (String fn1, String fn2) throws Exception {
    deleteIfExists (fn2);
    if (!new File (fn1).renameTo (new File (fn2))) {
      throw new Exception ("Could not rename file: " + fn1);
    }
  }
  
}
