package mark.core.math;
import java.util.*;

/**
 * Static utility methods that apply to arrays.
 *
 * @author Mark Chavira
 */

public abstract class DoubleArrays {

  /**
   * Java's array clone is very slow (see
   * http://forums.java.net/jive/thread.jspa?threadID=743&messageID=16562),
   * so for critical methods, use this method instead.
   *
   * @param a the array to clone.
   * @return the clone.
   */
    
  public static double[] clone (double[] a) {
    double[] ans = new double[a.length];
    System.arraycopy (a, 0, ans, 0, a.length);
    return ans;
  }

  /**
   * Retuns an array of doubles that is formed by concatenating the given
   * arrays of doubles.
   *
   * @param a the given arrays of doubles.
   * @return the single array of doubles.
   */
  
  public static double[] flatten (double[]... a) {
    int num = 0;
    for (int i = 0; i < a.length; i++) {
      num += a[i].length;
    }
    double[] ans = new double[num];
    int index = 0;
    for (int i = 0; i < a.length; i++) {
      for (int j = 0; j < a[i].length; j++) {
        ans[index++] = a[i][j];
      }
    }
    return ans;
  }

  /**
   * Retuns an array of doubles that is formed by concatenating the given
   * arrays of doubles.
   *
   * @param a the given arrays of doubles.
   * @return the single array of doubles.
   */
  
  public static double[] flatten (double[][]... a) {
    double[][] step1 = new double[a.length][];
    for (int i = 0; i < a.length; i++) {
      step1[i] = flatten (a[i]);
    }
    return flatten (step1);
  }

  /**
   * Adds the second given array into the first.  The arrays must have the
   * same dimesions.
   *
   * @param a1 the first array.
   * @param a2 the second array.
   */
  
  public static void addInto (double[] a1, double[] a2) {
    for (int i = 0; i < a1.length; i++) {
      a1[i] += a2[i];
    }
  }

  /**
   * Adds the second given array into the first.  The arrays must have the
   * same dimesions.
   *
   * @param a1 the first array.
   * @param a2 the second array.
   */
  
  public static void addInto (double[][] a1, double[][] a2) {
    for (int i = 0; i < a1.length; i++) {
      for (int j = 0; j < a1[i].length; j++) {
        a1[i][j] += a2[i][j];
      }
    }
  }

  /**
   * Adds the second given array into the first.  The arrays must have the
   * same dimesions.
   *
   * @param a1 the first array.
   * @param a2 the second array.
   */
  
  public static void addInto (double[][][] a1, double[][][] a2) {
    for (int i = 0; i < a1.length; i++) {
      for (int j = 0; j < a1[i].length; j++) {
        for (int k = 0; k < a1[i][j].length; k++) {
          a1[i][j][k] += a2[i][j][k];
        }
      }
    }
  }

  /**
   * Returns the sum of the two given arrays.  The arrays must have the
   * same dimesions.
   *
   * @param a1 the first array.
   * @param a2 the second array.
   */
  
  public static double[] add (double[] a1, double[] a2) {
    double[] ans = clone (a1);
    addInto (ans, a2);
    return ans;
  }

  /**
   * Returns the sum of the two given arrays.  The arrays must have the
   * same dimesions.
   *
   * @param a1 the first array.
   * @param a2 the second array.
   */
  
  public static double[][] add (double[][] a1, double[][] a2) {
    double[][] ans = clone (a1);
    addInto (ans, a2);
    return ans;
  }

  /**
   * Returns the sum of the two given arrays.  The arrays must have the
   * same dimesions.
   *
   * @param a1 the first array.
   * @param a2 the second array.
   */
  
  public static double[][][] add (double[][][] a1, double[][][] a2) {
    double[][][] ans = clone (a1);
    addInto (ans, a2);
    return ans;
  }

  /**
   * Subtracts the second given array from the first.  The arrays must have
   * the same dimesions.
   *
   * @param a1 the first array.
   * @param a2 the second array.
   */
  
  public static void subtractFrom (double[] a1, double[] a2) {
    for (int i = 0; i < a1.length; i++) {
      a1[i] -= a2[i];
    }
  }

  /**
   * Subtracts the second given array from the first.  The arrays must have
   * the same dimesions.
   *
   * @param a1 the first array.
   * @param a2 the second array.
   */
  
  public static void subtractFrom (double[][] a1, double[][] a2) {
    for (int i = 0; i < a1.length; i++) {
      for (int j = 0; j < a1[i].length; j++) {
        a1[i][j] -= a2[i][j];
      }
    }
  }

  /**
   * Subtracts the second given array from the first.  The arrays must have
   * the same dimesions.
   *
   * @param a1 the first array.
   * @param a2 the second array.
   */
  
  public static void subtractFrom (double[][][] a1, double[][][] a2) {
    for (int i = 0; i < a1.length; i++) {
      for (int j = 0; j < a1[i].length; j++) {
        for (int k = 0; k < a1[i][j].length; k++) {
          a1[i][j][k] -= a2[i][j][k];
        }
      }
    }
  }

  /**
   * Returns the difference of the two given arrays.  The arrays must have
   * the same dimesions.
   *
   * @param a1 the first array.
   * @param a2 the second array.
   */
  
  public static double[] subtract (double[] a1, double[] a2) {
    double[] ans = clone (a1);
    subtractFrom (ans, a2);
    return ans;
  }

  /**
   * Returns the difference of the two given arrays.  The arrays must have
   * the same dimesions.
   *
   * @param a1 the first array.
   * @param a2 the second array.
   */
  
  public static double[][] subtract (double[][] a1, double[][] a2) {
    double[][] ans = clone (a1);
    subtractFrom (ans, a2);
    return ans;
  }

  /**
   * Returns the difference of the two given arrays.  The arrays must have
   * the same dimesions.
   *
   * @param a1 the first array.
   * @param a2 the second array.
   */
  
  public static double[][][] subtract (double[][][] a1, double[][][] a2) {
    double[][][] ans = clone (a1);
    subtractFrom (ans, a2);
    return ans;
  }

  /**
   * Multiplies a scalar into the given array.
   *
   * @param a the array.
   * @param v the scalar.
   */
  
  public static void multiplyInto (double[] a, double v) {
    for (int i = 0; i < a.length; i++) {
      a[i] *= v;
    }
  }

  /**
   * Multiplies a scalar into the given array.
   *
   * @param a the array.
   * @param v the scalar.
   */
  
  public static void multiplyInto (double[][] a, double v) {
    for (int i = 0; i < a.length; i++) {
      for (int j = 0; j < a[i].length; j++) {
        a[i][j] *= v;
      }
    }
  }

  /**
   * Multiplies a scalar into the given array.
   *
   * @param a the array.
   * @param v the scalar.
   */
  
  public static void multiplyInto (double[][][] a, double v) {
    for (int i = 0; i < a.length; i++) {
      for (int j = 0; j < a[i].length; j++) {
        for (int k = 0; k < a[i][j].length; k++) {
          a[i][j][k] *= v;
        }
      }
    }
  }

  /**
   * Returns the product of the given array and scalar.
   *
   * @param a the array.
   * @param v the scalar.
   */
  
  public static double[] multiply (double[] a, double v) {
    double[] ans = clone (a);
    multiplyInto (ans, v);
    return ans;
  }

  /**
   * Returns the product of the given array and scalar.
   *
   * @param a the array.
   * @param v the scalar.
   */
  
  public static double[][] multiply (double[][] a, double v) {
    double[][] ans = clone (a);
    multiplyInto (ans, v);
    return ans;
  }

  /**
   * Returns the product of the given array and scalar.
   *
   * @param a the array.
   * @param v the scalar.
   */
  
  public static double[][][] multiply (double[][][] a, double v) {
    double[][][] ans = clone (a);
    multiplyInto (ans, v);
    return ans;
  }

  /**
   * Divides a scalar into the given array.
   *
   * @param a the array.
   * @param v the scalar.
   */
  
  public static void divideInto (double[] a, double v) {
    for (int i = 0; i < a.length; i++) {
      a[i] /= v;
    }
  }

  /**
   * Divides a scalar into the given array.
   *
   * @param a the array.
   * @param v the scalar.
   */
  
  public static void divideInto (double[][] a, double v) {
    for (int i = 0; i < a.length; i++) {
      for (int j = 0; j < a[i].length; j++) {
        a[i][j] /= v;
      }
    }
  }

  /**
   * Divides a scalar into the given array.
   *
   * @param a the array.
   * @param v the scalar.
   */
  
  public static void divideInto (double[][][] a, double v) {
    for (int i = 0; i < a.length; i++) {
      for (int j = 0; j < a[i].length; j++) {
        for (int k = 0; k < a[i][j].length; k++) {
          a[i][j][k] /= v;
        }
      }
    }
  }

  /**
   * Returns the result of dividing the the given array by the given scalar.
   *
   * @param a the array.
   * @param v the scalar.
   */
  
  public static double[] divide (double[] a, double v) {
    double[] ans = clone (a);
    divideInto (ans, v);
    return ans;
  }

  /**
   * Returns the result of dividing the the given array by the given scalar.
   *
   * @param a the array.
   * @param v the scalar.
   */
  
  public static double[][] divide (double[][] a, double v) {
    double[][] ans = clone (a);
    divideInto (ans, v);
    return ans;
  }

  /**
   * Returns the result of dividing the the given array by the given scalar.
   *
   * @param a the array.
   * @param v the scalar.
   */
  
  public static double[][][] divide (double[][][] a, double v) {
    double[][][] ans = clone (a);
    divideInto (ans, v);
    return ans;
  }


  /**
   * Treating the elements of the array as consisting of m sets of n
   * consecutive elements, normalizes each set to sum to 1.0.
   *
   * @param a the array.
   * @param n the number of elements in a set.
   */
  
  public static void normalizeInto (double[] a, int n) throws Exception {
    for (int i = 0; i < a.length; i += n) {
      double sum = 0;
      for (int j = 0; j < n; j++) {
        sum += a[i + j];
      }
      if (sum == 0) {
        throw new Exception ("Cannot normalize vector that sums to zero.");
      }
      for (int j = 0; j < n; j++) {
        a[i + j] /= sum;
      }
    }
  }

  /**
   * Normalizes the elements in the given array to sum to 1.0.
   *
   * @param a the array.
   */

  public static void normalizeInto (double[] a) throws Exception {
    double s = sum (a);
    if (s == 0) {
      throw new Exception ("Cannot normalize vector that sums to zero.");
    }
    divideInto (a, s);
  }

  /**
   * Normalizes the elements in the given array to sum to 1.0.
   *
   * @param a the array.
   */
   
  public static void normalizeInto (double[][] a) throws Exception {
    double s = sum (a);
    if (s == 0) {
      throw new Exception ("Cannot normalize vector that sums to zero.");
    }
    divideInto (a, s);
  }
  
  /**
   * Normalizes the elements in the given array to sum to 1.0.
   *
   * @param a the array.
   */
   
  public static void normalizeInto (double[][][] a) throws Exception {
    double s = sum (a);
    if (s == 0) {
      throw new Exception ("Cannot normalize vector that sums to zero.");
    }
    divideInto (a, s);
  }
  
  /**
   * Treating the elements of the array as consisting of m sets of n
   * consecutive elements, normalizes each set to sum to 1.0.
   *
   * @param a the array.
   * @param n the number of elements in a set.
   */
  
  public static double[] normalize (double[] a, int n) throws Exception {
    double[] ans = clone (a);
    normalizeInto (ans, n);
    return ans;
  }
  
  /**
   * Returns the result of normalizing the elements in the given array to
   * sum to 1.0.
   *
   * @param a the array.
   */

  public static double[] normalize (double[] a) throws Exception {
    double[] ans = clone (a);
    normalizeInto (ans);
    return ans;
  }

  /**
   * Returns the result of normalizing the elements in the given array to
   * sum to 1.0.
   *
   * @param a the array.
   */
   
  public static double[][] normalize (double[][] a) throws Exception {
    double[][] ans = clone (a);
    normalizeInto (ans);
    return ans;
  }
  
  /**
   * Returns the result of normalizing the elements in the given array to
   * sum to 1.0.
   *
   * @param a the array.
   */
   
  public static double[][][] normalize (double[][][] a) throws Exception {
    double[][][] ans = clone (a);
    normalizeInto (ans);
    return ans;
  }

  /**
   * Returns the dot product of the two given arrays.  The arrays must have
   * the same dimesions.
   *
   * @param a1 the first array.
   * @param a2 the second array.
   * @return the dot product.
   */

  public static double dot (double[] a1, double[] a2) {
    double ans = 0.0;
    for (int i = 0; i < a1.length; i++) {
      ans += a1[i] * a2[i];
    }
    return ans;
  }
  
  /**
   * Returns the dot product of the two given arrays.  The arrays must have
   * the same dimesions.
   *
   * @param a1 the first array.
   * @param a2 the second array.
   * @return the dot product.
   */

  public static double dot (double[][] a1, double[][] a2) {
    double ans = 0.0;
    for (int i = 0; i < a1.length; i++) {
      for (int j = 0; j < a1[i].length; j++) {
        ans += a1[i][j] * a2[i][j];
      }
    }
    return ans;
  }
  
  /**
   * Returns the dot product of the two given arrays.  The arrays must have
   * the same dimesions.
   *
   * @param a1 the first array.
   * @param a2 the second array.
   * @return the dot product.
   */

  public static double dot (double[][][] a1, double[][][] a2) {
    double ans = 0.0;
    for (int i = 0; i < a1.length; i++) {
      for (int j = 0; j < a1[i].length; j++) {
        for (int k = 0; k < a1[i][j].length; k++) {
          ans += a1[i][j][k] * a2[i][j][k];
        }
      }
    }
    return ans;
  }
  
  /**
   * Returns the sum of the given array's elements.
   *
   * @param a the array.
   * @return the sum.
   */

  public static double sum (double[] a) {
    double ans = 0.0;
    for (int x1 = 0; x1 < a.length; x1++) {
      ans += a[x1];
    }
    return ans;
  }

  /**
   * Returns the sum of the given array's elements.
   *
   * @param a the array.
   * @return the sum.
   */

  public static double sum (double[][] a) {
    double ans = 0.0;
    for (int x1 = 0; x1 < a.length; x1++) {
      ans += sum (a[x1]);
    }
    return ans;
  }

  /**
   * Returns the sum of the given array's elements.
   *
   * @param a the array.
   * @return the sum.
   */

  public static double sum (double[][][] a) {
    double ans = 0.0;
    for (int x1 = 0; x1 < a.length; x1++) {
      ans += sum (a[x1]);
    }
    return ans;
  }

  /**
   * Returns the indices of the minimum element in the given array.
   *
   * @param a the array.
   * @return the indices.
   */ 

  public static int argmin (int[] a) {
    int ans = 0;
    for (int c = 0; c < a.length; c++) {
      if (a[c] < a[ans]) {
        ans = c;
      }
    }
    return ans;
  }

  /**
   * Returns the indices of the minimum element in the given array.
   *
   * @param a the array.
   * @return the indices.
   */ 

  public static int[] argmin (int[][] a) {
    int ans0 = 0;
    int ans1 = 0;
    for (int c0 = 0; c0 < a.length; c0++) {
      for (int c1 = 0; c1 < a[c0].length; c1++) {
        if (a[c0][c1] < a[ans0][ans1]) {
          ans0 = c0;
          ans1 = c1;
        }
      }
    }
    return new int[] {ans0, ans1};
  }

  /**
   * Returns the indices of the minimum element in the given array.
   *
   * @param a the array.
   * @return the indices.
   */ 

  public static int[] argmin (int[][][] a) {
    int ans0 = 0;
    int ans1 = 0;
    int ans2 = 0;
    for (int c0 = 0; c0 < a.length; c0++) {
      for (int c1 = 0; c1 < a[c0].length; c1++) {
        for (int c2 = 0; c2 < a[c0][c1].length; c2++) {
          if (a[c0][c1][c2] < a[ans0][ans1][ans2]) {
            ans0 = c0;
            ans1 = c1;
            ans2 = c2;
          }
        }
      }
    }
    return new int[] {ans0, ans1, ans2};
  }

  /**
   * Returns the indices of the maximum element in the given array.
   *
   * @param a the array.
   * @return the indices.
   */ 

  public static int argmax (int[] a) {
    int ans = 0;
    for (int c = 0; c < a.length; c++) {
      if (a[c] > a[ans]) {
        ans = c;
      }
    }
    return ans;
  }

  /**
   * Returns the indices of the minimum element in the given array.
   *
   * @param a the array.
   * @return the indices.
   */ 

  public static int[] argmax (int[][] a) {
    int ans0 = 0;
    int ans1 = 0;
    for (int c0 = 0; c0 < a.length; c0++) {
      for (int c1 = 0; c1 < a[c0].length; c1++) {
        if (a[c0][c1] > a[ans0][ans1]) {
          ans0 = c0;
          ans1 = c1;
        }
      }
    }
    return new int[] {ans0, ans1};
  }

  /**
   * Returns the indices of the minimum element in the given array.
   *
   * @param a the array.
   * @return the indices.
   */ 

  public static int[] argmax (int[][][] a) {
    int ans0 = 0;
    int ans1 = 0;
    int ans2 = 0;
    for (int c0 = 0; c0 < a.length; c0++) {
      for (int c1 = 0; c1 < a[c0].length; c1++) {
        for (int c2 = 0; c2 < a[c0][c1].length; c2++) {
          if (a[c0][c1][c2] > a[ans0][ans1][ans2]) {
            ans0 = c0;
            ans1 = c1;
            ans2 = c2;
          }
        }
      }
    }
    return new int[] {ans0, ans1, ans2};
  }

  /**
   * Returns the minimum element in the given array.
   *
   * @param a the array.
   * @return the minimum element.
   */ 

  public static int min (int[] a) {
    return a[argmin (a)];
  }

  /**
   * Returns the minimum element in the given array.
   *
   * @param a the array.
   * @return the minimum element.
   */ 

  public static int min (int[][] a) {
    int[] indices = argmin (a);
    return a[indices[0]][indices[1]];
  }

  /**
   * Returns the minimum element in the given array.
   *
   * @param a the array.
   * @return the minimum element.
   */ 

  public static int min (int[][][] a) {
    int[] indices = argmin (a);
    return a[indices[0]][indices[1]][indices[2]];
  }

  /**
   * Returns the maximum element in the given array.
   *
   * @param a the array.
   * @return the minimum element.
   */ 

  public static int max (int[] a) {
    return a[argmax (a)];
  }

  /**
   * Returns the maximum element in the given array.
   *
   * @param a the array.
   * @return the minimum element.
   */ 

  public static int max (int[][] a) {
    int[] indices = argmax (a);
    return a[indices[0]][indices[1]];
  }

  /**
   * Returns the maximum element in the given array.
   *
   * @param a the array.
   * @return the minimum element.
   */ 

  public static int max (int[][][] a) {
    int[] indices = argmax (a);
    return a[indices[0]][indices[1]][indices[2]];
  }

  /**
   * Returns the indices of the minimum element in the given array.
   *
   * @param a the array.
   * @return the indices.
   */ 

  public static int argmin (double[] a) {
    int ans = 0;
    for (int c = 0; c < a.length; c++) {
      if (a[c] < a[ans]) {
        ans = c;
      }
    }
    return ans;
  }

  /**
   * Returns the indices of the minimum element in the given array.
   *
   * @param a the array.
   * @return the indices.
   */ 

  public static int[] argmin (double[][] a) {
    int ans0 = 0;
    int ans1 = 0;
    for (int c0 = 0; c0 < a.length; c0++) {
      for (int c1 = 0; c1 < a[c0].length; c1++) {
        if (a[c0][c1] < a[ans0][ans1]) {
          ans0 = c0;
          ans1 = c1;
        }
      }
    }
    return new int[] {ans0, ans1};
  }

  /**
   * Returns the indices of the minimum element in the given array.
   *
   * @param a the array.
   * @return the indices.
   */ 

  public static int[] argmin (double[][][] a) {
    int ans0 = 0;
    int ans1 = 0;
    int ans2 = 0;
    for (int c0 = 0; c0 < a.length; c0++) {
      for (int c1 = 0; c1 < a[c0].length; c1++) {
        for (int c2 = 0; c2 < a[c0][c1].length; c2++) {
          if (a[c0][c1][c2] < a[ans0][ans1][ans2]) {
            ans0 = c0;
            ans1 = c1;
            ans2 = c2;
          }
        }
      }
    }
    return new int[] {ans0, ans1, ans2};
  }

  /**
   * Returns the indices of the maximum element in the given array.
   *
   * @param a the array.
   * @return the indices.
   */ 

  public static int argmax (double[] a) {
    int ans = 0;
    for (int c = 0; c < a.length; c++) {
      if (a[c] > a[ans]) {
        ans = c;
      }
    }
    return ans;
  }

  /**
   * Returns the indices of the minimum element in the given array.
   *
   * @param a the array.
   * @return the indices.
   */ 

  public static int[] argmax (double[][] a) {
    int ans0 = 0;
    int ans1 = 0;
    for (int c0 = 0; c0 < a.length; c0++) {
      for (int c1 = 0; c1 < a[c0].length; c1++) {
        if (a[c0][c1] > a[ans0][ans1]) {
          ans0 = c0;
          ans1 = c1;
        }
      }
    }
    return new int[] {ans0, ans1};
  }

  /**
   * Returns the indices of the minimum element in the given array.
   *
   * @param a the array.
   * @return the indices.
   */ 

  public static int[] argmax (double[][][] a) {
    int ans0 = 0;
    int ans1 = 0;
    int ans2 = 0;
    for (int c0 = 0; c0 < a.length; c0++) {
      for (int c1 = 0; c1 < a[c0].length; c1++) {
        for (int c2 = 0; c2 < a[c0][c1].length; c2++) {
          if (a[c0][c1][c2] > a[ans0][ans1][ans2]) {
            ans0 = c0;
            ans1 = c1;
            ans2 = c2;
          }
        }
      }
    }
    return new int[] {ans0, ans1, ans2};
  }

  /**
   * Returns the minimum element in the given array.
   *
   * @param a the array.
   * @return the minimum element.
   */ 

  public static double min (double[] a) {
    return a[argmin (a)];
  }

  /**
   * Returns the minimum element in the given array.
   *
   * @param a the array.
   * @return the minimum element.
   */ 

  public static double min (double[][] a) {
    int[] indices = argmin (a);
    return a[indices[0]][indices[1]];
  }

  /**
   * Returns the minimum element in the given array.
   *
   * @param a the array.
   * @return the minimum element.
   */ 

  public static double min (double[][][] a) {
    int[] indices = argmin (a);
    return a[indices[0]][indices[1]][indices[2]];
  }

  /**
   * Returns the maximum element in the given array.
   *
   * @param a the array.
   * @return the minimum element.
   */ 

  public static double max (double[] a) {
    return a[argmax (a)];
  }

  /**
   * Returns the maximum element in the given array.
   *
   * @param a the array.
   * @return the minimum element.
   */ 

  public static double max (double[][] a) {
    int[] indices = argmax (a);
    return a[indices[0]][indices[1]];
  }

  /**
   * Returns the maximum element in the given array.
   *
   * @param a the array.
   * @return the minimum element.
   */ 

  public static double max (double[][][] a) {
    int[] indices = argmax (a);
    return a[indices[0]][indices[1]][indices[2]];
  }
  
  /**
   * A deep copy of a two-dimensional double array.
   *
   * @param a the array.
   * @return the copy.
   */
  
  public static double[][] clone (double[][] a) {
    double[][] ans = new double[a.length][];
    for (int i = 0; i < ans.length; i++) {
      ans[i] = clone (a[i]);
    }
    return ans;
  }
  
  /**
   * A deep copy of a three-dimensional double array.
   *
   * @param a the array.
   * @return the copy.
   */
  
  public static double[][][] clone (double[][][] a) {
    double[][][] ans = new double[a.length][][];
    for (int i = 0; i < ans.length; i++) {
      ans[i] = clone (a[i]);
    }
    return ans;
  }
  
  /**
   * Fills a double[][] with the given value.
   *
   * @param a the array.
   * @param val the value.
   */

  public static void fill (double[][] a, double val) {
    for (int x1 = 0; x1 < a.length; x1++) {
      Arrays.fill (a[x1], val);
    }
  }
 
  /**
   * Fills a double[][][] with the given value.
   *
   * @param a the array.
   * @param val the value.
   */

  public static void fill (double[][][] a, double val) {
    for (int x1 = 0; x1 < a.length; x1++) {
      fill (a[x1], val);
    }
  }

  /**
   * Returns the median of the elements in the given sorted array.
   *
   * @param a the array.
   * @return the median.
   */

  public static double median (double[] a) {
    return
    (a.length % 2 == 1) ?
    a[a.length / 2] :
    ((a[a.length / 2 - 1] + a[a.length / 2]) / 2.0);
  }

  /**
    * Returns the mean of the elements in the given array.
   *
   * @param a the array.
   * @return the mean.
   */

  public static double mean (double[] a) {
    double sum = 0.0;
    for (int i = 0; i < a.length; i++) {
      sum += a[i];
    }
    return sum / a.length;
  }

  /**
   * Returns the standard deviation of the elements in the given array.
   *
   * @param a the array.
   * @return the standard deviation.
   */

  public static double standardDeviation (double[] a) {
    double meanOfSquare = 0.0;
    double mean = 0.0;
    for (int i = 0; i < a.length; i++) {
      mean += a[i];
      meanOfSquare += a[i] * a[i];
    }
    meanOfSquare /= a.length;
    mean /= a.length;
    return Math.sqrt (meanOfSquare - mean * mean);
  }

  /**
   * Returns the magnitude of the given array.
   *
   * @return the magnitude of the given array.
   */
  
  public static double magnitude (double[] a) {
    return Math.sqrt (dot (a, a));
  }
  
  /**
   * Normalizes the magnitude of the given array.
   *
   * @return the old magnitude.
   */
  
  public static double normalizeMagnitude (double[] a) {
    double ans = magnitude (a);
    if (ans != 0.0) {
      divideInto (a, ans);
    }
    return ans;
  }

  /**
   * Returns the KL-divergence of two distributions over the same discrete
   * random variable.
   *
   * @param p the first distribution.
   * @param q the second distribution.
   * @return the KL divergence.
   */
  
  public static double klDivergence (double[] p, double[] q) {
    double ans = 0;
    for (int i = 0; i < p.length; i++) {
      if (p[i] != 0 && q[i] == 0) {
        return Double.POSITIVE_INFINITY;
      }
      if (p[i] != 0) {
        ans += p[i] * mark.core.math.Util.log2 (p[i] / q[i]);
      }
    }
    return ans;
  }

  /**
   * Returns the information radius of two distributions over the same
   * discrete random variable.  This metric measures the distance between the
   * distributions and ranges from 0 for identical distributions to 2 log_2
   * for most dissimilar.  It answers the question, how much information is
   * lost if we describe the random variable with the average of the two
   * distributions, assuming one of them is the correct distribution.  For
   * more information, see "Foundations of statistical natural language
   * processing" by Manning and Schutze page 304 for more information.
   *
   * @param p one of the distributions.
   * @param q the other distribution.
   * @return the information radius.
   */
  
  public static double informationRadius (double[] p, double[] q) {
    double ans = 0.0;
    for (int i = 0; i < p.length; i++) {
      if (p[i] != 0) {
        ans +=
          p[i] * mark.core.math.Util.log2 (p[i] / ((p[i] + q[i]) / 2.0));
      }
      if (q[i] != 0) {
        ans +=
          q[i] * mark.core.math.Util.log2 (q[i] / ((p[i] + q[i]) / 2.0));
      }
    }
    return ans;
  }

  /**
   * Computes the absolute value of the difference between each pair of
   * corresponding elements in the given arrays and returns the average.
   *
   * @param p the first given array.
   * @param q the second given array.
   * @return the sum.
   */
  
  public static double difference (double[] p, double[] q) {
    double ans = 0.0;
    for (int i = 0; i < p.length; i++) {
      ans += Math.abs (p[i] - q[i]);
    }
    return ans / p.length;
  }

  /**
   * Given a distribution on a finite set of values, samples a value
   * uniformly from those values that have non-zero probability.  The values
   * are assumed to be {0, 1, ..., L}, and the distribution is specified as a
   * double[], where indices are values and values are probabilities.  This
   * algorithm runs in O (L) time.  If sampling multiple times, then the
   * version of this method that generates multiple samples is more
   * efficient.
   *
   * @param r the random generator.
   * @param d the distribution.
   * @return the value sampled.
   */
  
  public static int sampleUniformly (Random r, double[] d) {
    int numNonzero = d.length - mark.core.util.Util.numValues (d, 0.0);
    int which = r.nextInt (numNonzero);
    int ans = -1;
    for (int i = 0; i < d.length; i++) {
      if (d[i] != 0) {
        if (which == 0) {
          ans = i;
          break;
        }
        --which;
      }
    }
    return ans;
  }

  /**
   * Given a distribution on a finite set of values, samples a value
   * according to the distribution.  The values are assumed to be {0, 1, ...,
   * L}, and the distribution is specified as a double[], where indices are
   * values and values are probabilities.  This algorithm runs in O (L) time.
   * If sampling multiple times, then the version of this method that
   * generates multiple samples is more efficient.
   *
   * @param r the random generator.
   * @param d the distribution.
   * @return the value sampled.
   */
  
  public static int sampleAccordingToModel (Random r, double[] d) {
    double s = r.nextDouble ();
    double mass = 0.0;
    for (int i = 0; i < d.length; i++) {
      mass += d[i];
      if (s < mass) {
        return i;
      }
    }
    return d.length - 1; // possible that we don't add up to precisely 1.0.
  }
  
  /**
   * Given a distribution on a finite set of values, samples a value
   * uniformly from those values that have non-zero probability N times.  The
   * values are assumed to be {0, 1, ..., L}, and the distribution is
   * specified as a double[], where indices are values and values are
   * probabilities.  This algorithm runs in O (N log (L) + L) time and is
   * therefore more efficent than calling the single-sample version of this
   * method N times.
   *
   * @param r the random generator.
   * @param d the distribution.
   * @param N the number of samples.
   * @return the value sampled.
   */
  
  public static int[] sampleUniformly (
   Random r, double[] d, int N) {
    int numNonzero = d.length - mark.core.util.Util.numValues (d, 0.0);
    int[] nonZero = new int[numNonzero];
    int index = 0;
    for (int i = 0; i < d.length; i++) {
      if (d[i] != 0) {
        nonZero[index++] = i;
      }
    }
    int[] ans = new int[N];
    for (int i = 0; i < ans.length; i++) {
      ans[i] = nonZero[r.nextInt (numNonzero)];
    }
    return ans;
  }

  /**
   * Given a distribution on a finite set of values, samples a value
   * according to the distribution N times.  The values are assumed to be
   * {0, 1, ..., L}, and the distribution is specified as a double[], where
   * indices are values and values are probabilities.  This algorithm runs in
   * O (N log (L) + L) time and is therefore more efficent than calling the
   * single-sample version of this method N times.
   *
   * @param r the random generator.
   * @param d the distribution.
   * @param N the number of samples.
   * @return the value sampled.
   */
  
  public static int[] sampleAccordingToModel (
   Random r, double[] d, int N) {
    int[] ans = new int[N];
    Partition p = new Partition (d);
    for (int i = 0; i < ans.length; i++) {
      ans[i] = p.select (r.nextDouble ());
    }
    return ans;
  }
  
}
