package mark.core.math;

/**
 * Defines a consecutive range of real numbers, where each end may be open
 * or closed.
 * <p>
 * This class conforms to the immutable design pattern.
 *
 * @author Mark Chavira
 */

public final class Interval implements Cloneable, Comparable {

  private static java.text.DecimalFormat fFormat;
  static {
    java.text.DecimalFormatSymbols s = new java.text.DecimalFormatSymbols ();
    s.setInfinity ("inf");
    fFormat = new java.text.DecimalFormat ("0.######", s);
  }

  private boolean fLowerOpen;
  private double fLower;
  private double fUpper;
  private boolean fUpperOpen;

  private static Interval fTrivialInterval =
    new Interval (
      true, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY, true);

  /**
   * @see java.lang.Object
   */

  public boolean equals (Object o) {
    if (!(o instanceof Interval)) {
      return false;
    }
    Interval i = (Interval)o;
    return
      fLowerOpen == i.fLowerOpen &&
      fLower == i.fLower &&
      fUpper == i.fUpper &&
      fUpperOpen == i.fUpperOpen;
  }

  /**
   * @see java.lang.Object
   */
  
  public String toString () {
    return
      (fLowerOpen ? "(" : "[") +
      fFormat.format (fLower) + "," +
      fFormat.format (fUpper) + (fUpperOpen ? ")" : "]");
  }

  /**
   * @see java.lang.Object
   */

  public Object clone () throws java.lang.CloneNotSupportedException {
    return this;
  }

  /**
   * @see java.lang.Object
   */

  public int hashCode() {
    return
      (Boolean.valueOf (fLowerOpen)).hashCode () ^
      (Double.valueOf (fLower)).hashCode () ^
      (Double.valueOf (fUpper)).hashCode () ^
      (Boolean.valueOf (fUpperOpen)).hashCode ();
  }

  /**
   * @see java.lang.Comparable
   */
   
  public int compareTo (Object o) {
    Interval i = (Interval)o;
    int t = compareLower (this, i);
    if (t != 0) {
      return t;
    }
    return compareUpper (this, i);
  }

  /**
   * Constructs an interval.
   *
   * @param lowerOpen whether the lower bound is open.
   * @param lower the lower bound.
   * @param upper the upper bound.
   * @param upperOpen whether the upper bound is open.
   */

  public Interval (
   boolean lowerOpen, double lower, double upper, boolean upperOpen) {
    fLowerOpen = lowerOpen;
    fLower = lower;
    fUpper = upper;
    fUpperOpen = upperOpen;
  }

  /**
   * Returns whether the interval is empty.
   *
   * @return whether the interval is empty.
   */

  public boolean empty () {
    return fLower > fUpper || (fLower == fUpper && !closedClosed ());
  }

  /**
   * Returns whether the interval is all real-numbers.
   *
   * @return whether the interval is all real-numbers.
   */

  public boolean full () {
    return fLower == Double.NEGATIVE_INFINITY &&
      fUpper == Double.POSITIVE_INFINITY;
  }
  
  /**
   * Returns whether the lower bound is open.
   *
   * @return whether the lower bound is open.
   */

  public boolean lowerOpen () {
    return fLowerOpen;
  }

  /**
   * Returns whether the upper bound is open.
   *
   * @return whether the upper bound is open.
   */

  public boolean upperOpen () {
    return fUpperOpen;
  }

  /**
   * Returns whether the interval is open on both sides.
   *
   * @return whether the interval is open on both sides.
   */
  
  public boolean openOpen () {
    return fLowerOpen & fUpperOpen;
  }

  /**
   * Returns whether the interval is open on the left and closed on the
   * right.
   *
   * @return whether the interval is open on the left and closed on the
   *   right.
   */

  public boolean openClosed () {
    return fLowerOpen & !fUpperOpen;
  }

  /**
   * Returns whether the interval is closed on the left and open on the
   * right.
   *
   * @return whether the interval is open on the left and closed on the
   *   right.
   */

  public boolean closedOpen () {
    return !fLowerOpen & fUpperOpen;
  }

  /**
   * Returns whether the interval is closed on both sides.
   *
   * @return whether the interval is closed on both sides.
   */

  public boolean closedClosed () {
    return !fLowerOpen & !fUpperOpen;
  }
  
  /**
   * Returns the lower bound of the interval.
   *
   * @return the lower bound of the interval.
   */

  public double lower () {return fLower;}

  /**
   * Returns the upper bound of the interval.
   *
   * @return the upper bound of the interval.
   */
  
  public double upper () {return fUpper;}

  /**
   * Returns a value in the interval.  Throws an exception if the interval is
   * empty.  Returns zero for the trivial interval.  Otherwise, if the lower
   * bound is -inf, returns the upper bound minus one.  Otherwise, if the
   * upper bound is inf, returns the lower bound plus one.  Otherwise,
   * returns the sum of the lower and upper bounds divided by two.
   *
   * @return a value in the interval.
   * @throws Exception if the interval is empty.
   */

  public double someValue () throws Exception {
    if (empty ()) {
      throw new Exception ("Interval is empty.");
    }
    if (lower () == Double.NEGATIVE_INFINITY) {
      return (upper () == Double.POSITIVE_INFINITY) ? 0.0 : upper () - 1.0;
    }
    return
      (upper () == Double.POSITIVE_INFINITY) ?
      fLower + 1.0 :
      ((fLower + fUpper) / 2.0);
  }

  /**
   * Returns whether the interval precedes the given value.  An empty
   * interval precedes nothing.
   *
   * @param value the value.
   * @return whether the upper bound precedes the given value.
   */

  public boolean precedes (double value) {
    return
      !empty () &&
      (upper () < value || upper () == value && upperOpen ());
  }

  /**
   * Returns whether the interval follows the given value.  An empty interval
   * follows nothing.
   *
   * @param value the value.
   * @return whether the lower bound follows the given value.
   */

  public boolean follows (double value) {
    return
      !empty () &&
      (lower () > value || lower () == value && lowerOpen ());
  }

  /**
   * Returns wether the interval contains the given value.
   *
   * @param value the value.
   * @return whether the interval contains the value.
   */
  
  public boolean contains (double value) {
    return !empty () && !precedes (value) && !follows (value);
  }
 
  /**
   * Returns the trivial interval that covers all reals.
   *
   * @return the interval.
   */
  
  public static Interval trivialInterval () {
    return fTrivialInterval;
  }

  /**
   * Returns whether two intervals are adjacent.  Two intervals are adjacent
   * if their intersection is empty and the smallest interval that contains
   * both intervals is their union.
   *
   * @param lhs the first interval.
   * @param rhs the secnd interval.
   * @return whether the two intervala are adjacent.
   */

  public static boolean adjacent (Interval lhs, Interval rhs) {
    if (lhs.empty () || rhs.empty ()) return false;
    return
      lhs.fUpper == rhs.fLower &&
      lhs.fUpperOpen != rhs.fLowerOpen ||
      rhs.fUpper == lhs.fLower &&
      rhs.fUpperOpen != lhs.fLowerOpen;
  }
  
  /**
   * Returns the intersection of two intervals.
   *
   * @param lhs the first interval.
   * @param rhs the secnd interval.
   * @return the intersection.
   */

  public static Interval intersection (Interval lhs, Interval rhs) {
    if (lhs.empty ()) {
      return lhs;
    }
    if (rhs.empty ()) {
      return rhs;
    }
    Interval lower = (compareLower (lhs, rhs) >= 0) ? lhs : rhs;
    Interval upper = (compareUpper (lhs, rhs) <= 0) ? lhs : rhs;
    return new Interval (
      lower.lowerOpen (), lower.lower (), upper.upper (),
      upper.upperOpen ());
  }

  /**
   * Returns the smallest interval that contains all numbers in two intervals.
   * intervals.
   *
   * @param lhs the first interval.
   * @param rhs the secnd interval.
   * @return the combined interval.
   */

  public static Interval combine (Interval lhs, Interval rhs) {
    if (lhs.empty ()) {
      return rhs;
    }
    if (rhs.empty ()) {
      return lhs;
    }
    Interval lower = (compareLower (lhs, rhs) <= 0) ? lhs : rhs;
    Interval upper = (compareUpper (lhs, rhs) >= 0) ? lhs : rhs;
    return new Interval (
      lower.lowerOpen (), lower.lower (), upper.upper (),
      upper.upperOpen ());
  }

  /**
   * Compares the lower bounds of two intervals.  Returns -1 if the lower
   * bound of the first is less than the lower bound of the second or if the
   * bounds are equal and the first is closed and the second is open.
   * Returns +1 if the lower bound of the first is greater than the lower
   * bound of the second or if the bounds are equal and the first is open
   * and the second is closed.  Otherwise, returns 0.
   *
   * @param lhs the first interval.
   * @param rhs the second interval.
   * @return the result of the comparison.
   */

  public static int compareLower (Interval lhs, Interval rhs) {
    if (lhs.lower () < rhs.lower ()) {
      return -1;
    }
    if (lhs.lower () > rhs.lower ()) {
      return +1;
    }
    if (!lhs.lowerOpen () && rhs.lowerOpen ()) {
      return -1;
    }
    if (lhs.lowerOpen () && !rhs.lowerOpen ()) {
      return +1;
    }
    return 0;
  }
  
  /**
   * Compares the upper bounds of two intervals.  Returns -1 if the upper
   * bound of the first is less than the upper bound of the second or if
   * the bounds are equal and the first is open and the second is closed.
   * Returns +1 if the upper bound of the first is greater than the upper
   * bound of the second or if the bounds are equal and the first is closed
   * and the second is open.  Otherwise, returns 0.
   *
   * @param lhs the first interval.
   * @param rhs the second interval.
   * @return the result of the comparison.
   */

  public static int compareUpper (Interval lhs, Interval rhs) {
    if (lhs.upper () < rhs.upper ()) {
      return -1;
    }
    if (lhs.upper () > rhs.upper ()) {
      return +1;
    }
    if (lhs.upperOpen () && !rhs.upperOpen ()) {
      return -1;
    }
    if (!lhs.upperOpen () && rhs.upperOpen ()) {
      return +1;
    }
    return 0;
  }

}
