package mark.core.math;
import java.util.*;

/**
 * Defines a partition of an interval of real numbers that uses significantly
 * less space space than storing an array of intervals.
 * <p>
 * This class conforms to the immutable design pattern.
 *
 * @author Mark Chavira
 */

public final class Partition implements Comparable {
  
  //=========================================================================
  // Classes:
  //=========================================================================

  /**
   * A class that represents an upper bound.  Assists construction
   * algorithms.
   *
   * @author Mark Chavira
   */
  
  private static class UpperBound implements Comparable {
    double fBound;
    boolean fOpen;
    public UpperBound (double bound, boolean open) {
      fBound = bound;
      fOpen = open;
    }
    public int compareTo (Object o) {
      UpperBound p = (UpperBound)o;
      return compareUpper (fBound, fOpen, p.fBound, p.fOpen);
    }
    public boolean equals (Object o) {
      UpperBound p = (UpperBound)o;
      return fBound == p.fBound && fOpen == p.fOpen;
    }
  }

  //=========================================================================
  // Fields
  //=========================================================================

  private static Partition fTrivialPartition = new Partition ();
  
  private boolean fOverallLowerOpen;
  private double fOverallLower;
  private double[] fUpper;
  private boolean[] fUpperOpen;

  //=========================================================================
  // Fulfillments
  //=========================================================================

  /**
   * @see java.lang.Object
   */
    
  public boolean equals (Object o) {
    if (!(o instanceof Partition)) {
      return false;
    }
    Partition p = (Partition)o;
    if (
     fOverallLower != p.fOverallLower ||
     fOverallLowerOpen != p.fOverallLowerOpen ||
     length () != p.length ()) {
      return false;
    }
    for (int i = 0; i < fUpper.length; i++) {
      if (
       fUpper[i] != p.fUpper[i] ||
       fUpperOpen[i] != p.fUpperOpen[i]) {
        return false;
      }
    }
    return true;
  }

  /**
   * @see java.lang.Object
   */
  
  public String toString () {
    return "{" + mark.core.text.Util.join (",", allIntervals ()) + "}";
  }

  /**
   * @see java.lang.Comparable
   */

  public int compareTo (Object o) {
    Partition p = (Partition)o;
    int temp =
      compareLower (
        fOverallLower, fOverallLowerOpen,
        p.fOverallLower, p.fOverallLowerOpen);
    if (temp != 0) {
      return temp;
    }
    temp =
      length () < p.length () ? -1 :
      length () > p.length () ? +1 :
      0;
    if (temp != 0) {
      return temp;
    }
    for (int i = 0; i < fUpper.length; i++) {
      temp =
        compareUpper (
          fUpper[i], fUpperOpen[i], p.fUpper[i], p.fUpperOpen[i]);
      if (temp != 0) {
        return temp;
      }
    }
    return 0;
  }

  //=========================================================================
  // Constructors
  //=========================================================================

  /**
   * Initializes the partition from the given overall-lower-bound and sorted
   * upper-bounds.  The object assumes ownership of the given arrays.
   *
   * @param overallLowerOpen whether the overall-lower-bound is open.
   * @param overallLower the overall-lower-bound.
   * @param upper the upper array.
   * @param upperOpen the upper open array.
   */
  
  private Partition (
   boolean overallLowerOpen, double overallLower,
   double[] upper, boolean[] upperOpen) {
    fOverallLowerOpen = overallLowerOpen;
    fOverallLower = overallLower;
    fUpper = upper;
    fUpperOpen = upperOpen;
  }

  /**
   * Initializes the partition to {(-inf, +inf)}.
   */
  
  private Partition () {
    this (
      true, Double.NEGATIVE_INFINITY,
      new double[] {Double.POSITIVE_INFINITY}, new boolean[] {true});
  }

  /**
   * Returns the partition that contains one interval with the given overall-
   * lower-bound and the given overall-upper-bound.
   *
   * @param olbo whether the overall-lower-bound is open.
   * @param olb the overall-lower-bound.
   * @param oub the overall-upper-bound.
   * @param oubo whether the overall-lower-bound is open.
   */
  
  public Partition (boolean olbo, double olb, double oub, boolean oubo) {
    this (olbo, olb, new double[] {oub}, new boolean[] {oubo});
  }

  /**
   * Initializes the partition from the given array of intervals.  Assumes
   * that the given array is a sorted partition.  Copies the data, so the
   * input parameter may be destroyed afterwards.
   *
   * @param partition the array of intervals.
   */

  public Partition (Interval[] partition) {
    fOverallLowerOpen = partition[0].lowerOpen ();
    fOverallLower = partition[0].lower ();
    fUpper = new double[partition.length];
    fUpperOpen = new boolean[partition.length];
    for (int i = 0; i < fUpper.length; i++) {
      fUpper[i] = partition[i].upper ();
      fUpperOpen[i] =  partition[i].upperOpen ();
    }
  }

  /**
   * Initializes the partition from the given list of intervals.  Assumes
   * that the given list is a sorted partition.  Copies the data, so the
   * input parameter may be destroyed afterwards.
   *
   * @param partition the array of intervals.
   */

  public Partition (List<Interval> partition) {
    fOverallLowerOpen = partition.get (0).lowerOpen ();
    fOverallLower = partition.get (0).lower ();
    fUpper = new double[partition.size ()];
    fUpperOpen = new boolean[partition.size ()];
    int index = 0;
    for (Interval interval : partition) {
      fUpper[index] = interval.upper ();
      fUpperOpen[index++] = interval.upperOpen ();
    }
  }
  
  /**
   * Initializes a partition from a distribution over a finite set of values.
   * The input distribution is in the form of a double[] where indices
   * denote different values and values denote the associated probability
   * mass.  The values must add to 1.0.  Suppose there are n values.  Then
   * there will be n intervals in the partition, and interval i will have
   * width corresponding to value i's probability mass.  The partition can
   * then be used for sampling by calling select () with randomly generated
   * numbers in [0, 1).
   *
   * @param d the distribution.
   */
  
  public Partition (double[] d) {
    fOverallLowerOpen = false;
    fOverallLower = 0.0;
    fUpper = new double[d.length];
    fUpper[0] = d[0];
    for (int i = 1; i < d.length - 2; i++) {
      fUpper[i] = fUpper[i - 1] + d[i];
    }
    fUpper[d.length - 1] = 1.0;
    fUpperOpen = new boolean[d.length];
    Arrays.fill (fUpperOpen, true);
  }
  
  /**
   * Returns the partition that denotes the single interval, from negative
   * infinity to positive infinity.
   *
   * @return the partition.
   */

  public static Partition trivialPartition () {
    return fTrivialPartition;
  }
  
  /**
   * Sorts the given array of upper-bounds and then constructs a partition
   * from the given overall-lower-bound and the sorted upper-bounds.  The
   * input upper-bounds may include duplicates.
   *
   * @param olbo whether the overall-lower-bound is open.
   * @param olb the overall-lower-bound.
   * @param ub the upper-bounds.
   * @param ubo whether the upper-bounds are open.
   * @param n the number of valid elements in ub.
   * @return the partition.
   */
  
  protected static Partition constructFromUpperBounds (
   boolean olbo, double olb, double[] ub, boolean[] ubo, int n) {

    // Sort the upper-bounds and count the number of unique upper-bounds.

    UpperBound[] ub2 = new UpperBound[n];
    for (int i = 0; i < n; i++) {
      ub2[i] = new UpperBound (ub[i], ubo[i]);
    }
    Arrays.sort (ub2);
    int numUnique = 1;
    for (int i = 1; i < n; i++) {
      if (!ub2[i].equals (ub2[i - 1])) {
        ++numUnique;
      }
    }

    // Put the unique upper-bounds into arrays.

    double[] unique = new double[numUnique];
    boolean[] uniqueOpen = new boolean[numUnique];
    int uniqueIndex = 0;
    unique[uniqueIndex] = ub2[0].fBound;
    uniqueOpen[uniqueIndex++] = ub2[0].fOpen;
    for (int i = 1; uniqueIndex < numUnique; i++) {
      if (!ub2[i].equals (ub2[i - 1])) {
        unique[uniqueIndex] = ub2[i].fBound;
        uniqueOpen[uniqueIndex++] = ub2[i].fOpen;
      }
    }

    // Return the result.
    
    return new Partition (olbo, olb, unique, uniqueOpen);

  }

  /**
   * Constructs a partition from two of partitions.  The partitions must have
   * the same overall-lower-bound and the same overall-upper-bound.  The
   * resulting partition will have the union of the upper bounds of the two
   * input partitions and corresponding lower bounds.
   *
   * @param lhs the first partition.
   * @param rhs the second partition.
   * @return the combined partition.
   */
   
  public static Partition constructFromPartitions (
   Partition lhs, Partition rhs) {
    int maxSize = lhs.fUpper.length + rhs.fUpper.length - 1;
    double[] upper = new double[maxSize];
    boolean[] upperOpen = new boolean[maxSize];
    int combinedIndex = 0;
    int leftIndex = 0;
    int rightIndex = 0;
    while (
     leftIndex < lhs.fUpper.length - 1 ||
     rightIndex < rhs.fUpper.length - 1) {
      boolean left;
      if (leftIndex == lhs.fUpper.length - 1) {
        left = false;
      } else if (rightIndex == rhs.fUpper.length - 1) {
        left = true;
      } else {
        int temp = compareUpper (
          lhs.fUpper[leftIndex], lhs.fUpperOpen[leftIndex],
          rhs.fUpper[rightIndex], rhs.fUpperOpen[rightIndex]);
        if (temp < 0) {
          left = true;
        } else if (temp > 0) {
          left = false;
        } else {
          left = true;
          ++rightIndex;
        }
      }
      if (left) {
        upper[combinedIndex] = lhs.fUpper[leftIndex];
        upperOpen[combinedIndex++] = lhs.fUpperOpen[leftIndex++];
      } else {
        upper[combinedIndex] = rhs.fUpper[rightIndex];
        upperOpen[combinedIndex++] = rhs.fUpperOpen[rightIndex++];
      }
    }
    upper[combinedIndex] = lhs.fUpper[lhs.fUpper.length - 1];
    upperOpen[combinedIndex++] = lhs.fUpperOpen[lhs.fUpperOpen.length - 1];
    if (combinedIndex == upper.length) {
      return new Partition (
        lhs.fOverallLowerOpen, lhs.fOverallLower, upper, upperOpen);
    }
    double[] upper2 = new double[combinedIndex];
    boolean[] upperOpen2 = new boolean[combinedIndex];
    System.arraycopy(upper, 0, upper2, 0, combinedIndex);
    System.arraycopy(upperOpen, 0, upperOpen2, 0, combinedIndex);
    return new Partition (
      lhs.fOverallLowerOpen, lhs.fOverallLower, upper2, upperOpen2);
  }

  /**
   * Constructs a partition from a list of partitions.  The partitions must
   * have the same overall-lower-bound and the same overall-upper-bound.  The
   * resulting partition p will be the partition with the fewest elements
   * such that for every for every interval i_1 in p and every interval i_2
   * an input partition, i_1 is disjoint from i_2 or i_1 is a subset of i_2.
   *
   * @param partitions the partitions.
   * @return the partition.
   */
  
  public static Partition constructFromPartitions (
   ArrayList<Partition> partitions) {

    // Count the number of upper-bounds.  We eventually want to eliminate all
    // duplicates.  Here, we make sure that there is only one occurrence of
    // the overall-upper-bound.
    
    int numUpperBounds = 1;
    for (Partition p : partitions) {
      numUpperBounds += p.length () - 1;
    }

    // Get the overall-lower-bound and the overall-upper-bound.
    
    Partition p = (Partition)partitions.get (0);
    boolean olbo = p.nthLowerOpen (0);
    double olb = p.nthLower (0);
    double[] ub = new double[numUpperBounds];
    boolean[] ubo = new boolean[numUpperBounds];
    ub[0] = p.nthUpper (p.length () - 1);
    ubo[0] = p.nthUpperOpen (p.length () - 1);

    // Get the other upper-bounds.

    int ubindex = 1;
    for (Partition pp : partitions) {
      for (int j = 0; j < pp.length () - 1; j++) {
        ub[ubindex] = pp.nthUpper (j);
        ubo[ubindex++] = pp.nthUpperOpen (j);
      }
    }
    
    // Return the result.
    
    return constructFromUpperBounds (olbo, olb, ub, ubo, ub.length);

  }

  /**
   * Constructs a partition from a given overall-lower-bound, overall-upper-
   * bound and list of intervals.  The intervals need not form a partition
   * but must all be within the overall-lower-bound and overall-upper-bound.
   *
   * @param olbo whether the overall-lower-bound is open.
   * @param olb the overall-lower-bound.
   * @param oub the overall-upper-bound.
   * @param oubo whether the overall-upper-bound is open.
   * @param intervals the intervals.
   * @return the partition.
   */
  
  public static Partition constructFromIntervals (
   boolean olbo, double olb, double oub, boolean oubo,
   ArrayList<Interval> intervals) {
    double[] ub = new double[2 * intervals.size () + 1];
    boolean[] ubo = new boolean[2 * intervals.size () + 1];
    ub[0] = oub;
    ubo[0] = oubo;
    int ubindex = 1;
    for (Interval l : intervals) {
      if (l.lower () != olb || l.lowerOpen () != olbo) {
        ub[ubindex] = l.lower ();
        ubo[ubindex++] = !l.lowerOpen ();
      }
      if (l.upper () != oub || l.upperOpen () != oubo) {
        ub[ubindex] = l.upper ();
        ubo[ubindex++] = l.upperOpen ();
      }
    }
    return constructFromUpperBounds (olbo, olb, ub, ubo, ubindex);
  }

  //=========================================================================
  // Methods
  //=========================================================================

  /**
   * Returns the overall-lower-bound.
   *
   * @return the overall-lower-bound.
   */
  
  public double overallLower () {
    return nthLower (0);
  }
  
  /**
   * Returns whether the overall-lower-bound is open.
   *
   * @return whether the overall-lower-bound is open.
   */
  
  public boolean overallLowerOpen () {
    return nthLowerOpen (0);
  }
  
  /**
   * Returns the overall-upper-bound.
   *
   * @return the overall-lower-bound.
   */
  
  public double overallUpper () {
    return nthUpper (length () - 1);
  }
  
  /**
   * Returns whether the overall-upper-bound is open.
   *
   * @return whether the overall-upper-bound is open.
   */
  
  public boolean overallUpperOpen () {
    return nthUpperOpen (length () - 1);
  }
  
  /**
   * Compares two lower bounds.
   *
   * @param lhsLower the first bound.
   * @param lhsLowerOpen whether the first bound is open.
   * @param rhsLower the second bound.
   * @param rhsLowerOpen whether the second bound is open.
   * @return -1, 0, or +1 depending on how they compare.
   */
  
  private static int compareLower (
   double lhsLower, boolean lhsLowerOpen, double rhsLower,
   boolean rhsLowerOpen) {
    int temp = Double.compare (lhsLower, rhsLower);
    if (temp != 0) {
      return temp;
    }
    return
      !lhsLowerOpen && rhsLowerOpen ? -1 :
      lhsLowerOpen && !rhsLowerOpen ? +1 :
      0;
  }

  /**
   * Compares two upper bounds.
   *
   * @param lhsUpper the first bound.
   * @param lhsUpperOpen whether the first bound is open.
   * @param rhsUpper the second bound.
   * @param rhsUpperOpen whether the second bound is open.
   * @return -1, 0, or +1 depending on how they compare.
   */
  
  private static int compareUpper (
   double lhsUpper, boolean lhsUpperOpen, double rhsUpper,
   boolean rhsUpperOpen) {
    int temp = Double.compare (lhsUpper, rhsUpper);
    if (temp != 0) {
      return temp;
    }
    return
      !rhsUpperOpen && lhsUpperOpen ? -1 :
      rhsUpperOpen && !lhsUpperOpen ? +1 :
      0;
  }

  /**
   * Returns the number of elements in the partition.
   *
   * @return the number of elements.
   */
   
  public int length () {
    return fUpper.length;
  }
  
  /**
   * Returns whether the lower bound is open on the nth interval.
   *
   * @param n identifies the interval.
   * @return whether it is open.
   */

  public boolean nthLowerOpen (int n) {
    return (n == 0) ? fOverallLowerOpen : !fUpperOpen[n - 1];
  }
  
  /**
   * Returns whether the lower bound on the nth interval.
   *
   * @param n identifies the interval.
   * @return the bound.
   */
   
  public double nthLower (int n) {
    return (n == 0) ? fOverallLower : fUpper[n - 1];
  }
  
  /**
   * Returns whether the upper bound on the nth interval.
   *
   * @param n identifies the interval.
   * @return the bound.
   */
   
  public double nthUpper (int n) {
    return fUpper[n];
  }
  
  /**
   * Returns whether the upper bound is open on the nth interval.
   *
   * @param n identifies the interval.
   * @return whether it is open.
   */
   
  public boolean nthUpperOpen (int n) {
    return fUpperOpen[n];
  }

  /**
   * Returns the nth interval in the partition.
   *
   * @return the nth interval.
   */
   
  public mark.core.math.Interval nthInterval (int n) {
    return new mark.core.math.Interval (
      nthLowerOpen (n), nthLower (n), nthUpper (n), nthUpperOpen (n));
  }

  /**
   * Returns a sorted array of all of the intervals in the partition.
   *
   * @return an array of all intervals in the partition.
   */

  public mark.core.math.Interval[] allIntervals () {
    mark.core.math.Interval[] ans =
      new mark.core.math.Interval[length ()];
    for (int i = 0; i < ans.length; i++) {
      ans[i] = nthInterval (i);
    }
    return ans;
  }

  /**
   * Given a real number, returns the index of the corresponding interval.
   *
   * @param value the real.
   * @return the index or -1 if there isn't any.
   */
   
  public int select (double value) {
    int l = 0;
    int u = length () - 1;
    while (l <= u) {
      int current = (l + u) / 2;
      boolean lowerOpen = nthLowerOpen (current);
      double lower = nthLower (current);
      double upper = nthUpper (current);
      boolean upperOpen = nthUpperOpen (current);
      if (lower > value || lower == value && lowerOpen) {
        u = current - 1;
      } else if (upper < value || upper == value && upperOpen) {
        l = current + 1;
      } else {
        return current;
      }
    }
    return -1;
  }
    
  /**
   * Returns the index of the interval that has the given lower bound.
   *
   * @param bound the lower bound.
   * @param open whether the lower bound is open.
   * @return the index or -1 if there isn't any.
   */
  
  public int lowerBoundToIndex (double bound, boolean open) {
    int start = 0;
    int end = length () - 1;
    while (start <= end) {
      int mid = (start + end) / 2;
      int c = compareLower (bound, open, nthLower (mid), nthLowerOpen (mid));
      if (c < 0) {
        end = mid - 1;
      } else if (c > 0) {
        start = mid + 1;
      } else {
        return mid;
      }
    }
    return -1;
  }
  
  /**
   * Returns the index of the interval that has the given upper bound.
   *
   * @param bound the upper bound.
   * @param open whether the upper bound is open.
   * @return the index or -1 if there isn't any.
   */
  
  public int upperBoundToIndex (double bound, boolean open) {
    int start = 0;
    int end = length () - 1;
    while (start <= end) {
      int mid = (start + end) / 2;
      int c = compareUpper (bound, open, nthUpper (mid), nthUpperOpen (mid));
      if (c < 0) {
        end = mid - 1;
      } else if (c > 0) {
        start = mid + 1;
      } else {
        return mid;
      }
    }
    return -1;
  }

}
