package mark.core.math;
import java.util.*;

/**
 * Static utility methods.
 *
 * @author Mark Chavira
 */

public class Util {

  /**
   * The constant ln (2).
   */

  private static double LOG2 = java.lang.Math.log (2);

  /**
   * Returns whether two doubles are close to equal.  This method was found
   * to be very important for efficiency.
   * 
   * @param p1 the first double.
   * @param p2 the second double.
   * @return whether they are close.
   */
  
  public static boolean close (double p1, double p2) {
    if (p1 == 0) {return p2 == 0;}
    if (p2 == 0) {return false;}
    double temp = 1.0 - p1/p2;
    return (temp > 0) ? (temp <= 0.00000000001) : (temp >= -0.00000000001);
  }

  /**
   * Checks whether each pair of slots of the two given arrays are close
   * as defined by close (double, double); if so, returns -1; if not, returns
   * the index of the first slot that isn't close.  Assumes that the arrays
   * have the same dimension.
   * 
   * @param a1 the first given array.
   * @param a2 the second given array.
   * @return whether they are close.
   */
  
  public static int close (double[] a1, double[] a2) {
    for (int i = 0; i < a1.length; i++) {
      if (!close (a1[i], a2[i])) {return i;}
    }
    return -1;
  }
  
  /**
   * Calculates the log base 2 of its argument.
   * 
   * @param a the argument
   * @return log base 2 of the argument
   */

  public static double log2 (double a) {
    return java.lang.Math.log (a) / LOG2;
  }

  /**
   * Fills an array with the low part of the binary representation of the
   * given number, in big-endian format (earlier elements in the arrray
   * denote higher bits).  For example, if n = 5 and d = 5, then returns
   * {false, false, true, false, true}.
   *
   * @param n the number.
   * @param d the number of binary digits to use.
   * @return the binary representation.
   */
  
  public static boolean[] toBinary (int n, int d) {
    boolean[] ans = new boolean[d];
    for (int i = ans.length - 1; i >= 0; i--) {
      ans[i] = (n % 2 == 1);
      n /= 2;
    }
    return ans;
  }

  /**
   * Constrains an int.  If the given int is in the given closed range,
   * then returns the given int.  If the given int is greater than the
   * largest value in the range, then returns the largest value in the range.
   * If the int is less than the smallest value in the range, then returns
   * the smallest value in the range.
   *
   * @param i the int.
   * @param min the smallest value in the range.
   * @param max the largest value in the range.
   * @return the constrained value.
   */
  
  public static int constrain (int i, int min, int max) {
    if (i < min) {
      return min;
    } else if (i > max) {
      return max;
    } else {
      return i;
    }
  }

  /**
   * Constrains a double.  If the given double is in the given closed range,
   * then returns the given double.  If the given double is greater than the
   * largest value in the range, then returns the largest value in the range.
   * If the double is less than the smallest value in the range, then returns
   * the smallest value in the range.
   *
   * @param d the double.
   * @param min the smallest value in the range.
   * @param max the largest value in the range.
   * @return the constrained value.
   */
  
  public static double constrain (double d, double min, double max) {
    if (d < min) {
      return min;
    } else if (d > max) {
      return max;
    } else {
      return d;
    }
  }

  /**
   * Returns a random int in the given range (inclusive).
   *
   * @param r the random generator.
   * @param low the start of the range.
   * @param high the end of the range.
   * @return the random number.
   */

  public static int random (Random r, int low, int high) {
    return r.nextInt (high - low + 1) + low;
  }

  /**
   * Returns the weighted pointwise mutual information of (X1 = x1, X2 = x2).
   * To get the mutual information, sum all of the pointwise values for all
   * instantiations.
   *
   * @param numX1X2 #(X1 = x1, X2 = x2).
   * @param numX1 #(X1 = x1).
   * @param numX2 #(X2 = x2).
   * @param num the total number of samples.
   * @return the weighted pointwise mutual information.
   */

  public static double weightedPointwiseMutualInformation (
  double numX1X2, double numX1, double numX2, double num) {
    if (numX1X2 == 0.0) {
      return 0.0;
    }
    double probX1X2 = numX1X2 / num;
    double probX1 = numX1 / num;
    double probX2 = numX2 / num;
    return probX1X2 * log2 (probX1X2 / probX1 / probX2);
  }

  /**
   * Returns the pointwise chi-squared value of (X1, X2).  To get the chi-
   * squared value, sum the pointwise chi-squared values for all
   * instantiations.
   *
   * @param numX1X2 #(X1 = x1, X2 = x2).
   * @param numX1 #(X1 = x1).
   * @param numX2 #(X2 = x2).
   * @param num the total number of samples.
   * @return the chi-squared value.
   */

  public static double pointwiseChiSquared (
  double numX1X2, double numX1, double numX2, double num) {
    double exp = numX1 * numX2 / num;
    double temp = numX1X2 - exp;
    return (temp * temp) / exp;
  }

  /**
   * Samples the given number of times from the given set of ints using a
   * uniform distribution.
   *
   * @param r the random generator.
   * @param n the given number.
   * @param s the given set.
   * @param replace whether to use replacement.
   * @return the sampled set.
   */
  
  public static int[] sampleUniform (
   Random r, int n, int[] s, boolean replace) {
    int unsampledSize = s.length;
    int ansSize = 0;
    int[] unsampled = mark.core.util.Util.clone (s);
    int[] ans = new int[n];
    while (ansSize < n) {
      int index = mark.core.math.Util.random (r, 0, unsampledSize - 1);
      ans[ansSize++] = unsampled[index];
      if (!replace) {
        unsampled[index] = unsampled[--unsampledSize];
      }
    }
    return ans;
  } 

  /**
   * Samples the given number of times from the given set using a uniform
   * distribution.
   *
   * @param r the random generator.
   * @param n the given number.
   * @param s the given set.
   * @param replace whether to use replacement.
   * @return the sampled set.
   */
  
  public static <T> ArrayList<T> sampleUniform (
   Random r, int n, Set<T> s, boolean replace) {
    ArrayList<T> unsampled = new ArrayList<T> (s);
    ArrayList<T> ans = new ArrayList<T> ();
    while (ans.size () < n) {
      int index = mark.core.math.Util.random (r, 0, unsampled.size () - 1);
      ans.add (unsampled.get (index));
      if (!replace) {
        unsampled.set (index, unsampled.get (unsampled.size () - 1));
        unsampled.remove (unsampled.size () - 1);
      }
    }
    return ans;
  } 

  /**
   * Generates a random double between the first given number (non-inclusive)
   * the second (non-inclusive).
   *
   * @param r the random generator.
   * @param min the first given number.
   * @param max the second given number.
   * @return the double.
   */

  public static double randomNonExtremeProbability (
   Random r, double min, double max) {
    return min + r.nextDouble () * (max - min);
  }
  
  /**
   * Generates an array of random doubles between the given number (non-
   * inclusive) and 1 (non-inclusive).
   *
   * @param r the random generator.
   * @param min the first given number.
   * @param max the second given number.
   * @param n the number of doubles.
   * @return the doubles.
   */
  
  public static double[] randomNonExtremeProbabilities (
   Random r, double min, double max, int n) {
    double[] ans = new double[n];
    for (int i = 0; i < n; i++) {
      ans[i] = randomNonExtremeProbability (r, min, max);
    }
    return ans;
  }
  
}
