package mark.core.rdag;
import java.util.*;

/**
 * A base class for a node in a dtree where each leaf stores an object of
 * type T.
 * 
 * @author Mark Chavira
 */

public class Dtree<T> extends Node {

  //=========================================================================
  // Protected:
  //=========================================================================

  /**
   * Whether the node is a leaf.
   */
  
  public boolean fIsLeaf;
  
  /**
   * The left child; null if a leaf.
   */
  
  protected Dtree<T> fLeft;
  
  /**
   * The right child; null if a leaf.
   */
  
  protected Dtree<T> fRight;
  
  //=========================================================================
  // Public:
  //=========================================================================

  /**
   * The object.
   */
  
  public T fObject;
  
  /**
   * The constructor for an internal node.
   * 
   * @param l the left child.
   * @param r the right child.
   */
  
  public Dtree (Dtree<T> l, Dtree<T> r) {
    fIsLeaf = false;
    fLeft = l;
    fRight = r;
  }
  
  /**
   * The constructor for a leaf.
   */
  
  public Dtree (T object) {
    fIsLeaf = true;
    fObject = object;
  }
  
  /**
   * Returns whether the node is a leaf.
   * 
   * @return the result of the test.
   */
  
  public boolean isLeaf () {
    return fIsLeaf;
  }
  
  /**
   * Returns whether the node is internal.
   * 
   * @return the result of the test.
   */

  public boolean isInternal () {
    return !isLeaf ();
  }
  
  /**
   * Returns the left child of an internal node.
   * 
   * @return the child.
   */
  
  public Dtree left () {
    return fLeft;
  }
  
  /**
   * Returns the right child of an internal node.
   * 
   * @return the child.
   */
  
  public Dtree right () {
    return fRight;
  }
  
  /**
   * @see Node
   */
  
  public int numOutgoingEdges () {
    return isLeaf () ? 0 : 2;
  }
  
  /**
   * @see Node
   */
  
  public Node nthChild (int n) {
    return (n == 0) ? fLeft : fRight;
  }
  
  /**
   * Constructs a full binary tree above the given nodes.
   *
   * @param nodes the given node.
   * @return the tree or null if the given collection is empty.
   */
  
  public static Dtree fullBinaryTree (Collection<Dtree> leaves)
   throws Exception {
    
    // If no nodes, then throw an exception.
    
    if (leaves.isEmpty ()) {
      throw new Exception ("Empty list of nodes!");
    }

    // Get an ordering on the leaves.  Also, create space for internal nodes.
    // If there are n leaves, there are n-1 internal nodes.
    
    Dtree[] ordered = new Dtree[2 * leaves.size () - 1];
    int index = 0;
    for (Dtree n : leaves) {ordered[index++] = n;}
    
    // Create the internal nodes and return the result.
    
    int size = leaves.size ();
    for (int i = 0; i < size - 1; i += 2) {
      ordered[size++] = new Dtree (ordered[i], ordered[i+1]);
    }
    return ordered[ordered.length - 1];

  }

  /**
   * Given a dtree T over source objects and a map M from source object to
   * set of target objects, creates another dtree T' over target objects that
   * has the same structure as T except that replaces each leaf in T having
   * source object S with another a full binary dtree over the target objects
   * M (S).
   * 
   * @param srcDtree the given source dtree.
   * @param srcLeafObjectToTargetSet the given map.
   * @return the new dtree.
   */
  
  public static Dtree targetDtreeExpand (
   Dtree srcDtree, Map srcLeafObjectToTargetSet) throws Exception {
      
    // Loop through the nodes in the source dtree and create a map from
    // each to a node in the target dtree.

    ArrayList<Dtree> nodes = (ArrayList)srcDtree.childrenBeforeParents ();
    ArrayList<Dtree> tgtNodes = new ArrayList<Dtree> ();
    HashMap<Dtree,Dtree> oldNodeToNewNode = new HashMap<Dtree,Dtree> ();
    for (Dtree oldNode : nodes) {

      // If the src dtree node N is a leaf corresponding to source object
      // D, then map N to a balanced dtree over the target set of D.  There
      // is one tricky part.  It is possible that D maps to an empty set.
      // In this case, map N to null, which will effectively delete that
      // part of the target dtree.
      
      if (oldNode.isLeaf ()) {
        Set targetSet =
          (Set)srcLeafObjectToTargetSet.get (oldNode.fObject);
        if (targetSet.isEmpty ()) {
          oldNodeToNewNode.put (oldNode, null);
        } else {
          tgtNodes.clear ();
          for (Object tgt : targetSet) {tgtNodes.add (new Dtree (tgt));}
          oldNodeToNewNode.put (oldNode, fullBinaryTree (tgtNodes));
        }
      
      // Otherwise, the src node N is internal.  Create a corresponding
      // node in the target dtree.  Again, there is one tricky part.
      // Either or both of the children may have been deleted.  We act
      // accordingly.

      } else {
        Dtree ch0 = oldNodeToNewNode.get (oldNode.nthChild (0));
        Dtree ch1 = oldNodeToNewNode.get (oldNode.nthChild (1));
        if (ch0 == null && ch1 == null) {
          oldNodeToNewNode.put (oldNode, null);
        } else if (ch0 == null) {
          oldNodeToNewNode.put (oldNode, ch1);
        } else if (ch1 == null) {
          oldNodeToNewNode.put (oldNode, ch0);
        } else {
          oldNodeToNewNode.put (oldNode, new Dtree (ch0, ch1));
        }
      }
      
    }
    
    // Return the target dtree.
    
    return  oldNodeToNewNode.get (nodes.get (nodes.size () - 1));

  }

  /**
   * Given a dtree T over source objects and a map M from source object to
   * target object, creates another dtree T' over target objects that has the
   * same structure as T except that each source object in a leaf is replaced
   * with the corresponding target object.
   * 
   * @param srcDtree the given source dtree.
   * @param srcLeafObjectToTarget the given map.
   * @return the new dtree.
   */
  
  public static Dtree targetDtree (
   Dtree srcDtree, Map srcLeafObjectToTarget) throws Exception {
      
    // Loop through the nodes in the source dtree and create a map from
    // each to a node in the target dtree.

    ArrayList<Dtree> nodes = (ArrayList)srcDtree.childrenBeforeParents ();
    HashMap<Dtree,Dtree> oldNodeToNewNode = new HashMap<Dtree,Dtree> ();
    for (Dtree oldNode : nodes) {

      // If the src dtree node N is a leaf corresponding to source object
      // D, then map N to the corresponding target object.
      
      if (oldNode.isLeaf ()) {
        oldNodeToNewNode.put (
          oldNode, new Dtree (srcLeafObjectToTarget.get (oldNode.fObject)));
      
      // Otherwise, the src node N is internal.  Create a corresponding
      // node in the target dtree.

      } else {
        oldNodeToNewNode.put (
          oldNode,
          new Dtree (
            oldNodeToNewNode.get (oldNode.nthChild (0)),
            oldNodeToNewNode.get (oldNode.nthChild (1))));
      }
      
    }
    
    // Return the target dtree.
    
    return  oldNodeToNewNode.get (nodes.get (nodes.size () - 1));

  }

}
