package mark.core.rdag;
import java.util.*;

/**
 * A base class for a node in a rooted dag.  A node has a set of outgoing
 * edges.  We view a node in two ways.  First, it is a node in a graph.
 * Hence, we can talk about its number of outgoing edges, etc.  Second,
 * because it is the root of exactly one graph, we view it as a graph.
 * Hence, we can talk about operations that apply to entire graphs.
 * <p>
 * Drawing: A node has a label, and for each edge, a label.
 * <p>
 * This class conforms to the immutable design pattern.
 *
 * @author Mark Chavira
 */

public abstract class Node {

  //=========================================================================
  // Basic methods that subclasses will often override.
  //=========================================================================

  /**
   * Returns the node's number of outgoing edges.
   *
   * @return the number outgoing edges.
   */
  
  public int numOutgoingEdges () {
    return 0;
  }
  
  /**
   * Returns the node's nth child, where n ranges from zero to
   * numOutgoingEdges () - 1.
   *
   * @param n identifies the child.
   * @return the child.
   */
  
  public Node nthChild (int n) {
    return null;
  }

  //=========================================================================
  // Drawing:
  //=========================================================================

  /**
   * The script header.
   */
  
  public static final String SCRIPT_HEADER =
    "on newLine(id1, id2, labelText, lineStyle)\n" +
    "  tell first document of application \"OmniGraffle Professional\"\n" +
    "    tell first canvas\n" +
    "      make new line at end of graphics with properties {line type:curved}\n" +
    "      set source of graphic -1 to shape id id1\n" +
    "      set destination of graphic -1 to shape id id2\n" +
    "      set stroke pattern of graphic -1 to lineStyle\n" +
    "      if labelText is not \"\" then\n" +
    "        make new shape at end of graphics with properties {text:{{size:9, text:labelText}}, autosizing:full, draws stroke:false, draws shadow:false}\n" +
    "        set labelConnection of graphic -1 to graphic -2\n" +
    "      end if\n" +
    "    end tell\n" +
    "  end tell\n" +
    "end newLine\n" +
    "tell first document of application \"OmniGraffle Professional\"\n" +
    "  activate\n" +
    "  tell first canvas\n"; 
  
  /**
   * The script footer.
   */
  
  public static final String SCRIPT_FOOTER =
    "    layout\n" +
    "  end tell\n" +
    "end tell";
  
  /**
   * Returns a string which if executed as an Applescript commands
   * OmniGraffle to draw the graph containing the given nodes and their
   * decendents.
   *
   * @param topNodes the given nodes.
   * @param si information needed to draw the graph.
   * @return the string.
   */

  public static String toGraffleScript (
   Collection<Node> topNodes, ScriptInformation si) throws Exception {
 
   // Collect the header.
 
   StringBuilder ans = new StringBuilder ();    
   ans.append (SCRIPT_HEADER);
   
   // Traverse information about the graph.

   ArrayList<Node> nodes = childrenBeforeParents (topNodes);
   HashMap<Node,Integer> nodeToId =
     new HashMap<Node,Integer> ();
   int index = -1;
   for (Node n : nodes) {
     ++index;
     ans.append (
         "    set node" + index + " to id of (" +
         "make new shape at end of graphics with properties {" +
         si.nodeProperties (n) + "})\n");
     Node[] children = new Node[n.numOutgoingEdges ()];
     for (int i = 0; i < children.length; i++) {
       children[i] = n.nthChild (i);
     }
     for (int j = 0; j < children.length; j++) {
       if (children[j] != null) {
         String label =
           si.drawOutgoingEdgeLabels (n) ?
           si.ithOutgoingEdgeLabel(n, j) : "";
         if (si.combineOutgoingEdges (n)) {
           for (int k = j + 1; k < children.length; k++) {
             if (children[j] == children[k]) {
               if (si.drawOutgoingEdgeLabels (n)) {
                 label += "U" + si.ithOutgoingEdgeLabel(n, k);
               }
               children[k] = null;
             }
           }
         }
         ans.append (
           "    tell me to newLine (node" + index + ", node" +
           nodeToId.get (children[j]) + ", \"" + label + "\"," +
           si.ithOutgoingEdgeStyle(n, j).omniId () + ")\n");
       }
     }
     nodeToId.put (n, index);
   }
     
   // Collect the footer and return the result.
   
   ans.append (SCRIPT_FOOTER);
   return ans.toString ();

 }
  
  /**
   * Returns a string which if executed as an Applescript commands
   * OmniGrafle to draw the graph containing the node and its decendents.
   *
   * @param si information needed to draw the graph.
   * @return the string.
   */
  
  public String toGraffleScript (ScriptInformation si) throws Exception {
    return toGraffleScript (Collections.singleton (this), si);
  }

  //=========================================================================
  // Constructor:
  //=========================================================================

  /**
   * The constructor.
   */
  
  protected Node () {}

  //=========================================================================
  // Map children to objects:
  //=========================================================================

  /**
   * Uses the children as keys into the given memo and returns the objects
   * referenced.
   * 
   * @param memo the given memo.
   * @return the objects.
   */
  
  public <T> ArrayList<T> traversalObjects (Map<Node,T> memo) {
    ArrayList<T> ans = new ArrayList<T> ();
    for (int i = 0; i < numOutgoingEdges (); i++) {
      ans.add (memo.get (nthChild (i)));
    }
    return ans;
  }
  
  /**
   * Uses the children as keys into the given memo and returns the booleans
   * referenced.
   *
   * @param memo the given memo
   * @return the booleans.
   */

  protected boolean[] traversalBooleans (Map<Node,Boolean> memo) {
    boolean[] ans = new boolean[numOutgoingEdges ()];
    for (int i = 0; i < ans.length; i++) {
      ans[i] = memo.get (nthChild (i));
    }
    return ans;
  }

  /**
   * Uses the children as keys into the given memo and returns the big
   * integers referenced.
   *
   * @param memo the given memo
   * @return the booleans.
   */

  protected java.math.BigInteger[] traversalBigInts (
   Map<Node,java.math.BigInteger> memo) {
    java.math.BigInteger[] ans = new java.math.BigInteger[numOutgoingEdges ()];
    for (int i = 0; i < ans.length; i++) {
      ans[i] = memo.get (nthChild (i));
    }
    return ans;    
  }

  /**
   * Uses the children as keys into the given memo and returns the ints
   * referenced.
   *
   * @param memo the given memo
   * @return the booleans.
   */

  protected int[] traversalInts (Map<Node,Integer> memo) {
    int[] ans = new int[numOutgoingEdges ()];
    for (int i = 0; i < ans.length; i++) {
      ans[i] = memo.get (nthChild (i));
    }
    return ans;    
  }

  /**
   * Uses the children as keys into the given memo and returns the doubles
   * referenced.
   *
   * @param memo the given memo
   * @return the booleans.
   */

  protected double[] traversalDoubles (Map<Node,Double> memo) {
    double[] ans = new double[numOutgoingEdges ()];
    for (int i = 0; i < ans.length; i++) {
      ans[i] = memo.get (nthChild (i));
    }
    return ans;    
  }

  //=========================================================================
  // Topological ordering:
  //=========================================================================

  /**
   * Traverses the graph bottom-up placing each node into the given list.
   *
   * @param memo a scratch area.
   * @param nodes the given list.
   */

  protected static <T extends Node> void childrenBeforeParentsHelper (
   HashSet<T> memo, T n, ArrayList<T> nodes) throws Exception {
    if (memo.contains (n)) {return;}
    memo.add (n);
    for (int i = 0; i < n.numOutgoingEdges (); i++) {
      childrenBeforeParentsHelper (memo, (T)n.nthChild (i), nodes);
    }
    nodes.add (n);
  }

  /**
   * Returns a list of the node and its decendents in bottom-up order.
   *
   * @return the list.
   */
  
  public ArrayList<Node> childrenBeforeParents ()
   throws Exception {
    return childrenBeforeParents (Collections.singleton (this));
  }
  
  /**
   * Returns a list of the given nodes and their decendents in bottom-up
   * order.
   *
   * @paam topNodes the given nodes.
   * @return the list.
   */

  public static <T extends Node> ArrayList<T> childrenBeforeParents (
   Collection<T> topNodes) throws Exception {
    ArrayList<T> ans = new ArrayList<T> ();
    HashSet<T> memo = new HashSet<T> ();
    for (T n : topNodes) {
      childrenBeforeParentsHelper (memo, n, ans);
    }
    return ans;
  }
  
  /**
   * Returns a list of the node and its decendents top-down order.
   *
   * @return the list.
   */
  
  public ArrayList<Node> parentsBeforeChildren ()
   throws Exception {
    return parentsBeforeChildren (Collections.singleton (this));
  }

  /**
   * Returns a list of the given nodes and their decendents in top-down
   * order.
   *
   * @paam topNodes the given nodes.
   * @return the list.
   */

  public static ArrayList<Node> parentsBeforeChildren (
   Collection<Node> topNodes) throws Exception {
    ArrayList<Node> ans = childrenBeforeParents (topNodes);
    Collections.reverse (ans);
    return ans;
  }

  //=========================================================================
  // Sets of nodes:
  //=========================================================================

  /**
   * Returns a map from each node in the given graph to one of its parents.
   * Does not contain a mapping for roots.
   * 
   * @param topNodes the given graph.
   * @return the map.
   */
  
  public static <T extends Node> HashMap<T,T> nodeToParentMap (
   Collection<T> topNodes)
   throws Exception {
    HashMap<T,T> ans = new HashMap<T,T> ();
    ArrayList<T> allNodes = childrenBeforeParents (topNodes);
    for (T n : allNodes) {
      for (int i = 0; i < n.numOutgoingEdges (); i++) {
        ans.put ((T)n.nthChild (i), n);
      }
    }
    return ans;
  }
    
  /**
   * Returns a map from each node in the given graph to one of its parents.
   * Does not contain a mapping for roots.
   * 
   * @param root the given graph.
   * @return the map.
   */
  
  public static <T extends Node> HashMap<T,T> nodeToParentMap (T root)
   throws Exception {
    return nodeToParentMap (Collections.singleton (root));
  }

  /**
   * Returns the subset of the first given set of nodes that have no parents
   * among the second set of nodes.
   *
   * @param possibleRoots the first set of given nodes.
   * @param allNodes the second set of given nodes.
   * @return the set.
   */
  
  static HashSet<Node> rootsHelper (
   Collection<Node> possibleRoots, Collection<Node> allNodes) {
    HashSet<Node> ans = new HashSet<Node> (possibleRoots);
    for (Node n : allNodes) {
      for (int j = 0; j < n.numOutgoingEdges (); j++) {
        ans.remove (n.nthChild (j));
      }
    }
    return ans;
  }

  /**
   * Returns the set of roots from the given nodes and their decendents.
   *
   * @paam topNodes the given nodes.
   * @return the set.
   */
  
  public static HashSet<Node> roots (Collection<Node> topNodes)
   throws Exception {
    return rootsHelper (topNodes, childrenBeforeParents (topNodes));
  }

  /**
   * Returns the subset of the given nodes that have no children.
   *
   * @param nodes the given nodes.
   * @return the set.
   */
  
  static HashSet<Node> sinksHelper (Collection<Node> nodes) {
    HashSet<Node> ans = new HashSet<Node> ();
    for (Node n : nodes) {
      if (n.numOutgoingEdges () == 0) {
        ans.add (n);
      }
    }
    return ans;
  }

  /**
   * Returns the set of sinks from the given nodes and their decendents.
   *
   * @paam topNodes the given nodes.
   * @return the set.
   */
  
  public static HashSet<Node> sinks (Collection<Node> topNodes)
   throws Exception {
    return sinksHelper (childrenBeforeParents (topNodes));
  }

  /**
   * Returns the set of roots from the node and its decendents.
   *
   * @return the list.
   */

  public HashSet<Node> sinks () throws Exception {
    return sinksHelper (childrenBeforeParents ());
  }

  /**
   * Returns a set of nodes consisting of the given bottom nodes and any
   * other node that lies on a path from one of the given top nodes to one
   * of the given bottom nodes.
   *
   * @param topNodes the given top nodes.
   * @param bottomNodes the given bottom nodes.
   * @return the set.
   */
  
  public static HashSet<Node> ancestors (
   Collection<Node> topNodes, Collection<Node> bottomNodes)
   throws Exception {
    HashSet<Node> ans = new HashSet<Node> (bottomNodes);
    ArrayList<Node> nodes = childrenBeforeParents (topNodes);
    for (Node p : nodes) {
      for (int j = 0; j < p.numOutgoingEdges (); j++) {
        if (ans.contains (p.nthChild (j))) {
          ans.add (p);
          break;
        }
      }
    }
    return ans;
  }

  /**
   * Returns the set of the nodes that lie on some path from the node to one
   * of the given nodes, including the node and the given nodes.
   *
   * @param bottomNodes the given nodes.
   * @return the set.
   */
  
  public HashSet<Node> ancestors (
   Collection<Node> bottomNodes) throws Exception {
    return ancestors (Collections.singleton (this), bottomNodes);
  }

  //=========================================================================
  // Getting information about the rdag.
  //=========================================================================

  /**
   * Returns an object containing graph statistics for the given nodes and
   * their decendents.
   *
   * @param topNodes the given nodes.
   * @return the statistics.
   */
  
  public static Statistics statistics (
   Collection<Node> topNodes) throws Exception {
    return new Statistics (topNodes);
  }
  
  /**
   * Returns an object containing graph statistics for the node and its
   * decendents.
   *
   * @return the statistics.
   */
  
  public Statistics statistics () throws Exception {
    return statistics (Collections.singleton (this));
  }
  
  /**
   * Returns the number of nodes in the graph containing the given nodes and
   * their decendents.
   *
   * @param topNodes the given Nodes.
   * @return the number of nodes.
   */
  
  public static int numNodes (Collection<Node> topNodes)
   throws Exception {
    return childrenBeforeParents (topNodes).size ();
  }
  
  /**
   * Returns the number of nodes in the graph containing the node and its
   * decendents.
   * 
   * @return the number of nodes.
   */
  
  public int numNodes () throws Exception {
    return numNodes (Collections.singleton (this));
  }
    
  /**
   * Returns the number of edges in the graph containing the given nodes.
   *
   * @param nodes the given nodes.
   * @return the number of edges.
   */
  
  public static int numEdgesHelper (Collection<Node> nodes) {
    int ans = 0;
    for (Node n : nodes) {
      ans += n.numOutgoingEdges ();
    }
    return ans;
  }

  /**
   * Returns the number of edges in the graph containing the given nodes and
   * their decendents.
   *
   * @param topNodes the given Nodes.
   * @return the number of edges.
   */
  
  public static int numEdges (Collection<Node> topNodes)
   throws Exception {
    return numEdgesHelper (childrenBeforeParents (topNodes));
  }

  /**
   * Returns the number of edges in the graph containing the node and its
   * decendents.
   *
   * @return the number of edges.
   */
  
  public int numEdges () throws Exception {
    return numEdges (Collections.singleton (this));
  }

  /**
   * Returns the number of nodes in a tree representation of the graph
   * containing the given nodes and their decendents.  Runs in time linear in
   * the size of the graph.
   *
   * @return the number of nodes.
   */
  
  public static int numTreeNodes (Collection<Node> topNodes)
   throws Exception {
    ArrayList<Node> nodes = childrenBeforeParents (topNodes);
    Map<Node,Integer> nodeToCount = new HashMap<Node,Integer> ();
    for (Node n : nodes) {
      int count = 1;
      for (int j = 0; j < n.numOutgoingEdges (); j++) {
        count += nodeToCount.get (n.nthChild (j));
      }
      nodeToCount.put (n, count);
    }
    int ans = 0;
    for (Node n : topNodes) {
      ans += nodeToCount.get (n);
    }
    return ans;
  }

  /**
   * Returns the number of nodes in a tree representation of the graph
   * containing the node and its decendents.  Runs in time linear in the size
   * of the graph.
   *
   * @return the number of nodes.
   */
  
  public int numTreeNodes () throws Exception {
    return numTreeNodes (Collections.singleton (this));
  }

  /**
   * Returns the number of edges in a tree representation of the graph
   * containing the given nodes and their decendents.  Runs in time linear in
   * the size of the graph.
   *
   * @return the number of edges.
   */

  public static int numTreeEdges (Collection<Node> topNodes)
   throws Exception {
    return numTreeNodes (topNodes) - 1;
  }

  /**
   * Returns the number of edges in a tree representation of the graph
   * containing the node and its decendents.  Runs in time linear in the size
   * of the graph.
   *
   * @return the number of edges.
   */
  
  public int numTreeEdges () throws Exception {
    return numTreeEdges (Collections.singleton (this));
  }

  /**
   * Returns an unmodifiable sorted map from the classes of the nodes in the
   * graph containing the given nodes and their decendents to the count of
   * the instances of that class.
   *
   * @param topNodes the given Nodes.
   * @return the map.
   */
  
  public static SortedMap<Class<?>,Integer> classMap (
   Collection<Node> topNodes) throws Exception {
    return
      mark.core.util.Util.typeStatistics (childrenBeforeParents (topNodes));
  }

  /**
   * Returns a map from the classes of the nodes in the graph containing the
   * node and its decendents to the count of the instances of that class.
   *
   * @return the map.
   */
  
  public SortedMap<Class<?>,Integer> classMap () throws Exception {
    return classMap (Collections.singleton (this));
  }

  //=========================================================================
  // Copying the dag.
  //=========================================================================

  /**
   * Returns a copy of the rdag.  This method only orchestrates the
   * operation; the NodeCopier actually copies the nodes, edges, and other
   * fields.
   *
   * @return the copy.
   */

  public Node copy (NodeCopier nc) throws Exception {
    return copy (Collections.singleton (this), nc).get (0);
  }
  
  /**
   * Returns a copy of the dag with the given roots.  This method only
   * orchestrates the operation; the NodeCopier actually copies the nodes,
   * edges, and other fields.
   *
   * @return the copy.
   */
  
  public static ArrayList<Node> copy (
   Collection<Node> topNodes, NodeCopier nc) throws Exception {
    ArrayList<Node> nodes = childrenBeforeParents (topNodes);
    HashMap<Node,Node> memo = new HashMap<Node,Node> ();
    for (Node oldNode : nodes) {
      memo.put (oldNode, nc.copy (oldNode, memo));
    }
    return mark.core.util.Util.imageAsList (topNodes, memo);
  }

}
