package mark.core.rdag;
import java.util.*;

/**
 * A node that allows clients to store an object and manipulate the list of
 * children.
 *
 * @author Mark Chavira
 */

public class SimpleNode extends Node {

  /**
   * A node copier that creates a copy of each node and edge but copies the
   * object shallowly.
   */

  public static final NodeCopier COPIER =
    new NodeCopier () {
      public Node copy (Node n, Map<Node,Node> memo)
       throws Exception {
        SimpleNode sn = (SimpleNode)n;
        SimpleNode ans = new SimpleNode (sn.fObject);
        for (int i = 0; i < sn.fChildren.size (); i++) {
          ans.fChildren.add (memo.get (sn.fChildren.get (i)));
        }
        return ans;
      }
    };

  public Object fObject;
  public ArrayList<Node> fChildren = new ArrayList<Node> ();

  /**
   * The constructor, which sets fObject to null.
   */
  
  public SimpleNode () {}
  
  /**
   * The constructor, which sets fObject to the given value.
   *
   * @param o the given value.
   */
  
  public SimpleNode (Object o) {
    fObject = o;
  }

  /**
   * The constructor, which gives the node the two given children and sets
   * fObject to the given value.
   *
   * @param ch1 the first given child.
   * @param ch2 the second given child.
   * @param o the give value.
   */
  
  public SimpleNode (Node ch1, Node ch2, Object o) {
    fChildren.add (ch1);
    fChildren.add (ch2);
    fObject = o;
  }

  /**
   * @see Node
   */
   
  public String label () {
    return fObject.toString ();
  }

  /**
   * @see Node
   */
  
  public int numOutgoingEdges () {
    return fChildren.size ();
  }
  
  /**
   * @see Node
   */
  
  public SimpleNode nthChild (int n) {
    return (SimpleNode)fChildren.get (n);
  }

  /**
   * Returns a list of nodes with fObject fields set to the given collection
   * of objects.
   *
   * @param c the given collection.
   * @return the list of nodes.
   */
  
  public static ArrayList<Node> collectionToNodeList (
   Collection<?> objects) {
    ArrayList<Node> ans = new ArrayList<Node> ();
    for (Object o : objects) {
      ans.add (new SimpleNode (o));
    }
    return ans;
  }

  /**
   * Returns a list of nodes with fObject fields set to the given collection
   * of objects.
   *
   * @param c the given collection.
   * @return the list of nodes.
   */
  
  public static HashSet<Node> collectionToNodeSet (
   Collection<?> objects) {
    HashSet<Node> ans = new HashSet<Node> ();
    for (Object o : objects) {
      ans.add (new SimpleNode (o));
    }
    return ans;
  }

  /**
   * Returns the set of objects in the given nodes.
   *
   * @param nodes the given nodes.
   * @return the objects.
   */
  
  public static HashSet<Object> nodesToSetOfObjects (
   Collection<Node> nodes) {
    HashSet<Object> ans = new HashSet<Object> ();
    for (Node n : nodes) {
      ans.add (((SimpleNode)n).fObject);
    }
    return ans;
  }

  /**
   * Returns the list of objects in the given nodes.
   *
   * @param nodes the given nodes.
   * @return the objects.
   */
  
  public static ArrayList<Object> nodesToListOfObjects (
   Collection<Node> nodes) {
    ArrayList<Object> ans = new ArrayList<Object> ();
    for (Node n : nodes) {
      ans.add (((SimpleNode)n).fObject);
    }
    return ans;
  }

  /**
   * Constructs a full binary tree above the given nodes.  The given nodes
   * need not be of type SimpleNode.
   *
   * @param nodes the given node.
   * @return the tree or null if the given collection is empty.
   */
  
  public static SimpleNode fullBinaryTree (Collection<Node> nodes)
   throws Exception {
    if (nodes.isEmpty ()) {
      throw new Exception ("Empty list of nodes!");
    }
    ArrayList<Node> nodeList = new ArrayList<Node> (nodes);
    /*
      SimpleNode n = (SimpleNode)nodeList.get (0);
      for (int i = 1; i < nodeList.size (); i++) {
      SimpleNode n2 = new SimpleNode ();
      n2.fChildren.add (n);
      n2.fChildren.add (nodeList.get (i));
      n = n2;
    }
    return n;*/
    for (int i = 0; i < nodeList.size () - 1; i += 2) {
      SimpleNode n = new SimpleNode ();
      n.fChildren.add (nodeList.get (i));
      n.fChildren.add (nodeList.get (i+1));
      nodeList.add (n);
    }
    return (SimpleNode)nodeList.get (nodeList.size () - 1);
  }

  /**
   * See java.lang.Object
   */
  
  public String toString () {
    return "(" + numOutgoingEdges () + "," + fObject + ")";
  }

}
