package mark.core.rdag;
import java.util.*;

/**
 * An object that gathers rdag statistics.
 * <p>
 * This class conforms to the immutable design pattern.
 *
 * @author Mark Chavira
 */

public class Statistics {

  //=========================================================================
  // Fields:
  //=========================================================================

  private int fNumNodes;
  private int fNumEdges;
  private int fNumRoots;
  private int fNumSinks;
  private int fNumNodesBothRootAndSink;
  private int fNumNodesNeitherRootNorSink;
  private SortedMap<Class<?>,Integer> fMap;

  /** 
   * Initializes the object.
   *
   * @param n the graph for which to gather statistics.
   */

  public Statistics (Collection<Node> topNodes) throws Exception {
    ArrayList<Node> nodes = Node.childrenBeforeParents (topNodes);
    HashSet<Node> roots = Node.rootsHelper (topNodes, nodes);
    HashSet<Node> sinks = Node.sinksHelper (nodes);
    fNumEdges = Node.numEdgesHelper (nodes);
    fNumNodes = nodes.size ();
    fNumRoots = roots.size ();
    fNumSinks = sinks.size ();
    HashSet<Node> temp = new HashSet<Node> (roots);
    temp.retainAll (sinks);
    fNumNodesBothRootAndSink = temp.size ();
    temp = new HashSet<Node> (nodes);
    temp.removeAll (roots);
    temp.removeAll (sinks);
    fNumNodesNeitherRootNorSink = temp.size ();
    fMap = mark.core.util.Util.typeStatistics (nodes);
  }

  //=========================================================================
  // Regular Methods:
  //=========================================================================

  /**
   * Returns the number of nodes in the graph.
   *
   * @return the number of nodes.
   */
  
  public int numNodes () {
    return fNumNodes;
  }

  /**
   * Returns the number of edges in the graph.
   *
   * @return the number of edges.
   */
  
  public int numEdges () {
    return fNumEdges;
  }

  /**
   * Returns the number of roots.
   *
   * @return the number of nodes.
   */
  
  public int numRoots () {
    return fNumRoots;
  }
  
  /**
   * Returns the number of sinks in the graph.
   *
   * @return the number of nodes.
   */
  
  public int numSinks () {
    return fNumSinks;
  }

  /**
   * Returns the number of nodes that are both roots and sinks.
   *
   * @return the number of nodes.
   */
  
  public int numBothRootAndSink () {
    return fNumNodesBothRootAndSink;
  }

  /**
   * Returns the number of nodes that are neither root nor sink.
   *
   * @return the number of nodes.
   */
  
  public int numNeitherRootNorSink () {
    return fNumNodesNeitherRootNorSink;
  }

  /**
   * Returns the number of internal nodes in the graph.
   *
   * @return the number of internal nodes.
   */
  
  public int numInternalNodes () {
    return fNumNodes - fNumSinks;
  }

  /**
   * Returns an unmodifiable sorted set of the node types.
   *
   * @return the list.
   */
  
  public SortedSet<Class<?>> nodeTypes () {
    return (SortedSet<Class<?>>)fMap.keySet ();
  }

  /**
   * Returns the number of nodes whose getClass () method returns the given
   * class.
   *
   * @param t identifies the class.
   * @return the count.
   */

  public int simpleCountOfClass (Class<?> t) {
    return mark.core.util.Util.simpleCountOfClass (fMap, t);
  }
   
  /**
   * Given a class, returns the number of nodes of that class in the dag.
   *
   * @param t the node type.
   * @return the count.
   */

  public int countOfClass (Class<?> t) {
    return mark.core.util.Util.countOfClass (fMap, t);
  }
  
  /**
   * @see java.lang.Object
   */
     
  public String toString () {
    StringBuilder b = new StringBuilder ();
    for (Class<?> c : fMap.keySet ()) {
      b.append (c.getName () + "\t: " + simpleCountOfClass (c) + "\n");
    }
    b.append ("Nodes\t: " + numNodes () + "\n");
    b.append ("Edges\t: " + numEdges () + "\n");
    b.append ("Roots\t: " + numRoots () + "\n");
    b.append ("Sinks\t: " + numSinks () + "\n");
    b.append ("BothRootAndSink\t: " + numBothRootAndSink () + "\n");
    b.append ("NeitherRootNorSink\t: " + numNeitherRootNorSink () + "\n");
    return mark.core.text.Util.alignFields (b.toString (), "\t", ' ', 2);
  }

}
