package mark.core.shell;
import java.util.*;

/**
 * An ArgHelper assists with processing command-line arguments.  Typically,
 * an application will initialize an ArgHelper with the arguments it receives
 * as parameters to main.  It then informs the ArgHelper which arguments it
 * expects, so that the ArgHelper can print a usage string should something
 * go wrong.  Thereafter, the application will call the ArgHelper to extract
 * arguments.  Sample usage follows:
 * <xmp>
 *   public static void main (String[] args) {
 *     ArgHelper ah = new ArgHelper (args, Ls.class);
 *     ah.augmentUsage ("-v", false);
 *     ah.augmentUsage ("-f <file>", true);
 *     boolean verbose = ah.extractMatched ("-v");
 *     File file = ah.extractMatchedPlusOne ("-f", true);
 *     ...
 * </xmp>
 *
 * @author Mark Chavira
 */

public class ArgHelper {

  private String fName = null;  
  private String fUsage = null;
  private LinkedList<String> fArgs = null;

  /**
   * Extracts the argument that matches the given string and returns an
   * iterator pointing to the following argument.  If no match occurs,
   * returns null.
   *
   * @param arg the arg to match (e.g. "-f").
   * @return an Iterator that points to the following argument or null if
   *   the argument did not exist.
   */
  
  private Iterator<String> extract (String arg) {
    for (Iterator<String> i = fArgs.iterator (); i.hasNext ();) {
      if (arg.equals (i.next ())) {
        i.remove ();
        return i;
      }
    }
    return null;
  }

  /**
   * Extracts the argument that matches the given string and returns an
   * iterator pointing to the following argument.  If no match occurs,
   * throws an exception.
   *
   * @param arg the arg to match (e.g. "-f").
   * @return an Iterator that points to the following argument.
   * @throws Exception if required and no match occurs.
   */
  
  private Iterator<String> extractRequired (String arg)
   throws Exception {
    Iterator<String> argsIter = extract (arg);
    if (argsIter == null) {
      error ("Argument " + arg + " required.");    
    }
    return argsIter;
  }

  /**
   * Initializes the ArgHelper.
   *
   * @param args the arguments.
   * @param programName the name of the program.
   */
  
  public ArgHelper (String[] args, String programName) {
    fName = programName;
    fUsage = "";
    reset (args);
  }
  
  /**
   * Initializes the ArgHelper.
   *
   * @param args the arguments.
   * @param programClass the class of the program, which can be obtained as
   *   <class>.class.
   */
  
  public ArgHelper (String[] args, Class programClass) {
    this (args, "java " + programClass.getName ());
  }
  

  /**
   * Resets the arg helper with new arguments.
   *
   * @param args the new arguments.
   */

  public void reset (String[] args) {
    fArgs = new LinkedList<String> ();
    for (String s : args) {
      fArgs.add (s);
    }
  }
  
  /**
   * Appends an option to the usage string.
   *
   * @param option the option to append (e.g. "-t").
   * @param required whether the option is required.
   */

  public void augmentUsage (String option, boolean required) {
    fUsage +=
      (fUsage.equals ("") ? "" : "\n") +
      (required ? "  " : "  [") +
      option.replaceAll ("\n", (required ? "\n  " : "\n   ")) +
      (required ? "" : "]");
  }

  /**
   * Returns the name.
   *
   * @return the name.
   */
  
  public String name () {
    return fName;
  }

  /**
   * Returns the usage string.
   *
   * @return the string.
   */
  
  public String usage () {
    return fUsage;
  }

  /**
   * Throws an exception.  Prepends usage to the error message.
   *
   * @param message the exception message.
   * @throws Exception always.
   */
   
  public void error (String message) throws Exception {
    throw new Exception (
      "\nusage - " + fName + "\n" + fUsage + "\n" + message + "\n");
  }

  /**
   * Returns whether the size is zero.
   *
   * @return whether the size is zero.
   */

  public boolean isEmpty () {
    return size () == 0;
  }
  
  /**
   * Returns the number of arguments remaining in the ArgHelper.
   *
   * @return the number of arguments.
   */
  
  public int size () {
    return fArgs.size ();
  }

  /**
   * Throws an exception if the given argument is in the ArgHelper.
   *
   * @param arg the argument to match (e.g. "-f").
   * @throws Exception if the argument is in the ArgHandler.
   */

  public void verifyNoMatch (String arg) throws Exception {
    if (extract (arg) != null) {
      error ("Argument " + arg + " is not legal.");
    }
  }
  
  /**
   * Extracts and returns the first argument in the ArgHelper.  If no
   * argument exists, returns null.
   *
   * @param required whether an argument is required.
   * @return the extracted argument or null if no argument exists.
   * @throws Exception if required and no argument exists.
   */

  public String extractFirst (boolean required) throws Exception {
    if (isEmpty ()) {
      if (required) {
        error ("Argument required.");
      } else {
        return null;
      }
    }
    return fArgs.removeFirst ();
  }    
  
  /**
   * Extracts and returns all remaining arguments.
   *
   * @param required whether an argument is required.
   * @return the arguments.
   * @throws Exception if required and no argument exists.
   */

  public String[] extractAll (boolean required) throws Exception {
    String[] ans = fArgs.toArray (new String[fArgs.size ()]);
    fArgs.clear ();
    return ans;
  }

  /**
   * Extracts the argument that matches the given String.  Returns whether
   * a match occurred.
   *
   * @param arg the argument to match (e.g. "-f").
   * @return whether the argument existed.
   */

  public boolean extractMatched (String arg) {
    return extract (arg) != null;
  }

  /**
   * Extracts the argument that matches the given String and the following
   * argument and returns the second extracted argument.  If no match occurs,
   * returns null.
   *
   * @param arg the argument to match (e.g. "-f").
   * @param required whether the match is required.
   * @return the second extracted argument or null if there is no match.
   * @throws Exception if required and no match occurs or if a match occurs on
   *   the last argument.
   */

  public String extractMatchedPlusOne (String arg, boolean required) 
   throws Exception {
    Iterator<String> iter =
      required ? extractRequired (arg) : extract (arg);
    if (iter == null) {
      return null;
    }
    if (!iter.hasNext ()) {
      error ("Argument " + arg + " requires a parameter.");
    }
    String ans = iter.next ();
    iter.remove ();
    return ans;
  }
  
  /**
   * Extracts the argument that matches the given String and the following
   * argument, converts the second extracted argument into a parameter list,
   * and returns the parameter list.  If no match occurs, returns null.
   * 
   * A parameter list is an array of arguments, constructed from a single
   * comma-separated argument.  The argument "a,b,c" converts into the
   * argument list {"a", "b", "c"}.
   *
   * @param arg the argument to match (e.g. "-f").
   * @param required whether the match is required.
   * @return the parameter list or null.
   * @throws Exception if required and no match occurs or if a match occurs
   *   on the last argument.
   */

  public String[] extractMatchedPlusMany (String arg, boolean required)
   throws Exception {
    String combined = extractMatchedPlusOne (arg, required);
    if (combined == null) {
      return null;
    }
    return combined.split (",");
  }

  /**
   * If an argument exists, throws an exception.
   *
   * @throws Exception if an argument exists.
   */

  public void verifyEmpty () throws Exception {
    if (!isEmpty ()) {
      error ("Too many arguments.");
    }
  }

  /**
   * If no argument exists, throws an exception.
   *
   * @throws Exception if no argument exists.
   */
   
  public void verifyNotEmpty () throws Exception {
    if (fArgs.isEmpty ()) {
      error ("Must specify argument.");
    }
  }

}
