package mark.core.shell;
import java.util.*;
import java.io.*;

/**
 * Static utility methods which can be used in conjunction with an ArgHelper
 * to process certain types of arguments.
 *
 * @author Mark Chavira
 */

public abstract class ArgUtil {

  /**
   * Replaces each arg string in the given array with a new arg string that
   * incorporates a prefix.  If the prefix is null or the empty string, does
   * nothing.  Otherwise, constructs each new arg string by capitalizing the
   * second character and inserting the prefix between the first and second
   * characters.
   * <p>
   * Examples:
   * <ul>
   *   <li> prefix="train", arg="-filter" -> result="-trainFilter"
   *   <li> prefix="test", arg="-dataSet" -> result="-testDataSet"
   * </ul>
   *
   * @param prefix the prefix string.
   * @param args the arg strings.
   */

  private static void addPrefix (String prefix, String[] args) {
    if (prefix != null && prefix.length () > 0) {
      for (int i = 0; i < args.length; i++) {
        args[i] =
          args[i].charAt (0) +
          prefix + 
          Character.toUpperCase (args[i].charAt (1)) + 
          args[i].substring (2);
      }
    }
  }

  /**
   * Within the context of processing mutually exclusive arguments, verifies
   * that at most one of the arguments is specified.  Optionally, also
   * verifies that at least one is specified.
   *
   * @param ah the ArgHandler.
   * @param possibilities an array containing a true value for each mutually
   *   exclusive argument specified and a false value for each that is not
   *   specified.
   * @param required whether at least one of the arguments must be specified.
   * @param name the name of the argument.
   * @param the argument's prefix.
   * @return whether there actually was one argument.
   * @throws Exception when the conditions are not met.
   */

  public static boolean checkNumMatches (
   ArgHelper ah, boolean required, String name,
   String prefix, boolean... possibilities) throws Exception {
    int num = mark.core.util.Util.numValues (possibilities, true);
    if (required && num != 1 || num != 0 && num != 1) {
      ah.error (
        "Must supply " + (required ? "exactly " : "zero or ") + "one " + 
        ((prefix == null || prefix.length () == 0) ? "" : prefix + " ") +
        name + " definition.");
    }
    return num == 1;
  }

  /**
   * Returns the usage string for a random generator.
   *
   * @param opt the switch string.
   * @return the string.
   */
  
  public static String randomGeneratorUsage (String opt) {
    return opt + " <seed>";
  }

  /**
   * Extracts a random generator from the given ArgHelper.  If not required
   * and the ArgHelper does not specify a seed, then generates a seed from
   * the system clock.
   *
   * @param ah the given ArgHelper.
   * @param opt the switch string.
   * @param required whether the argument is required.
   * @param o if non-null, the seed will be written here.
   * @return the random generator.
   */
  

  public static Random extractRandomGenerator (
   ArgHelper ah, String opt, boolean required, PrintStream o)
   throws Exception {
    String temp = ah.extractMatchedPlusOne (opt, required);
    long seed =
      (temp == null) ? System.currentTimeMillis () : Long.parseLong (temp);
    if (o != null) {
      o.println ("Seed : " + seed);
    }
    return new Random (seed);
  }

  /**
   * Returns the usage string for a FileFilter argument.
   *
   * @param prefix the name of the arguement (if more than one file filter).
   * @return the usage string.
   */

  public static String fileFilterUsage (String prefix) {
    String[] opts = new String[] {"-include", "-exclude", "-filter"};
    addPrefix (prefix, opts);
    return
      "(" + opts[0] + " <includeString> |\n" +
      " " + opts[1] + " <excludeString> |\n" +
      " " + opts[2] + " <filter>)";
  }

  /**
   * Extracts a FileFilter argument from the given ArgHelper.  The method
   * will match at most one of the following:
   * <ol>
   *   <li> -include <pattern> (e.g. -include /cornell/)
   *   <li> -exclude <pattern> (e.g. -exclude /cornell/)
   *   <li> -filter <FileFilter> (e.g. -filter
   *        edu.ucla.belief.HuginFilter)
   * </ol>
   * If the method matches form (1)/(2), then the filter will accept files
   * with canonical paths that do/(do not) contain the given pattern.  If the
   * method matches form (3), then it will instantiate the named class.
   *
   * @param ah the ArgHelper.
   * @param required whether the argument is required.
   * @param prefix the name of the argument (if more than one of same type).
   * @return the FileFilter or null if there is none.
   * @throws Exception if there is no filter and required.
   */

  public static FileFilter extractFileFilter (
   ArgHelper ah, boolean required, String prefix) throws Exception {
    String[] opts = new String[] {"-include", "-exclude", "-filter"};
    addPrefix (prefix, opts);
    String include = ah.extractMatchedPlusOne (opts[0], false);
    String exclude = ah.extractMatchedPlusOne (opts[1], false);
    String filter  = ah.extractMatchedPlusOne (opts[2], false);
    checkNumMatches (
      ah, required, "filter", prefix, include != null, exclude != null,
      filter != null);
    if (include != null) {
      return new mark.core.io.NamePatternFileFilter (include);
    } else if (exclude != null) {
      return
        new mark.core.io.NegatedFileFilter (
          new mark.core.io.NamePatternFileFilter (exclude));
    } else {
      return (FileFilter)Class.forName (filter).newInstance ();
    }
  }

  /**
   * Returns the usage string for a file set argument.
   *
   * @param prefix the name of the arguement (if more than file set).
   * @return the usage string.
   */

  public static String fileSetUsage (String prefix) {
    String[] opts = new String[] {"-files", "-from"};
    addPrefix (prefix, opts);
    return "(" + opts[0] + " <fileList> | " + opts[1] + " <file>)";
  }

  /**
   * Extracts a file set argument from the given ArgHelper.  The method will
   * match at most one of the following:
   * <ol>
   *   <li> -files <file>(,<file>)* (e.g. -files a.txt,b.txt)
   *   <li> -from <file> (e.g. -from a.txt)
   * </ol>
   * If the method matches form (1), then it obtains paths from the command
   * line.  If the method matches form (2), then it obtains paths from the
   * given file.  Either way, it processes each path obtained recursively to
   * generate a complete set of potential paths.  Each potential path is then
   * sent to the given file filter to determine whether it is included in the
   * set.  If the client specified no file filter, then the method includes
   * all potential paths.
   *
   * @param ah the ArgHelper.
   * @param required whether the argument is required.
   * @param prefix the name of the argument (if more than one of same type).
   * @param ff the file filter.
   * @return an ArrayList with Files in it or null if there is none.
   * @throws Exception if there is no file set and required.
   */

  public static ArrayList<File> extractFileSet (
   ArgHelper ah, boolean required, String prefix, final FileFilter ff)
   throws Exception {
    String[] opts = new String[] {"-files", "-from"};
    addPrefix (prefix, opts);
    String[] files = ah.extractMatchedPlusMany (opts[0], false);
    String from = ah.extractMatchedPlusOne (opts[1], false);
    checkNumMatches (
      ah, required, "file set", prefix, files != null, from != null);
    if (from != null) {
      files =
        mark.core.io.Util.stringFromReader (
          new BufferedReader (new FileReader (from))).split ("\\s");
    }
    final ArrayList<File> ans = new ArrayList<File> ();
    for (int i = 0; i < files.length; i++) {
      mark.core.io.Util.processFilesRecursively (
        new FileFilter () {
          public boolean accept (File f) {
            if (f.getName ().startsWith (".")) {return false;}
            if (f.isFile () && ff.accept (f)) {ans.add (f);}
            return true;
          }
        }, new File (files[i]));
    }
    return (ans.isEmpty ()) ? null : ans;
  }

  /**
   * Executes static void main (String[]) for the given class with the given
   * parameters, redirecting standad in, out, and error to/from files with
   * the given names.
   *
   * @param c the given class.
   * @param p the given parameters.
   * @param stdin the standard in or null if use default.
   * @param stdout the standard out or null if use default.
   * @param stderr the standard error or null if use default.
   */
  
  public static void executeMain (
   Class c, String[] p, InputStream stdin,
   OutputStream stdout, OutputStream stderr)
   throws Exception {
    java.lang.reflect.Method m = c.getMethod ("main", p.getClass ());
    if (!java.lang.reflect.Modifier.isStatic (m.getModifiers ())) {
      throw new Exception ("main is not static!");
    }
    InputStream savedStdIn = System.in;
    PrintStream savedStdOut = System.out;
    PrintStream savedStdErr = System.err;
    if (stdin != null) {
      System.setIn (new BufferedInputStream (stdin));
    }
    if (stdout != null) {
      System.setOut (
        new PrintStream (
          new BufferedOutputStream (stdout)));
    }
    if (stderr != null) {
      System.setErr (
        new PrintStream (
          new BufferedOutputStream (stderr)));
    }
    m.invoke (null, new Object[] {p});
    if (stdin != null) {
      System.in.close ();
      System.setIn (savedStdIn);
    }
    if (stdout != null) {
      System.out.close ();
      System.setOut (savedStdOut);
    }
    if (stderr != null) {
      System.err.close ();
      System.setErr (savedStdErr);
    }
  }

}
