package mark.core.shell;
import java.util.*;
import java.io.*;

/**
 * A class that can create objects partially or completely specified at the
 * command line.
 * <p>
 * The user's perspective.  The user invokes a program and passes as an
 * option a description that identifies the name of a class to instantiate
 * and certain "cli" parametrs in the format expected by ParameterParser.
 * <p>
 * The program's perspective.  The program reads the description.  It knows
 * the interface or superclass of the type to be created, passes the line and
 * certain "mandated" parameters to commandLineObject (), receives the
 * created object and casts it to the interface or superclass.
 * <p>
 * This class' perspective.  This class receives a description and mandated
 * parameters, calls ParameterParser to parse the line, which yields the name
 * of the class CLASS to instantiate and the cli parameters.  If CLASS is an
 * enumeration and there is a single CLI parameter, then returns the value
 * identified in the CLI parameter.  If CLASS is an enumeration and there are
 * no CLI parameters, then verifies that there is only one enumeration value
 * and returns it.  If the class is an instance of Singleton, then performs
 * similarly.  Otherwise, returns the result of invoking the following method
 * defined in CLASS:
 * <xmp>
 * public static Object create (Object[] mandated, Object[] cli)
 *   throws Exception
 * </xmp>
 * <p>
 * The instantiated object's perspective.  The instantiated object must be
 * a value of an enumeration or its class must define the create () method
 * shown above.  This method should verify the parameters, create the object,
 * and then return it.  If it needs to create other cli objects in the process,
 * it should call commandLineObject2 ().  The class must implement CliObject.
 * 
 * @author chavira
 */

public class CliUtil {

  /**
   * Creates a command line object from the given string specified at the
   * command line and given mandated parameters.
   *
   * @param line the given string.
   * @param mandated the given mandated parameters.  This parameter may be
   *   null, in which case, there are no mandated parameters.
   * @return the created object.
   */

  public static Object commandLineObject (String line, Object... mandated)
   throws Exception {
    return commandLineObject2 (
      (new mark.core.parse.ParameterParser (
        new StringReader (line))).termPlusEof (),
        mandated);
  }

  /**
   * Creates a command line object from the given parsed output of the
   * string specified at the command line and given mandated parameters.
   *
   * @param term the parsed output of the cl string. 
   * @param mandated the given mandated parameters.  This parameter may be
   *   null, in which case, there are no mandated parameters.
   * @return the created object.
   */

  public static Object commandLineObject2 (Object term, Object... mandated)
   throws Exception {
   
    // Ensure that mandated is not null.
    
    if (mandated == null) {
      mandated = new Object[0];
    }
    
    // If the term is just a string, then it is the name of the class of the
    // object to return.  Otherwise, the term is a linked list, where the
    // first element is the name of the class of the element to return and
    // the remaining elements are cli parameters.  Break the term into name
    // and parameter list now.
   
    String name;
    Object[] cli;
    if (term instanceof String) {
      name = (String)term;
      cli = new Object[0];
    } else {
      LinkedList<?> temp = (LinkedList<?>)term;
      name = (String)temp.removeFirst ();
      cli = temp.toArray (new Object[temp.size ()]);
    }
    
    // Get hold of the class object.  Verify that it implements CliObject.

    Class<?> cl = Class.forName (name);
    if (!CliObject.class.isAssignableFrom (cl)) {
      throw new Exception (
        "Class " + cl.getName () + " must implement " +
        CliObject.class.getName ());
    }

    // If the class is an enumeration, then we want to return one of its
    // values.  If there aren't any cli parms, then verify that there is just
    // one value and return it.  If there is exactly one cli parm p, then
    // return the value named p.  If there is more than one cli parm, then
    // throw an exception.

    if (cl.isEnum ()) {
      Object[] vals = cl.getEnumConstants ();
      if (cli.length == 0) {
        if (vals.length != 1) {
          throw new Exception ("Must specify value name for enumeration!");
        }
        return vals[0];
      } else if (cli.length == 1) {
        for (int i = 0; i < vals.length; i++) {
          if (((Enum)vals[i]).name ().equals (cli[0])) {
            return vals[i];
          }
        }
        throw new Exception (
          cli[0] + " is not an enumerated value of " + cl.getName ());
     } else {
        throw new Exception ("Enumeration has too many parameters!");
      }
    }

    // If the class conforms to the singleton design pattern, then we want to
    // return one of its values.  If there aren't any cli parms, then verify
    // that there is a static field named "INSTANCE" and return it.  If there
    // is exactly one cli parm p, then return the object stored in the static
    // field named p.  If there is more than one cli parm, then throw an
    // exception.
    
    if (mark.core.util.Singleton.class.isAssignableFrom (cl)) {
      if (cli.length == 0) {
        return cl.getField ("INSTANCE").get (null);
      } else if (cli.length == 1) {
        return cl.getField ((String)cli[0]).get (null);
      } else {
        throw new Exception (
          "Singleton class " + name + " cannot accept parameters!");
      }
    }

    // Obtain the creation method.  Verify that the creation method returns
    // an object.  Invoke the creation method and return the result.

    java.lang.reflect.Method m =
      cl.getDeclaredMethod ("create", mandated.getClass (), cli.getClass ());
    if (m.getReturnType () != Object.class) {
      throw new Exception ("Creation method must return type Object!");
    }
    return m.invoke (null, mandated, cli);

  }

}
