package mark.core.shell;
import java.util.*;

/**
 * A class that accumulates results into key-value descriptions.  A result
 * can have up to two descriptions.  A long description limits both key and
 * value to one line.  A short description limits both key and value to one
 * word.  Often, the same result will have both a short and long description.
 * An info object can itself be a value.
 * 
 * @author chavira
 */

public class Info {

  /**
   * A pattern that matches any space character or an equal sign.
   */
  
  private static final java.util.regex.Pattern EQUAL_OR_SPACE =
    java.util.regex.Pattern.compile ("\\s|=");
  
  /**
   * A map from short key to short value.
   */
  
  private HashMap<String,Object> fShortKeyToValue =
    new HashMap<String,Object> ();
  
  /**
   * An ordering on the short keys.
   */
  
  private ArrayList<String> fShortKeys = new ArrayList<String> ();
  
  /**
   * A map from long key to long value.
   */
  
  private HashMap<String,Object> fLongKeyToValue =
    new HashMap<String,Object> ();
  
  /**
   * An ordering on the long keys.
   */
  
  private ArrayList<String> fLongKeys = new ArrayList<String> ();
    
  /**
   * Flattens nested short reults.
   * 
   * @param keyPrefix the prefix to use for keys.
   * @param order on exit, an ordering on the flattened keys will be
   *   appended.
   * @param on exit, the flattened map from key to value will be added. 
   */
  
  private void flatten (
   String keyPrefix, ArrayList<String> order, Map<String,Object> m) {
   for (String k : fShortKeys) {
      Object v = fShortKeyToValue.get (k);
      if (v instanceof Info) {
        ((Info)v).flatten (keyPrefix + k + "-", order, m);
      } else {
        if (order != null) {order.add (keyPrefix + k);}
        m.put (keyPrefix + k, v); 
      }
    }
  }
 
  /**
   * Adds a short result having the given key and value.
   * 
   * @param k the given key.
   * @param v the given value.
   */
  
  public void addShort (String k, Object v) throws Exception {
    if (fShortKeyToValue.keySet ().contains (k)) {
      throw new Exception ("Short key " + k + " already exists!");
    }
    if (EQUAL_OR_SPACE.matcher (k).find ()) {
      throw
        new Exception ("Key \"" + k + "\" contains space or \"=\"!");
    }
    if (
     !(v instanceof Info) &&
     EQUAL_OR_SPACE.matcher (v.toString ()).find ()) {
      throw
        new Exception ("Value \"" + v + "\" contains space or \"=\"!");
    }
    fShortKeys.add (k);
    fShortKeyToValue.put (k, v);
  }
  
  /**
   * Adds a long result having the given key and value.
   * 
   * @param k the given key.
   * @param v the given value.
   */
  
  public void addLong (String k, Object v) throws Exception {
    if (fLongKeyToValue.keySet ().contains (k)) {
      throw new Exception ("Long key " + k + " already exists!");
    }
    fLongKeys.add (k);
    fLongKeyToValue.put (k, v);
  }
  
  /**
   * Adds short result having the given short key and given value and adds a
   * long description having the given long key and given value.
   * 
   * @param sk the given short key.
   * @param lk the given long key.
   * @param v the given value.  
   */
  
  public void add (String sk, String lk, Object v) throws Exception {
    addShort (sk, v);
    addLong (lk, v);
  }
  
  /**
   * Adds all results from the given result object.
   * 
   * @param r the given result object.
   */
  
  public void addAll (Info r) throws Exception {
    for (String k : r.fShortKeys) {
      addShort (k, r.fShortKeyToValue.get (k));
    }
    for (String k : r.fLongKeys) {
      addLong (k, r.fLongKeyToValue.get (k));
    }
  }
  
  /**
   * Returns long descriptions where every line begins with the given prefix.
   *  
   * @param prefix the given prefix.
   * @return the string.
   */
  
   public String longDesc (String prefix) {
     if (prefix == null) {prefix = "";}
     StringBuilder ans = new StringBuilder ();
     for (String k : fLongKeys) {
       ans.append (prefix);
       ans.append (k);
       ans.append (" :");
       Object v = fLongKeyToValue.get (k);
       if (v instanceof Info) {
         ans.append (mark.core.io.Util.NEWLINE);
         ans.append (((Info)v).longDesc (prefix + "  "));
       } else {
         ans.append (" ");
         ans.append (v);
         ans.append (mark.core.io.Util.NEWLINE);
       }
     }
     return ans.toString ();
   }
   
  /**
   * Returns a description of short results.
   *  
   * @return the string.
   */
    
   public String shortDesc () throws Exception {
     HashMap<String,Object> flattened = new HashMap<String,Object> ();
     ArrayList<String> order = new ArrayList<String> ();
     flatten ("", order, flattened);
     StringBuilder ans = new StringBuilder ();
     for (int i = 0; i < order.size (); i++) {
       String k = order.get (i);
       if (i != 0) {ans.append (" ");}
       ans.append (k);
       String v = flattened.get (k).toString ();
       ans.append ("=");
       ans.append (v);
     }
     //ans.append (mark.core.io.Util.NEWLINE);
     return ans.toString ();
   }
  
   /**
    * Returns the short results as a map.
    * 
    * @return the map.
    */
   
   public HashMap<String,Object> shortMap () {
     HashMap<String,Object> flattened = new HashMap<String,Object> ();
     flatten ("", null, flattened);
     return flattened;     
   }
   
   /**
    * Returns a description of each short result.  Every value must be a
    * Result.
    * 
    * @param mainKey the main key.
    * @return the string.
    */
   
   public String shortDesc (String mainKey) throws Exception {
     StringBuilder ans = new StringBuilder ();
     for (String k : fShortKeys) {
       Info r = new Info ();
       r.addShort (mainKey, k);
       r.addAll ((Info)fShortKeyToValue.get (k));
       ans.append (r.shortDesc ());
     }
     return ans.toString ();
   }
   
   /**
    * Reads back values written by short Desc.  Warning: does not yet work
    * for nested results.
    */

   public static Info undoShortDesc (String s) throws Exception {
     Info ans = new Info ();
     String[] fields = s.split ("\\s+");
     for (int i = 0; i < fields.length; i++) {
       ans.addShort (
         fields[i].replaceFirst ("=.*", ""),
         fields[i].replaceFirst (".*=", ""));
     }
     return ans;
   }
   
   /**
    * @see Object
    */

   public String toString () {
     return fShortKeyToValue.toString (); 
   }

}
