package mark.core.text;
import java.util.*;

/**
 * Util is a class that contains a number of simple static utility methods.
 *
 * @author Mark Chavira
 */

public class Util {

  /**
   * Converts an integer to a string, formatted to fill at least the
   * given number of digits, prepending zeros if necessary.
   *
   * @param n the int.
   * @param numDigits the number of digits.
   * @return the string.
   */
  
  public static String format (int n, int numDigits) {
    String ans = Integer.toString (n);
    if (n < 0) {
      ans = ans.substring (1);
      --numDigits;
    }
    numDigits -= ans.length ();
    ans = repeat ("0", numDigits) + ans;
    return (n < 0) ? ("-" + ans) : ans;
  }

  /**
   * Duplicates a string the given number of times.
   *
   * @param s the string.
   * @param n the number of times.
   * @return the result.
   */
  
  public static String repeat (String s, int n) {
    StringBuilder b = new StringBuilder ();
    for (int i = 0; i < n; i++) {
      b.append (s);
    }
    return b.toString ();
  }

  /**
   * Pads the given string on the right.
   *
   * @param s the string.
   * @param width the total width of the constructed string.
   * @param padCharacter the pad character.
   * @return the constructed string.
   */

  public static String padRight (String s, int width, char padCharacter) {
    return s + repeat ("" + padCharacter, width - s.length ());
  }

  /**
   * Pads the given string on the left.
   *
   * @param s the string.
   * @param width the total width of the constructed string.
   * @param padCharacter the pad character.
   * @return the constructed string.
   */

  public static String padLeft (String s, int width, char padCharacter) {
    return repeat ("" + padCharacter, width - s.length ()) + s;
  }

  /**
   * Given lines consisting of fields, aligns the fields by adding padding to
   * fields.
   *
   * @param lines the lines.
   * @param delim the java.util.regexp field delimiter to be replaced with padding.
   * @param padCharacter the pad character.
   * @param widthBetween the number of characters between full-length fields.
   * @return the new lines.
   */

  public static String alignFields (
   String lines, String delim, char padCharacter, int widthBetween) {
    return
      alignFields (
        lines, delim, padCharacter, widthBetween, null, null, null, null);
  }
  
  /**
   * Given lines consisting of fields, aligns the fields by adding padding to
   * fields.  Each field has a maximum width.  If all elements in the field
   * require less space than the maximum, then the field is shrunk.  If an
   * element in the field requires more space, then it is wrapped.  If after
   * all is done there are some fields that require less space than their
   * maximum, then the space is distributed first to other fields that were
   * wrapped and then to all fields.
   *
   * @param text the lines.
   * @param delim the java.util.regexp field delimiter.
   * @param padCharacter the pad character.
   * @param widthBetween the number of characters between fields.
   * @param fieldWidths specifies a field width.  If null, then all fields
   *   are as long as necessary.  If a column has a positive int n, then the
   *   field is exactly n characters wide; if negative int n, then the field
   *   is at most -n characters wide.  If 0, then the field is as as long as
   *   necessary.
   * @param grow specifies whether each field can grow when extra space is
   *   available.  If null, then no fields grow.  Otherwise, for each column,
   *   indicates whether the column can grow.
   * @param wrapDelim the regular expression delimiter of tokens within a
   *   field.
   * @param wrapSeparator the string to place between wrapped tokens within
   *   a field.
   * @return the new lines.
   */

  public static String alignFields (
   String text,
   String delim,
   char padCharacter,
   int widthBetween,
   int[] fieldWidths,
   boolean grow[],
   String wrapDelim,
   String wrapSeparator) {

    // Store away whether there is a final newline and get an array of lines
    // where a line is an array of fields.

    boolean finalNewline = text.charAt (text.length () - 1) == '\n';
    String[][] lines;
    int maxFields = 0;
    {
      String[] temp = text.split ("\n");
      lines = new String[temp.length][];
      for (int l = 0; l < lines.length; l++) {
        lines[l] = temp[l].split (delim);
        maxFields = Math.max (lines[l].length, maxFields);
      }
    }

    // For each field, determine its maximum size in any line.

    int[] desiredSizes = new int[maxFields];
    for (int l = 0; l < lines.length; l++) {
      for (int f = 0; f < lines[l].length; f++) {
        desiredSizes[f] =
          Math.max (desiredSizes[f], lines[l][f].length ());
      }
    }
  
    // Adjust the field sizes based on the input.
  
    int[] actualSizes = new int[maxFields];
    {
      int extraSpace = 0;
      for (int f = 0; f < maxFields; f++) {
        if (
         fieldWidths == null ||
         f >= fieldWidths.length ||
         fieldWidths[f] == 0) {
          actualSizes[f] = desiredSizes[f];
        } else if (fieldWidths[f] > 0) {
          actualSizes[f] = fieldWidths[f];
        } else { // fieldWidths[f] < 0
          actualSizes[f] = Math.min (desiredSizes[f], -fieldWidths[f]);
          if (actualSizes[f] < -fieldWidths[f]) {
            extraSpace += -fieldWidths[f] - actualSizes[f];
          }
        }
      }
      if (grow != null && mark.core.util.Util.numValues (grow, true) > 0) {
        boolean change = true;
        while (extraSpace > 0 && change) {
          change = false;
          for (int f = 0; extraSpace > 0 && f < grow.length; f++) {
            if (grow[f] && actualSizes[f] < desiredSizes[f]) {
              actualSizes[f]++;
              extraSpace--;
              change = true;
            }
          }
        }
      }
    }
  
    // Now build the new string.
    
    StringBuilder ans = new StringBuilder ();
    for (int l = 0; l < lines.length; l++) {

      // For each field in the line, split the field into multiple "sublines"
      // and figure out the maximum number of sublines for any field.

      int numSublines = 0;
      String[][] s = new String[maxFields][];
      for (int f = 0; f < maxFields; f++) {
        if (f >= lines[l].length) {
          s[f] = new String[0];
        } else if (desiredSizes[f] > actualSizes[f]) {
          s[f] =
            wrap (lines[l][f], actualSizes[f], wrapDelim, wrapSeparator).
              split ("\n");
        } else {
          s[f] = new String[] {lines[l][f]};
        }
        numSublines = Math.max (numSublines, s[f].length);  
      }
            
      // Now collect the answer string.
      
      for (int subline = 0; subline < numSublines; subline++) {
        for (int field = 0; field < maxFields; field++) {
          ans.append (
            padRight (
              ((subline < s[field].length) ? s[field][subline] : ""),
              actualSizes[field] + widthBetween, padCharacter));
        }
        ans.append ("\n");
      }
      
    }
    
    // Remove extra space from the end of lines and remove the final newline
    // if the input did not contain one.  Return the result.
    
    String temp = ans.toString ();
    temp = temp.replaceAll ("\\s+\\n", "\n");
    if (!finalNewline) {
      temp = temp.substring (0, temp.length () - 1);
    }
    return temp;
    
  }

  /**
   * Joins objects' string representations into a single string based on
   * a delimeter String.  Does what Perl's join function does.
   *
   * @param delimeter the delimiter string.
   * @param comps the objects to join.
   * @param first the index of the first element to join.
   * @param last the index of the last element to join plus one.
   * @return the joined string.
   */

  public static String join (
   String delimiter, Collection<?> c, int first, int last) {
    StringBuilder ans = new StringBuilder ();
    Iterator<?> iter = c.iterator ();
    if (last - first > 0) {
      ans.append (iter.next ());
    }
    for (int i = first + 1; i < last; i++) {
      ans.append (delimiter);
      ans.append (iter.next ());
    }
    return ans.toString ();
  }
  
  /**
   * Joins objects' string representations into a single string based on
   * a delimeter String.  Does what Perl's join function does.
   *
   * @param delimeter the delimiter string.
   * @param comps the objects to join.
   * @param first the index of the first element to join.
   * @param last the index of the last element to join plus one.
   * @return the joined string.
   */

  public static <T> String join (
   String delimiter, T[] comps, int first, int last) {
    StringBuilder ans = new StringBuilder ();
    if (last - first > 0) {
      ans.append (comps[first]);
    }
    for (int i = first + 1; i < last; i++) {
      ans.append (delimiter);
      ans.append (comps[i].toString ());
    }
    return ans.toString ();
  }

  /**
   * Joins integers' string representations into a single string based on
   * a delimeter String.  Does what Perl's join function does.
   *
   * @param delimeter the delimiter string.
   * @param comps the objects to join.
   * @param first the index of the first element to join.
   * @param last the index of the last element to join plus one.
   * @return the joined string.
   */

  public static String join (
   String delimiter, int[] comps, int first, int last) {
    StringBuilder ans = new StringBuilder ();
    if (last - first > 0) {
      ans.append (comps[first]);
    }
    for (int i = first + 1; i < last; i++) {
      ans.append (delimiter);
      ans.append (comps[i]);
    }
    return ans.toString ();
  }

  /**
   * Joins doubles' string representations into a single string based on
   * a delimeter String.  Does what Perl's join function does.
   *
   * @param delimeter the delimiter string.
   * @param comps the objects to join.
   * @param first the index of the first element to join.
   * @param last the index of the last element to join plus one.
   * @return the joined string.
   */

  public static String join (
   String delimiter, double[] comps, int first, int last) {
    StringBuilder ans = new StringBuilder ();
    if (last - first > 0) {
      ans.append (comps[first]);
    }
    for (int i = first + 1; i < last; i++) {
      ans.append (delimiter);
      ans.append (comps[i]);
    }
    return ans.toString ();
  }

  /**
   * Joins objects' string representations into a single string based on
   * a delimeter String.  Does what Perl's join function does.
   *
   * @param delimeter the delimiter string.
   * @param comps the objects to join.
   * @return the joined string.
   */

  public static String join (String delimiter, Collection<?> comps) {
    return join (delimiter, comps, 0, comps.size ());
  }

  /**
   * Joins objects' string representations into a single string based on
   * a delimeter String.  Does what Perl's join function does.
   *
   * @param delimeter the delimiter string.
   * @param comps the objects to join.
   * @return the joined string.
   */

  public static <T> String join (String delimiter, T[] comps) {
    return join (delimiter, comps, 0, comps.length);
  }

  /**
   * Joins ints' string representations into a single string based on
   * a delimeter String.  Does what Perl's join function does.
   *
   * @param delimeter the delimiter string.
   * @param comps the objects to join.
   * @return the joined string.
   */

  public static String join (String delimiter, int[] comps) {
    return join (delimiter, comps, 0, comps.length);
  }

  /**
   * Joins double' string representations into a single string based on
   * a delimeter String.  Does what Perl's join function does.
   *
   * @param delimeter the delimiter string.
   * @param comps the objects to join.
   * @return the joined string.
   */

  public static String join (String delimiter, double[] comps) {
    return join (delimiter, comps, 0, comps.length);
  }

  /**
   * Replaces every occurrence of one substring with another.  Faster and
   * easier than String.replaceAll () when the search and replace strings are
   * not expressions.
   *
   * @param s the string to replace.
   * @param replace replace this string.
   * @param with with this string.
   */
  
  public static String replace (String s, String replace, String with) {
    int index = 0;
    while ((index = s.indexOf (replace, index)) != -1) {
      s = 
        s.substring (0, index) + with + 
        s.substring (index + replace.length ());
      index += with.length ();
    }
    return s;
  }

  /**
   * Wraps tokens.  If the delimter appears one or more times at the
   * beginning, then those characters are prepended to the first token.
   *
   * @param text the text to wrap.
   * @param width the maximum width of a line in the resulting text not
   *   including the newline.
   * @param delim the regular expression delimiter of tokens.
   * @param separator the string to place between tokens.
   * @return the wrapped text.
   */
  
  public static String wrap (
   String text, int width, String delim, String separator) {
    java.util.regex.Pattern p =
      java.util.regex.Pattern.compile ("^" + "(" + delim + ")+");
    java.util.regex.Matcher m = p.matcher (text);
    String prefix = m.find () ? m.group () : "";
    text = m.replaceAll ("");
    String[] tokens = text.split (delim);
    tokens[0] = prefix + tokens[0];
    int currentLineWidth = 0;
    StringBuilder ans = new StringBuilder ();
    for (int i = 0; i < tokens.length; i++) {
      int tokenWidth =
        tokens[i].length () +
        ((currentLineWidth == 0) ? 0 : separator.length ());
      if (currentLineWidth + tokenWidth <= width) {
        if (currentLineWidth > 0) {
          ans.append (separator);
        }
        ans.append (tokens[i]);
        currentLineWidth += tokenWidth;
      } else if (currentLineWidth == 0) {
        ans.append (tokens[i].substring (0, width));
        tokens[i] = tokens[i].substring (width);
        ans.append ('\n');
        currentLineWidth = 0;
        i--;
      } else {
        ans.append ('\n');
        currentLineWidth = 0;
        i--;
      }
    }
    return ans.toString ();
  }

  /**
   * Converts an object into a string.  If the object is an array or a
   * collection, processes recursively.
   *
   * @param a the array.
   * @param delimeter the delimiter string.
   * @return the string.
   */
  
  public static String recursiveToString (Object a, String delimiter) {
    try {
      if (a instanceof Collection) {
        a = ((Collection)a).toArray ();
      }
      if (a.getClass ().isArray ()) {
        int length = java.lang.reflect.Array.getLength (a);
        String[] s = new String[length];
        for (int i = 0; i < length; i++) {
          s[i] =
            recursiveToString (
              java.lang.reflect.Array.get (a, i), delimiter);
        }
        return "(" + mark.core.text.Util.join (delimiter, s) + ")";
      }
      return a.toString ();
    } catch (Exception e) {
      System.out.println ("Got an unexpected error in recursiveToString ()");
      System.out.println (e);
      e.printStackTrace (System.err);
      return "";
    }
  }

  /**
   * A generalized toString method.  Will work on any object, so long as
   * the object's components have legitimate toString methods.  Copied from
   * Core Java 2 Volume 1 page 214.
   *
   * @param o the object to convert to a string.
   * @return the object as a string.
   */

 public static String genericToString(Object obj)
   {  Class cl = obj.getClass();
      String r = cl.getName() + "[";
      Class sc = cl.getSuperclass();
      if (!sc.equals(Object.class)) r += sc + ",";
      java.lang.reflect.Field[] fields = cl.getDeclaredFields();
      try
      {  java.lang.reflect.AccessibleObject.setAccessible(fields, true);
      }
      catch(SecurityException e) {}
      for (int i = 0; i < fields.length; i++)
      {  java.lang.reflect.Field f = fields[i];
         r += f.getName() + "=";
         try
         {  Object val = f.get(obj);
            r += val.toString ();
         } catch (IllegalAccessException e)
         {  r += "???";
         }
         if (i < fields.length - 1) 
            r += ","; 
         else
            r += "]";
      }
      return r;
   }

}
