package mark.core.util;
import java.util.*;

/**
 * A priority queue implemented with a heap where elements are ints in
 * [0,N-1] for some N, where keys are comparables, where compareTo () defines
 * priority, and where the queue may not contain duplicate items.  The size
 * of the queue is linear in N, so N should not be outrageous.
 *
 * @author Mark Chavira
 */

public class IntKeyPQ extends PriorityQueue {

  protected int fNumElements;
  protected int[] fElementToSlot;
  protected int[] fElements;
  protected Comparable[] fKeys;

  private void addElementsWhenEmpty (
   int[] elems, Comparable[] keys) throws Exception {
    fNumElements = elems.length;
    for (int i = 0; i < elems.length; i++) {
      create (i, elems[i], keys[i]);
    }
    for (int i = fNumElements / 2 - 1; i >= 0; i--) {
      pushDown (i);
    }
  }

  /**
   * @see PriorityQueue
   */
  
  protected void exchange (int s1, int s2) {
    int e = fElements[s1];
    Comparable k = fKeys[s1];
    set (s1, fElements[s2], fKeys[s2]);
    set (s2, e, k);
  }

  /**
   * @see PriorityQueue
   */
  
  protected int comparePriorities (int slot1, int slot2) {
    return fKeys[slot1].compareTo (fKeys[slot2]);
  }

  /**
   * @see PriorityQueue
   */
  
  public int size () {
    return fNumElements;
  }

  /**
   * @see PriorityQueue
   */
  
  public void clear () {
    for (int i = 0; i < size (); i++) {
      fElementToSlot[fElements[i]] = -1;
    }
    fNumElements = 0;
  }
  
  /**
   * Clears the priority queue and then adds the given elements.
   *
   * @param elems the given elements.
   * @param keys the given elements' keys.
   */
  
  public void clear (int[] elems, Comparable[] keys) throws Exception {
    clear ();
    addElementsWhenEmpty (elems, keys);
  }

  /**
   * Fills the given slot with the given element and key after verifying that
   * it is not already in the queue.
   *
   * @param s the given slot.
   * @param e the given element.
   * @param k the given key.
   */
  
  protected void create (int s, int e, Comparable k) throws Exception {
    if (contains (e)) {
      throw new Exception ("Attempt to insert element already in queue.");
    }
    set (s, e, k);
  }

  /**
   * Fills the given slot with the given element and key.
   *
   * @param s the given slot.
   * @param e the given element.
   * @param k the given element's key.
   */
  
  protected void set (int s, int e, Comparable k) {
    fElements[s] = e;
    fKeys[s] = k;
    fElementToSlot[e] = s;
  }
  
  /**
   * Initializes an empty priority queue.
   *
   * @param N defines the elements that may be entered into the queue.
   */
  
  public IntKeyPQ (int N) {
    fNumElements = 0;
    fElements = new int[N];
    fKeys = new Comparable[N];
    fElementToSlot = new int[N];
    Arrays.fill (fElementToSlot, -1);
  }

  /**
   * Initializes the priority queue with the given elements.  This method
   * runs in time linear in the number of elements, and is thus faster than
   * creating a priority queue and then adding the elements.
   *
   * @param N defines the elements that may be entered into the queue.
   * @param elems the given elements.
   * @param keys the given elements' keys.
   */

  public IntKeyPQ (int N, int[] elems, Comparable[] keys)
   throws Exception {
    this (N);
    clear (elems, keys);
  }

  /**
   * Inserts the given element into the queue.
   *
   * @param e the given element.
   * @param k the given element's key.
   */

  public void add (int e, Comparable k) throws Exception {
    ++fNumElements;
    create (fNumElements - 1, e, k);
    pushUp (fNumElements - 1);
  }

  /**
   * Returns the highest priority element.
   *
   * @return the highest priority element.
   */

  public int highest () {
    return fElements[0];
  }
  
  /**
   * Returns the highest priority key.
   *
   * @return the highest priority key.
   */
  
  public Comparable highestKey () {
    return fKeys[0];
  }

  /**
   * Returns whether the queue contains the given element.
   *
   * @param e the given element.
   * @return whether the queue constains the given element.
   */

  public boolean contains (int e) {
    return fElementToSlot[e] != -1;
  }
  
  /**
   * Sets the key of the given element to the given value.
   *
   * @param e the given element.
   * @param k the given key.
   */
  
  public void setKey (int e, Comparable k) throws Exception {
    int s = fElementToSlot[e];
    if (s == -1) {
      throw new Exception ("Element not a member of queue.");
    }
    Comparable oldKey = fKeys[s];
    fKeys[s] = k;
    if (k.compareTo (oldKey) < 0) {
      pushUp (s);
    } else {
      pushDown (s);
    }
  }

  /**
   * Returns the key of the given element.
   * 
   * @param e the given element.
   * @reuturn the key.
   */
  
  public Comparable getKey (int e) throws Exception {
    int s = fElementToSlot[e];
    if (s == -1) {
      throw new Exception ("Element not a member of queue.");
    }
    return fKeys[s];
  }

  /**
   * Adjusts the priority queue after the given element's key has been
   * changed.
   *
   * @param e the given element.
   */
  
  public void changeKey (int e) throws Exception {
    int s = fElementToSlot[e];
    if (s == -1) {
      throw new Exception ("Element not a member of queue.");
    }
    if (s > 0 && fKeys[s].compareTo (fKeys[parent (s)]) < 0) {
      pushUp (s);
    } else {
      pushDown (s);
    }
  }

  /**
   * Removes and returns the highest priority element.  This method runs in
   * time logarithmic in the size of the heap.
   *
   * @return the removed integer.
   */
  
  public int remove () {
    int ans = highest ();
    --fNumElements;
    set (0, fElements[fNumElements], fKeys[fNumElements]);
    fElementToSlot[ans] = -1;
    pushDown (0);
    return ans;
  }

  /**
   * Returns the nth element in the priority queue, where n indicates how
   * the elements are stored not their priority.
   *
   * @param n identifies the element.
   * @return the nth element.
   */
  
  public int nthElement (int n) {
    return fElements[n];
  }
  
  /**
   * Returns the nth key in the priority queue, where n indicates how the
   * elements are stored not their priority.
   *
   * @param n identifies the element.
   * @return the nth key.
   */
  
  public Comparable nthKey (int n) {
    return fKeys[n];
  }
  
  /**
   * @see java.lang.Object
   */
  
  public String toString () {
    StringBuilder ans = new StringBuilder ();
    ans.append ("[");
    for (int i = 0; i < size (); i++) {
      ans.append (i > 0 ? "," : "");
      ans.append ("(" + fElements[i] + "->" + fKeys[i] + ")");
    }
    ans.append ("]");
    return ans.toString ();
  }
  
}
