package mark.core.util;
import java.util.*;

/**
 * A base class for priority queues.
 *
 * @author Mark Chavira
 */

public abstract class PriorityQueue {

  /**
   * Exchanges the contents of the two given slots.
   *
   * @param s1 the first given slot.
   * @param s2 the second given slot.
   */
    
  protected abstract void exchange (int s1, int s2);

  /**
   * Compares the priorities of the contents of the given slots and returns
   * -1 if the first has higher priority, 0 if they have the same priority,
   * and +1 if the first has lower priority.
   *
   * @param s1 the first given slot.
   * @param s2 the second given slot.
   * @return the comparison.
   */
  
  protected abstract int comparePriorities (int s1, int s2);

  /**
   * Returns the number of ties in the priority queue.
   * 
   * @param s recursive calls use this parameter; call with 0 initiallly.
   * @param scratchSize recursive calls use this parameter; call with 0 initially.
   * @return the number of elements in fScratch.
   */
  
  private int countTies (int s, int num) {
    if (s < size () && comparePriorities (s, 0) == 0) {
      num++;
      num = countTies (leftChild (s), num);
      num = countTies (rightChild (s), num);
    }
    return num;
  }

  /**
   * Returns the slot of the nth tie (starting at 0) in the priority queue.
   */
  
  private int swapNthTie (int s, int num, int n) {
    if (s < size () && comparePriorities (s, 0) == 0) {
      if (n == num) {
        exchange (0, s);
        return -1;
      }
      num++;
      num = swapNthTie (leftChild (s), num, n);
      if (num == -1) {return -1;}
      num = swapNthTie (rightChild (s), num, n);
    }
    return num;
  }
  
  /**
   * Swaps the element at the top of the heap with an element chosen randomly
   * from equal priority elements.  This operation is potentially linear in
   * the size of the heap.
   * 
   * @param r the random genertor.
   */
  
  public void swapTies (Random r) {
    int numTies = countTies (0, 0);
    if (numTies == 1) {
      return;
    }
    int elect = r.nextInt (numTies);
    if (elect != 0) {
      swapNthTie (0, 0, elect);
    }
  }

  /**
   * Returns the parent slot of the given slot.
   *
   * @param slot the given slot.
   * @return the parent slot.
   */

  protected int parent (int s) {
    return (s - 1) / 2;
  }
  
  /**
   * Returns the left child slot of the given slot.
   *
   * @param slot the given slot.
   * @return the left child slot.
   */
  
  
  protected int leftChild (int s) {
    return (2 * s) + 1;
  }
  
  /**
   * Returns the right child slot of the given slot.
   *
   * @param slot the given slot.
   * @return the right child slot.
   */
  
  protected int rightChild (int s) {
    return (2 * s) + 2;
  }
  
  /**
   * Ensures that the tree is a heap, assuming that the only possible
   * violation is that the content of the given slot may need to be pushed
   * lower.
   *
   * @param s the given slot.
   */

  protected void pushDown (int s) {
    for (;;) {
      int l = leftChild (s);
      int highest = s;
      if (l >= size ()) {
        break;
      }
      if (comparePriorities (l, s) < 0) {
        highest = l;
      }
      int r = rightChild (s);
      if (r < size () && comparePriorities (r, highest) < 0) {
        highest = r;
      }
      if (highest == s) {
        break;
      }
      exchange (s, highest);
      s = highest;
    }
  }

  /**
   * Ensures that the tree is a heap assuming that the only possible
   * violation is that the content of the given slot may need to be pushed
   * higher.
   *
   * @param s the given slot.
   */
  
  protected void pushUp (int s) {
    for (;;) {
      if (s == 0) {
        break;
      }
      int p = parent (s);
      if (comparePriorities (s, p) >= 0) {
        break;
      }
      exchange (s, p);
      s = p;
    }
  }

  /**
   * Throws an exception if the heap property is not satisfied.  This method
   * is for debugging.
   */
  
  public void validateHeapProperty () throws Exception {
    if (size () == 0) {
      return;
    }
    for (int s = 0; ; s++) {
      int l = leftChild (s);
      if (l >= size ()) {
        break;
      }
      if (comparePriorities (l, s) < 0) {
        throw new Exception ("Heap property violated!");
      }
      int r = rightChild (s);
      if (r >= size ()) {
        break;
      }
      if (comparePriorities (r, s) < 0) {
        throw new Exception ("Heap property violated!");
      }
    }
  }

  /**
   * Returns the number of elements in the queue.
   *
   * @return the number of elements.
   */
  
  public abstract int size ();

  /**
   * Resets the priority queue.
   */
  
  public abstract void clear ();

}
