package mark.core.util;

/**
 * A class to use when timing parts of a program.
 * 
 * @author Mark Chavira
 */
 
public class Timer {

  private int fMessageNumber;
  private boolean fRunning;
  private long fStartTime;
  private long fTotalTime;
  
  /**
   * Returns the current time.  Subclasses may overrided to use a different
   * timing mechanism.
   *
   * @return the current time.
   */

  protected long currentTime () {
    return System.currentTimeMillis ();
  }

  /**
   * Initializes the timer.
   */
  
  public Timer () {
    reset ();
  }

  /**
   * Resets the timer.
   */
  
  public void reset () {
    fMessageNumber = 1;
    fRunning = false;
    fStartTime = 0;
    fTotalTime = 0;
  }

  /**
   * Resets the message number.
   */
  
  public void resetMessageNumber () {
    fMessageNumber = 1;
  }

  /**
   * Starts the timer.
   */

  public void start () throws Exception {
    if (fRunning) {
      throw new Exception ("Timer already running!");
    }
    fRunning = true;
    fStartTime = currentTime ();
  }
  
  /**
   * Stops the timer, returns the duration since the last start, and adds
   * this duration to the total time.
   *
   * @return the duration.
   */
  
  public long stop () throws Exception {
    long ans = currentTime () - fStartTime;
    if (!fRunning) {
      throw new Exception ("Timer not running!");
    }
    fRunning = false;
    fTotalTime += ans;
    return ans;
  }
  
  /**
   * Invokes stop () and then start () for convenience.
   *
   * @return the result of calling stop ().
   */
  
  public long stopAndStartAgain () throws Exception {
    long ans = stop ();
    start ();
    return ans;
  }
  
  /**
   * Returns the total time.
   *
   * @return the total time.
   */
  
  public long totalTime () {
    return fTotalTime;
  }
  
  /**
   * Stops the timer and prints to standard error the number of messages
   * displayed since the last message number reset, the given message, and
   * the number of seconds since the last start.
   *
   * @param message the message.
   * @return the number of milliseconds.
   */
  
  public long stopAndDisplay (String message) throws Exception {
    long newTime = stop ();
    System.err.println (
      "" + fMessageNumber++ + ". " + ((message == null) ? "" : message) +
      " " + ((double)newTime / 1000.0) + "s");
    return newTime;
  }
  
  /**
   * Invokes stopAndDisplay () and then start () for convenience;
   *
   * @param message the message.
   * @return the number of milliseconds.
   */
  
  public long stopDisplayAndStartAgain (String message) throws Exception {
    long ans = stopAndDisplay (message);
    start ();
    return ans;
  }
  
}
