package mark.core.util;
import java.util.*;

/**
 * Static utility methods.
 *
 * @author Mark Chavira
 */

public class Util {

  /**
   * Returns whether the first given set is a superset of the second given set.
   *  
   * @param lhs the first given set.
   * @param rhs the second given set.
   * @return whether lhs is a superset of rhs.
   */

  public static <E> boolean superset (Set<E> lhs, Set<E> rhs) {
    if (lhs.size () < rhs.size ()) {
      return false;
    }
    for (E e : rhs) {
      if (!lhs.contains (e)) {
        return false;
      }
    }
    return true;
   }
  
  /**
   * Returns whether the first given set is a subset of the second given set.
   *  
   * @param lhs the first given set.
   * @param rhs the second given set.
   * @return whether lhs is a subset of rhs.
   */

  public static <E> boolean subset (Set<E> lhs, Set<E> rhs) {
    return superset (rhs, lhs);
  }

  /**
   * Returns an int[] containing the same values as the given collection.
   * 
   * @param c the given collection.
   * @return the int[].
   */
 
  public static int[] toArray (Collection<Integer> c) {
    int[] ans = new int[c.size ()];
    int index = 0;
    for (int elem : c) {
      ans[index++] = elem;
    }
    return ans;
  }
  
  /**
   * An int[] comparator.
   */
  
  public static Comparator<int[]> INT_ARRAY_COMPARATOR =
    new Comparator<int[]> () {
      public int compare (int[] lhs, int[] rhs) {
        if (lhs.length < rhs.length) {return -1;}
        if (lhs.length > rhs.length) {return +1;}
        for (int i = 0; i < lhs.length; i++) {
          if (lhs[i] < rhs[i]) {return -1;}
          if (lhs[i] > rhs[i]) {return +1;}
        }
        return 0;
      }
    };

  /**
   * Java's array clone is very slow (see
   * http://forums.java.net/jive/thread.jspa?threadID=743&messageID=16562),
   * so for critical methods, use this method instead.
   *
   * @param a the array to clone.
   * @return the clone.
   */
    
  public static int[] clone (int[] a) {
    int[] ans = new int[a.length];
    System.arraycopy (a, 0, ans, 0, a.length);
    return ans;
  }
  
  /**
   * Clones a two dimensional array.
   * 
   * @apram a the array.
   * @return the clone.
   */
  
  public static int[][] clone (int[][] a) {
    int[][] ans = new int[a.length][];
    for (int i = 0; i < ans.length; i++) {
      ans[i] = clone (a[i]);
    }
    return ans;
  }
  
  /**
   * Java's array clone is very slow (see
   * http://forums.java.net/jive/thread.jspa?threadID=743&messageID=16562),
   * so for critical methods, use this method instead.
   *
   * @param a the array to clone.
   * @return the clone.
   */
    
  public static double[] clone (double[] a) {
    double[] ans = new double[a.length];
    System.arraycopy (a, 0, ans, 0, a.length);
    return ans;
  }
  
  /**
   * Clones a two dimensional array.
   * 
   * @apram a the array.
   * @return the clone.
   */
  
  public static double[][] clone (double[][] a) {
    double[][] ans = new double[a.length][];
    for (int i = 0; i < ans.length; i++) {
      ans[i] = clone (a[i]);
    }
    return ans;
  }

  /**
   * Clones the given array.
   * 
   * @param a the array.
   * @return the clone.
   */
  
  public static boolean[] clone (boolean[] a) {
    boolean[] ans = new boolean[a.length];
    System.arraycopy (a, 0, ans, 0, a.length);
    return ans;
  }
  
  /**
   * Clones a two dimensional array.
   * 
   * @apram a the array.
   * @return the clone.
   */
  
  public static boolean[][] clone (boolean[][] a) {
    boolean[][] ans = new boolean[a.length][];
    for (int i = 0; i < ans.length; i++) {
      ans[i] = clone (a[i]);
    }
    return ans;
  }
  
  /**
   * Retuns an array of ints that is formed by concatenating the given
   * arrays of ints.
   *
   * @param a the given arrays of ints.
   * @return the single array of ints.
   */
  
  public static int[] flatten (int[]... a) {
    int num = 0;
    for (int i = 0; i < a.length; i++) {
      num += a[i].length;
    }
    int[] ans = new int[num];
    int index = 0;
    for (int i = 0; i < a.length; i++) {
      for (int j = 0; j < a[i].length; j++) {
        ans[index++] = a[i][j];
      }
    }
    return ans;
  }

  /**
   * Creates a table with the given number of columns and the minimum number
   * of rows that will house the given items and fills the table first top to
   * bottom and then left to right.  For example, passing Integers 0..7 with
   * 3 columns produces:
   * <xmp>
   * 0 3 6
   * 1 4 7
   * 2 5 
   * </xmp>
   * Integers 0..22 with 4 colums produces:
   * <xmp>
   * 0 6 12 17
   * 1 7 13 18
   * 2 8 14 19
   * 3 9 15 20
   * 4 10 16 21
   * 5 11
   * </xmp>
   *
   * @param items the given items in order.
   * @param numCols the given number of columns.
   * @return the vertical fill table.
   */

  public static Object[][] verticalFillTable (Object[] items, int numCols) {
    int numFullRows = items.length / numCols;
    int numItemsInExtraRow = items.length % numCols;
    boolean extraRow = numItemsInExtraRow > 0;
    int numRows = numFullRows + (extraRow ? 1 : 0);
    int numFullCols = extraRow ? numItemsInExtraRow : numCols;
    Object[][] ans = new Object[numRows][];
    for (int r = 0; r < numRows; r++) {
      ans[r] = new Object[r < numFullRows ? numCols : numItemsInExtraRow];
    }
    int itemIndex = 0;
    for (int c = 0; c < numCols; c++) {
      int numRowsInThisColumn = numRows - (c < numFullCols ? 0 : 1);
      for (int r = 0; r < numRowsInThisColumn; r++) {
        ans[r][c] = items[itemIndex++];
      }
    }
    return ans;
  }

  /**
   * A list comparator that compares element by element using compareTo ()
   * and that places shorter lists first when all else is equal.
   */
  
  public static final <T extends Comparable<T>>
  Comparator<List<T>> listComparator () {
    return
      new Comparator<List<T>> () {
        public int compare (List<T> lhs, List<T> rhs) {
          Iterator<T> l = lhs.iterator ();
          Iterator<T> r = rhs.iterator ();
          for (;;) {
            if (!l.hasNext () && !r.hasNext ()) {
              return 0;
            }
            if (!l.hasNext ()) {
              return -1;
            }
            if (!r.hasNext ()) {
              return +1;
            }
            int temp = l.next ().compareTo (r.next ());
            if (temp != 0) {
              return temp;
            }
          }
        }
      };
  }

  /**
   * Returns a map that is the reverse of the given map.  If more than one
   * element in the domain maps to the same element in the range, then one
   * of the two domain elements will be lost.
   *
   * @param map the given map.
   * @return the reverse.
   */
  
  public static <K,V> HashMap<V,K> reverseMap (
   Map<K,V> map) {
    HashMap<V,K> ans = new HashMap<V,K> ();
    for (K oldKey : map.keySet ()) {
      ans.put (map.get (oldKey), oldKey);
    }
    return ans;
  }

  /**
   * Changes the pre-image of the given map.
   *
   * @param map the given map.
   * @param preImageTransform the transformation for the preimage.
   * @return the resulting map.
   */
  
  public static <K,V,K2>HashMap<K2,V> transformPreimage (
   Map<K,V> map, Map<K,K2> preImageTransform) {
    HashMap<K2,V> ans = new HashMap<K2,V> ();
    for (Map.Entry<K,V> e : map.entrySet ()) {
      K k = e.getKey ();
      V v = e.getValue ();
      ans.put (preImageTransform.get (k), v);
    }
    return ans;
  }

  /**
   * Changes the image of the given map.
   *
   * @param map the given map.
   * @param imageTransform the transformation for the image.
   * @return the resulting map.
   */
  
  public static <K,V,V2>HashMap<K,V2> transformImage (
   Map<K,V> map, Map<V,V2> imageTransform) {
    HashMap<K,V2> ans = new HashMap<K,V2> ();
    for (Map.Entry<K,V> e : map.entrySet ()) {
      K k = e.getKey ();
      V v = e.getValue ();
      ans.put (k, imageTransform.get (v));
    }
    return ans;
  }

  /**
   * Changes the pre-image and the image of the given map.
   *
   * @param map the given map.
   * @param preImageTransform the transformation for the preimage.  If null,
   *   then the pre-image not chagned.
   * @param imageTransform the transformation for the image.  If null, then
   *   image not changed.
   * @return the resulting map.
   */
  
  public static <K,V,K2,V2>HashMap<K2,V2> transformMap (
   Map<K,V> map,
   Map<K,K2> preImageTransform,
   Map<V,V2> imageTransform) {
    HashMap<K2,V2> ans = new HashMap<K2,V2> ();
    for (Map.Entry<K,V> e : map.entrySet ()) {
      K k = e.getKey ();
      V v = e.getValue ();
      ans.put (preImageTransform.get (k), imageTransform.get (v));
    }
    return ans;
  }
  /**
   * Returns the image of the given collection defined by the given map.
   *
   * @praam source the given set.
   * @param map the given map.
   * @return the image.
   */
  
  public static <K,V> HashSet<V> image (
   Collection<? extends K> source, Map<K,V> map) {
    HashSet<V> ans = new HashSet<V> ();
    for (K k : source) {
      ans.add (map.get (k));
    }
    return ans;
  }

  /**
   * Returns the image of the given collection defined by the given map as a
   * list.
   *
   * @param c the given collection.
   * @param the given map.
   * @return the image.
   */
  
  public static <K,V> ArrayList<V> imageAsList (
   Collection<? extends K> source, Map<K,V> map) {
    ArrayList<V> ans = new ArrayList<V> ();
    for (K k : source) {
      ans.add (map.get (k));
    }
    return ans;
  }

  /**
   * Returns a map that maps each given element into itself.
   *
   * @param s the given elements.
   * @return the map.
   */
  
  public static <T> HashMap<T,T> identityMap (
   Collection<T> s) {
    HashMap<T,T> ans = new HashMap<T,T> ();
    for (T t : s) {
      ans.put (t, t);
    }
    return ans;
  }
  
  /**
   * Returns the count of the given object in the given count map.
   *
   * @param cm the given count map.
   * @param o the given object.
   * @return the count.
   */
  
  public static <T> int cmCount (HashMap<T,Integer> cm, T t) {
    Integer temp = cm.get (t);
    return (temp == null) ? 0 : temp;
  }
  
  /**
   * Increments the count of the given object in the given count map by the
   * given amount.
   * 
   * @param cm the given count map.
   * @param t the given object.
   * @param inc the given amount.
   */
  
  public static <T> void cmIncrement (
   HashMap<T,Integer> cm, T t, int inc) {
    cmSet (cm, t, cmCount (cm, t) + inc);
  }
  
  /**
   * Replaces the count of the given object in the given count map by the
   * maximum of its current count and the given count.
   * 
   * @param cm the given count map.
   * @param o the given object.
   * @param c the given count.
   */
  
  public static <T> void cmMax (HashMap<T,Integer> cm, T t, int c) {
    if (cmCount (cm, t) < c) {cmSet (cm, t, c);}
  }
  
  /**
   * Sets the count of the given object in the given count map to the given
   * value.
   *
   * @param cm the given count map.
   * @param o the given object.
   * @param val the given value.
   */
  
  public static <T> void cmSet (
   HashMap<T,Integer> cm, T t, int val) {
    if (val == 0) {
      cm.remove (t);
    } else {
      cm.put (t, val);
    }
  }
  
  /**
   * Retruns a map from each object in the given list to its index.  If an
   * object appears more than once, it will receive its first index.
   *
   * @param l the given list.
   * @return the map.
   */
  
  public static <T> HashMap<T,Integer> indexMap (List<T> l) {
    HashMap<T,Integer> ans = new HashMap<T,Integer> ();
    int index = l.size ();
    for (ListIterator<T> i = l.listIterator (l.size ()); i.hasPrevious ();) {
      ans.put (i.previous (), --index);
    }
    return ans;
  }
   
  /**
   * Retruns a map from each object in the given list to its index.  If an
   * object appears more than once, it will receive its first index.
   *
   * @param l the given list.
   * @return the map.
   */
  
  public static <T> HashMap<T,Integer> indexMap (T[] l) {
    HashMap<T,Integer> ans = new HashMap<T,Integer> ();
    for (int index = l.length - 1; index >= 0; index--) {
      ans.put (l[index], index);
    }
    return ans;
  }
  
  /**
   * Returns whether the first given map agrees with the second on the given
   * keys.
   *
   * @param keys the given keys.
   * @param lhs the first given map.
   * @param rhs the second given map.
   * @return whether the two agree agree.
   */
  
  public static <K,V> boolean compatible (
   Collection<? extends K> keys,
   Map<? extends K,? extends V> lhs,
   Map<? extends K,? extends V> rhs) {
    for (K k : keys) {
      if (!lhs.get (k).equals (rhs.get (k))) {
        return false;
      }
    }
    return true;
  }
    
  /**
   * Projects the given map onto the given keys.
   *
   * @param map the given map.
   * @param keys the given keys.
   * @return the projection.
   */
   
  public static <K,V> HashMap<K,V> project (
   Map<K,V> map, Set<? extends K> keys) {
    HashMap<K,V> ans = new HashMap<K,V> ();
    for (K k : keys) {
      V v = map.get (k);
      if (v != null) {
        ans.put (k, v);
      }
    }
    return ans;
  }

  /**
   * Returns an array of evenly spaced integers.
   *
   * @param firstValue the value in the first slot.
   * @param increment the increment.
   * @param length the length of the array.
   */

  public static int[] increasingIntegers (
   int firstValue, int increment, int length) {
    int[] ans = new int[length];
    for (int i = 0; i < ans.length; i++) {
      ans[i] = firstValue;
      firstValue += increment;
    }
    return ans;
  }
  
  /**
   * Reverses an array of ints.
   *
   * @param a the array.
   */
  
  public static void reverse (int[] a) {
    for (int i = 0; i < a.length / 2; i++) {
      int temp = a[i];
      a[i] = a[a.length - 1 - i];
      a[a.length - 1 - i] = temp;
    }
  }

  /**
   * Returns whether the arrays are permutations of each other.
   *
   * @param lhs the first array.
   * @param rhs the second array.
   */

  public static boolean permutation (int[] lhs, int[] rhs) {
    if (lhs.length != rhs.length) {
      return false;
    }
    int[] d = clone (lhs);
    int[] p = clone (rhs);
    Arrays.sort (d);
    Arrays.sort (p);
    return Arrays.equals (d, p);
  }

  /**
   * Returns the number of consecutive pairs that differ.  The min is 0.
   * The max is the length of the array minus one.
   *
   * @param a the array.
   */
  
  public static int numConsecutiveDifferent (Object[] a) {
    int ans = 0;
    for (int i = 1; i < a.length; i++) {
      if (a[i - 1] != a[i]) {
        ++ans;
      }
    }
    return ans;
  }

  /**
   * Constructs a map from the given parallel arrays, which specify keys
   * and values.
   * 
   * @param k the keys.
   * @param v the values.
   */
 
  public static <K,V> HashMap<K,V> constructMap (
   ArrayList<K> k, ArrayList<V> v) {
    HashMap<K,V> ans = new HashMap<K,V> ();
    for (int i = 0; i < k.size (); i++) {
      ans.put (k.get (i), v.get (i));
    }
    return ans;
  }
  
  /**
   * Returns the number of occurrences of the given value within the given
   * array.
   *
   * @param a the array.
   * @param v the value.
   * @return the number of occurrences.
   */
  
  public static int numValues (boolean[] a, boolean v) {
    int ans = 0;
    for (int i = 0; i < a.length; i++) {
      if (a[i] == v) {
        ++ans;
      }
    }
    return ans;
  }

  /**
   * Returns the number of occurrences of the given value within the given
   * array.
   *
   * @param a the array.
   * @param v the value.
   * @return the number of occurrences.
   */
  
  public static int numValues (int[] a, int v) {
    int ans = 0;
    for (int i = 0; i < a.length; i++) {
      if (a[i] == v) {
        ++ans;
      }
    }
    return ans;
  }

  /**
   * Returns the number of occurrences of the given value within the given
   * array.
   *
   * @param a the array.
   * @param v the value.
   * @return the number of occurrences.
   */
  
  public static int numValues (double[] a, double v) {
    int ans = 0;
    for (int i = 0; i < a.length; i++) {
      if (a[i] == v) {
        ++ans;
      }
    }
    return ans;
  }

  /**
   * Returns whether two objects are equal.  If the objects are lists, then
   * processes recursively.
   *
   * @param l1 the first list.
   * @param l2 the second list.
   * @return whether the two are equal.
   */
  
  public static boolean recursiveEquals (Object o1, Object o2) {
    boolean list1 = o1 instanceof List<?>;
    boolean list2 = o2 instanceof List<?>;
    if (list1 != list2) {
      return false;
    }
    if (!list1) {
      return o1.equals (o2);
    }
    List<?> l1 = (List<?>)o1;
    List<?> l2 = (List<?>)o2;
    if (l1.size () != l2.size ()) {
      return false;
    }
    Iterator<?> i1 = l1.iterator ();
    Iterator<?> i2 = l2.iterator ();
    while (i1.hasNext ()) {
      if (!recursiveEquals (i1.next (), i2.next ())) {
        return false;
      }
    }
    return true;
  }

  /**
   * A generalized equals method.  Will work on any two objects, so long
   * as the objects components have legitimate equals methods.  Copied from
   * Core Java 2 Volume 1 page 214.
   *
   * @param o the object to convert to a string.
   * @return the object as a string.
   */

   public static boolean genericEquals(Object a, Object b)
   {  Class cl = a.getClass();
      if (!cl.equals(b.getClass())) return false;
      java.lang.reflect.Field[] fields = cl.getDeclaredFields();
      java.lang.reflect.AccessibleObject.setAccessible(fields, true);
      for (int i = 0; i < fields.length; i++)
      {  java.lang.reflect.Field f = fields[i];
         try
         {  if (!f.get(a).equals(f.get(b)))
               return false;
         }
         catch(IllegalAccessException e)
         {  return false;
         }
      }
      return true;
   }

  /**
   * Given a collection of objects, returns an unmodifiable sorted map from
   * the types of the objects to the number of objects of that type.  Keys
   * are therefore of type Class and values are of type Integer.  The map is
   * sorted in that the iterator will return the keys sorted by their names.
   *
   * @param c the objects.
   * @return the map.
   */

  public static SortedMap<Class<?>,Integer> typeStatistics (
   Collection<?> c) {
    HashMap<Class<?>,Integer> temp = new HashMap<Class<?>,Integer> ();
    for (Object o : c) {
      cmIncrement (temp, o.getClass (), 1);
    }
    TreeMap<Class<?>,Integer> ans =
      new TreeMap<Class<?>,Integer> (
        new Comparator<Class<?>> () {
          public int compare (Class<?> o1, Class<?> o2) {
            return (o1.getName ().compareTo (o2.getName ()));
          }
        });
    ans.putAll (temp);
    return Collections.unmodifiableSortedMap (ans);
  }

  /**
   * Assists childrenBeforeParents ().
   *
   * @param elms the graph specification.
   * @param order the order so far.
   * @param orderIndex the index of the next node in the order.
   * @param visited which nodes have been visited.
   * @param i the node to attempt to visit.
   * @return the updated orderIndex.
   */
   
  private static int childrenBeforeParentsHelper (
   Collection<Integer>[] elms, int[] order, int orderIndex, 
   boolean[] visited, int i) {
    if (visited[i]) {
      return orderIndex;
    }
    visited[i] = true;
    for (int el : elms[i]) {
      orderIndex =
        childrenBeforeParentsHelper (elms, order, orderIndex, visited, el);
    }
    order[orderIndex] = i;
    return orderIndex + 1;
  }

  /**
   * Given a specification for an acyclic graph, returns an ordering that
   * places children before parents.
   *
   * @param elms the graph.  The nodes are named 0..elms.length - 1.  For
   *   each, we specify a collection of its children, where each child is
   *   specified as an Integer.
   * @return the ordering.
   */

  public static int[] childrenBeforeParents (
   Collection<Integer>[] elms) throws Exception {
    boolean[] visited = new boolean[elms.length];
    int[] order = new int[elms.length];
    int orderIndex = 0;
    for (int i = 0; i < elms.length; i++) {
      orderIndex =
        childrenBeforeParentsHelper (elms, order, orderIndex, visited, i);
    }
    return order;
  }

  /**
   * Given a specification for an acyclic graph, returns an ordering that
   * places parents before children.
   *
   * @param elms the graph.  The nodes are named 0..elms.length - 1.  For
   *   each, we specify a collection of its children, where each child is
   *   specified as an Integer.
   * @return the ordering.
   */

  public static int[] parentsBeforeChildren (
   Collection<Integer>[] elms) throws Exception {
    int[] ans = childrenBeforeParents (elms);
    for (int i = 0; i < ans.length / 2; i++) {
      int t = ans[i];
      ans[i] = ans[ans.length - 1 - i];
      ans[ans.length - 1 - i] = t;
    }
    return ans;
  }

  /**
   * Returns the int corresponding to the given Class as defined by the given
   * map from Class to Integer or 0 if there is no mapping for the given
   * class.
   *
   * @param m the given map.
   * @param t the given class.
   * @return the count or zero.
   */

  public static int simpleCountOfClass (
   Map<Class<?>,Integer> m, Class t) {
    Integer ans = m.get (t);
    return ans == null ? 0 : ans;
  }
  
  /**
   * For the given class and all decendents of the given class, returns the
   * sum of the values returned by simpleCountOfClass.
   *
   * @param m the given map.
   * @param t the given class.
   */
  
  public static int countOfClass (
   Map<Class<?>,Integer> m, Class<?> t) {
    int ans = 0;
    for (Class<?> c : m.keySet ()) {
      if (t.isAssignableFrom (c)) {
        ans += simpleCountOfClass (m, c);
      }
    }
    return ans;
  }

  /**
   * Returns an absolute time given a base time and a relative time.
   *
   * @param startTime the given base time in milliseconds.
   * @param maxMillis the given relative time in milliseconds (
   *   Long.MAX_VALUE for forever).
   * @return the absolute time in milliseconds.
   */
  
  public static long absoluteTime (long startTime, long maxMillis) {
    long timeout = startTime + maxMillis;
    return timeout >= 0 ? timeout : Long.MAX_VALUE;  
  }

  /**
   * Returns an absolute time given a time relative to the current time.
   *
   * @param maxMillis the given relative time (Long.MAX_VALUE for forever).
   * @return the absolute time in milliseconds.
   */
  
  public static long absoluteTime (long maxMillis) {
    return absoluteTime (System.currentTimeMillis (), maxMillis);
  }

  /**
   * Given a list L=(0,...,K-1), a positive integer S, and a non-negative
   * integer I, returns the first element from a set T, where T is the
   * I'th set formed if we were to partition L into S approximately equal-
   * sized sets of consecutive elements from L.  To get the last element,
   * call with I set to one greater and subtract one from the result.  To
   * identify the last element in the identified set, call with I
   * incremented by 1 and subtract one from the answer.  If S is smaller
   * than K, then there will be empty sets, which can be identified
   * by the fact that the last element is smaller than the first.
   * 
   * @param K the number of elements in the list.
   * @param S the number of sets into which to partition the list.
   * @param I identifies the set of interest (starting at 0).
   * @return the first element in the identified set.
   */
  
  public static int equalSizeFirst (int K, int S, int I) {
    return K / S * I + Math.min (I, K % S);
  }

  /**
   * Shuffles the given array of ints.
   * 
   * @param a the given array.
   */
  
  public static void shuffle (int[] a) {
    ArrayList<Integer> temp = new ArrayList<Integer> ();
    for (int j = 0; j < a.length; j++) {
      temp.add (a[j]);
    }
    Collections.shuffle (temp);
    for (int j = 0; j < a.length; j++) {
      a[j] = temp.get (j);
    }
  }

  /**
   * Shuffles the given array of ints according to the given random
   * generator.
   * 
   * @param a the given array.
   * @param r the given random generator.
   */
  
  public static void shuffle (int[] a, Random r) {
    ArrayList<Integer> temp = new ArrayList<Integer> ();
    for (int j = 0; j < a.length; j++) {
      temp.add (a[j]);
    }
    Collections.shuffle (temp, r);
    for (int j = 0; j < a.length; j++) {
      a[j] = temp.get (j);
    }
  }

}
