package mark.inf.core;
import java.util.*;

/**
 * A base class for objects that perform bucket elimination.  Subclasses need
 * to worry about two types of data.  We eliminate "elements" from a set of
 * "groups", resulting in a set of groups.
 *
 * @author Mark Chavira
 */

public abstract class Be<E,G> {

  /**
   * Returns the group that results from combining the given set of groups
   * and eliminating the given element, which may be null if no element is
   * being eliminated.  The set is guaranteed to have at least one element.
   *
   * @param s the given set of groups.
   * @param e the given element.
   * @return the resulting group.
   */

  protected abstract G combine (
   HashSet<G> groups, E elem, int currentBucket, int totalBuckets)
   throws Exception;
  
  /**
   * Returns the set of elements in the given *initial* group.  The set will
   * not be modified.
   *
   * @param g the given initial group.
   */
  
  protected abstract Set<E> elements (G g);
  
  /**
   * Returns the index of the bucket in which the given set of elements
   * should reside according to the given elimination order.
   *
   * @param eo the given elimination order.
   * @param elems the given set of elements.
   * @return the index.
   */

  private int bucket (mark.core.util.OrderedSet<E> eo, HashSet<E> elems) {
    int ans = eo.size ();
    for (E e : elems) {
      int b = eo.indexOf (e);
      if (b != -1 && b < ans) {ans = b;}
    }
    return ans;
  }

  /**
   * Performs bucket elimination on the given groups according to the given
   * elimination order.
   *
   * @param eo the given elimination order.
   * @param groups the given groups.
   * @param combineLast whether to combine the last bucket.
   * @return the resulting set of groups.
   */

  public HashSet<G> eliminate (
   List<E> eo, Set<G> groups, boolean combineLast) throws Exception {
    
    // Replace the elimination order object with an object that allows for
    // fast mapping from element to position in the order.  Create and fill
    // the buckets.  We maintain two buckets for each element in the
    // elimination order.  One has groups.  The other contains all elements
    // in the groups.  Strictly, elemBuckets is not needed, because we could
    // call elements () to get the elements in a group throughout the
    // process.  However, elements () may be expensive (e.g., for dtrees), so
    // we keep track ourselves.

    mark.core.util.OrderedSet<E> eo2 = new mark.core.util.OrderedSet<E> (eo);    
    HashSet<E>[] elemBuckets = new HashSet[eo2.size () + 1];
    HashSet<G>[] groupBuckets = new HashSet[eo2.size () + 1];
    for (int i = 0; i < elemBuckets.length; i++) {
      elemBuckets[i] = new HashSet<E> ();
      groupBuckets[i] = new HashSet<G> ();
    }
    for (G g : groups) {
      HashSet<E> elems = new HashSet<E> (elements (g));
      int b = bucket (eo2, elems);
      elemBuckets[b].addAll (elems);
      groupBuckets[b].add (g);      
    }
    
    // Eliminate each element in the order, combine the result if
    // appropriate, and return the result.
    
    for (int i = 0; i < elemBuckets.length - 1; i++) {
      if (!elemBuckets[i].isEmpty ()) {
        elemBuckets[i].remove (eo2.get (i));
        int b = bucket (eo2, elemBuckets[i]);
        elemBuckets[b].addAll (elemBuckets[i]);
        groupBuckets[b].add (
          combine (groupBuckets[i], eo2.get (i), i + 1, elemBuckets.length));
        elemBuckets[i] = null;
        groupBuckets[i] = null;
      }
    }
    HashSet<G> ans = groupBuckets[groupBuckets.length - 1];
    if (!ans.isEmpty () && combineLast) {
       G g = combine (ans, null, elemBuckets.length, elemBuckets.length);
       ans.clear ();
       ans.add (g);
    }
    return ans;

  }

}
