package mark.inf.core;

/**
 * A class that knows how to iterate over all instantiations of a variable
 * list K and over compatible positions of another variable list L.
 * Variables in L not in K must be set with evidence E.  See classes
 * Instantiation and SingleIterator for some relevant definitions.  For
 * given instantiation k of K, we compute the corresopnding instantiation
 * l of L as follows.  If L[i] exists in K, then we set L[i] to the value
 * of L[i] in K.  Otherwise, we set L[i] to E[i].  Example usage follows:
 * <xmp>
 * DualIterator di = new DualIterator (KC, LC, E, M);
 * for (int kp = 0; kp < di.numInstantiations (); kp++) {
 *   int lp = di.next ();
 *   // di.k is the current K instantiation
 *   // kp is the current K position
 *   // lp is the current L position
 * }
 *
 * @author Mark Chavira
 */
 
public class DualIterator {
  
  /**
   * The K cardinality array.
   */
  
  private final int[] fKCardinalities;

  /**
   * The number of K instantiations.
   */
   
  private final int fNumKInstantiations;
  
  /**
   * The increment array from K to L.
   */
     
  private final int[] fIncVal;

  /**
   * The current L position.
   */
   
  private int fLp;
  
  /**
   * The current K instantaition.  This object is provided for efficiency
   * and should be considered read-only, even though the compiler does not
   * enforce this restriction.
   */

  public final int[] k;

  /**
   * Initializes the iterator.
   *
   * @param kCardinalities the K cardinality array.
   * @param lCardinalities the L cardinality array.
   * @param lEvidence the evidence involving variables in L not in K.
   * @param lMap the dual map from K to L.
   */
  
  public DualIterator (
   int[] kCardinalities, int[] lCardinalities, int[] lEvidence, int[] lMap) {

    // Store away K cardinalities and number of K instantiations.
    
    fKCardinalities = mark.core.util.Util.clone (kCardinalities);
    fNumKInstantiations =
      Instantiation.numCompatibleInstantiations (kCardinalities, null);
    
    // Compute the increment array.
    
    fIncVal = new int[kCardinalities.length];
    Instantiation.incArray (kCardinalities, lCardinalities, lMap, fIncVal);
    
    // Set k to the first instantiation.
    
    k = new int[kCardinalities.length];
    
    // Initilize L's position to the first position compatible with its
    // evidence.
    
    fLp =
      (lEvidence != null) ?
      Instantiation.position (lCardinalities, lEvidence) : 0; 
    
    // Finally, roll k and fLp back by one.
      
    if (k.length > 0) {
      k[k.length - 1] = -1;      
      fLp -= fIncVal[fIncVal.length - 1];
    }
    
  }

  /**
   * Returns the number of instantiations.
   *
   * @return the number of instantiations.
   */
   
  public int numInstantiations () {
    return fNumKInstantiations;
  }
  
  /**
   * Advances to the next instantiation.
   */
   
  public int next () {
    for (int i = k.length - 1; i >= 0; i--) {
      if (k[i] == fKCardinalities[i] - 1) {
        k[i] = 0;
      } else {
        k[i]++;
        fLp += fIncVal[i];
        return fLp;
      }
    }
    return fLp;
  }

}
