package mark.inf.core;
import java.util.*;

/**
 * A class for dealing with instantiations of a variable list.  For given
 * variable X, let D(X) represent X's domain.  We restrict D(X) to be
 * {0,1,...,|D(X)|-1}.  A variable list is a list of distinct variables.  For
 * a given variable list L:
 * <ul>
 * <li> The *cardinality array* C is an int[] parallel to L where C[i] =
 *      |D(L[i])|.
 * <li> The *product array* P is an int[] parallel to L where P[i] is the product
 *      of the cardinalities of variables succeeding L[i].
 * <li> An *instantiation* T is an int[] parallel to L where T[i] is in
 *      D(L[i]).
 * <li> The *instantiation list* is the list of instantiations of L ordered
 *      in the usual way: comparing first on first component, then on second,
 *      etc.
 * <li> *Evidence* E is an int[] parallel to L where E[i] is in D(L[i]) or
 *      E[i] = -1.
 * <li> Instantiation T and evidence E are *compatible* iff for each i,
 *      E[i] == -1 or E[i] == T[i].
 * <li> a *dual map* M from L1 to L2 is an int[] parallel to L1 where M[i] is
 *      the index of L1[i] in L2 or -1 if L1[i] does not exist in L2.
 * </ul>
 * </ul>
 */

public class Instantiation {

  /**
   * Returns the number of instantiations of the variable list with the
   * given cardinality array that are compatible with the given evidence.
   * 
   * @param c the given cardinality array.
   * @reutrn the number of instantiations.
   */
  
  public static int numCompatibleInstantiations (int[] c, int[] e) {
    int ans = 1;
    for (int i = 0; i < c.length; i++) {
      if (e == null || e[i] == -1) {
        ans *= c[i];
      }
    }
    return ans;
  }

  /**
   * For a variable list with the given cardinality array, returns the
   * product array.
   * 
   * @param c the given cardinality array.
   * @return the product array.
   */
  
  public static int[] productArray (int[] c) {
    int[] ans = new int[c.length];
    if (ans.length > 0) {ans[ans.length - 1] = 1;}
    for (int i = ans.length - 2; i >= 0; i--) {
      ans[i] = ans[i + 1] * c[i + 1];
    }
    return ans;
  }
  
  /**
   * For a variable list with the given cardinality array, returns the product
   * array.
   * 
   * @param c the given cardinality array.
   * @return the product array.
   */
  
  public static long[] longProductArray (long[] c) {
    long[] ans = new long[c.length];
    if (ans.length > 0) {
      ans[ans.length - 1] = 1L;
    }
    for (int i = ans.length - 2; i >= 0; i--) {
      ans[i] = ans[i + 1] * c[i + 1];
    }
    return ans;
  }

  /**
   * For a variable list with the given cardinality array, computes the
   * position in the instantiation list of the first instantiation compatible
   * with the given evidence.
   *
   * @param c the given cardinality array.
   * @param e the given evidence.  Only the first c.length slots of e are
   *   used even if e is larger.
   * @return the position.
   */
  
  public static int position (int[] c, int[] e) {
    int ans = 0;
    if (e != null) {
      int product = 1;
      for (int i = c.length - 1; i >= 0; i--) {
        if (e[i] != -1) {
          ans += e[i] * product;
        }
        product *= c[i];
      }
    }
    return ans;
  }

  /**
   * For a variable list with the given cardinality array, computes the
   * the fisrt instantiation compatible with the given evidence.
   *
   * @param c the given cardinality array.
   * @param e the given evidence.  Only the first c.length slots of e are
   *   used even if e is larger.
   * @return the instantiation.
   */
  
  public static int[] instantiation (int[] c, int[] e) {
    int[] ans = new int[c.length];
    if (e != null) {
      for (int i = 0; i < ans.length; i++) {
        ans[i] = e[i] == -1 ? 0 : e[i];
      }
    }
    return ans;
  }
  
  /**
   * For a variable list with the given cardinality array, computes the
   * position in the instantiation list of the first instantiation compatible
   * with the given evidence.
   *
   * @param c the given cardinality array.
   * @param e the given evidence.  Only the first c.length slots of e are
   *   used even if e is larger.
   * @return the position.
   */
  
  public static long longPosition (long[] c, long[] e) {
    long ans = 0;
    if (e != null) {
      long product = 1;
      for (int i = c.length - 1; i >= 0; i--) {
        if (e[i] != -1) {
          ans += e[i] * product;
        }
        product *= c[i];
      }
    }
    return ans;
  }

  /**
   * For a variable list with the given cardinality array, computes the
   * instantiation located at the given position in the instantiation list.
   *
   * @param c the cardinality array.
   * @param p the position.
   * @return the instantiation.
   */
  
  public static int[] instantiation (int[] c, int p) {
    int[] ans = new int[c.length];
    for (int i = ans.length - 1; i >= 0; i--) {
      ans[i] = p % c[i];
      p /= c[i];
    }
    return ans;
  }

  /**
   * For a variable list with the given cardinality array, computes the
   * instantiation located at the given position in the instantiation list.
   *
   * @param c the cardinality array.
   * @param p the position.
   * @return the instantiation.
   */
  
  public static long[] longInstantiation (long[] c, long p) {
    long[] ans = new long[c.length];
    for (int i = ans.length - 1; i >= 0; i--) {
      ans[i] = p % c[i];
      p /= c[i];
    }
    return ans;
  }

  /**
   * Returns a list of all positions compatible with the given evidence.
   *
   * @param c the cardinality array.
   * @param e the evidence.
   * @return the list.
   */

  public static int[] allCompatiblePositions (int[] c, int[] e) {
    SingleIterator si = new SingleIterator (c, e);
    int[] ans = new int[si.numCompatibleInstantiations ()];
    for (int i = 0; i < si.numCompatibleInstantiations (); i++) {
      int p = si.next ();
      ans[i] = p;
    }
    return ans;
  }

  /**
   * Constructs increment information from variable list L1 to variable list
   * L2.  Assume we have instantiation l1 of L1 and a compatible position p2
   * for L2.  Constructs an array incArrays such that incArray[i] is the
   * amount to adjust p2 when we add one to l1[i] and change each variable
   * following from its last value to 0.  Returns the amount to adjust p2
   * when we change each variable in L1 from its last value to 0.
   * 
   * @param c1 the cardinality array of L1.
   * @param c2 the cardinality array of L2.
   * @param dm the dual map from L1 to L2.
   * @return the increment array.
   */

  public static int incArray (
   int[] c1, int[] c2, int[] dm, int[] incArray) {
    int[] prod2 = productArray (c2);
    int s = 0;
    for (int i = c1.length - 1; i >= 0; i--) {
      if (dm[i] == -1) {
        incArray[i] = -s;
      } else {
        incArray[i] = prod2[dm[i]] - s;
        s += (c2[dm[i]] - 1) * prod2[dm[i]];
      }
    }
    return -s;
  }

  /**
   * Returns an the increment reset value from variable list L1 to variable
   * list L2.  Assume we have instantiation l1 of L1 that sets each variable
   * to its last value and a corresponding position p2 for L2.  Returns the
   * Returns the amount to adjust p2 when we change each variable in L1 from
   * its last value to 0.
   * 
   * @param c1 the cardinality array of L1.
   * @param c2 the cardinality array of L2.
   * @param dm the dual map from L1 to L2.
   * @return the increment array.
   */

  public static int incResetValue (int[] c1, int[] c2, int[] dm) {
    int[] prod2 = productArray (c2);
    int s = 0;
    for (int i = c1.length - 1; i >= 0; i--) {
      if (dm[i] != -1) {
        s += (c2[dm[i]] - 1) * prod2[dm[i]];
      }
    }
    return -s;
  }

  
  /**
   * Creates the dual mapping from the first given variable list to the
   * seconds given variable list.  Variables can be of any type but must
   * implement equals ().
   *
   * @param l1 the first given variable list.
   * @param l2 the second given variable list.
   * @return the dual mapping.
   */

  public static int[] dualMap (List l1, List l2) {
    int[] ans = new int[l1.size ()];
    for (int i = ans.length - 1; i >= 0; i--) {
      ans[i] = l2.indexOf (l1.get (i));
    }
    return ans;
  }

  /**
   * Creates the dual mapping from the first given variable list to the
   * seconds given variable list.  Variables can be of any type but must
   * implement equals ().
   *
   * @param l1 the first given variable list.
   * @param l2 the second given variable list.
   * @return the dual mapping.
   */

  public static <T> int[] dualMap (T[] l1, T[] l2) {
    int[] ans = new int[l1.length];
    for (int i = ans.length - 1; i >= 0; i--) {
      ans[i] = -1;
      for (int j = 0; j < l2.length; j++) {
        if (l2[j].equals (l1[i])) {ans[i] = j; break;}
      }
    }
    return ans;
  }

}
