package mark.inf.core;

/**
 * An object that defines how to map from a logical literal to information in
 * a source domain.  We solve some problems from a "source domain" by
 * converting them to logic, solving the problem in logic, and then
 * converting the answer back to the source domain.  The logic representation
 * includes n binary variables, called 1,2,..,n.  A literal is a negative or
 * positive occurrence of one of these variables.  The conversion back
 * requires keeping track of how to interpret the literals.  This class,
 * along with LitMap, helps to do so for a number of source domains.
 * <p>
 * Each literal has an associated weight.  There are three types of literals.
 * An indicator represents some value of some binary or multi-valued variable
 * in the source domain.  A parameter represents some set of positions in
 * some table (really an array, but we call it a table).  An auxiliary can
 * represent anything.  It is possible for an indicator (parameter) to be
 * <em>orphaned</em>.  In this case, the source varialbe (table) has been
 * lost, but some information about it has been retained.  This might happen,
 * for example, if we load a literal map from a file and the source objects
 * were not available beforehand.
 * <p>
 * If a variable in the source domain has N values, then those values are
 * 0,1,...,(N-1).  Each source variable, source value, and source table has
 * a name.
 * <p>
 * This class conforms to the immutable design pattern.
 * <p>
 * <h3>
 * Source Domain: Logic Circuits
 * </h3>
 * An indicator represents some value of some gate.  An auxiliary represents
 * some value of some auxiliary variable, or the negation of some literal
 * that is used for another purpose, or some source domain entity that we
 * need not remember.
 * <h3>
 * Source Domain: Probabilistic Inference by Weighted Model Counting (PIWMC)
 * </h3>
 * An indicator represents some value of some network variable.  A parameter
 * represents a set of positions in a table whose parameters are and will
 * always be equal.  An auxiliary represents the negation of some literal
 * that is used for another purpose, some value of some auxiliary variable,
 * or some source domain entity that we need not remember.

 * @author Mark Chavira
 */

public class LitDesc<V,T> {

  /**
   * An enumeration for literal types.
   */

  public enum LitType {INDICATOR, PARAMETER, AUXILIARY};

  /**
   * The literal type.
   */

  private LitType fType;
  
  /**
   * The source variable.  This field is null iff the literal is not an
   * indicator or the literal is orphaned.
   */
   
  private V fSrcVar;
  
  /**
   * The source variable name.  This field is null iff the literal is not an
   * indicator.
   */
   
  private String fSrcVarName;
  
  /**
   * The source value.  This field is -1 iff the literal is not an indicator.
   */
  
  private int fSrcVal;
  
  /**
   * The source value name.  This field is null iff the literal is not an
   * indicator.
   */
   
  private String fSrcValName;
  
  /**
   * The source table.  This field is null iff the literal is not a
   * parameter or the literal is orphaned.
   */
   
  private T fSrcTable;
  
  /**
   * The source table name.  This field will be null iff the literal is not
   * a parameter.
   */

  private String fSrcTableName;
  
  /**
   * The set of source positions.  This field will be null iff the literal is
   * not a parameter.
   */
   
  private int[] fSrcPositions;
  
  /**
   * The associated weight.
   */
  
  private double fWeight;
  
  /**
   * The constructor, which can only be invoked internally.
   *
   * @param t the literal type.
   * @param srcVar the source variable or null if none.
   * @param srcVarName the source variable name or null if none.
   * @param srcVal the source value or -1 if none.
   * @param srcValName the source value name or null none.
   * @param srcTable the source table or null if none.
   * @param srcTableName the source table name or null if none.
   * @param srcPositions the source positions or null if none.
   * @param number the associated number.
   */

  private LitDesc (
   LitType t, V srcVar, String srcVarName, int srcVal, String srcValName,
   T srcTable, String srcTableName, int[] srcPositions, double number) {
    fType = t;
    fSrcVar = srcVar;
    fSrcVarName = srcVarName;
    fSrcVal = srcVal;
    fSrcValName = srcValName;
    fSrcTable = srcTable;
    fSrcTableName = srcTableName;
    fSrcPositions = srcPositions;
    fWeight = number;
  }

  /**
   * Returns a new indicator description for the given source variable,
   * source variable name, source value, source value name, and weight.
   *
   * @param v the given source variable.
   * @param vn the given source variable's name.
   * @param u the given source value.
   * @param un the given source value's name.
   * @param w the given weight.
   * @return the description.
   */

  public static <V,T> LitDesc indDesc (
   V v, String vn, int u, String un, double w) {
    return
      new LitDesc<V,T> (
        LitType.INDICATOR, v, vn, u, un, null, null, null, w);
  }

  /**
   * Returns a new parameter description for the given source table, source
   * table name, set of source positions, and given weight.
   *
   * @param t the given source table.
   * @param tn the given source table's name.
   * @param ps the given position set.
   * @param w the given weight.
   * @return the description.
   */

  public static <V,T> LitDesc parmDesc (
   T t, String tn, int[] ps, double w) {
    return
      new LitDesc<V,T> (
        LitType.PARAMETER, null, null, -1, null, t, tn,
        ps == null ? new int[] {} : mark.core.util.Util.clone (ps), w);
  }

  /**
   * Returns an auxiliary description for the given weight.
   *
   * @param w the given weight.
   * @return the description.
   */

  public static <V,T> LitDesc constDesc (double w) {
    return
      new LitDesc<V,T> (
        LitType.AUXILIARY, null, null, -1, null, null, null, null, w);
  }
  
  /**
   * Returns a new description that is identical to this description, except
   * that it has the given weight.
   *
   * @param w the given weight.
   * @return the new description.
   */
  
  public LitDesc changeWeight (double w) throws Exception {
    return new LitDesc<V,T> (
      fType, fSrcVar, fSrcVarName, fSrcVal, fSrcValName, fSrcTable,
      fSrcTableName, fSrcPositions, w);
  }  

  /**
   * Returns a new description of a parameter literal equivalent to this
   * description of a parameter literal but with an additional position.
   * Throws an exception if the description is not a parameter literal.
   * 
   * @param pos the given position.
   */
  
  public LitDesc<V,T> addPosition (int pos) throws Exception {
    if (fType != LitType.PARAMETER) {
      throw new Exception ("Type not parameter!");
    }
    int[] positions = new int[fSrcPositions.length + 1];
    System.arraycopy (fSrcPositions, 0, positions, 0, fSrcPositions.length);
    positions[positions.length - 1] = pos;
    return
      (LitDesc<V,T>)parmDesc (
        srcTable (), srcTableName (), positions, weight ());
  }
  
  /**
   * Retruns the literal's type.
   *
   * @return the literal's type.
   */
  
  public LitType type () {
    return fType;
  }

  /**
   * Returns the literal's source variable if it exists and is available;
   * null otherwise.
   *
   * @return the literal's source variable.
   */
  
  public V srcVar () {
    return fSrcVar;
  }

  /**
   * Returns the literals's source variable name if it exists; null
   * otherwise.
   *
   * @return the source variable name.
   */
  
  public String srcVarName () {
    return fSrcVarName;
  }

  /**
   * Returns the literal's soruce value if it exists; -1 otherwise.
   * 
   * @return the source value.
   */
  
  public int srcVal () {
    return fSrcVal;
  }
  
  /**
   * Returns the literal's value name if it exists; null otherwise.
   *
   * @return the literal's value.
   */
  
  public String srcValName () {
    return fSrcValName;
  }

  /**
   * Returns the literal's source table if it exists and is available;
   * null otherwise.
   *
   * @return the literal's source variable.
   */
  
  public T srcTable () {
    return fSrcTable;
  }

  /**
   * Returns the literals's source variable name if it exists; null
   * otherwise.
   *
   * @return the source variable name.
   */
  
  public String srcTableName () {
    return fSrcTableName;
  }

  /**
   * Returns the number of source positions or 0 if none.
   * 
   * @return the number of positions.
   */
  
  public int numPositions () {
    return fSrcPositions != null ? fSrcPositions.length : 0;
  }
  
  /**
   * Returns the nth source position.
   * 
   * @param n identifies the source position.
   * @return the nth source position.
   */

  public int nthPosition (int n) {
    return fSrcPositions[n];
  }
  
  /**
   * Returns the literal's weight.
   *
   * @return the weight.
   */
  
  public double weight () {
    return fWeight;
  }

}
