package mark.inf.core;
import java.util.*;
import java.io.*;

/**
 * An object that assists in mapping between a "source domain" and
 * propositional logic over binary variables.  See LitDesc for many relevant
 * definitions.
 * <p>
 * A literal map file is a string that describes a collection of literals and
 * consists of five kinds of lines:
 * <ol>
 * <li> A line that precedes all other "cc" lines: "cc" "N" numLogicVars
 * <li> A table count line: "cc" "T" srcTabName parameterCount
 * <li> An indicator line: "cc" "I" lit weight srcVarName srcValName srcVal
 * <li> A parameter line: "cc" "P" lit weight srcTabName pos1,pos2,..
 * <li> An auxiliary line: "cc" "A" lit weight
 * <li> Other lines do not begin with "cc" and do not contribute meaning.
 * </ol>
 * <p>
 * This class conforms to the immutable design pattern.
 *
 * @author Mark Chavira
 */

public class LitMap<V,T> {

  //=========================================================================
  // Fields
  //=========================================================================
  
  /**
   * The pattern used to separate fields within a line when reading a literal
   * map file.
   */
  
  protected static final String READ_DELIMITER = "\\$";
  
  /**
   * The string to use to delimit fields within a line when writing a literal
   * map file.
   */
  
  protected static final String DELIMITER = "$";

  /**
   * A map from logic variable to the literal description of its negative
   * literal.  Maps 0 to null.
   */

  protected ArrayList<LitDesc<V,T>> fLogicVarToNegLitDesc;
  
  /**
   * A map from logic variable to the literal description of its positive
   * literal.  Maps 0 to null.
   */

  protected ArrayList<LitDesc<V,T>> fLogicVarToPosLitDesc;

  /**
   * A map from name to source variable.
   */
  
  protected HashMap<String,V> fNameToSrcVar;
  
  /**
   * A map from name to source table.
   */
  
  protected HashMap<String,T> fNameToSrcTab;
  
  /**
   * A map from logic variable to default weight of its negative literal.
   * Map from 0  is undefined.
   */
  
  protected double[] fLogicVarToDefaultNegWeight;

  /**
   * A map from logic variable to default weight of its positive literal.
   * Map from 0 is undefined.
   */
  
  protected double[] fLogicVarToDefaultPosWeight;

  /**
   * A map from source variable to value to indicator.
   */
  
  protected HashMap<V,int[]> fSrcVarToSrcValToIndicator;
  
  /**
   * A map from source table to position to parameter.
   */
  
  protected HashMap<T,int[]> fSrcTabToSrcPositionToParameter;
  
  /**
   * The collection of indicators in no particular order.
   */
  
  protected int[] fIndicators;
  /**
   * The collection of parameters in no particular order.
   */

  protected int[] fParameters;

  /**
   * The collection of auxiliaries in no particular order.
   */ 
  
  protected int[] fAuxiliaries;

  //=========================================================================
  // Private Methods
  //=========================================================================

  /**
   * Code common to the constructors.  Each constructor fills
   * fLogicVarToNegLitDesc and fLogicVarToPosLitDesc and then invokes this
   * method.
   * 
   * @param srcVarToName a map from source variable to name.
   * @param srcTabToName a map from table to name.
   * @param srcTabToNumPos a map from table to number of parameters.
   */
  
  private void init (
   Map<String,V> nameToSrcVar, Map<String,T> nameToSrcTab,
   Map<V,Integer> srcVarToNumVals, Map<T,Integer> srcTabToNumPos)
   throws Exception {

    // Store the maps from name to source variable and from name to source
    // table.
    
    fNameToSrcVar = new HashMap<String,V> (nameToSrcVar);
    fNameToSrcTab = new HashMap<String,T> (nameToSrcTab);
    
    // Create the maps from literal to default value.
    
    fLogicVarToDefaultNegWeight = new double[numLogicVars () + 1];
    fLogicVarToDefaultPosWeight = new double[numLogicVars () + 1];
    for (int i = 1; i <= numLogicVars (); i++) {
      fLogicVarToDefaultNegWeight[i] =
        fLogicVarToNegLitDesc.get (i).weight ();
      fLogicVarToDefaultPosWeight[i] =
        fLogicVarToPosLitDesc.get (i).weight ();
    }

    // Obtain the number of indicators, parameters, and auxiliaries.
    
    int numInds = 0, numParms = 0, numAux = 0;
    for (int l = -numLogicVars (); l <= numLogicVars (); l++) {
      if (l == 0) {continue;}
      LitDesc<V,T> ld = desc (l);
      if (ld.type () == LitDesc.LitType.INDICATOR) {
        ++numInds;
      } else if (ld.type () == LitDesc.LitType.PARAMETER) {
        ++numParms;
      } else if (ld.type () == LitDesc.LitType.AUXILIARY) {
        ++numAux;
      } else {
        throw new Exception ("Should never get here!");
      }
    }

    // Next, allocate the rest of the maps.
    
    fSrcVarToSrcValToIndicator = new HashMap<V,int[]> ();
    for (V v : srcVarToNumVals.keySet ()) {
      fSrcVarToSrcValToIndicator.put (v, new int[srcVarToNumVals.get (v)]);
    }
    fSrcTabToSrcPositionToParameter = new HashMap<T,int[]> ();
    for (T t : srcTabToNumPos.keySet ()) {
      fSrcTabToSrcPositionToParameter.put (
        t, new int[srcTabToNumPos.get (t)]);
    }
    fIndicators = new int[numInds];
    fParameters = new int[numParms];
    fAuxiliaries = new int[numAux];
    
    // Next, fill the rest of the maps.
    
    int indIndex = 0, parmIndex = 0, auxIndex = 0;
    for (int l = -numLogicVars (); l <= numLogicVars (); l++) {
      if (l == 0) {continue;}
      LitDesc<V,T> ld = desc (l);
      if (ld.type () == LitDesc.LitType.INDICATOR) {
        fSrcVarToSrcValToIndicator.get (ld.srcVar ())[ld.srcVal ()] = l;
        fIndicators[indIndex++] = l;
      } else if (ld.type () == LitDesc.LitType.PARAMETER) {
        int[] parms = fSrcTabToSrcPositionToParameter.get (ld.srcTable ());
        for (int pi = 0; pi < ld.numPositions (); pi++) {
          parms[ld.nthPosition (pi)] = l;
        }
        fParameters[parmIndex++] = l;
      } else if (ld.type () == LitDesc.LitType.AUXILIARY) {
        fAuxiliaries[auxIndex++] = l;
      } else {
        throw new Exception ("Should not get here!");
      }
    }
    
  }

  /**
   * Assists toString () by writing the information for the given literal
   * having the given description to the the given string builder.
   * 
   * @param l the given literal.
   * @param d the given literal description.
   * @param b the given string builder.
   */
  
  protected void literalToString (int l, LitDesc<V,T> d, StringBuilder b)
   throws Exception {
    b.append ("cc");
    b.append (DELIMITER);
    switch (d.type ()) {
      case INDICATOR:
        b.append ("I");
        b.append (DELIMITER);
        b.append (l);
        b.append (DELIMITER);
        b.append (d.weight ());
        b.append (DELIMITER);
        b.append (d.srcVarName ());
        b.append (DELIMITER);
        b.append (d.srcValName ());
        b.append (DELIMITER);
        b.append (d.srcVal ());
        break;
      case PARAMETER:
        b.append ("P");
        b.append (DELIMITER);
        b.append (l);
        b.append (DELIMITER);
        b.append (d.weight ());
        b.append (DELIMITER);
        b.append (d.srcTableName ());
        b.append (DELIMITER);
        b.append (d.nthPosition (0));
        for (int i = 1; i < d.numPositions (); i++) {
          b.append (",");
          b.append (d.nthPosition (i));
        }
        break;
      case AUXILIARY:
        b.append ("A");
        b.append (DELIMITER);
        b.append (l);
        b.append (DELIMITER);
        b.append (d.weight ());
        break;
      default:
        throw new Error ("Unknown literal type!");
    }
    b.append (mark.core.io.Util.NEWLINE);
  }

  /**
   * Code common to constructors that initialize a literal map from a reader.
   * 
   * @param nameToSrcVar a map from srouce variable name to source variable.
   * @param nameToSrcTab a map from srouce table name to source table.
   * @param r the given reader.
   */

  private void init (
   Map<String,V> nameToSrcVar, Map<String,T> nameToSrcTab, Reader r)
   throws Exception {
   
    // Prepare to parse.
   
    BufferedReader br =
      r instanceof BufferedReader ?
      (BufferedReader)r : new BufferedReader (r);
    java.util.regex.Pattern p =
      java.util.regex.Pattern.compile (READ_DELIMITER);
    int numLits = Integer.MAX_VALUE;
    int litsFinished = 0;
    HashMap<T,Integer> srcTabToNumPos = new HashMap<T,Integer> ();
    HashMap<V,Integer> srcVarToNumVals = new HashMap<V,Integer> ();
   
    // Process each line.
   
    while (litsFinished < numLits) {
     
      // Read the line.  Quit if eof.  Skip if comment (including blank
      // lines).  Otherwise, split into tokens.
     
      String line = br.readLine ();
      if (line == null) {break;} // eof
      if (!line.startsWith ("cc" + DELIMITER)) {continue;} // comment
      line = line.trim ();
      String[] tokens = p.split (line);
     
      // If the line is a header, it is of the form: "cc" "N" numVars.

      String type = tokens[1];
      if (type.equals ("N")) {
        int n = Integer.parseInt (tokens[2]);
        numLits = n * 2;
        fLogicVarToNegLitDesc = new ArrayList<LitDesc<V,T>> (n + 1);
        fLogicVarToPosLitDesc = new ArrayList<LitDesc<V,T>> (n + 1);
        for (int i = 0; i < n + 1; i++) {
          fLogicVarToNegLitDesc.add (null);
          fLogicVarToPosLitDesc.add (null);
        }
        continue;
      }
     
      // If the line is a table, then it has the following form:
      //   "cc" "T" srcTableName parameterCnt.
     
      if (type.equals ("T")) {
        String tn = tokens[2];
        int parmCount = Integer.parseInt (tokens[3]);
        T t = nameToSrcTab.get (tn);
        srcTabToNumPos.put (t, parmCount);
        continue;
      }
      
      // If the line is a table, then it has the following form:
      //   "cc" "V" srcVarName numVals (valName)+
      
      if (type.equals ("V")) {
        String vn = tokens[2];
        int valCount = Integer.parseInt (tokens[3]);
        V v = nameToSrcVar.get (vn);
        srcVarToNumVals.put (v, valCount);
        continue;
      }

      // This is not a header line or a table count line, so it must be a
      // literal description, which looks like one of the following:
      //   "cc" "I" literal weight srcVarName srcValName srcVal
      //   "cc" "P" literal weight srcTableName pos+
      //   "cc" "A" literal weight
     
      int l = Integer.parseInt (tokens[2]);
      double w = Double.parseDouble (tokens[3]);
      if (type.equals ("I")) {
        String vn = tokens[4];
        V v = nameToSrcVar.get (vn);
        String un = tokens[5];
        int u = Integer.parseInt (tokens[6]);
        (l < 0 ? fLogicVarToNegLitDesc : fLogicVarToPosLitDesc).set (
          Math.abs (l), LitDesc.indDesc (v, vn, u, un, w));
      } else if (type.equals ("P")) {
        String tn = tokens[4];
        T t = nameToSrcTab.get (tn);
        String[] posStrings = tokens[5].split (",");
        int[] positions = new int[posStrings.length];
        for (int i = 0; i < positions.length; i++) {
          positions[i] = Integer.parseInt (posStrings[i]);
        }
        (l < 0 ? fLogicVarToNegLitDesc : fLogicVarToPosLitDesc).set (
          Math.abs (l), LitDesc.parmDesc (t, tn, positions, w));
      } else if (type.equals ("A")) {
        (l < 0 ? fLogicVarToNegLitDesc : fLogicVarToPosLitDesc).set (
          Math.abs (l), LitDesc.constDesc (w));
      } else {
        throw new Exception (
          "\"cc\" must be followed by " +
          "\"N\", \"V\", \"T\", \"I\", \"P\", or \"A\"");
      }
      ++litsFinished;
    }
   
    // Invoke the helper to do the rest of the work.
   
    init (nameToSrcVar, nameToSrcTab, srcVarToNumVals, srcTabToNumPos);

  }
  
  //=========================================================================
  // Construction
  //=========================================================================

  /**
   * Constructs a literal map from the given text file.  Does not close the
   * reader.
   *
   * @param nameToSrcVar a map from srouce variable name to source variable.
   * @param nameToSrcTab a map from srouce table name to source table.
   * @param r the given reader.
   */

  public LitMap (
   Map<String,V> nameToSrcVar, Map<String,T> nameToSrcTab, String filename)
   throws Exception {
    BufferedReader r = new BufferedReader (new FileReader (filename));
    init (nameToSrcVar, nameToSrcTab, r);
    r.close ();
  }
  
  
  /**
   * Constructs a literal map from the given text file.  Does not close the
   * reader.
   *
   * @param nameToSrcVar a map from srouce variable name to source variable.
   * @param nameToSrcTab a map from srouce table name to source table.
   * @param r the given reader.
   */

  public LitMap (
   Map<String,V> nameToSrcVar, Map<String,T> nameToSrcTab, Reader r)
   throws Exception {
    init (nameToSrcVar, nameToSrcTab, r);
  }

  /**
   * Constructs a literal map from the given map from logic variable to
   * negative literal description, the given map from logic variable to
   * positive literal description, the given map from name to source
   * variable, the given map from name to source table, and the given map
   * from source table to number of positions.
   *
   * @param logicVarToNegLitDesc the given map from logic variable to
   *   negative literal description.
   * @param logicVarToPosLitDesc the given map from logic variable to
   *   positive literal description.
   * @param nameToSrcVar the given map from name to source variable.
   * @param srcTabToName the given map from name to source table.
   * @param nameToSrcTab the given map from srouce table to number of
   *   positions.
   */
  
  public LitMap (
   List<LitDesc<V,T>> logicVarToNegLitDesc,
   List<LitDesc<V,T>> logicVarToPosLitDesc,
   Map<String,V> nameToSrcVar, Map<String,T> nameToSrcTab,
   Map<V,Integer> srcVarToNumVals, Map<T,Integer> srcTabToNumPos)
   throws Exception {
    fLogicVarToNegLitDesc =
      new ArrayList<LitDesc<V,T>> (logicVarToNegLitDesc);
    fLogicVarToPosLitDesc =
      new ArrayList<LitDesc<V,T>> (logicVarToPosLitDesc);
    init (nameToSrcVar, nameToSrcTab, srcVarToNumVals, srcTabToNumPos);
  }

  /**
   * Constructs a new literal by replacing each constant with the log of
   * the constant.
   * 
   * @return the new literal map.
   */
  
  public LitMap<V,T> toLog () throws Exception {
    ArrayList<LitDesc<V,T>> logicVarToNegLitDesc =
      new ArrayList<LitDesc<V,T>> ();
    ArrayList<LitDesc<V,T>> logicVarToPosLitDesc =
      new ArrayList<LitDesc<V,T>> ();
    for (LitDesc<V,T> d : fLogicVarToNegLitDesc) {
      logicVarToNegLitDesc.add (
        d == null ? null : d.changeWeight (Math.log (d.weight ())));
    }
    for (LitDesc<V,T> d : fLogicVarToPosLitDesc) {
      logicVarToPosLitDesc.add (
        d == null ? null : d.changeWeight (Math.log (d.weight ())));
    }
    HashMap<V,Integer> varToNumVals = new HashMap<V,Integer> ();
    for (V v : fSrcVarToSrcValToIndicator.keySet ()) {
      varToNumVals.put (v, fSrcVarToSrcValToIndicator.get (v).length);
    }
    HashMap<T,Integer> tableToNumPos = new HashMap<T,Integer> ();
    for (T t : fSrcTabToSrcPositionToParameter.keySet ()) {
      tableToNumPos.put (t, fSrcTabToSrcPositionToParameter.get (t).length);
    }
    return
      new LitMap<V,T> (
        logicVarToNegLitDesc, logicVarToPosLitDesc, fNameToSrcVar,
        fNameToSrcTab, varToNumVals, tableToNumPos); 
  }
  
  //=========================================================================
  // Retrieve information.
  //=========================================================================

  /**
   * Returns the number of source variables in the literal map.
   * 
   * @return the number of source variables.
   */
  
  public int numSrcVars () {
    return fSrcVarToSrcValToIndicator.size ();
  }
  
  /**
   * Returns whether the given src variable is associated with literals
   * in the literal map.
   * 
   * @param v the given src variable.
   * @return whether the literal map knows about the src variable.
   */
  
  public boolean srcVariableDefined (V v) {
    return fSrcVarToSrcValToIndicator.containsKey (v);
  }
  
  /**
   * Returns whether the given src table is associated with literals
   * in the literal map.
   * 
   * @param v the given src variable.
   * @return whether the literal map knows about the src variable.
   */
  
  public boolean srcTableDefined (T t) {
    return fSrcTabToSrcPositionToParameter.containsKey (t);
  }

  /**
   * Returns the indicators belonging to the given src variable or null if
   * there is no such mapping.
   *
   * @param v the given source variable.
   * @return the indicators.
   */
  
  public int[] inds (V v) {
    return mark.core.util.Util.clone (fSrcVarToSrcValToIndicator.get (v));
  }

  /**
   * Returns the logic literal that corresponds to the given src variable and
   * and value.
   *
   * @param v the given src variable.
   * @param u the given src value.
   * @return the literal.
   */
  
  public int srcVarValToLiteral (V v, int u) {
    return fSrcVarToSrcValToIndicator.get (v)[u];
  }  
  
  /**
   * Returns the logic literal that corresponds to the given src potential
   * and position.
   * 
   * @param p the given src potentail.
   * @param pos the given src position.
   * @return the literal.
   */
  
  public int srcPotPositionToLiteral (T p, int pos) {
    return fSrcTabToSrcPositionToParameter.get (p)[pos];  
  }
  
  /**
   * Returns the number of logic variables in the literal map.
   *
   * @return the nubmer of variables.
   */
  
  public int numLogicVars () {
    return fLogicVarToNegLitDesc.size () - 1;
  }

  /**
   * Returns the given literal's description.
   *
   * @param l the given literal.
   * @return the type.
   */

  public LitDesc<V,T> desc (int l) {
    return
      l < 0 ? fLogicVarToNegLitDesc.get (-l) : fLogicVarToPosLitDesc.get (l);
  }

  /**
   * Returns a sorted list all variables whose positive literal is a
   * parameter.
   *
   * @return the list.
   */
  
  public ArrayList<Integer> parmLogicVars () {
    ArrayList<Integer> ans = new ArrayList<Integer> ();
    for (int l : fParameters) {if (l > 0) {ans.add (l);}}
    Collections.sort (ans);
    return ans;
  }

  /**
   * Returns a sorted list of the indicator variables whose negative and
   * positive literals correspond to the two values of a binary source
   * variable.
   *
   * @return the list.
   */
  
  public ArrayList<Integer> binaryIndicatorVars () {
    ArrayList<Integer> ans = new ArrayList<Integer> ();
    for (int[] lits : fSrcVarToSrcValToIndicator.values ()) {
      if (lits.length == 2 && lits[0] == -lits[1]) {
        ans.add (Math.abs (lits[0]));
      }
    }
    Collections.sort (ans);
    return ans;
  }

  /**
   * Returns a list of the positive indictors corresponding to the given
   * source variables.
   *
   * @param srcVars the given source variables.
   * @return the list.
   */

  public ArrayList<Integer> correspondingLogicVars (Collection<V> srcVars) {
    ArrayList<Integer> ans = new ArrayList<Integer> ();
    for (V v : srcVars) {
      for (int l : fSrcVarToSrcValToIndicator.get (v)) {
        if (l > 0) {ans.add (l);}
      }
    }
    return ans;
  }
  
  /**
   * Returns a map from source variable to source value corresponding to
   * 
   * @param logicEvidence the given logic evidence.
   * @return the described map.
   */
  
  public HashMap<V,Integer> correspondingSrcEvidence (int[] logicEvidence) {
    HashMap<V,Integer> ans = new HashMap<V,Integer> ();
    for (int l : logicEvidence) {
      LitDesc<V,T> desc = desc (l);
      if (desc.type () == LitDesc.LitType.INDICATOR) {
        ans.put (desc.srcVar (), desc.srcVal ());
      }
    }
    return ans;
  }
      
  
  /**
   * Returns a set of literals corresponding to the given source evidence.
   * 
   * @param srcEvidence the given src evidence.
   * @return the logic evidence.
   */
  
  public HashSet<Integer> correspondingLogicEvidence (
   Map<V,Integer> srcEvidence) {
    HashSet<Integer> ans = new HashSet<Integer> ();
    if (srcEvidence != null) {      
      for (V v : srcEvidence.keySet ()) {
        int trueLit = fSrcVarToSrcValToIndicator.get (v)[srcEvidence.get (v)];
        ArrayList<Integer> logicVars =
          correspondingLogicVars (Collections.singleton (v));
        for (int logicVar : logicVars) {
          ans.add (trueLit == logicVar ? logicVar : -logicVar);
        }
      }
    }
    return ans;
  }

  /**
   * Returns source evidence given a set of logic evidence.
   * 
   * @param logicEvidence the logic evidence.
   * @return the source evidence.
   */
  
  public HashMap<V,Integer> correspondingSrcEvidence (
   HashSet<Integer> logicEvidence) {
    HashMap<V,Integer> ans = new HashMap<V,Integer> ();
    for (int lit : logicEvidence) {
      LitDesc<V,T> d = desc (lit);
      if (d.type () == LitDesc.LitType.INDICATOR) {
        ans.put (d.srcVar (), d.srcVal ());
      }
    }
    return ans;
  }
  
  /**
   * Writes the given literals to the given writer as expected by the c2d
   * compiler: number of literals followed by the literals, all delimited by
   * a single space character.
   *
   * @param lits the given literals.
   * @param pw the given writer.
   */
  
  public static void writeLitFile (List<Integer> lits, PrintWriter pw) {
    pw.print (lits.size ());
    for (int l : lits) {pw.print (" " + l);}
    pw.println ();
  }
  
  /**
   * @see java.lang.Object
   */
  
  public String toString () {
    StringBuilder ans = new StringBuilder ();
    String NL = mark.core.io.Util.NEWLINE;
    ans.append ("c Following is the literal map:" + NL);
    ans.append ("c" + NL);
    ans.append ("cc" + DELIMITER + "N" + DELIMITER + numLogicVars () + NL);
    ArrayList<String> varNamesInOrder =
      new ArrayList<String> (fNameToSrcVar.keySet ());
    Collections.sort (varNamesInOrder);
    for (String vn : varNamesInOrder) {
      V v = fNameToSrcVar.get (vn);
      int[] inds = fSrcVarToSrcValToIndicator.get (v);
      ans.append (
        "cc" + DELIMITER + "V" + DELIMITER + vn + DELIMITER + inds.length);
      for (int l : inds) {ans.append (DELIMITER + desc (l).srcValName ());}
      ans.append (NL);
    }
    ArrayList<String> tabNamesInOrder =
      new ArrayList<String> (fNameToSrcTab.keySet ());
    Collections.sort (tabNamesInOrder);
    for (String tn : tabNamesInOrder) {
      T t = fNameToSrcTab.get (tn);
      ans.append (
        "cc" + DELIMITER + "T" + DELIMITER + tn + DELIMITER +
        fSrcTabToSrcPositionToParameter.get (t).length + NL);
    }
    for (int i = 1; i <= numLogicVars (); i++) {
      try {
        literalToString (-i, fLogicVarToNegLitDesc.get (i), ans);
        literalToString (i, fLogicVarToPosLitDesc.get (i), ans);
      } catch (Exception e) {
        throw new Error (e);
      }
    }
    return ans.toString ();
  }
  
  /**
   * Writes a block order to the given writer according to the elimination
   * order defined by the given list of variable names.
   *
   * @param pw the given writer.
   * @param eo the given list of variable names.
   */
   
  /*
  public void writeBlockOrder (PrintWriter pw, List<V> eo) {
    
    // Create a map from network variable to indicator block.  If the
    // variable generates parameters, then there will be another block
    // immediately preceding it for the parameters.
    
    HashMap<V,Integer> srcVarToBlock = new HashMap<V,Integer> ();
    int currentBlock = 0;
    for (V srcVar : eo) {
      if (fVarToParameters.containsKey (srcVar)) {
        currentBlock++;
      }
      srcVarToBlock.put (srcVar, currentBlock++);
    }
    
    // Now write the file.
    
    for (int i = 1; i <= numLogicVars (); i++) {
      pw.print (i == 1 ? "" : " ");
      pw.print (
        srcVarToBlock.get (fPos.get (i).variable ()) -
        (fPos.get (i).type () == LitDesc.LitType.PARAMETER ? 1 : 0));
    }
    pw.println ();

  }
  */

}
