package mark.inf.core;

/**
 * A class that knows how to iterate over all instantiations of a variable
 * list K and over compatible positions of other variable lists Li. 
 * Variables in Li that are not in K must be set with evidence Ei.  See
 * classes Instantiation and SingleIterator for some relevant definitions.
 * For given instantiation k of K, we compute the corresponding instantiation
 * li of Li as follows.  If Li[i] exists in K, then we set Li[i] to the value
 * of Li[i] in K.  Otherwise, we set Li[i] to Ei[i].  Example usage follows:
 * <xmp>
 * MultiIterator mi =
 *   new MultiIterator (KC, {C0,C1,..,Cn}, {E0,E1,..,En}, {M0,M1,...,Mn});
 * for (int kp = 0; kp < mi.numInstantiations (); kp++) {
 *   mi.next ();
 *   // mi.k is the current K instantiation
 *   // kp is the current K position
 *   // mi.lp[i] is the current Li position
 * }
 *
 * @author Mark Chavira
 */
 
public class MultiIterator {
  
  /**
   * The K cardinality array.
   */
  
  private final int[] fKCardinalities;

  /**
   * The number of K instantiations.
   */
   
  private final int fNumKInstantiations;
  
  /**
   * For each Li, an increment array from K to Li.
   */
     
  private final int[][] fIncVal;

  /**
   * The current K instantaition.  This object is provided for efficiency
   * and should be considered read-only, even though the compiler does not
   * enforce this restriction.
   */

  public final int[] k;

  /**
   * For each Li, its current position.  This object is provided for
   * efficiency and should be considered read-only, even though the compiler
   * does not enforce this restriction.
   */
   
  public final int[] lp;
  
  /**
   * Initializes the iterator.
   *
   * @param kCardinalities the K cardinality array.
   * @param lCardinalities for each L, the cardinality array.
   * @param lEvidence for each L, evidence for variables not in K.
   * @param lMaps for each L, the dual map from K to L.
   */
  
  public MultiIterator (
   int[] kCardinalities, int[][] lCardinalities,
   int[][] lEvidence, int[][] lMaps) {

    // Store away K cardinalities and number of K instantiations.
    
    fKCardinalities = mark.core.util.Util.clone (kCardinalities);
    fNumKInstantiations =
      Instantiation.numCompatibleInstantiations (kCardinalities, null);

    // For each Li, compute an increment array.

    int numL = lCardinalities.length;
    fIncVal = new int[numL][];
    for (int i = 0; i < numL; i++) {
      fIncVal[i] = new int[kCardinalities.length];
      Instantiation.incArray (
        kCardinalities, lCardinalities[i], lMaps[i], fIncVal[i]);
    }

    // Set k to the first instantiation.
    
    k = new int[kCardinalities.length];

    // For each Li, we initialize its position to the first position
    // compatible with its evidence.
    
    lp = new int[numL];
    if (lEvidence != null) {
      for (int i = 0; i < numL; i++) {
        if (lEvidence[i] != null) {
          lp[i] = Instantiation.position (lCardinalities[i], lEvidence[i]);
        }
      }
    }    
    
    // Finally, roll k and each lp[i] back by one.
 
    if (k.length > 0) {
      k[k.length - 1] = -1;      
      for (int i = 0; i < numL; i++) {
        lp[i] -= fIncVal[i][fIncVal[i].length - 1];
      }
    }
    
  }

  /**
   * Returns the number of instantiations.
   *
   * @return the number of instantiations.
   */
   
  public int numInstantiations () {
    return fNumKInstantiations;
  }
  
  /**
   * Advances to the next instantiation.
   */
   
  public void next () {
    for (int i = k.length - 1; i >= 0; i--) {
      if (k[i] == fKCardinalities[i] - 1) {
        k[i] = 0;
      } else {
        k[i]++;
        for (int j = 0; j < lp.length; j++) {
          lp[j] += fIncVal[j][i];
        }
        return;
      }
    }
  }

}
