package mark.inf.core;
import java.util.*;

/**
 * A class that knows how to iterate over instantiations of a variable list K
 * that are compatible with given evidence E.  See class Instantiation for
 * some relevant definitions.  Example usage follows:
 * <xmp>
 * SingleIterator si = new SingleIterator (C, E);
 * for (int i = 0; i < si.numCompatibleInstantiations (); i++) {
 *   int kp = si.next ();
 *   // si.k is the current instantiation
 *   // kp is the current position
 * }
 * </xmp>
 *
 * @author Mark Chavira
 */

public class SingleIterator {

  /**
   * The K cardinality array.
   */

  protected final int[] fKCardinalities;
  
  /**
   * The number of compatible K instantiations.
   */
  
  protected final int fNumCompatibleKInstantiations;
  
  /**
   * Parallel to L, an array where fInc[i] is whether E does *not* involve
   * L[i].  If fInc[i] then we say that L[i] is "incrementing".
   */
   
  protected final boolean[] fInc;
  
  /**
   * Parallel to K, for each i, the amount to adjust the position when we
   * add 1 to K[i] and change each subsequent variable from its maximum
   * value to 0.
   */
   
  protected final int[] fIncVal;
  
  /**
   * The current K position.
   */
   
  protected int fKp;
  
  /**
   * The largest i such that fInc[i] or -1 if none.
   */
   
  protected final int fFirst;

  /**
   * The current instantiation.  This object is provided for efficiency and
   * should be considered read-only, even though java does not enforce this
   * restriction.
   */
  
  public final int[] k;
  
  /**
   * Constructs an int[] of the given length filled with the given value.
   *
   * @param l the given length.
   * @param v the give value.
   */
  
  protected static int[] filledIntArray (int l, int v) {
    int[] ans = new int[l];
    Arrays.fill (ans, v);
    return ans;
  }
  
  /**
   * Initializes the iterator when L.length == n and where |D(L[i])| = c for
   * all i.
   *
   * @param n the number of varibles.
   * @param c the common cardinality.
   * @param e the evidence.
   */
  
  public SingleIterator (int n, int c, int[] e) {
    this (filledIntArray (n, c), e);
  }
    
  /**
   * Initializes the iterator.
   *
   * @param c the cardinality array.
   * @param e the evidence.
   */
  
  public SingleIterator (int[] kCardinalities, int[] kEvidence) {
  
    // Store away K cardinalities and number of compatible K instantiations.
    
    fKCardinalities = mark.core.util.Util.clone (kCardinalities);
    fNumCompatibleKInstantiations =
      Instantiation.numCompatibleInstantiations (kCardinalities, kEvidence);
    
    // Compute the increment arrays.
    
    fInc = new boolean[kCardinalities.length];
    fIncVal = new int[kCardinalities.length];
    int s = 0;
    int[] prod = Instantiation.productArray (kCardinalities);
    for (int i = kCardinalities.length - 1; i >= 0; i--) {
      fInc[i] = kEvidence == null || kEvidence[i] == -1;
      if (fInc[i]) {
        fIncVal[i] = prod[i] - s;
        s += (kCardinalities[i] - 1) * prod[i];
      }
    }
    
    // Set k to the first compatible instantiation.
    
    k = Instantiation.instantiation (kCardinalities, kEvidence);
    
    // Initilaize the position to the first compatible position.
    
    fKp = Instantiation.position (kCardinalities, kEvidence);
    
    // Roll k and fKp back by one and set fFirst.

    int first = -1;
    for (int i = fKCardinalities.length - 1; i >= 0; i--) {
      if (fInc[i]) {
        k[i]--;
        fKp -= prod[i];
        first = i;
        break;
      }
    }
    fFirst = first;
    
  }

  /**
   * Returns the number of compatible instantiations.
   *
   * @return the number of compatible instantiations.
   */
   
  public int numCompatibleInstantiations () {
    return fNumCompatibleKInstantiations;
  }
  
  /**
   * Advances to the next instantiation.
   */
   
  public int next () {
    for (int i = fFirst; i >= 0; i--) {
      if (fInc[i]) {
        if (k[i] == fKCardinalities[i] - 1) {
          k[i] = 0;
        } else {
          k[i]++;
          fKp += fIncVal[i];
          return fKp;
        }
      }
    }
    return fKp;
  }
    
}
