package mark.inf.eo;
import java.util.*;

/**
 * A base class for algorithms that produce an elimination order over a
 * subset of a set of nodes.  The nodes are consecutive integers starting at
 * zero.  Each node has an associated cardinality which is an interger >= 2.
 * <p>
 * To use an instance of a subclass, create the instance and call order ().
 * Multiple calls to order () are permissible, but the class is not thread
 * safe, so they should be sequential instead of simultaneous.
 *
 * @author Mark Chavira
 */

public abstract class Eoe {
  
  /**
   * A map from each node to its cardinality.
   */
  
  protected int[] fNodeToCardinality;
  
  /**
   * A map from each node to the log of its cardinality.
   */
  
  protected double[] fNodeToLogCardinality;

  /**
   * Initializes the subclass.  Subclass may override to perform
   * initializations.
   * 
   * @param N the number of nodes.
   */
  
  protected void init (int N) {}
  
  /**
   * Computes the cost of the given node using brute-force.  For our
   * purposes, the cost is simply an object that can be compared to another
   * cost, thereby imposing an ordering on the nodes.
   *
   * @param nodeToAdj a map from node to adjacency list.
   * @param nodeToAdjSize a map from node to adjacency list size.
   * @param n the given node.
   * @param marked an empty one hop neighbor map.  The subclass may use as
   *   scratch but must restore upon exit.
   * @return the cost.
   */

  protected abstract Comparable bruteCost (
   int[][] nodeToAdj, int[] nodeToAdjSize, int n, boolean[] marked);
  
  /**
   * Updates the costs of nodes in the priority queue in preparation for a
   * subsequent removal of the given node.  Nodes are stored in a priority
   * queue according to cost.
   *
   * @param nodeToAdj a map from node to adjacency list.
   * @param nodeToAdjSize a map from node to adjacency list size.
   * @param removed the node that will be removed.
   * @param marked a zero/one hop neighbor map of removed.
   * @param marked2 the two hop neighbor map of removed.
   * @param pq the priority queue.
   */
  
  protected abstract void updateCosts (
   int[][] nodeToAdj, int[] nodeToAdjSize, int removed,
   boolean[] marked, boolean[][] marked2, mark.core.util.IntKeyPQ pq)
   throws Exception;

  /**
   * Throws an exception if some element has a score that is not the expected
   * score.  This method has high overhead and should only be called while
   * debugging.
   *
   * @param nodeToAdj the map from node to adjacency list.
   * @param nodeToAdjSize the map from node to adjacency list size.
   * @param marked on entry and exit, an empty one hop neighbor map; may be
   *   used as scratch but must be restored upon exit.
   * @param pq the priority queue.
   */
  
  protected void validateScores (
   int[][] nodeToAdj, int[] nodeToAdjSize, boolean[] marked,
   mark.core.util.IntKeyPQ pq) throws Exception {
    for (int nodeIndex = 0; nodeIndex < pq.size (); nodeIndex++) {
      int n = pq.nthElement (nodeIndex);
      Comparable expectedCost =
        bruteCost (nodeToAdj, nodeToAdjSize, n, marked);
      Comparable actualCost = pq.nthKey (nodeIndex);
      if (expectedCost.compareTo (actualCost) != 0) {
        throw new Exception (
          "Score for " + n + " is " + actualCost + " but expected " +
          expectedCost);
      }
    }
  }

  /**
   * Computes a "restricted" elimination order for a subset of nodes.
   * <p>
   * The parameter nodes specifies a partition of the subset and defines the
   * restriction as follows.  Variables in nodes[0] appear first in the
   * order, in nodes[1] second, etc.  Nodes within nodes[i] are ordered
   * according to the subclass.  For a normal elimination order, nodes will
   * have only one dimension.  For MAP, it will have two dimensions.  For
   * something like an influence diagram, there may be many dimensions.
   * <p>
   * The method allows the client to specify a random generator.  This
   * generator is used to randomly break ties within the subclass.  The
   * random generator may be specified as null, in which case the first node
   * available in the priority queue is always taken.
   * <p>
   * The method allows the client to specify a prefix which will be prepended
   * to the order.  Nodes in nodes should not also appear in prefix.
   * <p>
   * Finally, the method can compute multiple orders and take the best, where
   * best is defined by the client by passing in an implementation of an
   * interface.
   * 
   * @param r a random generator used to break ties; may be null to always
   *   take the first node avaiable.
   * @param nodeToCardinality maps each node to its cardinality.
   * @param g maps each node to its adjacency list.
   * @param nodes an ordered partition of the nodes to order.
   * @param iterations the number of orders to compute ultimately returning
   *   the best.
   * @param ss the scoring strategy; if this field is null, then
   *   SsMaxClusterSize will be used by default.
   * @param scoreToBeat if cannot find an order with smaller score, then
   *   returns a null order.  A smaller score is considered better.
   * @return (order,logMaxCluster,score), where order/logMaxCluster/score is
   *   null/infinity/null if could not beat given score.
   */
   
  public Result<int[]> order (
   Random r, int[] nodeToCardinality, int[][] g, int[][] nodes,
   int iterations, int[] prefix, ScoringStrategy ss, Comparable scoreToBeat,
   boolean combineAtEnd)
   throws Exception {

    // If no strategy specified, use defautl
    
    if (ss == null) {ss = SsMaxClusterSize.INSTANCE;}

    // Define the number of nodes, the size of the resulting order, a map
    // from node to cardinality, a map from node to log cardinality, and the
    // log_2 cluster size of the cluster formed by variables not in the
    // prefix and not in the returned order.  These objects will not change
    // after this block.

    int N = g.length;
    int ansOrderSize = (prefix == null) ? 0 : prefix.length;
    for (int[] pe : nodes) {ansOrderSize += pe.length;}
    fNodeToCardinality = nodeToCardinality;
    fNodeToLogCardinality = Util.logCardinalities (nodeToCardinality);
    double log2MaxClusterUnelminatedVars = 0.0;
    if (combineAtEnd) {
      boolean[] eliminating = new boolean[N];
      if (prefix != null) {for (int v : prefix) {eliminating[v] = true;}}
      for (int[] pe : nodes) {for (int v : pe) {eliminating[v] = true;}}
      log2MaxClusterUnelminatedVars = 0.0;
      for (int v = 0; v < eliminating.length; v++) {
        if (!eliminating[v]) {
          log2MaxClusterUnelminatedVars += fNodeToLogCardinality[v];
        }
      }
    }
    
    // Initialize the subclass.

    init (N);
    
    // Scratch to store zero/one hop neighbor maps.  Scratch to store two hop
    // neighbor maps.  Create the priority queue.
    
    boolean[] marked = new boolean[N];
    boolean[][] marked2 = new boolean[Util.MAX_VARS_IN_CLUSTER - 1][N];
    mark.core.util.IntKeyPQ pq = new mark.core.util.IntKeyPQ (N);

    // Crate the initial representation of the graph and the log max cluster
    // incurred by the prefix.  Compute the score so far and see whether we
    // can already reject.  Compute the initial score.
    
    int[][] iNodeToAdj = new int[N][];
    int[] iNodeToAdjSize = new int[N];
    Util.initGraph (g, iNodeToAdj, iNodeToAdjSize);
    double iLog2MaxCluster =
      combineAtEnd ? log2MaxClusterUnelminatedVars : 0.0;
    if (prefix != null) {
      for (int prefixIndex = 0; prefixIndex < prefix.length; prefixIndex++) {
        int removed = prefix[prefixIndex];
        iLog2MaxCluster =
          Math.max (
            iLog2MaxCluster,
            Util.prepareForUpdate (
                iNodeToAdj, iNodeToAdjSize, removed, marked, marked2,
                fNodeToLogCardinality));
        Util.update (iNodeToAdj, iNodeToAdjSize, removed, marked, marked2);
      }
    }
    Comparable iScore = ss.startOverWithNewPrefix (prefix, iLog2MaxCluster);
    if (scoreToBeat != null && iScore.compareTo (scoreToBeat) == 1) {
      return null;
    }
    
    // Compute the initial best order, best score, and best log_2 max
    // cluster.
    
    int[] bestOrder = null;
    Comparable bestScore = scoreToBeat;
    double bestLog2MaxCluster = Double.POSITIVE_INFINITY;
    
    // If the random generator is null, then set the number of iterations to
    // 1, since there is no point to iterating more than once.
    
    if (r == null) {iterations = 1;}

    // Iterate.

    for (int iter = 0; iter < iterations; iter++) {

      // Reset the prefix scoring strategy.
      
      ss.revertToPrefix ();
      
      // Create a version of the graph and log max cluster size to modify.
      // We can just modify the version we already have if this is the last
      // iteration.
      
      int[][] nodeToAdj =
        (iter == iterations - 1) ? iNodeToAdj :
        mark.core.util.Util.clone (iNodeToAdj);
      int[] nodeToAdjSize =
        (iter == iterations - 1) ? iNodeToAdjSize :
        mark.core.util.Util.clone (iNodeToAdjSize);
      double log2MaxCluster = iLog2MaxCluster;
      Comparable score = iScore;
      int[] order = new int[ansOrderSize];
      if (prefix != null) {
        System.arraycopy (prefix, 0, order, 0, prefix.length);
      } 
      int orderIndex = (prefix == null) ? 0 : prefix.length;
    
      // For each element of the partition of the subset of nodes we are to
      // order, first add the nodes of the element to a priority queue.  Then
      // enter another loop.  For each interation of the inner loop, choose a
      // variable to remove, prepare to remove and connect, update the costs,
      // and finally remove and connect.  After this loop, if the order
      // should be rejected, then score will be null; otherwise, score will
      // be the score of the order and log2MaxCluster the log_2 of the
      // maximum cluster size or null if the order should be rejected.
      
      for (int[] pe : nodes) {
      
        // Compute the initial costs.  initialCosts is parallel to pe.
        
        Comparable[] initialCosts = new Comparable[pe.length];
        for (int nodeIndex = 0; nodeIndex < pe.length; nodeIndex++) {
          int n = pe[nodeIndex];
          initialCosts[nodeIndex] =
            bruteCost (nodeToAdj, nodeToAdjSize, n, marked);
        }
        pq.clear (pe, initialCosts);
        
        //debugging:
        //pq.validateHeapProperty ();
        //validateScores (nodeToAdj, nodeToAdjSize, marked, pq);

        // Enter the inner loop.
  
        while (pq.size () > 0) {
          if (r != null) {pq.swapTies (r);}
          int removed = pq.remove ();
          order[orderIndex++] = removed;
          log2MaxCluster =
            Math.max (
              log2MaxCluster,
              Util.prepareForUpdate (
                  nodeToAdj, nodeToAdjSize, removed, marked, marked2,
                  fNodeToLogCardinality));
          score = ss.addVar (order, orderIndex, log2MaxCluster);
          if (bestScore != null && score.compareTo (bestScore) == 1) {
            score = null; break;
          }
          updateCosts (
            nodeToAdj, nodeToAdjSize, removed, marked, marked2, pq);
          Util.update (nodeToAdj, nodeToAdjSize, removed, marked, marked2);
          //debugging:
          //pq.validateHeapProperty ();
          //validateScores (nodeToAdj, nodeToAdjSize, marked, pq);
        }
      
        if (score == null) {break;}
        
      }

      // Combine if appropriate.
      
      if (score != null && combineAtEnd) {
        score = ss.combineAtEnd (log2MaxCluster);
        if (bestScore != null && score.compareTo (bestScore) == 1) {
          score = null;
        }
      }
      
      // Update the best order and log max cluster.
      
      if (score != null) {
        bestOrder = order;
        bestLog2MaxCluster = log2MaxCluster;
        bestScore = score;
      }

    }
      
    // Return the result.
//System.out.println ("best: " + bestLog2MaxCluster);
    return
      bestOrder == null ? null :
      new Result<int[]> (bestOrder, bestLog2MaxCluster, bestScore);
    
  }

}
