package mark.inf.eo;

/**
 * An EliminationOrderEngine that uses minfill.
 *
 * @author Mark Chavira
 */

public abstract class EoeMinfillBase extends Eoe {

  /**
   * Keeps track of changes to the fillin score.  The only reason this object
   * is not inside updateCosts () is so that it doesn't need to be
   * reallocated and reinitialized repeatedly.
   */
  
  public double[] deltaFillins;
  
  /**
   * Keeps track of changes to the log cluster score.  The only reason this
   * object is not inside updateCosts () is so that it doesn't need to be
   * reallocated and reinitialized repeatedly.
   */
  
  public double[] deltaLogClusterSize;
  
  /**
   * @see Eoe
   */
  
  public void init (int N) {
    deltaFillins = new double[N];
    deltaLogClusterSize = new double[N];
  }
  
  /**
   * Returns the cost of an edge between the two given nodes.
   * 
   * @param n1 the first given node.
   * @param n2 the second given node.
   * @return the cost.
   */
  
  public abstract int edgeCost (int n1, int n2);
  
  /**
   * The pq element.
   * 
   * @author Mark Chavira
   */
  
  private static class Cost implements Comparable<Cost> {
    public int fillins;
    public double logClusterSize;
    public int compareTo (Cost c) {
      return
        (fillins < c.fillins) ? -1 :
        (fillins > c.fillins) ? +1 :
        mark.core.math.Util.close (logClusterSize, c.logClusterSize) ? 0 :
        logClusterSize < c.logClusterSize ? -1 : +1;
    }
  }

  /**
   * @see EliminationOrderAlgorithm
   */

  protected Comparable bruteCost (
   int[][] nodeToAdj, int[] nodeToAdjSize, int n, boolean[] marked) {
    Cost ans = new Cost ();
    ans.fillins = 0;
    ans.logClusterSize =
      Util.logClusterSize (
        nodeToAdj, nodeToAdjSize, n, fNodeToLogCardinality);
    for (
     int oneHopIndex1 = 0; oneHopIndex1 < nodeToAdjSize[n]; oneHopIndex1++) {
      int oneHop1 = nodeToAdj[n][oneHopIndex1];
      Util.fillOneHopNeighborMap (
        nodeToAdj, nodeToAdjSize, oneHop1, marked, true);
      for (
       int oneHopIndex2 = oneHopIndex1 + 1; oneHopIndex2 < nodeToAdjSize[n];
       oneHopIndex2++) {
        int oneHop2 = nodeToAdj[n][oneHopIndex2];
        if (!marked[oneHop2]) {
          ans.fillins += edgeCost (oneHop1, oneHop2);
        }
      }
      Util.fillOneHopNeighborMap (
        nodeToAdj, nodeToAdjSize, oneHop1, marked, false);
    }
    return ans;
  }
  
  /**
   * @see EliminationOrderAlgorithm
   */
  
  protected void updateCosts (
   int[][] nodeToAdj, int[] nodeToAdjSize, int removed,
   boolean[] marked, boolean[][] marked2, mark.core.util.IntKeyPQ pq)
   throws Exception {

    // Let G be a graph.  To compute a node D's score within G, it is
    // sufficient to consider the projectgion of G onto the one hop neighbors
    // of D.  Call this projection P(G,D).
    
    // Let G0 be the original graph, and let E be the node we are
    // eliminating.  The question arises, if we eliminate E, then what nodes
    // might have their scores affected?  Eliminating E will potentially
    // changes edges of E and its one hop neighbors.  Therefore, the only
    // nodes whose scores can be affected are the union of E's one hop
    // neighbors and E's two hop neighbors.

    // Let us first consider a two hop neighbor THN that is not also a one
    // hop neighbor.  Let G1 be G0 after eliminating E.  Consider P(G0,THN)
    // and P(G1,THN).  The nodes NS in P(G0,THN) are the same as the nodes in
    // P(G1,THN), since THN's edges are unaffected by the elimination.
    // However, the edges in P(G0,THN) and P(G1,THN) may be different.  In
    // particular, all nodes in NS that are one hop neighbors of E in G will
    // have edges between them in P(G1,THN), and some of those edges may not
    // exist in P(G0,THN).  Therefore, the only way that eliminating E can
    // affect the score of THN is as follows.  For every pair of nodes that
    // are one hop nieghbors of E in G0, if the pair are both one hop
    // neihbors of THN in G0, and if there is not an edge between the pair
    // in G0, then there we subtract the value of the edge from THN.  
    
    // Now let us consider a one hop neighbor OHN of E.  P(G1,OHN) will not
    // contain E, and P(G0,OHN) will.  Any edges added to E when eliminating
    // OHN from P(G0,OHN) will no longer be added when eliminating OHN from
    // P(G1,OHN).  So we subtract from OHN's score in the following
    // circumstance: if THN is a one hop neighbor of OHN and THN is not E and
    // THN is not a one hop neighbor of E.
    
    // Let OHN be a one hop neighbor of the node E being eliminated.  If
    // there is a two hop neigbhor THN of E through OHN, and if there is no
    // no edge between E and THN, then eliminating OHN would put an edge
    // between E and THN.  Since we are eliminating E, this is no longer the
    // case.  So we subtract from OHN's score.
    //
    // E----OHN----THN
    
    for (
     int oneHopIndex = 0; oneHopIndex < nodeToAdjSize[removed]; oneHopIndex++) {
      int oneHop = nodeToAdj[removed][oneHopIndex];
      if (pq.contains (oneHop)) {
        for (
         int twoHopIndex = 0; twoHopIndex < nodeToAdjSize[oneHop];
         twoHopIndex++) {
          int twoHop = nodeToAdj[oneHop][twoHopIndex];
          if (!marked[twoHop]) {
            deltaFillins[oneHop] -= edgeCost (removed, twoHop);
          }
        }
      }
    }
    
    // For each pair of one hop neighbors OHN1 and OHN2 of the node E being
    // eliminated, if there is no edge between them, then there will be soon,
    // and we take action...
    
    for (
     int oneHopIndex1 = 0; oneHopIndex1 < nodeToAdjSize[removed]; oneHopIndex1++) {
      int oneHop1 = nodeToAdj[removed][oneHopIndex1];
      for (
       int oneHopIndex2 = oneHopIndex1 + 1; oneHopIndex2 < nodeToAdjSize[removed];
       oneHopIndex2++) {
        int oneHop2 = nodeToAdj[removed][oneHopIndex2];
        if (!marked2[oneHopIndex1][oneHop2]) {
          
          // If there is a node THN that is a two hop neighbor of E through
          // OHN1 and a two hop neighbor of E through OHN2, then eliminating
          // THN will put an edge between OHN1 and OHN2.  However, since we
          // are alrady doing so, we subtract from THN's score.
          //
          //  /---OHN1---\           /---OHN1---\
          // E           THN  -or-  E-----------THN
          //  \---OHN2---/           \---OHN2---/
          
          for (
           int twoHopIndex = 0; twoHopIndex < nodeToAdjSize[oneHop1];
           twoHopIndex++) {
            int twoHop = nodeToAdj[oneHop1][twoHopIndex];
            if (marked2[oneHopIndex2][twoHop] && pq.contains (twoHop)) {
              deltaFillins[twoHop] -= edgeCost (oneHop1, oneHop2); 
            }
          }
          
          // If there is a node THN that is a two hop neighbor of E through
          // OHN1, and if there is no edge between THN and E, and if there
          // is no edge between THN and OHN2, then eliminating OHN1 would put
          // an edge between E and THN and would not put an edge between OHN2
          // and THN.  After eliminating E, eliminating OHN1 will not put an
          // edge between E and THN and will between OHN2 and THN.  So it is
          // a wash.  However, we already accounted for the loss the edge
          // between E and THN, so we increment OH1's score.
          //
          //  /---OHN1---\
          // E           THN
          //  \---OHN2
          
          if (pq.contains (oneHop1)) {
            for (
             int twoHopIndex = 0; twoHopIndex < nodeToAdjSize[oneHop1];
             twoHopIndex++) {
              int twoHop = nodeToAdj[oneHop1][twoHopIndex];
              if (!marked[twoHop] && !marked2[oneHopIndex2][twoHop]) {
                deltaFillins[oneHop1] += edgeCost (twoHop, oneHop2);
              }
            }
          }
          
          // If there is a node THN that is a two hop neighbor of E through
          // OHN2, and if there is no edge between THN and E, and if there
          // is no edge between THN and OHN1, then eliminating OHN2 would put
          // an edge between E and THN and would not put an edge between OHN1
          // and THN.  After eliminating E, eliminating OHN2 will not put an
          // edge between E and THN and will between OHN1 and THN.  So it is
          // a wash.  However, we already accounted for the loss the edge
          // between E and THN, so we increment OH1's score.
          //
          //  /---OHN1
          // E           THN
          //  \---OHN2---/
          
          if (pq.contains (oneHop2)) {
            for (
             int twoHopIndex = 0; twoHopIndex < nodeToAdjSize[oneHop2];
             twoHopIndex++) {
              int twoHop = nodeToAdj[oneHop2][twoHopIndex];
              if (!marked[twoHop] && !marked2[oneHopIndex1][twoHop]) {
                deltaFillins[oneHop2] += edgeCost (twoHop, oneHop1);
              }
            }
          }
          
        }
      }
    }

    // Now update the log cluster sizes.
    
    for (
     int oneHopIndex1 = 0; oneHopIndex1 < nodeToAdjSize[removed]; oneHopIndex1++) {
      int oneHop1 = nodeToAdj[removed][oneHopIndex1];
      if (pq.contains (oneHop1)) {
        deltaLogClusterSize[oneHop1] -= fNodeToLogCardinality[removed];
      }
      for (
       int oneHopIndex2 = oneHopIndex1 + 1; oneHopIndex2 < nodeToAdjSize[removed];
       oneHopIndex2++) {
        int oneHop2 = nodeToAdj[removed][oneHopIndex2];
        if (!marked2[oneHopIndex1][oneHop2]) {
          if (pq.contains (oneHop1)) {
            deltaLogClusterSize[oneHop1] += fNodeToLogCardinality[oneHop2];
          }
          if (pq.contains (oneHop2)) {
            deltaLogClusterSize[oneHop2]  += fNodeToLogCardinality[oneHop1];
          }
        }
      }
    }
    
    // Now actually update the queue.
    
    for (int i = 0; i < pq.size (); i++) {
      int n = pq.nthElement (i);
      if (deltaFillins[n] != 0.0 || deltaLogClusterSize[n] != 0.0) {
        Cost c = (Cost)pq.getKey (n);
        c.fillins += deltaFillins[n];
        c.logClusterSize += deltaLogClusterSize[n];
        pq.changeKey (n);
        deltaFillins[n] = 0.0;
        deltaLogClusterSize[n] = 0.0;
      }
    }
    
  }

}
