package mark.inf.eo;

/**
 * Static utility routines.
 * <p>
 * We represent nodes as integers 0..N-1.  We represent a graph as an array
 * that maps each node to its adjacency list.  We represent an adjacency as
 * an int array containing one-hop neighbors in no particular order.
 * <p>
 * Internally, we work with a different graph representation.  This
 * representation is the same as the previous, except that the length of each
 * adjacency list may be larger than the number of elements in the list.  The
 * second representation can be identified, because it is accompanied by an
 * array that maps from each node to the size of its adjacency list.
 * <p>
 * A one hop neighbor map for node n maps each node to whether it is a one
 * hop neighbor of n.  An empty one hop neighbor map is a map from each node
 * to false.
 * <p>
 * A zero/one hop neighbor map for node n maps each node to whether it is n
 * or a one-hop neighbor of n.  
 * <p>
 * A two hop neighbor map for node n maps is a
 * boolean[BIGGEST_POSSIBLE_CLUSTER - 1][N], where the first index is
 * parallel to the adjacency list for n and the second dimension maps the one
 * hop neighbor to its one hop neighbor map.  An empty two hop neighbor map
 * sets all values to false.
 *  
 * @author Mark Chavira
 */

public class Util {

  /**
   * A constant for the maximum number of variables in a cluster.  Algoithms
   * will fail if a cluster is produced with more variables.
   */

 public static final int MAX_VARS_IN_CLUSTER = 4000;

  /**
   * Returns a map from each variable to the log base 2 of its cardinality.
   *
   * @param cardinalities a map from each variable to its cardinality.
   * @return the map from each variable to the log of its cardinality.
   */
  
  public static double[] logCardinalities (int[] cardinalities) {
    double[] ans = new double[cardinalities.length];
    for (int n = 0; n < ans.length; n++) {
      ans[n] = mark.core.math.Util.log2 (cardinalities[n]);
    }
    return ans;
  }
  
  /**
   * Returns the log base two of the maximum cluster size of the given
   * elimination order when applied to the given graph.
   *
   * @param eo the given elimination order.
   * @param cardinalities a map from each node to its cardinality.
   * @param g the given graph.
   * @param combineLast whether to combine the nodes after eliminating.
   * @return the log base two of the maximum cluster size.
   */
  
  public static double logMaxClusterSize (
   int[] eo, int[] cardinalities, int[][] g, boolean combineLast) {
    int N = g.length;
    boolean[] marked = new boolean[N];
    boolean[][] marked2 = new boolean[MAX_VARS_IN_CLUSTER - 1][N];
    int[][] nodeToAdj = new int[N][];
    int[] nodeToAdjSize = new int[N];
    initGraph (g, nodeToAdj, nodeToAdjSize);
    double[] logCardinalities = logCardinalities (cardinalities);
    double ans = Double.NEGATIVE_INFINITY;
    for (int eoIndex = 0; eoIndex < eo.length; eoIndex++) {
      int removed = eo[eoIndex];
      ans =
        Math.max (
          Util.prepareForUpdate (
            nodeToAdj, nodeToAdjSize, removed, marked,
            marked2, logCardinalities),
          ans);
      Util.update (nodeToAdj, nodeToAdjSize, removed, marked, marked2);
    }
    if (combineLast) {
      double logClusterSize = 0.0; 
      boolean[] done = new boolean[g.length];
      for (int i = 0; i < eo.length; i++) {
        done[eo[i]] = true;
      }
      for (int i = 0; i < g.length; i++) {
        if (!done[i]) {
          logClusterSize += logCardinalities[i];
        }
      }
      ans = Math.max (ans, logClusterSize);
    }
    return ans;
  }

  /**
   * A utility routine that helps in converting between an empty one hop
   * neighbor map and a one hop neighbor map.  Its purpose is to improve
   * efficiency by allowing the same data structure to be reused as a map
   * for differing nodes over time.
   * 
   * @param nodeToAdj a map from each node to its adjacency list.
   * @param nodeToAdjSize a map from each node to the size of its adjacency list.
   * @param n the given node.
   * @param marked2 on entry, an empty one hop neighbor map or a one hop
   *   neighbor map for n; on exit, the opposite.
   * @param val whether the given map is empty.
   */

  public static void fillOneHopNeighborMap (
   int[][] nodeToAdj, int[] nodeToAdjSize, int n, boolean[] map,
   boolean val) {
    for (
     int oneHopIndex = 0; oneHopIndex < nodeToAdjSize[n]; oneHopIndex++) {
      int oneHop = nodeToAdj[n][oneHopIndex];
      map[oneHop] = val;
    }
  }

  /**
   * A utility routine that helps in converting between an empty two hop
   * neighbor map and a two hop neighbor map.  Its purpose is to improve
   * efficiency by allowing the same data structure to be reused as a map
   * for differing nodes over time.
   * 
   * @param nodeToAdj a map from each node to its adjacency list.
   * @param nodeToAdjSize a map from each node to the size of its adjacency
   *   list.
   * @param n the given node.
   * @param marked2 on entry, an empty two hop neighbor map or a two hop
   *   neighbor map for n; on exit, the opposite.
   * @param val whether the given map is empty.
   */
   
  public static void fillTwoHopNeighborMap (
   int[][] nodeToAdj, int[] nodeToAdjSize, int n, boolean[][] map,
   boolean val) {
    for (
     int oneHopIndex = 0; oneHopIndex < nodeToAdjSize[n]; oneHopIndex++) {
      int oneHop = nodeToAdj[n][oneHopIndex];
      fillOneHopNeighborMap (
        nodeToAdj, nodeToAdjSize, oneHop, map[oneHopIndex], val);
    }
  }
  
  /**
   * Initializes a graph in the internal representation from a graph in the
   * external representation.  Initially, all adjacency lists are two times
   * the number of neighbors.  However, update () below facilitates the
   * growth of an adjacency list should its length be insufficient.
   * 
   * @param g the given external graph.
   * @param on entry, an array of the same dimension as g; on exit, a map
   *   from each node to its adjacency list.
   * @param nodeToAdjSize on entry, an array of the same dimension as g; on
   *   exit, a map from each node to the size of its adjacency list.
   */
  
  public static void initGraph (
   int[][] g, int[][] nodeToAdj, int[] nodeToAdjSize) {
    for (int n = 0; n < nodeToAdj.length; n++) {
      nodeToAdj[n] = new int[g[n].length * 2];
      nodeToAdjSize[n] = g[n].length;
      System.arraycopy (g[n], 0, nodeToAdj[n], 0, nodeToAdjSize[n]);
    }
  }

  /**
   * Prepares for the removal of one of the nodes from the graph.  The method
   * constructs a zero/one hop map and a two hop map for the given node.
   *  
   * @param nodeToAdj a map from each node to its adjacency list.
   * @param nodeToAdjSize a map from each node to the size of its adjacency
   *   list.
   * @param removed the given node.
   * @param marked on entry, an empty one hop neighbor map; on exit, a
   *   zero/one hop neighbor map for removed.
   * @param marked2 on entry, an empty two hop neighbor map; on exit, a two
   *   hop neighbor map for removed.
   * @param logCardinalities a map from node to log of its cardinality.
   * @return the cluster size created by the update.
   */
  
  public static double prepareForUpdate (
   int[][] nodeToAdj, int[] nodeToAdjSize, int removed, boolean[] marked,
   boolean[][] marked2, double[] logCardinalities) {
    fillOneHopNeighborMap (nodeToAdj, nodeToAdjSize, removed, marked, true);
    marked[removed] = true;
    fillTwoHopNeighborMap (nodeToAdj, nodeToAdjSize, removed, marked2, true);
    return
      logClusterSize (nodeToAdj, nodeToAdjSize, removed, logCardinalities);
  }

  /**
   * Returns the log cluster size of the cluster created when removing the
   * given node from the given graph.
   * 
   * @param nodeToAdj a map from each node to its adjacency list.
   * @param nodeToAdjSize a map from each node to the size of its adjacency
   *   list.
   * @param removed the given node.
   * @param logCardinalities a map from node to log of its cardinality.
   */

  public static double logClusterSize (
   int[][] nodeToAdj, int[] nodeToAdjSize, int removed,
   double[] logCardinalities) {
    double ans = logCardinalities[removed];
    for (
     int oneHopIndex = 0; oneHopIndex < nodeToAdjSize[removed];
     oneHopIndex++) {
      int oneHop = nodeToAdj[removed][oneHopIndex];
      ans += logCardinalities[oneHop];
    }
    return ans;
  }
  
  /**
   * Removes edges from each one-hop neighbor to the given node and connects
   * all one hop neighbors; also restores the given zero one hop neighbor map
   * and given two hop neighbor map.  Doesn't bother removing edges from
   * given node to its one hop neighbors.
   *
   * @param nodeToAdj a map from each node to its adjacency list.
   * @param nodeToAdjSize a map from each node to the size of its adjacency
   *   list.
   * @param removed the given node.
   * @param marked on entry, the zero/one hop neighbor map for removed; on
   *   exit, an empty one hop neighbor map.
   * @param marked2 on entry, the two hop neighbor map for removed; on exit,
   *   an empty two hop neighbor map.
   */
  
  public static void update (
   int[][] nodeToAdj, int[] nodeToAdjSize, int removed, boolean[] marked,
   boolean[][] marked2) {
    
    // Restore marked from zero/one hop neighbor map to all false.
    
    marked[removed] = false;
    fillOneHopNeighborMap (nodeToAdj, nodeToAdjSize, removed, marked, false);
    
    // Remove edges from one-hop neighbors to the removed node.  Also
    // update the two hop neighbor map for this change.
    
    for (
     int oneHopIndex = 0; oneHopIndex < nodeToAdjSize[removed];
     oneHopIndex++) {
      int oneHop = nodeToAdj[removed][oneHopIndex];
      for (int j = 0; j < nodeToAdjSize[oneHop]; j++) {
        if (nodeToAdj[oneHop][j] == removed) {
          nodeToAdj[oneHop][j] = nodeToAdj[oneHop][--nodeToAdjSize[oneHop]];
          break;
        }
      }
      marked2[oneHopIndex][removed] = false;
    }
    
    // Now, iterate through the one hop neighbors of the removed node, and
    // for each, restore its part of marked2 and connect new neighbors.
    
    for (
     int oneHopIndex1 = 0; oneHopIndex1 < nodeToAdjSize[removed];
     oneHopIndex1++) {
      
      // Restore the neighbor's part of maked2 to all false.
      
      int oneHop1 = nodeToAdj[removed][oneHopIndex1];
      fillOneHopNeighborMap (
        nodeToAdj, nodeToAdjSize, oneHop1, marked2[oneHopIndex1], false);
      
      // For each other neighbor that hasn't been taken into account, if
      // it is not a neighbor of this neighbor, then first ensure that the
      // two adjacency lists are large enough and then make the two
      // neighbors.
      
      for (
       int oneHopIndex2 = oneHopIndex1 + 1;
       oneHopIndex2 < nodeToAdjSize[removed]; oneHopIndex2++) {
        int oneHop2 = nodeToAdj[removed][oneHopIndex2];
        if (!marked2[oneHopIndex2][oneHop1]) {
          if (nodeToAdjSize[oneHop1] == nodeToAdj[oneHop1].length) {
            int[] scratch = new int[nodeToAdjSize[oneHop1] * 3 / 2];
            System.arraycopy (
              nodeToAdj[oneHop1], 0, scratch, 0, nodeToAdjSize[oneHop1]);
            nodeToAdj[oneHop1] = scratch;
          }
          if (nodeToAdjSize[oneHop2] == nodeToAdj[oneHop2].length) {
            int[] scratch = new int[nodeToAdjSize[oneHop2] * 3 / 2];
            System.arraycopy (
              nodeToAdj[oneHop2], 0, scratch, 0, nodeToAdjSize[oneHop2]);
            nodeToAdj[oneHop2] = scratch;
          }
          nodeToAdj[oneHop1][nodeToAdjSize[oneHop1]++] = oneHop2;
          nodeToAdj[oneHop2][nodeToAdjSize[oneHop2]++] = oneHop1;
        }
      }
      
    }
    
  }

}
