package mark.inf.gate;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.Reader;
import java.util.*;

/**
 * A representation of a logic gate.
 * 
 * @author chavira
 */

public class Gate {

  /**
   * A string indicating healthy.
   */
  
  public static final String HEALTHY_STRING = "healthy";

  /**
   * A string indicating stuck at zero.
   */
  
  public static final String STUCK_AT_ZERO_STRING = "stuckAtZero";

  /**
   * A string indicating stuck at one.
   */

  public static final String STUCK_AT_ONE_STRING = "stuckAtOne";
  
  /**
   * The false domain name.
   */
  
  public static final String FALSE_NAME = "F";
  
  /**
   * The true domain name.
   */
  
  public static final String TRUE_NAME = "T";  
  
  /**
   * Constructs a set of gates from the given reader using the given parser.
   * Does not close the reader.
   * 
   * @param r the given reader.
   * @param p the given parser.
   */
  
  public static HashSet<Gate> constructGates (Reader r, Gater p)
   throws Exception {
    return p.read(r);
  }
  
  /**
   * Constructs a set of gates from the file with the given name using the
   * given parser.
   * 
   * @param n the given name.
   * @param p the given parser.
   */
  
  public static HashSet<Gate> constructGates (String n, Gater p)
   throws Exception {
    BufferedReader r = new BufferedReader (new FileReader (n));
    HashSet<Gate> ans = constructGates (r, p);
    r.close ();
    return ans;
  }

  /**
   * A list containing the domain names.
   */
  
  public static final List<String> DOMAIN_NAMES;
  static {
    ArrayList<String> temp = new ArrayList<String> ();
    temp.add ("F");
    temp.add ("T");
    DOMAIN_NAMES = Collections.unmodifiableList (temp);
  }

  /**
   * A comparator based on gate name.
   */
  
  public static final Comparator<Gate> NAME_COMPARATOR =
    new Comparator<Gate> () {
      public int compare (Gate lhs, Gate rhs) {
        return lhs.name ().compareTo (rhs.name ());
      }
    };
  
  /**
   * The gate's name.  This information may be meaningful, or it may just
   * be a unique string.
   */
  
  private String fName;
  
  /**
   * The gate type.
   */
  
  private GateType fType;
  
  /**
   * An unmodifiable array list of gate inputs.
   */
  
  private List<Gate> fInputs;
  
  /**
   * The health set.
   */
  
  private GateHealth fHealthSet;

  /**
   * A string that identifies a health variable.
   */
  
  static public final String VAR_NAME_HEALTH_STRING = "VarIsHealth";

  /**
   * Returns the gate's name.
   */
  
  public String name () {
    return fName;
  }
  
  /**
   * Returns the gate's type.
   */
  
  public GateType type () {
    return fType;
  }
  
  /**
   * Returns an unmodifiable array list of the gate's inputs.
   */
  
  public List<Gate> inputs () {
    return fInputs;
  }
  
  /**
   * Returns the gate's health set.
   */
  
  public GateHealth healthSet () {
    return fHealthSet;
  }
 
  /**
   * The constructor.
   * 
   * @param name the gate's name.
   * @param type the gate's type.
   * @param inputs the gate's inputs.
   * @param healthSet the gate's health set.
   */
  
  public Gate (
   String name, GateType type, List<Gate> inputs, GateHealth healthSet) {
    fName = name;
    fType = type;
    fInputs = Collections.unmodifiableList (new ArrayList (inputs));
    fHealthSet = healthSet;
  }
  
  /**
   * Given a set of gates that comprise a circuit, returns those gates that
   * are inputs.
   * 
   * @param gates the given set of gates.
   * @return the inputs.
   */
  
  public static HashSet<Gate> inputs (Collection<Gate> gates) {
    HashSet<Gate> ans = new HashSet<Gate> (gates);
    for (Gate g : gates) {
      if (g.inputs ().size () != 0) {
        ans.remove (g);
      }
    }
    return ans;
  }
  
  /**
   * Given a set of gates that comprise a circuit, returns those gates that
   * are outputs.
   * 
   * @param gates the given set of gates.
   * @return the inputs.
   */
  
  public static HashSet<Gate> outputs (Collection<Gate> gates) {
    HashSet<Gate> ans = new HashSet<Gate> (gates);
    for (Gate g1 : gates) {
      for (Gate g2 : g1.inputs ()) {
        ans.remove (g2);
      }
    }
    return ans;
  }

  /**
   * Strips from and buff gates from a collection of gates and rewires
   * inputs as necessary.
   * 
   * @param gates the input gates.
   * @return the modified gates.
   */
  
  public static HashSet<Gate> stripFromAndBuff (Collection<Gate> gates) {

    // Create a map from gate to effective gate.

    HashMap<Gate,Gate> gateToEffectiveGate = new HashMap<Gate,Gate> ();
    for (Gate g : gates) {
      Gate eg = g;
      while (
       eg.fType == mark.inf.gate.GateType.FROM ||
       eg.fType == mark.inf.gate.GateType.BUFF) {
        eg = eg.fInputs.get (0);
      }
      gateToEffectiveGate.put (g, eg);
    }
  
    // Replace references to ids with references to effective ids, construct
    // the culled list of gates, and return the answer.
    
    HashMap<Gate,Gate> oldToNew = new HashMap<Gate,Gate> ();
    for (Gate g : gates) {
      if (g.fType == mark.inf.gate.GateType.FROM ||
          g.fType == mark.inf.gate.GateType.BUFF) {continue;}
      ArrayList<Gate> newInputs = new ArrayList<Gate> ();
      for (Gate input : g.fInputs) {
        newInputs.add (gateToEffectiveGate.get (input));
      }
      oldToNew.put (g, new Gate (g.name (), g.type (), newInputs, g.healthSet ()));
    }
    return new HashSet<Gate> (oldToNew.values ());
    
  }
  
  /**
   * Given an instantiation of inputs, returns the output of the gate with
   * the given type.
   * 
   * @param inst the instantiation of the intputs.
   * @param type the given type.
   * @return the gate's output.
   */
  
  public static int evaluate (GateType type, int... inputs)
   throws Exception {
    int ans;
    switch (type) {
      case INPUT: throw new Exception ("Cannot evaluate input!");
      case LOW: return 0;
      case HIGH: return 1;
      case AND: for (int b : inputs) {if (b == 0) {return 0;}} return 1;
      case NAND: for (int b : inputs) {if (b == 0) {return 1;}} return 0;
      case OR: for (int b : inputs) {if (b == 1) {return 1;}} return 0;
      case NOR: for (int b : inputs) {if (b == 1) {return 0;}} return 1;
      case XOR: ans = 0; for (int b : inputs) {if (b == 1) {ans = 1 - ans;}} return ans;
      case XNOR: ans = 1; for (int b : inputs) {if (b == 1) {ans = 1 - ans;}} return ans;
      case BUFF: return inputs[0];
      case NOT: return 1 - inputs[0];
      case FROM: return inputs[0];
    }
    throw new Exception ("Unknown gate type!");
  }
  
  /**
   * Given an instantiation of its inputs, returns the output for gate.
   * Throws an exception if the gate is an input.
   * 
   * @param inst the instantiation of the inputs.
   * @return the gate's output.
   */

  public int evaluate (int... inputs) throws Exception {
    return evaluate (type (), inputs);
  }

  /**
   * Returns the name of the health variable belonging to the gate variable
   * with the given name.
   * 
   * @param n the given name.
   * @return the name of the health variable.
   */
  
  public static String healthVarName (String n) {
    return n + "_" + VAR_NAME_HEALTH_STRING;
  }

}
