package mark.inf.wcnf;
import java.util.*;
import java.io.*;

import mark.inf.core.LitMap;

/**
 * A representation of a cnf.  Variables are 1..numVars.
 *
 * @author Mark Chavira
 */
 
public class WCnf {

  /**
   * The number of variables.
   */

  public int fNumVars = -1;
  
  /**
   * The clauses.  Each clause is an array of literals.  Negative number
   * means negative literal.
   */

  public double[] fWeights = null;
  public int[][] fClauses = null;

  /**
   * Constructs a wcnf from the given reader using the given parser.  Does not
   * close the reader.
   * 
   * @param r the given reader.
   * @param p the given parser.
   */
  
  public static WCnf constructWCnf (Reader r, WCnfr p) throws Exception {
    return p.wcnf (r);
  }
  
  /**
   * Constructs a wcnf from the file with the given name using the given
   * parser.
   * 
   * @param n the given name.
   * @param p the given parser.
   */
  
  public static WCnf constructWCnf (String n, WCnfr p) throws Exception {
    BufferedReader r = new BufferedReader (new FileReader (n));
    WCnf ans = constructWCnf (r, p);
    r.close ();
    return ans;
  }

  /**
   * Returns a copy of the wcnf with additional clauses.
   *
   * @param frontClauses clauses added to the beginning.
   * @param backClauses clauses added to the end.
   * @return the new cnf.
   */

/*
  public Cnf addClauses (
   ArrayList<int[]> frontClauses, ArrayList<int[]> backClauses) {
    Cnf ans = new Cnf ();
    ans.fNumVars = fNumVars;
    int numClauses =
      fClauses.length + 
      (frontClauses != null ? frontClauses.size () : 0) +
      (backClauses != null ? backClauses.size () : 0);
    ans.fClauses = new int[numClauses][];
    int index = 0;
    if (frontClauses != null) {
      for (int i = 0; i < frontClauses.size (); i++) {
        ans.fClauses[index++] = (int[])frontClauses.get (i);
      }
    }
    System.arraycopy (fClauses, 0, ans.fClauses, index, fClauses.length);
    index += fClauses.length;
    if (backClauses != null) {    
      for (int i = 0; i < backClauses.size (); i++) {
        ans.fClauses[index++] = (int[])backClauses.get (i);
      }
    }
    return ans;
  }
*/

  /**
   * Returns the number of occurrences of literals.
   *
   * @return the number of occurrences.
   */
  
  public int numOccurrencesOfLiterals () {
    int ans = 0;
    for (int i = 0; i < fClauses.length; i++) {
      ans += fClauses[i].length;
    }
    return ans;
  }

  /**
   * Sorts the literals within each clause and then sorts the clauses.
   */

/*
  public void sortLiteralsAndClauses () {
    sortLiterals ();
    Arrays.sort (
      fClauses,
      new Comparator () {
        public int compare (Object o1, Object o2) {
          int[] a1 = (int[])o1;
          int[] a2 = (int[])o2;
          for (int i = 0; ; i++) {
            if (i == a1.length && i == a2.length) return 0;
            if (i == a1.length) return -1;
            if (i == a2.length) return +1;
            if (a1[i] < a2[i]) return -1;
            if (a1[i] > a2[i]) return +1;
          }
        }
      });
  }
*/

  /**
   * Sorts the literals within each clause.
   */

/*
  public void sortLiterals () {
    for (int i = 0; i < fClauses.length; i++) {
      Arrays.sort (fClauses[i]);
    }
  }
*/

  /**
   * Writes the given cnf to the file having the given name in dimacs cnf
   * format. If the given number of e-clauses is > 0, then adds an "eclauses"
   * line after "p cnf".
   *
   * @param suppressTimestamp whether to suppress the timestamp in the dimacs
   *    file.
   * @param header a string to prepend to the dimacs file.  May be null if
   *   none.
   * @param afterPString a string to be placed between the problem line and
   *   the first clause.  May be null if none.
   * @param numEclauses if >= 0, the prints an eclauses line with this
   *   many eclauses.
   * @param lm an optional literal map which can also be written.
   * @param parallel to the clauses, the weights if writing a weighted cnf;
   *   null otherwise.
   * @param fn the given name.
   */

  public void writeDimacsString (
   boolean suppressTimestamp, String header, String afterPString,
   int numEClauses, LitMap lm, double[] weights, String fn)
   throws Exception {
    PrintWriter pw =
      new PrintWriter (new BufferedWriter (new FileWriter (fn)));
    writeDimacsString (
      suppressTimestamp, header, afterPString, numEClauses, lm, weights, pw);
    pw.close ();
  }

  /**
   * Writes the given cnf to the given print writer in dimacs cnf format.
   * If the given number of e-clauses is > 0, then adds an "eclauses" line
   * after "p wcnf".  Does not close the given writer.
   *
   * @param suppressTimestamp whether to suppress the timestamp in the dimacs
   *    file.
   * @param header a string to prepend to the dimacs file.  May be null if
   *   none.
   * @param afterPString a string to be placed between the problem line and
   *   the first clause.  May be null if none.
   * @param numEclauses if >= 0, the prints an eclauses line with this
   *   many eclauses.
   * @param lm an optional literal map which can also be written.
   * @param parallel to the clauses, the weights if writing a weighted cnf;
   *   null otherwise.
   * @param pw the given print writer.
   */

  public void writeDimacsString (
   boolean suppressTimestamp, String header, String afterPString,
   int numEClauses, LitMap lm, double[] weights, PrintWriter pw)
   throws Exception {
    if (header != null) {pw.println (header);}
    if (!suppressTimestamp) {
      pw.println ("c CNF generated on " + new Date () + ".");
    }
    if (lm != null) {
      pw.println ("c");
      pw.println (lm);
    }
    String e =
      numEClauses > 0 ?
      ("eclauses " + numEClauses + mark.core.io.Util.NEWLINE) : "";
    pw.print (weights == null ? "p cnf " : "p wcnf ");
    if (fClauses.length == numEClauses) {
      pw.println (fNumVars + " " + (fClauses.length + 1));
      pw.print (e);
      pw.println (weights == null ? "1 -1 0" : "0.0 1 -1 0");
    } else {
      pw.println (fNumVars + " " + fClauses.length);
      pw.print (e);
    }
    if (afterPString != null) {pw.println (afterPString);}
    for (int i = 0; i < fClauses.length; i++) {
      if (weights != null) {pw.print (weights[i] + " ");}
      for (int j = 0; j < fClauses[i].length; j++) {
        pw.print (fClauses[i][j]);
        pw.print (" ");
      }
      pw.println ("0");
    }
  }

  /**
   * Writes the given cnf to a temporary file and returns the file object.
   *
   * @param cnf the cnf.
   * @param numEClauses the number of eclauses.
   * @return the name of the temp file.
   */

  public String writeDimacsToTempFile (int numEClauses) throws Exception {
    String ans =
      File.createTempFile ("cnf", ".cnf").getAbsolutePath ();
    writeDimacsString (true, null, null, numEClauses, null, null, ans);
    return ans;
  }

}
