package mark.reason.apps;
import java.util.*;
import java.io.*;
import mark.reason.core.*;

/**
 * A program that inputs a CNF and a maximum cluster size, splits variables
 * until the maximum cluster size is the given size or less, and then outputs
 * the resulting CNF. 
 * <xmp>
 * Usage
 *   [-seed <seed>] <log2ClusterLimit> <wcnfIn> <cnfOut>
 * where:
 *   <seed>             : specifies the seed for the random generator
 *   <log2ClusterLimit> : the given cluster limit.
 *   <wcnfIn>            : the input WCNF.
 *   <cnfOut>           : the output CNF.
 * </xmp>
 * 
 * @author Mark Chavira
 */

public class WCnfSplit {

  /**
   * A scope that carries around an integer (the clause index).
   * 
   * @author Mark Chavira
   */
  
  private static class ClauseScope extends Scope {
    public int fClauseIndex;
    public double fClauseWeight;
    public ClauseScope (Scope s, int clIndex, double clWeight) {
      super (s);
      fClauseIndex = clIndex;
      fClauseWeight = clWeight;
    }
    public Scope replaceVariables (Map<Var,Var> variableMap) {
      return new ClauseScope (replaceScope (variableMap), fClauseIndex, fClauseWeight);
    }
  }
    
  /**
   * The main program.
   * 
   * @param args the command line arguments.
   */
  
  public static void main (String[] args) throws Exception {
    
    // Start the timer.
    
    mark.core.util.Timer timer = new mark.core.util.Timer ();
    timer.start ();

    // Extract the arguments.
    
    mark.core.shell.ArgHelper ah =
      new mark.core.shell.ArgHelper (args, WCnfSplit.class);
    ah.augmentUsage ("-seed <seed>", false);
    ah.augmentUsage ("-minsize", false);
    ah.augmentUsage ("-maxcluster", false);
    ah.augmentUsage ("-weighted", false);
    ah.augmentUsage ("-clauselength", false);
    ah.augmentUsage ("-dtreeFrequency <frequency>", false);
    ah.augmentUsage ("-numOrders <num>", false);
    ah.augmentUsage ("<log2ClusterLimit>", true);
    ah.augmentUsage ("<wcnfIn>", true);
    ah.augmentUsage ("<outPrefix>", true);
    Random r =
      mark.core.shell.ArgUtil.extractRandomGenerator (
        ah, "-seed", false, null);
    boolean minsize = ah.extractMatched ("-minsize");
    boolean maxcluster = ah.extractMatched ("-maxcluster");
    boolean weighted = ah.extractMatched ("-weighted");
    boolean clauselength = ah.extractMatched ("-clauselength");
    int dtreeFrequency;
    {
      String temp = ah.extractMatchedPlusOne ("-dtreeFrequency", false);
      dtreeFrequency =
        temp == null ? Integer.MAX_VALUE : Integer.parseInt (temp);
    }
    int numOrders;
    {
      String temp = ah.extractMatchedPlusOne ("-numOrders", false);
      numOrders = temp == null ? 1 : Integer.parseInt (temp);
    }
    double log2ClusterLimit = Double.parseDouble (ah.extractFirst (true));
    mark.inf.wcnf.WCnf wcnf =
      mark.inf.wcnf.WCnf.constructWCnf (
        ah.extractFirst (true), mark.inf.wcnf.WCnfrDimacs.INSTANCE);
    String outPrefix = ah.extractFirst (true);
    String cnfOut = outPrefix + ".cnf";
    String mapOut = outPrefix + ".map";
    String dtreeOut = outPrefix + ".dtree";
    ah.verifyEmpty ();
    
    // Create the orig reason variables.
    
    HashMap<Var,Integer> reasonOrigVarToDimacsOrigVar =
      new HashMap<Var,Integer> ();
    Var[] dimacsOrigVarToReasonOrigVar = new Var[wcnf.fNumVars + 1];
    for (
     int origDimacsVar = 1; origDimacsVar <= wcnf.fNumVars; origDimacsVar++) {
      dimacsOrigVarToReasonOrigVar[origDimacsVar] =
        new VarBinary ("X" + origDimacsVar);
      reasonOrigVarToDimacsOrigVar.put (
        dimacsOrigVarToReasonOrigVar[origDimacsVar], origDimacsVar);
    }
    
    // Create the orig scopes, which correspond to the clauses.  Also, find
    // out which variables do not appear in the clauses.

    boolean[] dimacsOrigVarToAppearsInCluase = new boolean[wcnf.fNumVars + 1];
    HashSet<ClauseScope> origScopes = new HashSet<ClauseScope> ();
    for (int clIndex = 0; clIndex < wcnf.fClauses.length; clIndex++) {
      int[] cl = wcnf.fClauses[clIndex];
      HashSet<Var> origScopeVars = new HashSet<Var> ();
      for (int lit : cl) {
        int v = Math.abs (lit);
        dimacsOrigVarToAppearsInCluase[v] = true;
        origScopeVars.add (dimacsOrigVarToReasonOrigVar[v]);
      }
      origScopes.add (
        new ClauseScope (
          new Scope (
            origScopeVars.toArray (
              new Var[origScopeVars.size ()])), clIndex, wcnf.fWeights[clIndex]));
    }

    // Calculate the total clause weight score for each variable
    HashMap<Var,Double> reasonOrigVarToWeightScore = null;
    double totalClauseWeight = 0.0;

    if (weighted) {
      reasonOrigVarToWeightScore = new HashMap<Var,Double>();
      for (ClauseScope cl  : origScopes) {
        for (Var v : cl.variables()) {
          Double wscore = reasonOrigVarToWeightScore.get(v);
          if (wscore == null) {
	    wscore = new Double(0.0);
	  }
          if (!clauselength)
            reasonOrigVarToWeightScore.put(v, wscore + cl.fClauseWeight);
          else
            reasonOrigVarToWeightScore.put(v, wscore + cl.fClauseWeight*Math.pow(2.0, -1.0*cl.variables().size()));
        }
        if (!clauselength)
          totalClauseWeight += cl.fClauseWeight;
        else
          totalClauseWeight += cl.fClauseWeight*Math.pow(2.0, -1.0*cl.variables().size());
      }
    }
    
    // Determine which variables to split and define mappings between
    // original variables and clones.  Make sure to put mappings for
    // variables that don't appear in the clauses too.
    
    HashMap<Var,Var[]> reasonOrigVarToReasonCloneVars =
      new HashMap<Var,Var[]> ();
    mark.core.util.Pair<
     mark.reason.struct.SplittingStrategy.Result<ClauseScope>,
     mark.core.rdag.Dtree<ClauseScope>> result =
      new mark.reason.struct.WeightedSpDtree (
        minsize ?
          new mark.inf.eo.Eoe[] {new mark.inf.eo.EoeMinsize ()} :
          new mark.inf.eo.Eoe[] {new mark.inf.eo.EoeMinfillUnweighted ()},
        new int[] {numOrders}).constructApproximateScopes (
      r, origScopes, reasonOrigVarToReasonCloneVars, log2ClusterLimit,
      dtreeFrequency, reasonOrigVarToWeightScore, totalClauseWeight, maxcluster);
    HashMap<Var,Var> reasonCloneVarToReasonOrigVar = new HashMap<Var,Var> ();
    for (Var orig : reasonOrigVarToReasonCloneVars.keySet ()) {
      for (Var clone : reasonOrigVarToReasonCloneVars.get (orig)) {
        reasonCloneVarToReasonOrigVar.put (clone, orig);
      }
    }
    for (
     int dimacsOrigVar = 1; dimacsOrigVar <= wcnf.fNumVars; dimacsOrigVar++) {
      if (!dimacsOrigVarToAppearsInCluase[dimacsOrigVar]) {
        Var reasonOrigVar = dimacsOrigVarToReasonOrigVar[dimacsOrigVar]; 
        reasonOrigVarToReasonCloneVars.put (
          reasonOrigVar, new Var[] {reasonOrigVar});
        reasonCloneVarToReasonOrigVar.put (reasonOrigVar, reasonOrigVar);
      }
    }

/*
    for (Var var : reasonOrigVarToReasonCloneVars.keySet()) {
      if (reasonOrigVarToReasonCloneVars.get(var).length > 1) {
        System.out.print(reasonOrigVarToDimacsOrigVar.get(var));
	System.out.print(" ");
      }
    }
    System.out.println("");
*/

    // Count the number of clone variables (the number of variables in the
    // new cnf.
    
    int numCloneVars = 0;
    for (Var[] clones : reasonOrigVarToReasonCloneVars.values ()) {
      numCloneVars += clones.length;
    }
    
    // Create the clone dimacs variables.
    
    int nextDimacsCloneVar = wcnf.fNumVars + 1;
    HashMap<Var,Integer> reasonCloneVarToDimacsCloneVar =
      new HashMap<Var,Integer> ();
    Var[] dimacsCloneVarToReasonCloneVar = new Var[numCloneVars + 1];
    int[][] dimacsOrigVarToDimacsCloneVars = new int[wcnf.fNumVars + 1][];
    for (Var reasonOrigVar : reasonOrigVarToReasonCloneVars.keySet ()) {
      int dimacsOrigVar = reasonOrigVarToDimacsOrigVar.get (reasonOrigVar);
      Var[] reasonCloneVars =
        reasonOrigVarToReasonCloneVars.get (reasonOrigVar);
      dimacsOrigVarToDimacsCloneVars[dimacsOrigVar] =
        new int[reasonCloneVars.length];
      for (int i = 0; i < reasonCloneVars.length; i++) {
        Var reasonClone = reasonCloneVars[i];
        int dimacsClone = i == 0 ? dimacsOrigVar : nextDimacsCloneVar++;
        dimacsOrigVarToDimacsCloneVars[dimacsOrigVar][i] = dimacsClone;
        reasonCloneVarToDimacsCloneVar.put (reasonClone, dimacsClone);
        dimacsCloneVarToReasonCloneVar[dimacsClone] = reasonClone;
      }
    }
    
    // Now create the new cnf.
    
    wcnf.fNumVars = numCloneVars;
    for (ClauseScope s : result.elem0.fScopes) {
      int[] cl = wcnf.fClauses[s.fClauseIndex];
      for (int i = 0; i < cl.length; i++) {
        int origDimacsVar = Math.abs (cl[i]);
        Var origReasonVar = dimacsOrigVarToReasonOrigVar[origDimacsVar];
        Var cloneReasonVar = null;
        for (Var currentCloneReasonVar : s.variables ()) {
          if (reasonCloneVarToReasonOrigVar.get (currentCloneReasonVar) == origReasonVar) {
            cloneReasonVar = currentCloneReasonVar;
            break;
          }
        }
        int cloneDimacsVar =
          reasonCloneVarToDimacsCloneVar.get (cloneReasonVar);
        cl[i] = cl[i] < 0 ? -cloneDimacsVar : cloneDimacsVar;
      }
    }

    // Write out the new cnf.
    
    wcnf.writeDimacsString (
      false, "c CNF generated by " + WCnfSplit.class.getName () +
      " by Mark Chavira", null, 0, null, null, cnfOut);
    
    // Write out the dtree.
    
    HashMap<ClauseScope,Integer> m = new HashMap<ClauseScope,Integer> ();
    ArrayList<mark.core.rdag.Dtree<ClauseScope>> nodes =
      (ArrayList)result.elem1.childrenBeforeParents ();
    for (mark.core.rdag.Dtree<ClauseScope> n : nodes) {
      if (n.isInternal ()) {continue;}
      m.put (n.fObject, n.fObject.fClauseIndex);
    }
    mark.reason.struct.DtreeClauseUtils.writeClauseDtree (
      mark.core.rdag.Dtree.targetDtree (result.elem1, m), dtreeOut);
    
    // Write out the map file.
    
    PrintWriter out =
      new PrintWriter (new BufferedWriter (new FileWriter (mapOut)));
    for (int i = 1; i < dimacsOrigVarToDimacsCloneVars.length; i++) {
      out.println (
        i + " " + Arrays.toString (dimacsOrigVarToDimacsCloneVars[i]));
    }
    out.close ();
    
    // Output the time.
    
    timer.stop ();
    int numVarsSplit = 0;
    int numClones = 0;
    for (Var[] clones : reasonOrigVarToReasonCloneVars.values ()) {
      if (clones.length > 1) {numVarsSplit++;}
      numClones += clones.length;
    }
    mark.core.shell.Info info = new mark.core.shell.Info ();
    info.add ("totalTimeSeconds", "Total Time (s)", timer.totalTime () / 1000.0);
    info.add ("numVarsSplit", "Number of variables split", numVarsSplit);
    info.add ("numVarsNotSplit", "Number of variables not split", reasonOrigVarToReasonCloneVars.size () - numVarsSplit);
    info.add ("numClones", "Number of clones", numClones);
    info.add ("log2MaxCluster", "Log_2 max cluster achieved", result.elem0.fLog2MaxClusterAchieved);
    System.out.println (info.longDesc (null));
    
  }
  
}
