package mark.reason.bn;
import java.util.*;
import java.io.*;
import mark.reason.core.*;

/**
 * Static utility methods.
 *
 * @author Mark Chavira
 */

public class BnUtil {

  /**
   * A format for probabilities.
   */
   
  public static final java.text.DecimalFormat PROB_FORMAT =
    new java.text.DecimalFormat ("0.######E0");

  //=========================================================================
  // Methods that act on sets of potentials.
  //=========================================================================

  /**
   * Performs a generalization of classical pruning that applies to any set
   * of potentials, not just a Bayesian network.  The filter first shrinks
   * the given potentials on the evidence and then enters a loop, where it
   * does two things: identifies a variable not in the query that appears in
   * only one potential, sums or maxes out that variable, and then if the
   * resulting potential is a constant, replaces the potential with the
   * constant.  The loop ends when there is no non-query variable that
   * appears in only one potential.  At the end, if a query variable does not
   * appear, which will hapen if a variable is both query and evidence, adds
   * an indicator for the query variable.  Also at the end, combines all
   * constants into a single constant.  If the set of potentials is a
   * Bayesian network, then this procedure reduces to classical pruning.
   *
   * @param pots the given potentials.
   * @param queryVars the given query variables.
   * @param evidence the given evidence.
   * @param maximize whether to maximize as opposed to summing.
   * @return the resulting set of potentials.
   */

  public static HashSet<mark.reason.net.Pot> orient (
   Set<mark.reason.net.Pot> pots, Set<Var> queryVars,
   Map<Var,Double> evidence, boolean maximize) throws Exception {

    // First, perform some preliminaries.  Shrink based on the evidence.
    // Shrinking may get rid of some query variables.  If so, then we will
    // add them back at the end.  Create a map from variable to set of
    // potentials containing that variable.  Identify which variables are not
    // in query and exist in only one potential.  These variables will be
    // maximized out shortly.
    
    HashSet<mark.reason.net.Pot> ans =
      mark.reason.flt.FltShrink.shrink (pots, evidence);
    HashMap<Var,HashSet<mark.reason.net.Pot>> varToPotSet =
      CoreUtil.varToScopeMap (ans);
    HashSet<Var> nonQueryAndExactlyOne = new HashSet<Var> ();
    for (Var v : varToPotSet.keySet ()) {
      if (
       (queryVars == null || !queryVars.contains (v)) &&
       varToPotSet.get (v).size () == 1) {
        nonQueryAndExactlyOne.add (v);
      }
    }
    
    // Enter the loop.  In each iteration, process a variable that is not a
    // query variable and that exists in only one potential.  The loop
    // invariant at the top is that the mpe probability for ans and
    // pots/evidence is the same; varToPotSet is a map from each variable in
    // ans to the set of potentials in ans containing the variable; and
    // nonQueryAndOnlyOne is a set of the variables in ans that do not appear
    // in the query and that appear in exactly one potential in ans.
    
    while (!nonQueryAndExactlyOne.isEmpty ()) {
      Iterator<Var> iter = nonQueryAndExactlyOne.iterator ();
      Var chosenVar = iter.next ();
      iter.remove ();
      
      // Max out the variable from its one potential and determine whether
      // the result is a constant.
      
      mark.reason.net.Pot chosenPot =
        varToPotSet.get (chosenVar).iterator ().next ();
      varToPotSet.remove (chosenVar);
      mark.reason.net.Pot newPot =
        maximize ?
        chosenPot.maxOut (Collections.singleton (chosenVar)) :
        chosenPot.sumOut (Collections.singleton (chosenVar));
      boolean constant = newPot.almostEquivalentToConstant ();
      
      // If the new potential is a zero constant, then remove all potentials
      // from the answer, add a zero constant to the answer, and exit the
      // loop.
      
      if (constant && newPot.getCp (0) == 0.0) {
        ans.clear ();
        ans.add (new mark.reason.net.PotTable (0.0));
        varToPotSet.clear ();
        nonQueryAndExactlyOne.clear ();
        break;
      }
        
      // If the new potential is a non-zero constant, then do the following.
      // (1) Remove the old potential from ans, (2) add the constant to ans,
      // and (3) process each variable in the old potential.  To process a
      // variable in the old potential, (3a) remove the old potential from
      // the variable's set of potentials, (3b) if the variable then has no
      // potentials, remove it from the list of variables with one potential,
      // and (3c) if the variable is non-query and now exists in only a
      // single potential, then add it to the list of such variables.
                
      if (constant) {
        ans.remove (chosenPot);
        ans.add (new mark.reason.net.PotTable (newPot.getCp (0)));
        for (Var v : chosenPot.variables ()) {
          if (v != chosenVar) {
            HashSet<mark.reason.net.Pot> s = varToPotSet.get (v);
            s.remove (chosenPot);
            if (s.size () == 0) {
              nonQueryAndExactlyOne.remove (v);
            } else if (
             (queryVars == null || !queryVars.contains (v)) &&
             s.size () == 1) {
              nonQueryAndExactlyOne.add (v);
            }
          }
        }
      
      // Otherwise the new potential is not a constant.  Just replace the old
      // potential with the new in the answer and in the set of potentials
      // for each affected variable.
        
      } else {
        ans.remove (chosenPot);
        ans.add (newPot);
        for (Var v : chosenPot.variables ()) {
          if (v != chosenVar) {
            HashSet<mark.reason.net.Pot> s = varToPotSet.get (v);
            s.remove (chosenPot);
            s.add (newPot);
          }
        }
      }

    }
    
    // If there are constants, replace the set of constants with a single
    // constant (or no constants if the product is 1).  Ensure that each
    // query variable appears.  Return the answer.

    ans = mark.reason.flt.FltCombineConstants.combine (ans);
    HashSet<Var> appearingVars = CoreUtil.variables (ans);    
    if (queryVars != null) {
      for (Var v : queryVars) {
        if (evidence != null && evidence.keySet ().contains (v)) {
          ans.add (
            mark.reason.net.PotTable.genIndicatorTable (
              v, (int)(double)evidence.get (v)));
        } else if (!appearingVars.contains (v)) {
          ans.add (mark.reason.net.PotTable.genIndicatorTable (v));
        }
      }
    }
    return ans;

  }

  /**
   * Creates a name for each potential according to its signature and returns
   * a map from name to potential.  A potential's signature is the name of
   * its last variable or "__constant__" if the table has no variables.  If a
   * potential's signature is unique, then its name will be its singature.
   * Otherwise, all potentials with the same signature will be sorted
   * according to the total ordering and named <signature>-000000,
   * <signature>-000001, <signature-000002>, etc. 
   * 
   * @param pots the input potentials.
   */
  
  public static HashMap<String,mark.reason.net.Pot> potNameToPotMap (
   Set<mark.reason.net.Pot> pots)
   throws Exception {
    
    // Create a map from last variable to set of potentials.  For any set
    // of size 1, use its signature for its name.  For any set larger than
    // 1, postfix with a number.
    
    HashMap<String,mark.reason.net.Pot> ans =
      new HashMap<String,mark.reason.net.Pot> ();
    HashMap<Var,HashSet<mark.reason.net.Pot>> lastVarToPot =
      BnUtil.lastVarToScopeSetMap (pots);
    for (Var v : lastVarToPot.keySet ()) {
      String signature = (v == null) ? "__constant__" : v.name ();
      HashSet<mark.reason.net.Pot> set = lastVarToPot.get (v);
      if (set.size () == 1) {
        ans.put (signature, set.iterator ().next ());
      } else {
        ArrayList<mark.reason.net.Pot> sorted =
          mark.reason.net.Pot.TOTAL_SORTER.sort (set);
        for (int i = 0; i < sorted.size (); i++) {
          ans.put (
            signature +
            mark.core.text.Util.padLeft (Integer.toString (i), 6, '0'),
            sorted.get (i));
        }
      }
    }
    return ans;
  
  }
      
  /**
   * Reads and returns potentials from the given potential reader and file
   * with the given filename.
   * 
   * @param pr the given reader.
   * @param fn the given filename.
   * @return the potentials.
   */
  
  public static HashSet<mark.reason.net.Pot> potentialsFromFile (
   mark.reason.net.Pr pr, String fn, boolean normalize) throws Exception {
    HashSet<mark.reason.net.Pot> ans = pr.read (new File (fn));
    return
      normalize ?
      mark.reason.flt.FltNormalizeLast.normalizeLastVar (ans) : ans;
  }

  /**
   * Changes variables in the given potentials.
   *
   * @param potentials the given potentials.
   * @param variableMap the mapping from old variables to new.
   * @return the new potentials.
   */

  public static <T extends Scope> HashSet<T> replaceVariables (
   Set<T> scopes, Map<Var,Var> variableMap) throws Exception {
    HashSet<T> ans = new HashSet<T> ();
    for (T s : scopes) {
      ans.add ((T)s.replaceVariables (variableMap));
    }
    return ans;
  }

  //=========================================================================
  // Methods that deal with topological order.
  //=========================================================================
  
  public static <T extends Scope> HashMap<Var,HashSet<Var>>
   lastVarToSetOfChildren (Set<T> scopes) {
    HashMap<Var,HashSet<Var>> ans = new HashMap<Var,HashSet<Var>> ();
    for (Var v : CoreUtil.variables (scopes)) {
      ans.put (v, new HashSet<Var> ());
    }
    for (Scope p : scopes) {
      for (Var v : p.variables ()) {
        if (v != p.lastVariable ()) {
          ans.get (v).add (p.lastVariable ());
        }
      }
    }
    return ans;
  }

  /**
   * Using the given potentials, returns a map from last variable to set of
   * potentials that have that last variable.  Also maps null if there are
   * potentials with no last variable.
   * 
   * @param pots the given potentials.
   * @return the map.
   */
  
  public static <T extends Scope> HashMap<Var,HashSet<T>>
   lastVarToScopeSetMap (Set<T> scopes) {
    HashMap<Var,HashSet<T>> ans = new HashMap<Var,HashSet<T>> ();
    for (T s : scopes) {
      Var v = s.variables ().size () > 0 ? s.lastVariable () : null;
      HashSet<T> set = ans.get (v);
      if (set == null) {
        set = new HashSet<T> ();
        ans.put (v, set);
      }
      set.add (s);
    }
    return ans;
  }
  
  /**
   * Returns a map from last variable to potential for the given potentials.
   *
   * @param potentials the given potentials.
   * @return the map.
   */
  
  public static <T extends Scope> HashMap<Var,T> lastVarToScopeMap (
   Set<T> scopes) {
    HashMap<Var,T> ans = new HashMap<Var,T> ();
    for (T s : scopes) {
      ans.put (s.lastVariable (), s);
    }
    return ans;
  }

  /**
   * Returns a map from last variable name to potential for the given potentials.
   *
   * @param potentials the given potentials.
   * @return the map.
   */
  
  public static <T extends Scope> HashMap<String,T> lastVarNameToScopeMap (
   Set<T> scopes) {
    HashMap<String,T> ans = new HashMap<String,T> ();
    for (T s : scopes) {
      ans.put (s.lastVariable ().name (), s);
    }
    return ans;
  }

  /**
   * Returns the number of edges in the given set of potentials.
   *
   * @return the number of edges.
   */
  
  public static int numEdges (Set<mark.reason.net.Pot> potentials) {
    int ans = 0;
    for (mark.reason.net.Pot d : potentials) {
      ans += Math.max (0, d.variables ().size () - 1); // accnt for constants
    }
    return ans;
  }

  /**
   * Returns a topological ordering on the last variables of the given
   * potentials. The method works even if the given potentials are a
   * subset of all potentials in the BN, so long as there are no paths
   * that exit and re-enter the potentials' last variables.  In other
   * words, there shouldn't be a node outside of the given nodes that lies
   * between two of the given nodes.  Because rdag () produces a an rdag in
   * a canonical way, this algorithm is such that calling it multiple times
   * with the same set of potentials produces the same order, even if the
   * set data structure used to contain the potentials is different and
   * even though there may be multiple toplogical orders.
   *
   * @param potentials the potentials.
   * @return the ordering.
   */
  
  public static ArrayList varTopologicalOrder (
   Set<mark.reason.net.Pot> potentials) throws Exception {
    return
      mark.core.rdag.SimpleNode.nodesToListOfObjects (
        mark.core.rdag.Node.parentsBeforeChildren (rdag (potentials)));
  }
  
  /**
   * Returns an rdag that describes the given potentials.  The rdag may
   * have multiple roots.  Nodes are of type mark.core.rdag.SimpleNode, and
   * the fObject field of each node contains the corresponding variable.
   * Each node has children sorted by variable name, and the list of roots is
   * sorted by variable name.  The sorting ensures getting exactly the same
   * graph each time the same potentials are input.
   *
   * @param potentials the roots.
   * @return the rdag.
   */
  
  public static ArrayList rdag (
    Set<mark.reason.net.Pot> potentials) throws Exception {
   
    // First, we create a node for each variable, define a map from variable
    // to node, and create the set of roots.
    
    HashMap<Var,mark.core.rdag.SimpleNode> nodes =
      new HashMap<Var,mark.core.rdag.SimpleNode> ();
    ArrayList<mark.core.rdag.SimpleNode> roots =
      new ArrayList<mark.core.rdag.SimpleNode> ();
    for (mark.reason.net.Pot d : potentials) {
      Var primaryVariable = d.lastVariable ();
      mark.core.rdag.SimpleNode n = new mark.core.rdag.SimpleNode ();
      n.fObject = primaryVariable;
      nodes.put (primaryVariable, n);
      if (d.variables ().size () <= 1) {
        roots.add (n);
      }
    }
    
    // Now we put edges from parents to children.

    for (mark.reason.net.Pot d : potentials) {
      Var primaryVariable = d.lastVariable ();
      mark.core.rdag.SimpleNode primaryNode = nodes.get (primaryVariable);
      for (Var parentVariable : d.variables ()) {
        if (parentVariable != d.lastVariable ()) {
          mark.core.rdag.SimpleNode parentNode = nodes.get (parentVariable);
          parentNode.fChildren.add (primaryNode);
        }
      }
    }

    // Now sort the roots and children.
    
    Comparator c =
      new mark.core.rdag.SimpleNodeComparator (Var.NAME_COMPARATOR);
    Collections.sort (roots, c);
    for (mark.core.rdag.SimpleNode n : nodes.values ()) {
      Collections.sort (n.fChildren, c);
    }

    // Return the result.
    
    return roots;
   
  }

  //=========================================================================
  // Marginal error methods:
  //=========================================================================

  /**
   * Compares two sets of marginals according to the absolute error
   * criterion.  There is a score for each variable.  The score is equal to
   * the sum of the absolute values of the differences of the parameters.
   *
   * @param lhs the baseline map from variable to marginal.
   * @param rhs the comparison map from variable to marginal.
   * @return the scores.
   */
  
  public static double[] scoreAbsoluteError (
   HashMap<Var,mark.reason.net.Pot> lhs,
   HashMap<Var,mark.reason.net.Pot> rhs) {
    double[] ans = new double[lhs.size ()];
    int scoreIndex = 0;
    for (Var v : lhs.keySet ()) {
      mark.reason.net.Pot lhsMarginal = lhs.get (v);
      mark.reason.net.Pot rhsMarginal = rhs.get (v);
      for (int i = 0; i < v.domainSize (); i++) {
        ans[scoreIndex] +=
          Math.abs (lhsMarginal.getCp (i) - rhsMarginal.getCp (i));
      }
      ++scoreIndex;
    }
    return ans;
  }

  /**
   * Compares two sets of marginals according to the absolute error
   * criterion.  There is a score for each variable.  The score is equal to
   * the sum of the absolute values of the differences of the parameters.
   *
   * @param lhs the baseline map from variable to marginal.
   * @param rhs the comparison map from variable to marginal.
   * @return two double[], where the first is the sorted parameter scores and
   *   the second is null.
   */
   
  public static double[][] scoreAbsoluteErrorSortedWithBadScores (
   HashMap<Var,mark.reason.net.Pot> lhs,
   HashMap<Var,mark.reason.net.Pot> rhs) {
    double[] scores = scoreAbsoluteError (lhs, rhs);
    Arrays.sort (scores);
    return new double[][] {scores, null};
  }

  /**
   * Compares two sets of marginals according to the darwiche error
   * criterion.  There is a score for each parameter.  If the baseline
   * parameter is 0 and the comparison parameter is nonzero, then the score
   * is a "bad" score equal to the comparison parameter.  If both parameters
   * are zero, then the score is a "good" score of 0.  Otherwise, the score
   * is a good score equal to the absolute value of the difference divided
   * by the baseline parameter.
   *
   * @param lhs the baseline map from variable to marginal.
   * @param rhs the comparison map from variable to marginal.
   * @return two double[] where the first is the sorted good scores and the
   *   second is the sorted bad scores.
   */

  public static double[][] scoreDarwicheSortedWithBadScores (
   HashMap<Var,mark.reason.net.Pot> lhs,
   HashMap<Var,mark.reason.net.Pot> rhs) {
   
    // Count the total number of parameters.
    
    int numParameters = 0;
    for (Var v : lhs.keySet ()) {
      numParameters += v.domainSize ();
    }
    
    // Obtain the parameters.
    
    double[] lhsParameters = new double[numParameters];
    double[] rhsParameters = new double[numParameters];
    int parameterIndex = 0;
    for (Var v : lhs.keySet ()) {
      mark.reason.net.Pot lhsPotential = lhs.get (v);
      mark.reason.net.Pot rhsPotential = rhs.get (v);
      for (int j = 0; j < v.domainSize (); j++) {
        lhsParameters[parameterIndex] = lhsPotential.getCp (j);
        rhsParameters[parameterIndex++] = rhsPotential.getCp (j);
      }
    }
    
    // Count the number of parameters where actual is zero and comparison
    // is nonzero.
    
    int numProblemParameters = 0;
    for (int i = 0; i < numParameters; i++) {
      if (lhsParameters[i] == 0 && rhsParameters[i] != 0) {
        ++numProblemParameters;
      }
    }
    
    // Obtain the score for each parameter.
    
    double[] goodScores = new double[numParameters - numProblemParameters];
    double[] problemScores = new double[numProblemParameters];
    int goodIndex = 0;
    int problemIndex = 0;
    for (int i = 0; i < numParameters; i++) {
      if (lhsParameters[i] == 0 && rhsParameters[i] != 0) {
        problemScores[problemIndex++] = rhsParameters[i];
      } else if (lhsParameters[i] == 0) {
        goodScores[goodIndex++] = 0;
      } else {
        goodScores[goodIndex++] =
          Math.abs (lhsParameters[i] - rhsParameters[i]) / lhsParameters[i];
      }
    }
    Arrays.sort (goodScores);
    Arrays.sort (problemScores);
    return new double[][] {goodScores, problemScores};
    
  }
  
  //=========================================================================
  // Creating parameters:
  //=========================================================================

  /**
   * Returns uniform parameters for the given index.
   *
   * @param index the given index.
   * @reutrn the parameters.
   */
  
  public static double[] uniformParameters (Scope x) {
    double[] ans = new double[x.numInstantiations ()];
    Arrays.fill (ans, 1.0 / x.lastVariable ().domainSize ());
    return ans;
  }

  /**
   * Returns evidence parameters for the given variable and value: parameters
   * over a single variable where all parameters are 0 except the one
   * corresponding to the given value.
   *
   * @param index the given index.
   * @reutrn the parameters.
   */
  
  public static double[] evidenceParameters (Var v, int l) {
    double[] ans = new double[v.domainSize ()];
    ans[l] = 1.0;
    return ans;
  }
  
  /**
   * Computes some parameters from a plg.  For each instantiation that models
   * the first given plg, sets the corresponding parameter.  If the
   * instantiation models the second given plg, then set to 1.0; otherwise,
   * set to 0.0.
   *
   * @param x the index of the parameters to set.
   * @param parameters the parameters to set.
   * @param setThese the first given plg.
   * @param toThis the second given plg.
   */

  public static void someParametersFromPlg (
   Scope x, double[] parameters, mark.reason.plg.Node setThese,
   mark.reason.plg.Node toThis) throws Exception {
   SingleIterator si = x.siMap (null);
   for (int i = 0; i < si.numCompatibleInstantiations (); i++) {
     int p = si.next ();
     if (setThese.modelTest (si.w)) {
        parameters[p] = toThis.modelTest (si.w) ? 1.0 : 0.0;
      }
    }
  }

  /**
   * Computes parameters from the given plg.  For each instantiation, if the
   * instantiation models the plg, then set the corresponding parameter to
   * 1.0; otherwise, set to 0.0.
   *
   * @param index the index.
   * @param toThis the given plg.
   * @return the parameters.
   */

  public static double[] parametersFromPlg (
   Scope x, mark.reason.plg.Node toThis) throws Exception {
    double[] ans = new double[x.numInstantiations ()];
    someParametersFromPlg (
      x, ans, mark.reason.plg.NodeTrue.INSTANCE, toThis);
    return ans;
  }

  /**
   * Constructs a map from a set of potentials and a text file. 
   * 
   * @param pots the potentials with which to associate the map.
   * @param filename the name of the file from which to read the map.
   */
  
  public static mark.inf.core.LitMap<Var,mark.reason.net.Pot> constructLitMap (
   Set<mark.reason.net.Pot> pots, String mapFilename) throws Exception {
    return
      new mark.inf.core.LitMap<Var,mark.reason.net.Pot> (
        mark.reason.core.CoreUtil.varNameToVarMap (
          mark.reason.core.CoreUtil.variables (pots)),
        mark.reason.bn.BnUtil.potNameToPotMap (pots), mapFilename);
  }

}
