package mark.reason.core;
import java.util.*;

/**
 * A class for converting among reason variables and an integer variable
 * representation.
 * <p>
 * This class conforms to the immutable design pattern.
 *
 * @author Mark Chavira
 */

public class Converter {

  private mark.core.util.OrderedSet<Var> fReasonVars;
  private int[] fCardinalities;

  /**
   * Initializes the converter by creating a set of integer variables that
   * match the given set of reason variables.
   *
   * @param reasonVars the samiam variables.
   */
  
  public Converter (Collection<Var> reasonVars) {
    fReasonVars = new mark.core.util.OrderedSet<Var> (reasonVars);
    fCardinalities = new int[fReasonVars.size ()];
    for (int i = 0; i < fCardinalities.length; i++) {
      fCardinalities[i] = reasonVariable (i).domainSize ();
    }
  }

  /**
   * Returns the number of variables.
   *
   * @return the number of variables.
   */
  
  public int length () {
    return fReasonVars.size ();
  }

  /**
   * Returns an unmodifiable map from reason variables to samiam variables.
   *
   * @return the map.
   */
  
  public mark.core.util.OrderedSet<Var> map () {
    return fReasonVars;
  }

  /**
   * Returns the integer variable correponding to the given reason variable.
   *
   * @param rv the given reason variable.
   * @return the corresponding integer.
   */

  public int integerVariable (Var rv) {
    return fReasonVars.indexOf (rv);
  }
  
  /**
   * Returns the reason variable correponding to the given samiam variable.
   *
   * @param sv the given samiam variable.
   */

  public Var reasonVariable (int iv) {
    return fReasonVars.get (iv);
  }
  
  /**
   * Returns the integer variables corresponding to the given reason
   * variables as an int[].
   *
   * @param reasonVariables the given reason variables.
   * @return the integer variables.
   */
   
  public int[] integerVariableArray (Collection<Var> reasonVariables) {
    int[] ans = new int[reasonVariables.size ()];
    int index = 0;
    for (Var v : reasonVariables) {
      ans[index++] = fReasonVars.indexOf (v);
    }
    return ans;
  }
  
  /**
   * Returns the integer variables corresponding to the given reason
   * variables as a set.
   *
   * @param reasonVariables the given reason variables.
   * @return the integer variables.
   */
   
  public HashSet<Integer> integerVariableSet (Collection<Var> reasonVariables) {
    HashSet<Integer> ans = new HashSet<Integer> ();
    for (Var v : reasonVariables) {
      ans.add (fReasonVars.indexOf (v));
    }
    return ans;
  }

  /**
   * Returns the reason variable set corresponding to the given integer
   * variables.
   *
   * @param integerVariables the given integer variables.
   * @return the reason variables.
   */
   
  public HashSet<Var> reasonVariableSet (int[] integerVariables) {
    HashSet<Var> ans =
      new HashSet<Var> ();
    for (int i = 0; i < integerVariables.length; i++) {
      ans.add (reasonVariable (integerVariables[i]));
    }
    return ans;
  }
  
  /**
   * Returns the reason variable list corresponding to the given integer
   * variables.
   *
   * @param integerVariables the given integer variables.
   * @return the reason variables.
   */
   
  public ArrayList<Var> reasonVariableList (int[] integerVariables) {
    ArrayList<Var> ans = new ArrayList<Var> ();
    for (int i = 0; i < integerVariables.length; i++) {
      ans.add (reasonVariable (integerVariables[i]));
    }
    return ans;
  }

  /**
   * Returns the cardinality array for the given list of integer variables.
   *
   * @param integerVariables the given integer variables.
   * @return the cardinality array.
   */
  
  public int[] cardinalities (int[] integerVariables) {
    int[] ans = new int[integerVariables.length];
    for (int i = 0; i < ans.length; i++) {
      ans[i] = fCardinalities[integerVariables[i]];
    }
    return ans;
  }
  
  /**
   * Returns the cardinality array for all integer variables.
   *
   * @return the cardinality array.
   */
  
  public int[] cardinalities () {
    return mark.core.util.Util.clone (fCardinalities);
  }

}
