package mark.reason.core;
import java.io.*;
import java.util.*;

/**
 * Static utility routines.
 * 
 * @author chavira
 */

public class CoreUtil {

  //=========================================================================
  // Opeations on variables:
  //
  // Variable lists are saved to file as a list of newline separated
  // variable names.
  //=========================================================================

  /**
   * Returns a map from the names of the given variables to the variables.
   *
   * @param vs the given variables.
   * @return the map.
   */
  
  public static HashMap<String,Var> varNameToVarMap (Collection<Var> vs) {
    HashMap<String,Var> ans = new HashMap<String,Var> ();
    for (Var v : vs) {
      ans.put (v.name (), v);
    }
    return ans;
  }

  /**
   * Returns a map from variable to map from value name to value.
   *
   * @param vs the given variables.
   * @return the map.
   */
  
  public static HashMap<Var,HashMap<String,Integer>>
   varToValNameToValMap (Set<Var> vs) {
    HashMap<Var,HashMap<String,Integer>> ans =
      new HashMap<Var,HashMap<String,Integer>> ();
    for (Var v : vs) {
      HashMap<String,Integer> valNameToVal =
        new HashMap<String,Integer> ();
      for (int j = 0; j < v.domainSize (); j++) {
        valNameToVal.put (v.domainElementToString (j), j);
      }
      ans.put (v, valNameToVal);
    }
    return ans;
  }

  /**
   * Reads a list of the given variables from the given reader.
   *
   * @param b the given reader.
   * @param vars the given variables.
   * @return the list.
   */
  
  public static ArrayList<Var> readVariableList (
   BufferedReader b, Collection<Var> vars) throws Exception {
    HashMap<String,Var> varMap = varNameToVarMap (vars);
    ArrayList<Var> ans = new ArrayList<Var> ();
    for (;;) {
      String line = b.readLine ().trim ();
      if (line.equals (".")) {break;}
      Var v = varMap.get (line);
      if (v == null) {
        throw new Exception ("Variable " + line + " is not valid!");
      }
      ans.add (v);
    }
    return ans;
  }

  /**
   * Reads a set of variables from a reader.
   *
   * @param r the given reader.
   * @param potentials the order's potentials.
   * @return the set.
   */
  
  public static HashSet<Var> readVariableSet (
   BufferedReader b, Set<Var> vars) throws Exception {
    return new HashSet<Var> (readVariableList (b, vars));
  }

  /**
   * Returns a string representation of the given variable list.
   *
   * @param l the given list.
   * @return the string.
   */
  
  public static String writeVariableList (List<Var> l) throws Exception {
    StringBuilder ans = new StringBuilder ();
    for (int i = 0; i < l.size (); i++) {
      ans.append (l.get (i).name ());
      ans.append (mark.core.io.Util.NEWLINE);
    }
    ans.append (".");
    ans.append (mark.core.io.Util.NEWLINE);
    return ans.toString ();
  }

  /**
   * Returns a string representation of the given variable set.
   *
   * @param s the given set.
   * @return the string.
   */
  
  public static String writeVariableCollection (Collection<Var> s)
   throws Exception {
    ArrayList<Var> o = new ArrayList<Var> (s);
    Collections.sort (o, Var.NAME_COMPARATOR);
    return writeVariableList (o);
  }

  //=========================================================================
  // Evidence:
  //
  // Evidence is saved to a file as a list of lines where line 1 is the
  // number of variables involved, line 2 is the first variable name, line 3
  // is the first variable value, etc.  We use lines rather than whitespace,
  // because some network formats allow names to have whitespace in them.
  //=========================================================================

  /**
   * Converts the given double evidence into integer evidence.
   * 
   * @param doubleEvidence the given double evidence.
   * @return the integer evidence.
   */
  
  public static HashMap<Var,Integer> doubleEvidenceToIntEvidence (
   Map<Var,Double> doubleEvidence) {
    if (doubleEvidence == null) {
      return null;
    }
    HashMap<Var,Integer> ans = new HashMap<Var,Integer> ();
    for (Var v : doubleEvidence.keySet ()) {
      ans.put (v, doubleEvidence.get (v).intValue ());
    }
    return ans;
  }

  /**
   * Converts the given int evidence into double evidence.
   * 
   * @param intEvidence the given int evidence.
   * @return the double evidence.
   */
  
  public static HashMap<Var,Double> intEvidenceToDoubleEvidence (
    Map<Var,Integer> intEvidence) {
     if (intEvidence == null) {
       return null;
     }
     HashMap<Var,Double> ans = new HashMap<Var,Double> ();
     for (Var v : intEvidence.keySet ()) {
       ans.put (v, (double)intEvidence.get (v));
     }
     return ans;
   }

  /**
   * Returns a map from variable name to value name corresponding to the
   * given evidence.
   *
   * @param pw the given instantiation.
   * @return the map.
   */
  
  public static HashMap<String,String> varNameToValName (
   Map<Var,Double> pw) {
    HashMap<String,String> ans =
      new HashMap<String,String> ();
    for (Var v : pw.keySet ()) {
      ans.put (v.name (), v.domainElementToString (pw.get (v)));
    }
    return ans;
  }

  /**
   * Converts bn evidence to logic evidence according to the given literal
   * map.
   * 
   * @param bnEvidence the bn evidence.
   * @param lm the given literal map.
   * @return the logic evidence.
   */
  
  public static HashSet<Integer> logicEvidence (
    Map<Var,Double> bnEvidence, mark.inf.core.LitMap lm) {
    return
      lm.correspondingLogicEvidence (
        mark.reason.core.CoreUtil.doubleEvidenceToIntEvidence (bnEvidence));
  }

  /**
   * Converts logic evidence into bn evidence according to the given literal
   * map.
   * 
   * @param logicEvidence the logic evidence.
   * @param lm the given literal map.
   * @return the bn evidence.
   */
  
  public static HashMap<Var,Double> bnEvidence (
   int[] logicEvidence, mark.inf.core.LitMap lm) {
    return
      intEvidenceToDoubleEvidence (
        lm.correspondingSrcEvidence (logicEvidence));
  }
  
  /**
   * Converts evidence from Map<String,Integer> to Map<Var,Double>.
   * 
   * @param e the input evidnece.
   * @return the output evidence.
   */

  public static HashMap<Var,Double> stringEvidenceToDoubleEvidence (
   Set<Var> vars, Map<String,Integer> e) throws Exception {
    HashMap<Var,Double> ans = new HashMap<Var,Double> ();
    HashMap<String,Var> nameToVar = CoreUtil.varNameToVarMap (vars);
    for (String s : e.keySet ()) {
      Var v = nameToVar.get (s);
      if (v == null) {
        throw new Exception (
          "Variable with name " + s + " does not exist in the network!");
      }
      ans.put (v, (double)e.get (s));
    }
    return ans;
  }

  /**
   * Converts evidence from Map<Var,Double> to Map<String,Integer>.
   * 
   * @param e the input evidnece.
   * @return the output evidence.
   */

  public static HashMap<String,Integer> doubleEvidenceToStringEvidence (
   Map<Var,Double> e) {
    HashMap<String,Integer> ans = new HashMap<String,Integer> ();
    for (Var v : e.keySet ()) {
      ans.put (v.name (), e.get (v).intValue ());
    }
    return ans;
  }

  //=========================================================================
  // Methods that act on scopes.
  //=========================================================================

  /**
   * Returns all variables mentioned in the given potentials.
   *
   * @param distribtuions the given distribtuions.
   * @return the variables.
   */
  
  public static HashSet<Var> variables (Collection<? extends Scope> scopes) {
    HashSet<Var> ans = new HashSet<Var> ();
    for (Scope s : scopes) {
      ans.addAll (s.variables ());
    }
    return ans;
  }

  /**
   * Returns the log base two of the maximum cluster size of the given
   * potentials.
   *
   * @param potentials the potentials.
   * @param order the given order.
   * @return the log base two of the maximum cluster size.
   */
  
  public static double log2MaxClusterSize (
   Set<? extends Scope> scopes, List<Var> order) {
    Converter c = new Converter (variables (scopes));
    return
      mark.inf.eo.Util.logMaxClusterSize (
        c.integerVariableArray (order),
        c.cardinalities (),
        integerMoralGraph (c, moralGraph (scopes)), true);
  }

  /**
   * Creates the moral graph of the given set of scopes.
   *
   * @param scopes the given set of scopes.
   * @return a map from each variable to its one-hop neighbors in the
   *   moral graph.
   */
  
  public static HashMap<Var,HashSet<Var>> moralGraph (
   Set<? extends Scope> scopes) {
    HashMap<Var,HashSet<Var>> ans = new HashMap<Var,HashSet<Var>> ();
    for (Var v : variables (scopes)) {
      ans.put (v, new HashSet<Var> ());
    }
    for (Scope s : scopes) {
      mark.core.util.OrderedSet<Var> vars =  s.variables ();
      for (int j = 0; j < vars.size (); j++) {
        Var v1 = vars.get (j);
        HashSet<Var> neighbors1 = ans.get (v1);
        for (int k = j + 1; k < vars.size (); k++) {
          Var v2 = vars.get (k);
          HashSet<Var> neighbors2 = ans.get (v2);
          neighbors1.add (v2);
          neighbors2.add (v1);
        }
      }
    }
    return ans;
  }

  /**
   * Converts from a set of potentials to an integer moral graph
   * mark.reason.eo.Util.logMaxClusterSize ().
   *
   * @param potentials the potentials.
   */
  
  public static int[][] integerMoralGraph (
   Converter c, HashMap<Var,HashSet<Var>> reasonMoralGraph) {
    int[][] ans = new int[c.length ()][];
    for (int i = 0; i < ans.length; i++) {
      ans[i] =
        c.integerVariableArray (reasonMoralGraph.get (c.reasonVariable (i)));
    }
    return ans;
  }

  /**
   * Invokes the given elimination order engine n times on the given
   * potentials to produce a segmented ordering on the given segmented
   * variables and returns the best order found.  A segmented ordering has
   * variables in the first segment precede those in the second which
   * precende those in the third, etc.  Within a segment, the eoe has
   * discretion to order variables.
   * 
   * @param eoe a set of elimination order engines to apply.
   * @param n parallel to eoe, the number of times to apply the corresponding
   *   engine.
   * @param r a random number generator.
   * @param potentials the given potentials.
   * @param variables the given variables, segmented appropriately.
   * @param eoe the given elimination order engine.
   * @param ss the scoring strategy or null for default.
   * @param scoreToBeat the score to beat.
   * @param c the converter to use; may be null in which case will create
   *   own converter.
   * @return (log2MaxCluster,order,score) or null if no order found that is
   *   better than given scoreToBeat.
   */
  
  public static mark.inf.eo.Result<List<Var>> eliminationOrder (
   mark.inf.eo.Eoe[] eoe, int[] n, Random r,
   Set<? extends Scope> scopes, List<Set<Var>> variables, 
   mark.inf.eo.ScoringStrategy ss, Comparable scoreToBeat,
   boolean combineLast, Converter c) throws Exception {
    
    // Convert the segmented variables into integer format.
    
    if (c == null) {c = new Converter (variables (scopes));}
    int[][] convertedVars = new int[variables.size ()][];
    int index = 0;
    for (Set<Var> varSet : variables) {
      convertedVars[index++] = c.integerVariableArray (varSet);
    }
  
    // Invoke the helper function to do the work and package the result.

    mark.inf.eo.Result<int[]> best = null;
    for (int i = 0; i < eoe.length; i++) {
      mark.inf.eo.Result<int[]> current =
        eoe[i].order (
          r, c.cardinalities (),
          mark.reason.core.CoreUtil.integerMoralGraph (c, moralGraph (scopes)),
          convertedVars, n[i], null, ss, scoreToBeat, combineLast);
      if (
       best == null ||
       (current != null && current.fScore.compareTo (best.fScore) == -1)) {
        best = current;
      }
    }
    return
      best == null ? null :
      new mark.inf.eo.Result<List<Var>> (
        best, c.reasonVariableList (best.fOrder));
  
  }

  /**
   * Returns a map from variable to set of the given potentials containing
   * the variable.
   * 
   * @param distriubutions the given potentials.
   * @return the map.
   */
  
  public static <T extends Scope> HashMap<Var,HashSet<T>> varToScopeMap (
   Set<T> scopes) {
    HashMap<Var,HashSet<T>> ans = new HashMap<Var,HashSet<T>> ();
    for (Var v : variables (scopes)) {
      ans.put(v, new HashSet<T> ());
    }
    for (T s : scopes) {
      for (Var v : s.variables()) {
        ans.get(v).add(s);
      }
    }
    return ans;
  }

}
