package mark.reason.core;
import java.util.*;

/**
 * Utilities for instantiations.
 * <p>
 * There are four instantiation types:
 * <ol>
 * <li> The int instantiation is the position of the instantiation in the
 *      instantiation list.
 * <li> The int[] instantiation is an int[] from each variable index to
 *      domain object index.
 * <li> A map instantiation is a Map from each variable to
 *      java.lang.Double.
 * <li> The Double[] instantiation is a Double[] from each variable index to
 *      java.lang.Double.
 * </ol>
 * Currently, support for the Double[] instantiation is not implemented,
 * because I have not yet had a need for it.
 *
 * @author Mark Chavira
 */

public class Instantiation {

  /**
   * Returns a cardinality array for the given variable list.
   *
   * @param l the given variable list.
   * @return the cardinality array.
   */
  
  public static int[] cardinalityArray (Var[] l) {
    int[] ans = new int[l.length];
    for (int i = l.length - 1; i >= 0; i--) {
      ans[i] = l[i].domainSize ();
    }
    return ans;
  }

  /**
   * Returns a cardinality array for the given variable list.
   *
   * @param l the given variable list.
   * @return the cardinality array.
   */
  
  public static int[] cardinalityArray (List<Var> l) {
    return cardinalityArray (l.toArray (new Var[l.size ()]));
  }

  /**
   * For the given variable list, returns the number of instantiations.
   *
   * @param l the given variable list.
   * @return the number of instantiations.
   */
  
  public static int numInstantiations (Var[] l) {
    int ans = 1;
    for (Var v : l) {
      ans *= v.domainSize ();
    }
    return ans;
  }
  
  /**
   * For the given variable list, returns the log base 2 of the number of
   * instantiations.
   *
   * @return the log base 2 of the number of instantiations.
   */
  
  public static double logNumInstantiations (Var[] l) {
    double ans = 0.0;
    for (Var v : l) {
      ans += mark.core.math.Util.log2 (v.domainSize ());
    }
    return ans;
  }

  /**
   * For the given variable list, computes the position in the instantiation
   * list of the first instantition compatible with the given evidence.
   *
   * @param l the given variable list.
   * @param e the given evidence.
   * @return the position.
   */

  public static int position (Var[] l, int[] e) {
    int ans = 0;
    if (e != null) {
      int product = 1;
      for (int i = l.length - 1; i >= 0; i--) {
        if (e[i] != -1) {
          ans += e[i] * product;
        }
        product *= l[i].domainSize ();
      }
    }
    return ans;
  }

  /**
   * For the given variable list, computes the position in the instantiation
   * list of the first instantition compatible with the given evidence.
   *
   * @param l the given variable list.
   * @param e the given evidence.
   * @return the position.
   */
  
  public static int position (Var[] l, Map<Var,Double> inst) {
    int ans = 0;
    if (inst != null) {
      int product = 1;
      for (int i = l.length - 1; i >= 0; i--) {
        Double d = inst.get (l[i]);
        if (d != null) {
          ans += (int)(double)d * product;
        }
        product *= l[i].domainSize ();
      }
    }
    return ans;
  }
  
  /**
   * For the given variable list, computes the instantiation located at the
   * given position in the instantiation list.
   *
   * @param l the given variable list.
   * @param p the given position.
   * @return the instantiation.
   */
  
  public static int[] intArrayInstantiation (Var[] l, int p) {
    int[] ans = new int[l.length];
    for (int i = l.length - 1; i >= 0; i--) {
      int ds = l[i].domainSize ();
      ans[i] = p % ds;
      p /= ds;
    }
    return ans;
  }

  /**
   * For the given variable list, computes the instantiation located at the
   * given position in the instantiation list.
   *
   * @param l the given variable list.
   * @param p the given position.
   * @return the instantiation.
   */
  
  public static HashMap<Var,Double> mapInstantiation (Var[] l, int p) {
    HashMap ans = new HashMap ();
    for (int i = l.length - 1; i >= 0; i--) {
      int ds = l[i].domainSize ();
      ans.put (l[i], (double)(p % ds));
      p /= ds;
    }
    return ans;
  }

  /**
   * For the given variable list, converts the given evidence to int[]
   * format.
   *
   * @param l the given variable list.
   * @param e the given evidence.
   * @return the converted evidence.
   */
  
  public static int[] convertEvidenceToIntArray (Var[] l, Map<Var,Double> e) {
    int[] ans = new int[l.length];
    if (e != null) {
      for (int i = l.length - 1; i >= 0; i--) {
        Double d = e.get (l[i]);
        ans[i] = d == null ? -1 : (int)(double)d;
      }
    }
    return ans;
  }

  /**
   * For the given variable list, converts the given evidence to map format.
   *
   * @param l the given variable list.
   * @param e the given evidence.
   * @return the converted evidence.
   */
   
  public static HashMap<Var,Double> convertEvidenceToMap (Var[] l, int[] e) {
    HashMap<Var,Double> ans = new HashMap<Var,Double> ();
    if (e != null) {
      for (int i = l.length - 1; i >= 0; i--) {
        if (e [i] != -1) {
          ans.put (l[i], (double)(e[i] + l[i].minDomainElement ()));
        }
      }
    }
    return ans;
  }
  
  /**
   * Converts a partial world into a string.
   * 
   * @param the given partial world.
   * @return the string.
   */
  
  public static String partialWorldToString (Map<Var,Double> pw) {
    StringBuffer ans = new StringBuffer ();
    ArrayList<Var> order = new ArrayList<Var> (pw.keySet ());
    Collections.sort (order, Var.NAME_COMPARATOR);
    ans.append ("{");
    for (int i = 0; i < order.size (); i++) {
      if (i > 0) {ans.append (",");}
      Var v = order.get (i);
      ans.append (v + "=" + v.domainElementToString (pw.get (v)));
    }
    ans.append ("}");
    return ans.toString ();
  }

}
