package mark.reason.core;
import java.util.*;

/**
 * An object which defines an ordered set of variables, which knows how to
 * convert among different types of instantiations of these variables, and
 * which can return iterators.
 * <p>
 * There are some redundant fields throughout this class, because it is very
 * important that we squeeze every ounce of efficiency out of this code: the
 * efficiency of probabilistic reasoning relies heavily on this code.
 * <p>
 * This class conforms to the immutable design pattern.
 *
 * @author Mark Chavira
 */

public class Scope {

  /**
   * An unmodifiable list of variables to return to clients.
   */

  protected mark.core.util.OrderedSet<Var> fVariableList;

  /**
   * Convenient access to the variables.  fVariables is redundant, since the
   * information is already in fVariableList.
   */
    
  protected Var[] fVariables;

  /**
   * The cardinalities of the variables.  This field is redundant, since we
   * can get the information from the variables themselves.  Profiling
   * revealed this field to be very useful.  My code was running about 30%
   * slower than samiam code prior to adding this field, and following, it
   * was neck-and-neck with samiam code.
   */

  protected int[] fCardinalities;

  /**
   * The total number of instantiations.
   */

  protected int fNumInstantiations;

  /**
   * Code common to the constructors.
   *
   * @param vars the variables.
   */
  
  protected void init (Var[] vars) {
    fVariables = vars;
    fVariableList = new mark.core.util.OrderedSet<Var> (vars);
    fCardinalities = Instantiation.cardinalityArray (vars);
    fNumInstantiations = Instantiation.numInstantiations (vars);
  }
  
  /**
   * An empty scope.
   */
  
  public static final Scope EMPTY_SCOPE = new Scope ();
  
  /**
   * Initializes the scope.
   *
   * @param vars the variables.
   */
  
  // Java array clone is very slow, and this method is critical, so clone
  // by hand.
  
  private Var[] cl (Var[] a) {
    Var[] ans = new Var[a.length];
    System.arraycopy (a, 0, ans, 0, a.length);
    return ans;
  }

  /**
   * Initializes the scope.
   * 
   * @param vars the variables.
   */
  
  public Scope (Var... vars) {
    init (cl (vars));
  }
  
  /**
   * Initializes the scope.
   *
   * @param vars the variables.
   */
  
  public Scope (List<Var> vars) {
    init (vars.toArray (new Var[vars.size ()]));
  }

  /**
   * Initializes the scope.
   *
   */
  
  public Scope (Scope x) {
    fVariableList = x.fVariableList;
    fVariables = x.fVariables;
    fCardinalities = x.fCardinalities;
    fNumInstantiations = x.fNumInstantiations;
  }

  /**
   * Returns a single iterator over instantiations.
   *
   * @param e the evidence.
   * @return the iterator.
   */
  
  public mark.inf.core.SingleIterator si (Map<Var,Double> e) {
    return
      new mark.inf.core.SingleIterator (
        fCardinalities, e == null ? null : convertEvidenceToIntArray (e));
  }

  /**
   * Returns a single iterator that also iterates through map instantiations.
   *
   * @param e the evidence.
   * @return the iterator.
   */
  
  public SingleIterator siMap (Map<Var,Double> e) {
    return new SingleIterator (fVariables, e);
  }
  
  /**
   * Returns a dual iterator over this and the given scopes.
   *
   * @param x the given scope.
   * @return the iterator.
   */
  
  public mark.inf.core.DualIterator di (
   Scope x, Map<Var,Double> e) {
    return
      new mark.inf.core.DualIterator (
        fCardinalities, x.fCardinalities,
        e == null ? null : x.convertEvidenceToIntArray (e),
        mark.inf.core.Instantiation.dualMap (
          fVariableList, x.fVariableList));
  }

  /**
   * Returns a multi iterator over this and the given scopes.
   * 
   * @param x the given scopes.
   * @return the iterator.
   */
  
  public mark.inf.core.MultiIterator mi (
   Scope[] scopes, Map<Var,Double> e) {
    int[][] lCardinalities = new int[scopes.length][];
    int[][] lEvidence = e == null ? null : new int[scopes.length][];
    int[][] lMaps = new int[scopes.length][];
    for (int i = 0; i < lCardinalities.length; i++) {
      lCardinalities[i] = scopes[i].fCardinalities;
      if (e != null) {
        lEvidence[i] = scopes[i].convertEvidenceToIntArray (e);
      }
      lMaps[i] =
        mark.inf.core.Instantiation.dualMap (
          fVariableList, scopes[i].fVariableList);
    }
    return
      new mark.inf.core.MultiIterator (
        fCardinalities, lCardinalities, lEvidence, lMaps);
  }
  
  /**
   * Returns an unmodifiable list of the variables.
   */

  public final mark.core.util.OrderedSet<Var> variables () {
    return fVariableList;
  }
  
  /**
   * Returns the last variable.
   */

  public final Var lastVariable () {
    return fVariables[fVariables.length - 1];
  }

  /**
   * Returns the number of instantiations of the variables.
   *
   * @return the number of instantiations.
   */
  
  public final int numInstantiations () {
    return fNumInstantiations;
  }

  /**
   * Returns the correct number of instantiations of the variables, even
   * if this number overflows an int.
   * 
   * @return the number of instantiations.
   */
  
  public final java.math.BigInteger numBigInstantiations () {
    java.math.BigInteger ans = java.math.BigInteger.ONE;
    for (Var v : fVariables) {
      ans = ans.multiply (java.math.BigInteger.valueOf (v.domainSize ()));
    }
    return ans;
  }
  
  /**
   * Returns the log base 2 of the number of instantiations.
   * 
   * @return the log base 2 of the number of instantiations.
   */
  
  public final double logNumInstantiations () {
    return Instantiation.logNumInstantiations (fVariables);
  }
  
  /**
   * Computes the position in the instantiation list of the first
   * instantition compatible with the given evidence.
   *
   * @param e the given evidence.
   * @return the position.
   */
  
  public final int position (int[] e) {
    return mark.inf.core.Instantiation.position (fCardinalities, e);
  } 

  /**
   * Computes the position in the instantiation list of the first
   * instantition compatible with the given evidence.
   *
   * @param e the given evidence.
   * @return the position.
   */
  
  public final int position (Map<Var,Double> e) {
    return Instantiation.position (fVariables, e);
  }

  /**
   * Computes the instantiation located at the given position in the
   * instantiation list.
   *
   * @param p the given position.
   * @return the instantiation.
   */
  
  public final int[] intArrayInstantiation (int p) {
    return mark.inf.core.Instantiation.instantiation (fCardinalities, p);
  }

  /**
   * Computes the instantiation located at the given position in the
   * instantiation list.
   *
   * @param p the given position.
   * @return the instantiation.
   */
  
  public final HashMap<Var,Double> mapInstantiation (int p) {
    return Instantiation.mapInstantiation (fVariables, p);
  }

  /**
   * Converts the given evidence to int[] format.
   *
   * @param e the given evidence.
   * @return the converted evidence.
   */
  
  public final int[] convertEvidenceToIntArray (Map<Var,Double> e) {
    return Instantiation.convertEvidenceToIntArray (fVariables, e);
  }
  

  /**
   * Converts the given evidence to map format.
   *
   * @param e the given evidence.
   * @return the converted evidence.
   */
  
  public final HashMap<Var,Double> convertEvidenceToMap (int[] e) {
    return Instantiation.convertEvidenceToMap (fVariables, e);
  }

  /**
   * Returns a clone of the cardinality array.
   *
   * @return the array.
   */
  
  public int[] cardinalities () {
    return mark.core.util.Util.clone (fCardinalities);
  }

  /**
   * Returns whether all variables are binary.
   *
   * @return whether all variables are binary.
   */
  
  public boolean binary () {
    for (int i = 0; i < variables ().size (); i++) {
      if (variables ().get (i).domainSize () != 2) {
        return false;
      }
    }
    return true;
  }

  /**
   * Returns a scope that can be used for the result of combining this scope
   * with the given other scopes
   * 
   * @param others the given other scopes.
   * @return the scope.
   */
  
  public static <T extends Scope> Scope staticCombineScope (Set<T> scopes) {
    HashSet<Var> vars = new HashSet ();
    for (Scope s : scopes) {
      vars.addAll (s.variables ());
    }
    ArrayList<Var> ordered = new ArrayList<Var> (vars);
    Collections.sort (ordered, Var.NAME_COMPARATOR);
    return new Scope (ordered);
  }

  /**
   * Returns a scope that can be used for the result of combining this scope
   * with the given other scopes.
   * 
   * @param others the given other scopes.
   * @return the scope.
   */
  
  public <T extends Scope> Scope combineScope (Set<T> others) {
    HashSet<Var> vars = new HashSet (variables ());
    for (Scope s : others) {
      vars.addAll (s.variables ());
    }
    ArrayList<Var> ordered = new ArrayList<Var> (vars);
    Collections.sort (ordered, Var.NAME_COMPARATOR);
    return new Scope (ordered);
  }
  
  /**
   * Returns a scope that can be used for the result of combining this scope
   * with the given other scopes making sure the give variables are last in
   * the order.
   * 
   * @param others the given other scopes.
   * @param vars the given variables.
   * @return the scope.
   */

  public <T extends Scope> Scope combineScope (
   Set<T> others, Set<Var> vars) {
    HashSet<Var> firstVars = new HashSet<Var> (variables ());
    for (Scope s : others) {
      firstVars.addAll (s.variables ());
    }
    HashSet<Var> secondVars = new HashSet<Var> (firstVars);
    secondVars.retainAll (vars);
    firstVars.removeAll (vars);
    ArrayList<Var> firstList = new ArrayList<Var> (firstVars);
    ArrayList<Var> secondList = new ArrayList<Var> (secondVars);
    Collections.sort (firstList, Var.NAME_COMPARATOR);
    Collections.sort (secondList, Var.NAME_COMPARATOR);
    firstList.addAll (secondList);
    return new Scope (firstList);
  }

  /**
   * Returns a scope that can be used for the result of combining this scope
   * with the given other scopes and then forgetting the given variables.
   * 
   * @param others the given other scopes.
   * @param forgetVars the given variables.
   * @return the scope.
   */
  
  public <T extends Scope> Scope combineAndForgetScope (
   Set<T> others, Set<Var> forgetVars) {
    HashSet<Var> vars = new HashSet<Var> (variables ());
    for (Scope s : others) {
      vars.addAll (s.variables ());
    }
    vars.removeAll (forgetVars);
    ArrayList<Var> ordered = new ArrayList<Var> (vars);
    Collections.sort (ordered, Var.NAME_COMPARATOR);
    return new Scope (ordered);
  }
    
  
  /**
   * Returns a scope that can be used for the result of forgetting the given
   * variables.
   * 
   * @param forgetVars the given variables.
   * @return the scope.
   */
  
  public Scope forgetScope (Set<Var> forgetVars) {
    ArrayList<Var> ordered = new ArrayList<Var> ();
    for (Var v : fVariables) {
      if (!forgetVars.contains (v)) {
        ordered.add (v);
      }
    }
    //Collections.sort (ordered, Var.NAME_COMPARATOR);
    return new Scope (ordered);
  }
  
  /**
   * Returns a scope suitable for replacing variables according to the given
   * map.
   * 
   * @param variableMap the given map.
   * @return the scope.
   */
  
  public final Scope replaceScope (Map<Var,Var> variableMap) {
    Var[] newVariables = new Var[fVariables.length];
    for (int i = 0; i < fVariables.length; i++) {
      newVariables[i] = variableMap.get (fVariables[i]);
      if (newVariables[i] == null) {newVariables[i] = fVariables[i];}
    }
    return new Scope (newVariables);
  }
  
  
  /**
   * Returns a scope suitable for replacing variables according to the given
   * map.
   * 
   * @param variableMap the given map.
   * @return the scope.
   */
  
  public Scope replaceVariables (Map<Var,Var> variableMap)
  throws Exception {
    return replaceScope (variableMap);
  }

}
