package mark.reason.core;
import java.util.*;

/**
 * A base class for a variable.  A variable has a name which should be unique
 * among variables that will be used together and a domain which is a subset
 * of the set of real-numbers.  The domain satisifes the following
 * requirements:
 * <ul>
 * <li> may include either or both of -inf and +inf but not other strange
 *      numbers like NaN.
 * <li> may contain a finite or infinite number of elements.
 * <li> is guaranteed to contain some non-infinite number.
 * <li> is guaranteed to contain at least two elements.
 * </ul>
 * In addition, each subclass:
 * <ul>
 * <li> defines a string representation of each domain element.
 * <li> defines the notion of a descriptor, which is an object that
 *      represents a set of domain elements.  A descriptor need not be able
 *      to describe any set, but only those sets that are useful.  A
 *      descriptor is essentially a "nice" way of specifying commonly used
 *      sets of domain elements.  Descriptors are to be used to label edges
 *      in RODDs and possibly for other purposes.
 * <li> optionally defines the notion of a partition, which is a collection
 *      of descriptors that partition the variable's domain.
 * </ul>
 * <p>
 * This class conforms to the immutable design pattern.
 *
 * @author Mark Chavira
 */

public abstract class Var {
 
  //=========================================================================
  // Fields:
  //=========================================================================

  /**
   * The variable's name.
   */
  
  private String fName;
  
  /**
   * A comparator that compares variables by name.
   */
  
  public static Comparator<Var> NAME_COMPARATOR =
    new Comparator<Var> () {
      public int compare (Var lhs, Var rhs) {
        return lhs.name ().compareTo (rhs.name ());
      }
    };
    
  /**
   * A comparator that compares variabls by domain size.
   */

  public static Comparator<Var> DOMAIN_SIZE_COMPARATOR =
    new Comparator<Var> () {
      public int compare (Var v1, Var v2) {
        if (v1.domainSize () < v2.domainSize ()) {return -1;}
        if (v1.domainSize () > v2.domainSize ()) {return +1;}
        return 0;
      }
    };

  //=========================================================================
  // Fulfillments.
  //=========================================================================

  /**
   * @see Object
   */
    
  public String toString () {
    return fName;
  }
    
  //=========================================================================
  // Regular Methods:
  //=========================================================================

  /**
   * Initializes the variable.
   *
   * @param name the variable's name.
   */
  
  protected Var (String name) {
    fName = name;
  }

  /**
   * Returns the variable's name.
   *
   * @return the variable's name.
   */

  public String name () {
    return fName;
  }

  //=========================================================================
  // Abstract Methods:
  //=========================================================================

  /**
   * Returns a string representing the definition of the varible.
   * 
   * @return the string.
   */
  
  public abstract String definitionString ();
  
  /**
   * Returns a variable of the same type and domain with the given name.
   *
   * @param newName the new name.
   * @return the new variable.
   */
  
  public abstract Var clone (String newName);

  /**
   * Returns whether the given variable has the same domain.  The given
   * variable must be of the same concrete class.
   *
   * @param v the given variable.
   * @return whether the domains are equal.
   */
  
  public abstract boolean sameDomain (Var v);

  /**
   * For variable types that cannot have more domain elements than an int
   * can store, returns the number of values in the domain; for other
   * variables, returns -1.
   *
   * @return the domain size or -1.
   */
  
  public abstract int domainSize ();

  /**
   * Returns the number of values in the domain or -1 if the domain is
   * infinite.
   *
   * @return the domain size or -1.
   */
  
  public abstract java.math.BigInteger domainSizeAsBigInt ();

  /**
   * Returns the smallest domain element, which could be -inf.
   *
   * @return the element.
   */
  
  public abstract double minDomainElement ();

  /**
   * Retuns a domain element guaranteed not to be -inf or +inf.
   *
   * @return the element.
   */
  
  public abstract double someDomainElement ();

  /**
   * Returns the largest domain element, which could be +inf.
   *
   * @return the element.
   */
  
  public abstract double maxDomainElement ();

  /**
   * Returns the string corresponding to the given domain element.
   *
   * @param v the given domain element.
   * @return the string.
   */
  
  public abstract String domainElementToString (double v);

  /**
   * Returns the domain element corresponding to the given string.
   *
   * @param s the given string.
   * @return the domain element.
   */

  public abstract double stringToDomainElement (String s) throws Exception;

  /**
   * Given a domain element, returns a string that asserts that the variable
   * be equal to the element.
   *
   * @param v the element.
   * @return the string.
   */
  
  public abstract String assertEqualString (double v);

  /**
   * Converts a descriptor to a string.
   *
   * @param d the descriptor.
   * @return the string.
   */
   
  public abstract String descriptorToString (Object d);
  
  /**
   * Given a descriptor, returns a string that asserts that the variable
   * be a member of the elements described by the descriptor.
   *
   * @param d the descriptor.
   * @return the string.
   */

  public abstract String assertMemberString (Object d);

  /**
   * For variable types that cannot have more domain elements than an int
   * can store, returns the names of the domain elements as an unmodifiable
   * list; otherwise returns null.
   *
   * @return the domain elements.
   */
  
  public abstract mark.core.util.OrderedSet<String> domainNames ();

}
