package mark.reason.core;

/**
 * A binary variable.  The domain is {0.0, 1.0}.  A descriptor describes one
 * value and is "F" for 0 and "T" for 1.  There is no partition type, because
 * the partition os always {{0},{1}}.
 * <p>
 * This class conforms to the immutable design pattern.
 *
 * @author Mark Chavira
 */

public class VarBinary extends Var {

  /**
   * The name of the first value.
   */

  public static final String FALSE_NAME = "F";
  
  /**
   * The name of the second value.
   */
  public static final String TRUE_NAME = "T";
  
  /**
   * The domain names as an ordered set.
   */
   
  public static final mark.core.util.OrderedSet<String> DOMAIN_NAMES =
    new mark.core.util.OrderedSet<String> (FALSE_NAME, TRUE_NAME);

  //=========================================================================
  // Constructors:
  //=========================================================================
  
  /**
   * Initializes the variable.
   *
   * @param name the variable's name.
   */

  public VarBinary (String name) {
    super (name);
  }

  //=========================================================================
  // Static:
  //=========================================================================

  /**
   * Returns the domain element corresponding to the given string.
   *
   * @param double s the given string.
   * @return the domain element.
   */
  
  public static double stringToDomainElementStatic (String s)
   throws Exception {
    if (s.equals (FALSE_NAME)) {
      return 0.0;
    }
    if (s.equals (TRUE_NAME)) {
      return 1.0;
    }
    throw new Exception (s + " is not a domain element.");
  }

  /**
   * Retruns the string corresponding to the given domain element.
   *
   * @praram v the given domain element.
   * @return the string.
   */
  
  public static String domainElementToStringStatic (double v) {
    return (v == 0.0) ? FALSE_NAME : TRUE_NAME;
  }

  //=========================================================================
  // Fulfillments:
  //=========================================================================
  
  /**
   * @see Var
   */
  
  public String definitionString () {
    return name () + " : binary;";
  }

  /**
   * @see Var
   */

  public Var clone (String newName) {
    return new VarBinary (newName);
  }

  /**
   * @see Var
   */

  public boolean sameDomain (Var v) {
    return true;
  }

  /**
   * @see Var
   */

  public int domainSize () {
    return 2;
  }

  /**
   * @see Var
   */

  public java.math.BigInteger domainSizeAsBigInt () {
    return java.math.BigInteger.valueOf (2);
  }

  /**
   * @see Var
   */

  public double minDomainElement () {
    return 0.0;
  }

  /**
   * @see Var
   */

  public double someDomainElement () {
    return 0.0;
  }

  /**
   * @see Var
   */

  public double maxDomainElement () {
    return 1.0;
  }

  /**
   * @see Var
   */

  public String domainElementToString (double v) {
    return domainElementToStringStatic (v);
  }

  /**
   * @see Var
   */
  
  public double stringToDomainElement (String s) throws Exception {
    return stringToDomainElementStatic (s);
  }

  /**
   * @see Var
   */
  
  public String assertEqualString (double e) {
    return (e == 0.0) ? ("~" + name ()) : name ();
  }

  /**
   * @see Var
   */

  public String descriptorToString (Object d) {
    return (String)d;
  }

  /**
   * @see Var
   */

  public String assertMemberString (Object d) {
    return d.equals (FALSE_NAME) ? ("~" + name ()) : name ();
  }

  /**
   * @see Var
   */
  
  public mark.core.util.OrderedSet<String> domainNames () {
    return DOMAIN_NAMES;
  }

}
