package mark.reason.core;
import java.util.*;

import mark.inf.gate.Gate;

/**
 * A discrete variable.  The domain is {0.0, 1.0, ..., n.0}, where n is the
 * finite size of the domain minus one.  A descriptor describes one value and
 * is any unique string.  There is no partition type, because the partition
 * is alwyas {{0},{1},...,{n}}.
 * <p>
 * This class conforms to the immutable design pattern.
 *
 * @author Mark Chavira
 */

public class VarDiscrete extends Var {

  //=========================================================================
  // Fields:
  //=========================================================================
  
  /**
   * Names for the constructor that creates a variable with given domain
   * size.  Starts out null and is allocated when needed.
   */
  
  private static ArrayList<String> fNames = null;
  
  /**
   * The names of domain elements.
   */
  
  private mark.core.util.OrderedSet<String> fDomainNames;
  
  //=========================================================================
  // Constructors:
  //=========================================================================

  /**
   * Initializes the variable.
   *
   * @param name the variable's name.
   * @param domainNames the names of the domain elements.
   */
  
  public VarDiscrete (
   String name, mark.core.util.OrderedSet<String> domainNames) {
    super (name);
    fDomainNames = domainNames;
  }

  /**
   * Initializes the variable.
   *
   * @param name the variable's name.
   * @param domainNames the names of the domain elements.
   */
  
  public VarDiscrete (String name, Collection<String> domainNames) {
    this (name, new mark.core.util.OrderedSet<String> (domainNames));
  }

  /**
   * Initializes the variable.
   *
   * @param name the variable's name.
   * @param domainNames the names of the domain elements.
   */
  
  public VarDiscrete (String name, String... domainNames) {
    this (name, new mark.core.util.OrderedSet<String> (domainNames));
  }

  /**
   * Initializes the variable with the given name and domain size, using as
   * value names "s0", "s1", "s2", ...
   * 
   * @param name the variable's name.
   * @param domainSize the given domain size.
   */
  
  public VarDiscrete (String name, int domainSize) {
    super (name);
    if (fNames == null) {
      fNames = new ArrayList<String> (domainSize);
    }
    while (fNames.size () < domainSize) {
      fNames.add (new String ("s" + Integer.toString (fNames.size ())));
    }
    ArrayList<String> domainNames = new ArrayList<String> ();
    for (int i = 0; i < domainSize; i++) {
      domainNames.add (fNames.get (i));
    }
    fDomainNames = new mark.core.util.OrderedSet<String> (domainNames); 
  }
  
  //=========================================================================
  // Fulfillments:
  //=========================================================================

  /**
   * @see Var
   */
  
  public String definitionString () {
    return
      name () + " : {" +
      mark.core.text.Util.join (",", fDomainNames.toArray ()) + "};";
  }

  /**
   * @see Var
   */

  public Var clone (String newName) {
    return new VarDiscrete (newName, fDomainNames);
  }

  /**
   * @see Var
   */

  public boolean sameDomain (Var v) {
    return fDomainNames.equals (((VarDiscrete)v).fDomainNames);
  }

  /**
   * @see Var
   */

  public int domainSize () {
    return domainNames ().size ();
  }

  /**
   * @see Var
   */

  public java.math.BigInteger domainSizeAsBigInt () {
    return java.math.BigInteger.valueOf (domainSize ());
  }

  /**
   * @see Var
   */

  public double minDomainElement () {
    return 0.0;
  }

  /**
   * @see Var
   */

  public double someDomainElement () {
    return 0.0;
  }

  /**
   * @see Var
   */

  public double maxDomainElement () {
    return domainSize () - 1.0;
  }

  /**
   * @see Var
   */

  public String domainElementToString (double v) {
    return (String)fDomainNames.get ((int)v);
  }

  /**
   * @see Var
   */

  public double stringToDomainElement (String s) throws Exception {
    double ans = (double)fDomainNames.indexOf (s);
    if (ans < 0) {
      throw new Exception (
        s + " is not in the domain of " + name () + "; the domain is " +
        fDomainNames);
    }
    return ans;
  }

  /**
   * @see Var
   */

  public String assertEqualString (double v) {
    return name () + "=" + domainElementToString (v);
  }  

  /**
   * @see Var
   */

  public String descriptorToString (Object d) {
    return (String)d;
  }

  /**
   * @see Var
   */

  public String assertMemberString (Object d) {
    return name () + "=" + d;
  }

  /**
   * @see Var
   */
  
  public mark.core.util.OrderedSet<String> domainNames () {
    return fDomainNames;
  }

  /**
   * Creates a health variable with the given health states for the gate
   * variable with the given name.  Throws an exception if the health set is
   * {healthy}.
   *
   * @param n the given name.
   * @param healthSet specifies a health set other than {healthy}.
   * @return the health variable.
   */
  
  public static VarDiscrete rep2CreateHealthVariable (
   String n, mark.inf.gate.GateHealth healthSet) throws Exception {
    if (healthSet == mark.inf.gate.GateHealth.ALWAYS_HEALTHY) {
      throw new Exception (
        "Cannot create health variable when there are no failure modes!");
    }
    return new VarDiscrete (Gate.healthVarName (n), healthSet.domain ());
  }

}
