package mark.reason.core;
import java.util.*;

/**
 * Maintains a ordered collection of variables.  Allows one to map a variable
 * in constant time among object, index, and name.
 * <p>
 * This class conforms to the immutable design pattern.
 *
 * @author Mark Chavira
 */

public final class VariableTable {
  
  //=========================================================================
  // Fields:
  //
  // We map vaiables, names, and indices.
  //=========================================================================

  private mark.core.util.OrderedSet<String> fNames;
  private mark.core.util.OrderedSet<Var> fVariables;

  private final Comparator<Var> fLevelComparator =
    new Comparator<Var> () {
      public int compare (Var v1, Var v2) {
        int i1 = mapToIndex (v1);
        int i2 = mapToIndex (v2);
        return i1 < i2 ? -1 : i1 > i2 ? +1 : 0;
      }
    };

  //=========================================================================
  // Constructors:
  //=========================================================================

  /**
   * Initializes the variable table.
   *
   * @param variables the variables.
   */
  
  public VariableTable (Collection variables) throws Exception {
    this ((Var[])variables.toArray (new Var[variables.size ()]));
  }

  /**
   * Initializes the variable table.
   *
   * @param variables the variables.
   */
  
  public VariableTable (Var[] variables) throws Exception {
    fVariables = new mark.core.util.OrderedSet<Var> (variables);
    String[] names = new String[variables.length];
    for (int i = 0; i < names.length; i++) {
      names[i] = variables[i].name ();
    }
    fNames = new mark.core.util.OrderedSet<String> (names);
  }

  //=========================================================================
  // Regular Methods:
  //=========================================================================

  /**
   * @see java.lang.Object
   */
  
  public boolean equals (Object o) {
    if (this == o) {
      return true;
    }
    if (!(o instanceof VariableTable)) {
      return false;
    }
    VariableTable vt = (VariableTable)o;
    if (length () != vt.length ()) {
      return false;
    }
    for (int i = 0; i < length (); i++) {
      if (mapToVariable (i) != vt.mapToVariable (i)) {
        return false;
      }
    }
    return true;
  }

  /**
   * @see java.lang.Object
   */
   
  public final String toString () {
    StringBuffer ans = new StringBuffer ();
    for (int i = 0; i < fVariables.size (); i++) {
      if (i > 0) {ans.append ("\n");}
      ans.append (fVariables.get (i).definitionString());
    }
    return ans.toString ();
  }

  /**
   * Looks up an index by name.
   *
   * @param name the variable's name.
   * @return the variable's index or -1 if the variable does not exist.
   */
  
  public final int mapToIndex (String name) {
    return fNames.indexOf (name);
  }

  /**
   * Looks up an index by variable.
   *
   * @param v the variable.
   * @return the variable's index.
   */
  
  public final int mapToIndex (Var v) {
    return fVariables.indexOf (v);
  }

  /**
   * Looks up a variable by name.
   *
   * @param name the variable's name.
   * @return the variable or null if it does not exist.
   */
  
  public final Var mapToVariable (String name) {
    int index = fNames.indexOf (name);
    return index == -1 ? null : fVariables.get (index);
  }

  /**
   * Looks up a variable by index.
   *
   * @param index the variable's index.
   * @return the variable or null if it does not exist.
   */

  public final Var mapToVariable (int index) {
    return fVariables.get (index);
  }

  /**
   * Looks up a name by index.
   *
   * @param index the variable's index.
   * @return the variable's name.
   */

  public final String mapToName (int index) {
    return fNames.get (index);
  }

  /**
   * Looks up a name by variable.
   *
   * @param v the variable.
   * @return the variable's name.
   */
  
  public final String mapToName (Var v) {
    return v.name ();
  }

  /**
   * Returns the number of elements in the variable table.
   *
   * @return the number of elements.
   */
  
  public final int length () {
    return fVariables.size ();
  }
  
  /**
   * Returns a new variable table that reorders the variables.
   *
   * @param order variable with index i now has index order[i].
   */

  public final VariableTable reorder (int[] order) throws Exception {
    return new VariableTable (variableArray (order));
  }

  /**
   * Returns a new variable table that replaces the variables according to
   * the given map.
   *
   * @param map the map.
   * @return the new variable table.
   */
  
  public final VariableTable rename (
   Map<Var,Var> map) throws Exception {
    Var[] newVars = new Var[length ()];
    for (int i = 0; i < newVars.length; i++) {
      newVars[i] = map.get (fVariables.get (i));
    }
    return new VariableTable (newVars);
  }

  /**
   * Returns a comparator that sorts variables by level.
   *
   * @return the comparator.
   */

  public final Comparator levelComparator () {
    return fLevelComparator;
  }

  /**
   * Given an array of indices, returns a parallel array of variables.
   *
   * @param indices the indices.
   * @return the variables.
   */
  
  public final Var[] variableArray (int[] indices) {
    Var[] ans = new Var[indices.length];
    for (int i = 0; i < ans.length; i++) {
      ans[i] = mapToVariable (indices[i]);
    }
    return ans;
  }

  /**
   * Returns the variables as an unmodifiable set.
   *
   * @return the variables.
   */
  
  public final Set variables () {
    return fVariables;
  }

  /**
   * Returns whether for every variable in the table there is a corresponding
   * variable in the given table that is of the same type, has the same name,
   * and has the same domain.
   *
   * @param vt the given table.
   */
  
  public final boolean subsetSame (VariableTable vt) {
    if (length () > vt.length ()) {
      return false;
    }
    for (int i = 0; i < length (); i++) {
      Var v1 = mapToVariable (i);
      Var v2 = vt.mapToVariable (v1.name ());
      if (
       v2 == null || v1.getClass () != v2.getClass () ||
       !v1.sameDomain (v2)) {
        return false;
      }
    }
    return true;
  }
  
  /**
   * Returns a variable table over the given variables sorted by variable
   * name.
   *
   * @param s the given variables.
   * @return the variable table.
   */

  public static VariableTable sortedVariableTable (Set s)
   throws Exception {
    ArrayList varList = new ArrayList (s);
    Collections.sort (varList, Var.NAME_COMPARATOR);
    return new VariableTable (varList);
  }

}
