package mark.reason.inf;
import java.util.*;
import mark.reason.core.*;

/**
 * An interface for an object that represents a propositional logic theory
 * and performs inference.  A class that implements this interface will
 * implement a subset of the operations defined here, throw Exception for the
 * rest, and clearly document which are implemented.  An algorithm that
 * requires a theory specifies the operations it requires.  In general, a
 * theory is compatible only with theories of the same concrete class.
 * That is, one cannot mix theories of two different classes in operations
 * like conjoin, etc.
 *
 * @author Mark Chavira
 */

public interface Theory {

  /**
   * Returns the variables mentioned in the theory.  This set is often
   * smaller than the entire set of variables.
   *
   * @return the variables.
   */

  Set<Var> variables () throws Exception;

  /**
   * Returns whether the theory is consistent.
   *
   * @return whether the theory is consistent.
   */
  
  boolean consistent () throws Exception;

  /**
   * Returns whether the theory is valid.
   *
   * @return whether the theory is valid.
   */

  boolean valid () throws Exception;
  
  /**
   * Returns whether the given world is a model of the theory.
   *
   * @param w the world.
   * @return whether the world is a model.
   */
  
  boolean modelTest (Map<Var,Double> w) throws Exception;
   
  /**
   * Returns one of the models of the given theory or null if there aren't
   * any.
   *
   * @return the model or null if none exists.
   */
  
  Map<Var,Double> modelFind () throws Exception;
  
  /**
   * Returns whether the given partial-world entails the theory.
   *
   * @param pw the parital-world.
   * @return whether the partial world is an implicant.
   */
  
  boolean implicant (Map<Var,Double> pw) throws Exception;

  /**
   * Returns whether the theory entails the given theory.
   *
   * @param rhs the second theory.
   * @return whether the theory entails rhs.
   */

  boolean entails (Theory rhs) throws Exception;

  /**
   * Returns whether the theory is equivalent to the given theory.
   *
   * @param rhs the second theory.
   * @return whether the theories are equivalent.
   */

  boolean equivalent (Theory rhs) throws Exception;

  /**
   * When all variables have finite domains, returns the number of models in
   * the theory; otherwise, returns -1.
   *
   * @return the number of models.
   */

  java.math.BigInteger modelCount () throws Exception;

  /**
   * Conjoins the theory with another theory.
   *
   * @param rhs the second theory.
   * @return the conjunction.
   */

  Theory conjoin (Theory rhs) throws Exception;

  /**
   * Disjoins the theory with another theory.
   *
   * @param rhs the second theory.
   * @return the conjunction.
   */

  Theory disjoin (Theory rhs) throws Exception;

  /**
   * Negates the theory.
   *
   * @return the negation.
   */

  Theory negate () throws Exception;

  /**
   * Conditions (aka restriction) the theory on the given partial world.
   *
   * @param pw the parital-world.
   * @return the resulting theory.
   */
  
  Theory condition (Map<Var,Double> pw) throws Exception;

  /**
   * Existentially quantifies the given variables in the given theory.
   *
   * @param v the variables.
   * @return the resulting theory.
   */

  Theory eQuantify (Set<Var> v) throws Exception;

  /**
   * Universally quantifies the given variables in the given theory.
   *
   * @param v the variables.
   * @return the resulting theory.
   */

  Theory aQuantify (Set<Var> v) throws Exception;

  /**
   * Returns a theory that places the same constraints on a modified set of
   * variables.  The map must pair variables of the same type and domain.
   * This operation may return a theory that is not compatible with the
   * original theory.
   *
   * @param v a mapping from old variables to new variables.
   * @return the new plg.
   */
  
  Theory rename (Map<Var,Var> v) throws Exception;

  /**
   * Returns the minimum cardinality or -1 if the theory is inconsistent.
   * Throws an exception if the variables are not all binary.
   *
   * @return the minimum cardinality or -1.
   */
  
  int minimumCardinality () throws Exception;

  /**
   * Returns a theory that consists only of the minimum cardinality models.
   * Throws an exception if the variables are not all binary.
   *
   * @return the resulting theory.
   */
  
  Theory minimizeCardinality () throws Exception;

  /**
   * Sends each model to the given partial world processor.  This method is
   * appropriate when the domain of every variable is a contiguous sequence
   * of ints.
   *
   * @param p the given processor.
   */

  void modelEnumerate (Pwp p) throws Exception;

}
