package mark.reason.net;
import java.util.*;
import mark.reason.core.*;

/**
 * A base class for a potential.  Each potential is of a certain type.
 * Currently, we have types tables, noisymax, and min.
 * <p>
 * This class conforms to the immutable design pattern.
 *
 * @author Mark Chavira
 */

public abstract class Pot extends Scope {
  
  //=========================================================================
  // Comparing potentials.
  //=========================================================================
  
  /**
   * A comparator that compares potentials by name of last variable.
   */
  
  public static Comparator<Pot> LAST_NAME_COMPARATOR =
    new Comparator<Pot> () {
      public int compare (Pot lhs, Pot rhs) {
        return
          lhs.lastVariable ().name ().compareTo (
            rhs.lastVariable ().name ());
      }
    };
    
  /**
   * A comparator that compares potentials in the following order: number of
   * variables, potential type, number modeling parameters, name of first
   * variable, name of second variable, ..., name of last variable, first
   * modeling parameter, second modeling parameter, ..., last modeling
   * parameter.
   */
  
  public static Comparator<Pot> TOTAL_COMPARATOR =
    new Comparator<Pot> () {
      public int compare (Pot l, Pot r) {
        if (l.variables ().size () < r.variables ().size ()) {return -1;}
        if (l.variables ().size () > r.variables ().size ()) {return +1;}
        int temp = l.type ().compareTo (r.type ());
        if (temp != 0) {return temp;}
        if (l.numModelingParms () < r.numModelingParms ()) {return -1;}
        if (l.numModelingParms () > r.numModelingParms ()) {return +1;}
        Iterator<Var> lit = l.variables ().iterator ();
        Iterator<Var> rit = r.variables ().iterator ();
        while (lit.hasNext ()) {
          Var lv = lit.next ();
          Var rv = rit.next ();
          temp = lv.name ().compareTo (rv.name ());
          if (temp != 0) {return temp;}
        }
        for (int i = 0; i < l.numModelingParms (); i++) {
          double lp = l.getCp (i);
          double rp = r.getCp (i);
          if (lp < rp) {return -1;}
          if (lp > rp) {return +1;}
        }
        return 0;
      }
    };

  /**
   * Returns whether the potential is equivalent to the given potential, in
   * type, variables, and modeling parameters.
   *
   * @param d the given potential.
   * @return whether they are equivalent.
   */

  public boolean equivalent (Pot d) {
    if (type () != d.type ()) {return false;}
    if (!variables ().equals (d.variables ())) {return false;}
    if (numModelingParms () != d.numModelingParms ()) {return false;}
    for (int i = 0; i < numModelingParms (); i++) {
      if (nthModelingParm (i) != d.nthModelingParm (i)) {return false;}
    }
    return true;
  }
  
  /**
   * Returns whether the potential has the given parameters.
   *
   * @param parms the given parameters.
   * @return whether the parameters match.
   */

  public boolean matches (double[] parms) {
    if (numInstantiations () != parms.length) {
      return false;
    }
    for (int i = 0; i < parms.length; i++) {
      if (getCp (i) != parms[i]) {
        return false;
      }
    }
    return true;
  }

  /**
   * Returns whether the given potential is the same potential as this
   * potential, ignoring differing ordering of variables.  The potentials are
   * assumed to have the same variables.
   * 
   * @param p the given potential.
   * @return the result of the comparison.
   */
  
  public boolean samePotentialPossiblyDifferentOrders (
   mark.reason.net.Pot p) {
    mark.inf.core.DualIterator di = di (p, null);
    for (int kp = 0; kp < di.numInstantiations (); kp++) {
      int lp = di.next ();
      if (!mark.core.math.Util.close (getCp (kp), p.getCp (lp))) {
        return false;
      }
    }
    return true;
  }

  /**
   * Throws an exception if it is not the case that the given potential
   * represents the same potential as this potential, but with a possibly
   * different ordering.
   * 
   * @param p the given potential.
   */
  
  public void assertSamePotentialPossiblyDifferentOrders (
   mark.reason.net.Pot p) throws Exception {
    mark.inf.core.DualIterator di = di (p, null);
    for (int kp = 0; kp < di.numInstantiations (); kp++) {
      int lp = di.next ();
      if (!mark.core.math.Util.close (getCp (kp), p.getCp (lp))) {
        System.err.println (toCanonicalString ());
        System.err.println ("===================");
        System.err.println (p.toCanonicalString ());
        throw new Error (kp + " " + getCp (kp) + " " + p.getCp (lp));
      }
    }
  }

  //=========================================================================
  // Ordering potentials.
  //=========================================================================

  /**
   * A PotentialSorter that uses a comparator.
   */
  
  public static class ComparatorSorter implements PotSorter {
    private Comparator fComparator;
    public ComparatorSorter (Comparator c) {fComparator = c;}
    public ArrayList<Pot> sort (Set<Pot> potentials) {
      ArrayList<Pot> ans = new ArrayList<Pot> (potentials);
      Collections.sort (ans, fComparator);
      return ans;
    }
  }

  /**
   * A PotentialSorter that uses the NAME_COMPARATOR.
   */
  
  public static final PotSorter LAST_NAME_SORTER =
    new ComparatorSorter (LAST_NAME_COMPARATOR);
  
  /**
   * A PotentialSorter that uses the TOTAL_COMPARATOR.
   */
  
  public static final PotSorter TOTAL_SORTER =
    new ComparatorSorter (TOTAL_COMPARATOR);  
  
  /**
   * A potential sorter that generates a reproduceable topological ordering.
   * Because this method gets Util.varTopological order to do much of the
   * work, this algorithm is such that calling it multiple times with the
   * same set of potentials produces the same order, even if the set data
   * structure used to contain the potentials is different and even though
   * there may be multiple toplogical orders.
   *
   * @param potentials the given potentials.
   * @return the topological ordering.
   */
    
  public static final PotSorter TOPOLOGICAL_SORTER =
    new PotSorter () {
      public ArrayList<Pot> sort (Set<Pot> potentials) throws Exception {
        return
          mark.core.util.Util.imageAsList (
            mark.reason.bn.BnUtil.varTopologicalOrder (potentials),
            mark.reason.bn.BnUtil.lastVarToScopeMap (potentials));
      }
    };

  //=========================================================================
  // Obtaining information.
  //=========================================================================

  /**
   * An enumeration defining potential types.
   */
  
  public enum Type {TABLE, NOISYMAX, MIN}
  
  /**
   * Returns the potential type.
   *
   * @return the type.
   */
   
  public abstract Type type ();

  /**
   * Returns the probability associated with the given instantiation.
   *
   * @param instantiation the instantiation.
   * @return the probability.
   */

  public abstract double getCP (Map<Var,Double> instantiation);

  /**
   * Returns the probability associated with the given instantiation.
   *
   * @param instantiation the instantiation.
   * @return the probability.
   */

  public abstract double getCP (int[] instantiation);

  /**
   * Returns the probability associated with the given instantiation.
   *
   * @param instantiation the instantiation.
   * @return the probability.
   */

  public abstract double getCp (int instantiation);

  /**
   * Returns a copy of the potential's parameters.
   *
   * @return the data.
   */
  
  public double[] parametersClone () {
    double[] ans = new double[numInstantiations ()];
    for (int i = 0; i < ans.length; i++) {
      ans[i] = getCp (i);
    }
    return ans;
  }
  
  /**
   * Returns the number of parameters that are specified at the modeling
   * level.  For example, a table has one modeling parameter per row; a
   * noisy-or has one modeling parameter for the leak and one for each
   * parent.
   *
   * @return the number of modeling parameters.
   */
  
  public abstract int numModelingParms ();
  
  /**
   * Returns the nth modeling parameter.
   *
   * @param n identifies the modeling parameter to return.
   * @return the modeling parameter.
   */
  
  public abstract double nthModelingParm (int n);

  /**
   * Returns a copy of the potential's modeling parameters.
   *
   * @return the data.
   */
  
  public double[] modelingParmsClone () {
    double[] ans = new double[numModelingParms ()];
    for (int i = 0; i < ans.length; i++) {
      ans[i] = nthModelingParm (i);
    }
    return ans;
  }

  //=========================================================================
  // Converting to string:
  //=========================================================================
  
  /**
   * Assists various toString () methods.
   *
   * @param index defines the ordering of the variables.
   */

  private String toString (Scope x) {
    StringBuilder ans = new StringBuilder ();
    for (int i = 0; i < x.variables ().size (); i++) {
      ans.append (x.variables ().get (i).name ());
      ans.append (
        i == x.variables ().size () - 1 ? mark.core.io.Util.NEWLINE : "\t");
    }
    for (int i = 0; i < x.numInstantiations (); i++) {
      if (i > 0) {
        ans.append (mark.core.io.Util.NEWLINE);
      }
      Map<Var,Double> mapInstantiation = x.mapInstantiation (i);
      for (int j = 0; j < x.variables ().size (); j++) {
        Var v = x.variables ().get (j);
        ans.append (
          v.domainElementToString (mapInstantiation.get (v)) + "\t");
      }
      ans.append (
        mark.reason.bn.BnUtil.PROB_FORMAT.format (getCP (mapInstantiation)));
    }
    return ans.toString ();
  }

  /**
   * @see java.lang.Object
   */
  
  public String toString () {
    return toString (this);
  }

  /**
   * Returns a canoncial string representation of the potential.
   *
   * @return the string.
   */
    
  public String toCanonicalString () {
    ArrayList sortedVars = new ArrayList (variables ());
    Collections.sort (sortedVars, Var.NAME_COMPARATOR);
    return toString (new Scope (sortedVars));
  }

  //=========================================================================
  // Constructor:
  //=========================================================================

  /**
   * Initializes the potential.
   *
   * @param index the index.
   */
  
  protected Pot (Scope index) {
    super (index);
  }

  //=========================================================================
  // Queries:
  //=========================================================================

  /**
   * If exactly one of the parameters is 1 and the rest are zero, then
   * returns the position corresponding to the 1; otherwise, returns -1.
   * 
   * @return the position or -1.
   */
  
  public int oneOne () {
    int ans = -1;
    for (int p = 0; p < numInstantiations (); p++) {
      double parm = getCp (p);
      if (parm == 1.0) {
        if (ans != -1) {return -1;}
        ans = p;
      } else if (parm != 0) {
        return -1;
      }
    }
    return ans;
  }
  
  /**
   * Returns whether the potential contains no 0.0 parameters.
   *
   * @return whether the potential contains no 0.0 parameters.
   */
  
  public boolean noZeros () {
    for (int i = 0; i < numInstantiations (); i++) {
      if (getCp (i) == 0.0) {
        return false;
      }
    }
    return true;
  } 

  /**
   * Returns whether the zero clause set applies to a given row.
   *
   * @param appliedPreviously whether the zero clause set applied in the
   *   previous row or false if this row is the first row.
   * @param r the row.
   * @return whether the zero clause set applies.
   */
  
  public boolean rowInZeroSet (boolean appliedPreviously, int r) {
    int ds = lastVariable ().domainSize ();
    if (r % ds != 0) {
      return appliedPreviously;
    }
    for (int k = 0; k < ds; k++) {
      if (getCp (r + k) != 0) {
        return false;
      }
    }
    return true;
  }  

  /**
   * Returns whether the one clause set applies to a given row.
   *
   * @param appliedPreviously whether the one clause set applied in the
   *   previous row or false if this row is the first row.
   * @param r the row.
   * @return whether the one clause set applies.
   */
   
  public boolean rowInOneSet (boolean appliedPreviously, int r) {
    int ds = lastVariable ().domainSize ();
    if (r % ds != 0) {
      return appliedPreviously;
    }
    int numOne = 0;
    for (int k = 0; k < ds; k++) {
      double prob = getCp (r + k);
      if (prob == 1) {
        ++numOne;
      } else if (prob != 0) {
        return false;
      }
    }
    return numOne == 1;
  }

  /**
   * Returns whether each instantiation of the parents yields probabilities
   * equal to 0.0, except for one probability equal to 1.0.
   *
   * @return whether the potential is zero-one.
   */

  public boolean allRowsInOneSet () {
    int step = lastVariable ().domainSize ();
    for (int i = 0; i < numInstantiations (); i += step) {
      if (!rowInOneSet (false, i)) {
        return false;
      }
    }
    return true;
  }

  /**
   * Returns whether the potential consists of all 0.0 and 1.0 parameters.
   *
   * @return whether the potential is 0-1.
   */
  
  public boolean allProbabilitiesZeroOrOne () {
    for (int i = 0; i < numInstantiations (); i++) {
      double cp = getCp (i);
      if (cp != 0.0 && cp != 1.0) {
        return false;
      }
    }
    return true;
  }
  
  /**
   * Returns whether the distribuion is uniform : all parameters are equal.
   *
   * @return whether the potential is uniform.
   */
  
  public boolean uniform () {
    double cp = getCp (0);
    for (int i = 0; i < numInstantiations (); i++) {
      if (getCp (i) != cp) {
        return false;
      }
    }
    return true;
  }
  
  /**
   * Returns whether all parameters are equal to one.
   * 
   * @return whether all parameters equal to one.
   */
  
  public boolean one () {
    return getCp (0) == 1.0 && uniform ();
  }
  
  /**
   * Returns whether all parameters are equal.
   * 
   * @return whether all parameters are equal.
   */
  
  public boolean equivalentToConstant () {
    double first = getCp (0);
    for (int i = 1; i < numInstantiations (); i++) {
      if (getCp (i) != first) {
        return false;
      }
    }
    return true;
  }
  
  /**
   * Retuns whether all parameters are close to the first parameter.  This
   * can result when a variable is a constant but because of rounding errors
   * has achieved slightly different values.
   * 
   * @return whether all parameters are equal.
   */
  
  public boolean almostEquivalentToConstant () {
    double first = getCp (0);
    for (int i = 1; i < numInstantiations (); i++) {
      if (!mark.core.math.Util.close (getCp (i), first)) {
        return false;
      }
    }
    return true;
  }

  //=========================================================================
  // Operations:
  //=========================================================================

  /**
   * Returns the potential where each parameter is log_e of the original
   * parameter.
   * 
   * @return the new potential.
   */
  
  public abstract Pot logBaseE () throws Exception;
  
  /**
   * Asserts evidence and sums out the evidence variables.  If the evidence
   * is null, then will return the given potential.
   *
   * @param evidence the evidence.
   * @return the resulting potential.
   */

  public abstract Pot shrink (Map<Var,Double> evidence) throws Exception;

  /**
   * For each instantaition of parents, zeroes out rows that are non-maximal
   * and then sums out the evidence veriables.
   */
  
  /**
   * Asserts evidence.
   *
   * @param evidence the evidence.
   * @return the resulting potential.
   */
  
  public abstract Pot assertEvidence (Map<Var,Double> evidence)
   throws Exception;
   
  /**
   * Maxes out variables and returns the result.  Does not modify the
   * potential.
   *
   * @param v the set of variables.
   * @return the resulting potential.
   */

  public abstract Pot maxOut (Set<Var> v) throws Exception;

  /**
   * Sums out variables and returns the result.  Does not modify the
   * potential.
   *
   * @param v the set of variables.
   * @return the resulting potential.
   */

  public abstract Pot sumOut (Set<Var> v) throws Exception;
  
  /**
   * Projects onto the given set of variables.
   *
   * @param v the given set of variables.
   * @return the resulting potential.
   */

  public Pot project (Set v) throws Exception {
    HashSet variables = new HashSet (variables ());
    variables.removeAll (v);
    return sumOut (variables);
  }

  /**
   * Normalizes each instantiation of the parents.  Does not modify the
   * potential.  If all parameters corresponding to an instantiation of
   * the parents are zero, throws an Exception.  Otherwise, divides each
   * parameter by the sum of the parameters sharing the parent instantiation.
   *
   * @return the resulting potential.
   */
  
  public abstract Pot normalizeLast () throws Exception;

  /**
   * Replaces variables and returns the result.  Does not modify the
   * potential.
   *
   * @param variableMap a map from the old variables to the new.  If a 
   *   variable is not a key, then it will not be replaced.
   * @return the resulting potential.
   */

  public abstract Scope replaceVariables (Map<Var,Var> variableMap)
   throws Exception;


  /**
   * Multiplies all of the given potentials, sums out the given variables,
   * and reutrns the result.
   *
   * @param pots the given potentials.
   * @param sumOutVars the given variables.
   * @return the result.
   */

  public abstract Pot multiplyAndSumOut (Set<Pot> pots, Set<Var> sumOutVars)
   throws Exception;
  
  /**
   * Multiplies all of the given potentials, maximizes out the given
   * variable, for each instantiation, remembers which value of the variable
   * was maximal, and returns the resulting potential.
   * 
   * @param v the given variable.
   * @param remember on entry, an array of size equal to the number of
   *   instantiations of the original potential divided by the domain size
   *   of the given variable; on exit, an array parallel to the resulting
   *   potential that indicates, for each instantiation, which value of the
   *   given variable was maximial.
   * @return the resulting potential and an array parallel to the resulting
   *   potential that indicates, for each instantiation, which value of the
   *   given variable was maximal.
   */
  
  public abstract mark.core.util.Pair<Pot,int[]>
   multiplyAndMaximizeOutAndRemember (
   Set<Pot> pots, Var maximizeVar) throws Exception;
  
   /**
    * Multiplies the given potentials, maximizes out the given variable, and
    * then returns the result and the maximizing values.  Throws an exception
    * if there are no potentials given.
    * 
    * @param pots the given potentials.
    * @param v the given varaible.
    * @param f the potential factory, which may be null if pots is guaranteed
    *   to contain elements.
    * @return the resuting potential and maximizing values.
    */
   
   public static mark.core.util.Pair<Pot,int[]> multiplyAndMaximizeOut (
    Set<Pot> pots, Var v, PotFact f) throws Exception {
     HashSet<Pot> others = new HashSet<Pot> (pots);
     Iterator<Pot> iter = others.iterator ();
     Pot p = iter.next ();
     iter.remove ();
     return p.multiplyAndMaximizeOutAndRemember (others, v);
   }

  /**
   * Multiples all of the given potentials into this potential and returns
   * the result.
   *
   * @param d the given potentials.
   * @return the resuting distribuiton.
   */
  
  public Pot multiply (Pot... d) throws Exception {
    HashSet<Pot> s = new HashSet<Pot> ();
    for (Pot n : d) {s.add (n);}
    return multiply (s);
  }

  /**
   * Adds all of the given potentials, maximizes out the given variable, for
   * each instantiation, remembers which value of the variable was maximal,
   * and returns the resulting potential.
   * 
   * @param v the given variable.
   * @param remember on entry, an array of size equal to the number of
   *   instantiations of the original potential divided by the domain size
   *   of the given variable; on exit, an array parallel to the resulting
   *   potential that indicates, for each instantiation, which value of the
   *   given variable was maximial.
   * @return the resulting potential and an array parallel to the resulting
   *   potential that indicates, for each instantiation, which value of the
   *   given variable was maximal.
   */
  
  public abstract mark.core.util.Pair<Pot,int[]>
   addAndMaximizeOutAndRemember (
   Set<Pot> pots, Var maximizeVar) throws Exception;
  
   /**
    * Adds the given potentials, maximizes out the given variable, and then
    * returns the result and the maximizing values.  Throws an exception
    * if there are no potentials given.
    * 
    * @param pots the given potentials.
    * @param v the given varaible.
    * @param f the potential factory, which may be null if pots is guaranteed
    *   to contain elements.
    * @return the resuting potential and maximizing values.
    */
   
   public static mark.core.util.Pair<Pot,int[]> addAndMaximizeOut (
    Set<Pot> pots, Var v, PotFact f) throws Exception {
     HashSet<Pot> others = new HashSet<Pot> (pots);
     Iterator<Pot> iter = others.iterator ();
     Pot p = iter.next ();
     iter.remove ();
     return p.addAndMaximizeOutAndRemember (others, v);
   }

   /**
   * Adds all of the given potentials into this potential and returns the
   * result.
   *
   * @param d the given potentials.
   * @return the resuting distribuiton.
   */
  
  public Pot add (Pot... d) throws Exception {
    HashSet<Pot> s = new HashSet<Pot> ();
    for (Pot n : d) {s.add (n);}
    return add (s);
  }

  /**
   * Multiples all of the given potentials into this potential and returns
   * the result.
   *
   * @param d the given potentials.
   * @return the resuting distribuiton.
   */
  
  public abstract Pot multiply (Set<Pot> d) throws Exception;

  /**
   * Adds all of the given potentials into this potential and returns the
   * result.
   *
   * @param d the given potentials.
   * @return the resuting distribuiton.
   */
  
  public abstract Pot add (Set<Pot> d) throws Exception;

}
