package mark.reason.net;
import java.util.*;

import mark.reason.core.*;

/**
 * My own implementation of a tabular potential.
 * <p>
 * This class conforms to the immutable design pattern.
 *
 * @author Mark Chavira
 */

public class PotTable extends Pot {

  //=========================================================================
  // Fields:
  //=========================================================================

  private double[] fParameters;

  /**
   * A constant one potential.
   */
  
  public static final PotTable ONE = new PotTable (1.0);
  
  //=========================================================================
  // Creating potentials.
  //=========================================================================

  /**
   * Initializes the table, optionally assuming ownership of the given
   * parameters. The purpose of this constructor is to let trusted methods
   * that create large arrays of parameters many times do so without
   * duplicating those large arrays.  We cannot allow non-trusted code to
   * call this method, because doing so would prevent the class from
   * conforming to the immutable design pattern.  This limitation is not
   * serious, however, because non-trusted usually only creates small tables
   * and create them rarely.
   *
   * @param parameters the parameters.
   * @param assumeOwnership whether to assume ownership of the parameters.
   * @param index the index.
   */
  
  private PotTable (
   double[] parameters, boolean assumeOwnership, Scope index) {
    super (index);
    fParameters =
      assumeOwnership ? parameters :
      mark.core.math.DoubleArrays.clone (parameters);
  }

  /**
   * Intializes the table with the given parameters and index.
   *
   * @param parms the parameters.
   * @param index the index.
   */
  
  public PotTable (double[] parms, Scope index) {
    this (parms, false, index);
  }

  /**
   * Initializes the table from another potential.
   * 
   * @param p the given potential.
   */
  
  public PotTable (Pot p) {
    this (p.parametersClone (), true, p);
  }

  /**
   * Initializes the table.
   *
   * @param parms the parameters.
   * @param vars the variables.
   */
  
  public PotTable (double[] parms, Var... vars) {
    this (parms, false, new Scope (vars));
  }

  /**
   * Initializes the table.
   * 
   * @param parms the parameters.
   * @param vars the variables.
   */
  
  public PotTable (double[] parms, List<Var> vars) {
    this (parms, false, new Scope (vars));
  }
  
  /**
   * Initializes a constant table.
   *
   * @param d the constant.
   */
  
  public PotTable (double d) {
    this (new double[] {d}, true, Scope.EMPTY_SCOPE);
  }

  /**
   * Returns a uniform table.
   *
   * @param index the index.
   */
  
  public static PotTable genUniformTable (Var... v) {
    Scope x = new Scope (v);
    return new PotTable (
      mark.reason.bn.BnUtil.uniformParameters (x), true, x);
  }

  /**
   * Returns an indicator potential allowing all values for the given
   * variable.
   * 
   * @param v the given variable.
   * @return the indicator.
   */
  
  public static PotTable genIndicatorTable (Var v) {
    double[] parms = new double[v.domainSize ()];
    Arrays.fill (parms, 1.0);
    return new PotTable (parms, v);
  }

  /**
   * Returns an indicator potential for the given variable and value.
   * 
   * @param v the given variable.
   * @param u the given value.
   * @return the indicator.
   */
  
  public static PotTable genIndicatorTable (Var v, int u) {
    Scope x = new Scope (v);
    return
      new PotTable (
        mark.reason.bn.BnUtil.evidenceParameters (v, u), true, x);
  }
  
  /**
   * Returns a gate health table over the given variable having the given
   * health set.
   * 
   * @param healthVar the given health variable.
   * @param healthSet the given health set.
   * @return the table.
   */
  
  public static PotTable genHealthTable (
   Var healthVar, mark.inf.gate.GateHealth healthSet) {
    double[] parms = new double[healthVar.domainSize ()];
    Arrays.fill (parms, 0.0001 / (healthVar.domainSize () - 1));
    parms[healthSet.healthyIndex ()] = 0.9999;
    return new PotTable (parms, healthVar);
  }

  /**
   * Returns a table that encodes a logic gate of the given type over the
   * given output and given inputs.
   * 
   * @param type the given type.
   * @param output the given output.
   * @param inputs the given inputs.
   * @return the table.
   */
  
  public static PotTable genGateTable (
   mark.inf.gate.GateType type, Var output, Var... inputs) throws Exception {
    return
      genGateTable (
        type, mark.inf.gate.GateHealth.ALWAYS_HEALTHY, output, null, inputs);
  }

  /**
   * Returns a table that encodes a logic gate of the given type over the
   * given output and given inputs and having the given health set and
   * given health variable.  The health variable should be null iff the
   * health state is always healthy.  All variables must be binary except
   * possibly the health variable, which must have as many states as the
   * number of health states.
   * 
   * @param type the given type.
   * @param healthSet the given helath set.
   * @param output the given output.
   * @param healthVar the given health variable.
   * @param inputs the given inputs.
   * @return the table.
   */
  
  public static PotTable genGateTable (
   mark.inf.gate.GateType type, mark.inf.gate.GateHealth healthSet,
   Var output, Var healthVar, Var... inputs) throws Exception {

    // Construct the scope for the gate's CPT.
    
    boolean alwaysHealthy =
      healthSet == mark.inf.gate.GateHealth.ALWAYS_HEALTHY;
    ArrayList<Var> vars = new ArrayList<Var> ();
    if (healthVar != null) {vars.add (healthVar);}
    for (Var v : inputs) {
      vars.add (v);
    }
    vars.add (output);
    Scope x = new Scope (vars);
    
    // Construct the parameters.  First, take care of stuck at zero
    // parameters.
    
    double[] parms = new double[x.numInstantiations()];
    int stuckZeroIndex = healthSet.stuckZeroIndex ();
    if (stuckZeroIndex != -1) {
      mark.inf.core.SingleIterator si =
        x.si (Collections.singletonMap (healthVar, (double)stuckZeroIndex));
      for (int i = 0; i < si.numCompatibleInstantiations (); i++) {
        int p = si.next ();
        parms[p] = si.k[si.k.length - 1] == 0 ? 1 : 0;
      }
    }
    
    // Next, handle stuck at one parameters.
    
    int stuckOneIndex = healthSet.stuckOneIndex ();
    if (stuckOneIndex != -1) {
      mark.inf.core.SingleIterator si =
        x.si (Collections.singletonMap (healthVar, (double)stuckOneIndex));
      for (int i = 0; i < si.numCompatibleInstantiations (); i++) {
        int p = si.next ();
        parms[p] = si.k[si.k.length - 1] == 1 ? 1 : 0;
      }
    }
  
    // Now, handle healthy parameters.
    
    Map<Var,Double> e =
      alwaysHealthy ? null :
      Collections.singletonMap (
        healthVar, (double)healthSet.healthyIndex ());
    mark.inf.core.SingleIterator si = x.si (e);
    for (int i = 0; i < si.numCompatibleInstantiations (); i++) {
      int p = si.next ();
      if (type == mark.inf.gate.GateType.INPUT) {
        parms[p] = 0.5;
      } else {
        int[] temp = new int[si.k.length - 1 - (alwaysHealthy ? 0 : 1)];
        System.arraycopy (
          si.k, alwaysHealthy ? 0 : 1, temp, 0, temp.length);
        parms[p] =
          si.k[si.k.length - 1] == mark.inf.gate.Gate.evaluate (type, temp) ?
          1 : 0;
      }
    }

    // Create the table.
    
    return new PotTable (parms, x);

  }
  
  /**
   * Returns a table asserting that the given variables must be equivalent.
   * All variables must have the same domain size.
   *  
   * @param vars the given variables.
   */

  public static PotTable genEquivalentTable (Var... vars) throws Exception {
    
    // Verify that there are at least two variables.
    
    if (vars.length < 2) {
      throw new Exception ("Need at least two variables!");
    }
    
    // Check the domain sizes.
    
    int domainSize = vars[0].domainSize ();
    for (int i = 1; i < vars.length; i++) {
      if (vars[i].domainSize () != domainSize) {
        throw new Exception ("Domain sizes don't match!");
      }
    }
    
    // Construct and return the table.
    
    Scope s = new Scope (vars);
    double[] parms = new double[s.numInstantiations ()];
    mark.inf.core.SingleIterator si = s.si (null);
    for (int i = 0; i < si.numCompatibleInstantiations (); i++) {
      int p = si.next ();
      parms[p] =
        mark.core.util.Util.numValues (si.k, si.k[0]) == si.k.length ?
        1.0 : 0.0;
    }
    return new PotTable (parms, vars);
    
  }
  
  /**
   * Returns a table that asserts that the last of the given variables has
   * minimum value of the rest of the given variables.  All variables must
   * have the same domain size.
   *  
   * @param vars the given variables.
   */

  public static PotTable genMinTable (Var... vars) throws Exception {
    
    // Verify that there are at least two variables.
    
    if (vars.length < 2) {
      throw new Exception ("Need at least two variables!");
    }
    
    // Check the domain sizes.
    
    int domainSize = vars[0].domainSize ();
    for (int i = 1; i < vars.length; i++) {
      if (vars[i].domainSize () != domainSize) {
        throw new Exception ("Domain sizes don't match!");
      }
    }
    
    // Construct and return the table.
    
    Scope s = new Scope (vars);
    double[] parms = new double[s.numInstantiations ()];
    mark.inf.core.SingleIterator si = s.si (null);
    for (int i = 0; i < si.numCompatibleInstantiations (); i++) {
      int p = si.next ();
      int min = Integer.MAX_VALUE;
      for (int j = 0; j < si.k.length - 1; j++) {
        min = Math.min (si.k[j], min);  
      }
      parms[p] = si.k[si.k.length - 1] == min ? 1.0 : 0.0;
    }
    return new PotTable (parms, vars);
    
  }

  /**
   * Returns a table corresponding to the given potential.  If the potential
   * is already a table, then just returns the potential.
   *
   * @param p the given potential.
   * @return the table.
   */
  
  public static PotTable genTable (Pot p) {
    return (p.type () == Pot.Type.TABLE) ? (PotTable)p : new PotTable (p);
  }

  //=========================================================================
  // Obtaining information:
  //=========================================================================

  /**
   * @see Pot
   */
   
  public Type type () {
    return Type.TABLE;
  }

  /**
   * @see Pot
   */
   
  public int numModelingParms () {
    return numInstantiations ();
  }
  
  /**
   * @see Pot
   */
   
  public double nthModelingParm (int n) {
    return getCp (n);
  }
  

  /**
   * @see Pot
   */

  public double getCP (Map<Var,Double> instantiation) {
    return getCp (position (instantiation));
  }

  /**
   * @see Pot
   */

  public double getCP (int[] instantiation) {
    return getCp (position (instantiation));
  }

  /**
   * @see Pot
   */

  public double getCp (int instantiation) {
    return fParameters[instantiation];
  }
  
  //=========================================================================
  // Transforming the potential.
  //=========================================================================

  /**
   * @see Pot
   */

  //public Pot normalize () throws Exception {
  //  return
  //    new PotTable (
  //      null, mark.core.math.DoubleArrays.normalize (fParameters),
  //      true, this);
  //}

  /**
   * @see Pot
   */

  public Pot normalizeLast () throws Exception {
    return new PotTable (
      mark.core.math.DoubleArrays.normalize (
        fParameters, lastVariable ().domainSize ()), true, this);
  }

  /**
   * @see Pot
   */

  public Pot replaceVariables (Map<Var,Var> variableMap) throws Exception {
    Scope newScope = replaceScope (variableMap);
    if (variables ().equals (newScope.variables ())) {return this;}
    return new PotTable (fParameters, true, newScope);
  }

  /**
   * @see Pot
   */

  public Pot logBaseE () throws Exception {
    double[] newParameters = new double[numInstantiations ()];
    for (int p = 0; p < newParameters.length; p++) {
      newParameters[p] = Math.log (fParameters[p]);
    }
    return new PotTable (newParameters, true, this);
  }

  /**
   * @see Pot
   */

  public Pot shrink (Map<Var,Double> evidence) {
    if (evidence == null || evidence.isEmpty ()) {
      return this;
    }
    Scope newScope = forgetScope (evidence.keySet ());
    if (newScope.variables ().size () == variables ().size ()) {
      return this;
    }
    double[] newParameters = new double[newScope.numInstantiations ()];
    mark.inf.core.DualIterator di = newScope.di (this, evidence);
    for (int p1 = 0; p1 < di.numInstantiations (); p1++) {
      int p2 = di.next ();
      newParameters[p1] = fParameters[p2];
    }
    return new PotTable (newParameters, true, newScope);
  }

  /**
   * @see Pot
   */

  public Pot assertEvidence (Map<Var,Double> evidence) {
    double[] newParameters = new double[numInstantiations ()];
    mark.inf.core.SingleIterator si =
      new mark.inf.core.SingleIterator (
        fCardinalities, convertEvidenceToIntArray (evidence));
    for (int i = 0; i < si.numCompatibleInstantiations (); i++) {
      int p = si.next ();
      newParameters[p] = fParameters[p];
    }
    return new PotTable (newParameters, true, this);
  }

  /**
   * @see Pot
   */

  public Pot maxOut (Set<Var> maxOutVars) {
    if (maxOutVars == null || maxOutVars.isEmpty ()) {
      return this;
    }
    Scope newScope = forgetScope (maxOutVars);
    if (newScope.variables ().size () == variables ().size ()) {
      return this;
    }
    double[] newParameters = new double[newScope.numInstantiations ()];
    Arrays.fill (newParameters, Double.NEGATIVE_INFINITY);
    mark.inf.core.DualIterator di = di (newScope, null);
    for (int p1 = 0; p1 < di.numInstantiations (); p1++) {
      int newIndex = di.next ();
      double oldVal = fParameters[p1];
      if (oldVal > newParameters[newIndex]) {
        newParameters[newIndex] = oldVal;
      }
    }
    return new PotTable (newParameters, true, newScope);
  }

  /**
   * @see Pot
   */

  public Pot sumOut (Set<Var> sumOutVars) {
    if (sumOutVars == null || sumOutVars.isEmpty ()) {
      return this;
    }
    Scope newScope = forgetScope (sumOutVars);
    if (newScope.variables ().size () == variables ().size ()) {
      return this;
    }
    double[] newParameters = new double[newScope.numInstantiations ()];
    mark.inf.core.DualIterator di = di (newScope, null);
    for (int p1 = 0; p1 < di.numInstantiations (); p1++) {
      newParameters[di.next ()] += fParameters[p1];
    }
    return new PotTable (newParameters, true, newScope);
  }
 
  /**
   * One would think that the below would be faster, because it writes to
   * the resulting potentail much less often, but experimentation shows that
   * no.  It may be faster however when summing out multiple variables, so
   * keeping it around.
   */
  
  public Pot multiplyAndSumOut1 (Set<Pot> pots, Set<Var> sumOutVars)
   throws Exception {

    // Some simple cases.
    
    if (pots.isEmpty ()) {
      return sumOut (sumOutVars);
    }
    if (sumOutVars.isEmpty ()) {
      return multiply (pots);
    }
    
    // Create the scope for the result.  Sort the variables by name.  Create
    // a scope for all variables.  Sort the variablers by name, except put
    // sum out variables last.  Create parameters for the result.
    
    Scope resultScope = combineAndForgetScope (pots, sumOutVars);
    Scope allScope = combineScope (pots, sumOutVars);
    double[] resultParms = new double[resultScope.numInstantiations ()];
    
    // Create an ordering of the scopes involved.  Also create an array of
    // the parameters for fast access.
    
    Scope[] orderedScopes = new Scope[1 + pots.size ()];
    orderedScopes[0] = this;
    int index = 1;
    for (Scope s : pots) {
      orderedScopes[index++] = s;
    }
    double[][] allParms = new double[orderedScopes.length][];
    for (int i = 0; i < allParms.length; i++) {
      allParms[i] = ((PotTable)orderedScopes[i]).fParameters;
    }

    // Compute the effective value of the first keep variable.
    
    int val = 1;
    for (Var v : sumOutVars) {
      val *= v.domainSize ();
    }
    
    // Iterate over instantiations.

    int resultIndex = 0;
    mark.inf.core.MultiIterator mi = allScope.mi (orderedScopes, null);
    double sum = 0.0;
    for (int kp = 0; kp < mi.numInstantiations (); kp++) {
      mi.next ();
      double prod = 1.0;
      for (int i = 0; i < allParms.length; i++) {
        prod *= allParms[i][mi.lp[i]];
      }
      sum += prod;
      if ((kp + 1) % val == 0) {
        resultParms[resultIndex++] = sum;
        sum = 0.0;
      }
    }
    return new PotTable (resultParms, resultScope);

  }

  /**
   * @see Pot
   */
  
  public Pot multiplyAndSumOut (Set<Pot> pots, Set<Var> sumOutVars)
   throws Exception {

    // Some simple cases.
    
    if (pots.isEmpty ()) {
      return sumOut (sumOutVars);
    }
    if (sumOutVars.isEmpty ()) {
      return multiply (pots);
    }
    
    // Create the scope for the result.  Sort the variables by name.  Create
    // a scope for all variables.  Sort the variablers by name, except put
    // sum out variables last.
    
    Scope resultScope = combineAndForgetScope (pots, sumOutVars);
    Scope allScope = combineScope (pots);
    
    // Create an ordering of the scopes involved.  Also create an array of
    // the parameters for fast access.
    
    Scope[] orderedScopes = new Scope[2 + pots.size ()];
    orderedScopes[0] = resultScope;
    orderedScopes[1] = this;
    int index = 2;
    for (Scope s : pots) {
      orderedScopes[index++] = s;
    }
    double[][] allParms = new double[orderedScopes.length][];
    allParms[0] = new double[resultScope.numInstantiations ()];
    for (int i = 1; i < allParms.length; i++) {
      allParms[i] = ((PotTable)orderedScopes[i]).fParameters;
    }

    // Iterate over instantiations.

    mark.inf.core.MultiIterator mi = allScope.mi (orderedScopes, null);
    for (int kp = 0; kp < mi.numInstantiations (); kp++) {
      mi.next ();
      double prod = 1.0;
      for (int i = 1; i < allParms.length; i++) {
        prod *= allParms[i][mi.lp[i]];
      }
      allParms[0][mi.lp[0]] += prod;
    }
    return new PotTable (allParms[0], resultScope);

  }
  
  /**
   * @see Pot
   */
  
  public mark.core.util.Pair<Pot,int[]>
   multiplyAndMaximizeOutAndRemember (
   Set<Pot> pots, Var maximizeVar) throws Exception {

     // Create the scope for the result.  Create a scope for all variables.
     // Allocate the array holding maximizing values.
    
    Scope resultScope =
      combineAndForgetScope (pots, Collections.singleton (maximizeVar));
    Scope allScope = combineScope (pots);
    int[] remember = new int[resultScope.numInstantiations ()];
    int maximizeVarIndex = allScope.variables ().indexOf (maximizeVar);
    if (!allScope.variables ().contains (maximizeVar)) {
      throw
        new Exception ("Potentials do not contain the maximize variable");
    }
    
    // Create an ordering of the scopes involved.  Also create an array of
    // the parameters for fast access.
    
    Scope[] orderedScopes = new Scope[2 + pots.size ()];
    orderedScopes[0] = resultScope;
    orderedScopes[1] = this;
    int index = 2;
    for (Scope s : pots) {
      orderedScopes[index++] = s;
    }
    double[][] allParms = new double[orderedScopes.length][];
    allParms[0] = new double[resultScope.numInstantiations ()];
    Arrays.fill (allParms[0], Double.NEGATIVE_INFINITY);
    for (int i = 1; i < allParms.length; i++) {
      allParms[i] = ((PotTable)orderedScopes[i]).fParameters;
    }

    // Iterate over instantiations.

    mark.inf.core.MultiIterator mi = allScope.mi (orderedScopes, null);
    for (int kp = 0; kp < mi.numInstantiations (); kp++) {
      mi.next ();
      double prod = 1.0;
      for (int i = 1; i < allParms.length; i++) {
        prod *= allParms[i][mi.lp[i]];
      }
      int secondIndex = mi.lp[0];
      double oldVal = allParms[0][secondIndex];
      if (prod > oldVal) {
        allParms[0][secondIndex] = prod;
        remember[secondIndex] = mi.k[maximizeVarIndex]; 
      }
    }
    return
      new mark.core.util.Pair<Pot,int[]> (
        new PotTable (allParms[0], resultScope), remember);

  }

  /**
   * @see Pot
   */

  public Pot multiply (Set<Pot> pots) throws Exception {

    // Simple case.
    
    if (pots == null || pots.isEmpty ()) {
      return this;
    }

    // Create the scope for the result.  Sort the variables by name.
    // Also create result parameters.
    
    Scope resultScope = combineScope (pots);
    double[] resultParms = new double[resultScope.numInstantiations ()];
    
    // Create an ordering of the scopes involved.  Also create an array of
    // the parameters for fast access.
    
    Scope[] orderedScopes = new Scope[1 + pots.size ()];
    orderedScopes[0] = this;
    int index = 1;
    for (Scope s : pots) {orderedScopes[index++] = s;}
    double[][] allParms = new double[orderedScopes.length][];
    for (int i = 0; i < allParms.length; i++) {
      allParms[i] = ((PotTable)orderedScopes[i]).fParameters;
    }

    // Iterate over instantiations.

    mark.inf.core.MultiIterator mi = resultScope.mi (orderedScopes, null);
    for (int kp = 0; kp < mi.numInstantiations (); kp++) {
      mi.next ();
      double prod = 1.0;
      for (int i = 0; i < allParms.length; i++) {
        prod *= allParms[i][mi.lp[i]];
      }
      resultParms[kp] = prod;
    }
    return new PotTable (resultParms, resultScope);

  }

  /**
   * @see Pot
   */
  
  public mark.core.util.Pair<Pot,int[]>
   addAndMaximizeOutAndRemember (
   Set<Pot> pots, Var maximizeVar) throws Exception {

    // Create the scope for the result.  Create a scope for all variables.
    // Allocate the array holding maximizing values.
    //System.out.println (":" + maximizeVar);
    Scope resultScope =
      combineAndForgetScope (pots, Collections.singleton (maximizeVar));
    Scope allScope = combineScope (pots);
    int[] remember = new int[resultScope.numInstantiations ()];
    int maximizeVarIndex = allScope.variables ().indexOf (maximizeVar);
    if (!allScope.variables ().contains (maximizeVar)) {
      throw
        new Exception ("Potentials do not contain the maximize variable");
    }
    
    // Create an ordering of the scopes involved.  Also create an array of
    // the parameters for fast access.
    
    Scope[] orderedScopes = new Scope[2 + pots.size ()];
    orderedScopes[0] = resultScope;
    orderedScopes[1] = this;
    int index = 2;
    for (Scope s : pots) {
      orderedScopes[index++] = s;
    }
    double[][] allParms = new double[orderedScopes.length][];
    allParms[0] = new double[resultScope.numInstantiations ()];
    Arrays.fill (allParms[0], Double.NEGATIVE_INFINITY);
    for (int i = 1; i < allParms.length; i++) {
      allParms[i] = ((PotTable)orderedScopes[i]).fParameters;
    }

    // Iterate over instantiations.

    mark.inf.core.MultiIterator mi = allScope.mi (orderedScopes, null);
    for (int kp = 0; kp < mi.numInstantiations (); kp++) {
      mi.next ();
      double sum = 0.0;
      for (int i = 1; i < allParms.length; i++) {
        sum += allParms[i][mi.lp[i]];
      }
      int secondIndex = mi.lp[0];
      double oldVal = allParms[0][secondIndex];
      if (sum > oldVal) {
        allParms[0][secondIndex] = sum;
        remember[secondIndex] = mi.k[maximizeVarIndex]; 
      }
    }
    return
      new mark.core.util.Pair<Pot,int[]> (
        new PotTable (allParms[0], resultScope), remember);

  }

  /**
   * @see Pot
   */

  public Pot add (Set<Pot> pots) throws Exception {

    // Simple case.
    
    if (pots == null || pots.isEmpty ()) {
      return this;
    }

    // Create the scope for the result.  Sort the variables by name.
    // Also create result parameters.
    
    Scope resultScope = combineScope (pots);
    double[] resultParms = new double[resultScope.numInstantiations ()];
    
    // Create an ordering of the scopes involved.  Also create an array of
    // the parameters for fast access.
    
    Scope[] orderedScopes = new Scope[1 + pots.size ()];
    orderedScopes[0] = this;
    int index = 1;
    for (Scope s : pots) {orderedScopes[index++] = s;}
    double[][] allParms = new double[orderedScopes.length][];
    for (int i = 0; i < allParms.length; i++) {
      allParms[i] = ((PotTable)orderedScopes[i]).fParameters;
    }

    // Iterate over instantiations.

    mark.inf.core.MultiIterator mi = resultScope.mi (orderedScopes, null);
    for (int kp = 0; kp < mi.numInstantiations (); kp++) {
      mi.next ();
      double sum = 0.0;
      for (int i = 0; i < allParms.length; i++) {
        sum += allParms[i][mi.lp[i]];
      }
      resultParms[kp] = sum;
    }
    return new PotTable (resultParms, resultScope);

  }

  /**
   * Returns a new table formed by changing all nonzero parameters to 1.0.
   * 
   * @return the table.
   */

  public PotTable makeBoolean () {
    double[] parms = parametersClone ();
    for (int i = 0; i < parms.length; i++) {
      parms[i] = parms[i] == 0 ? 0 : 1.0;
    }
    return new PotTable (parms, this);
  }

  /**
   * @see Pot
   */

  public HashMap<Var,PotTable> learnEvidence (
   Map<Var,PotTable> tEvidence) throws Exception {
    
    // Copy the parameters potential.  Then, for each outlawed value of
    // each variable in the potential, zero out corresponding parameters.
    // Then create a new potential from the new parameters.
    
    double[] parms = parametersClone ();
    for (Var v : variables ()) {
      PotTable t = tEvidence.get (v);
      for (int u = 0; u < v.domainSize (); u++) {
        if (t.getCp (u) != 0) {continue;}
        mark.inf.core.SingleIterator si =
          si (Collections.singletonMap (v, (double)u));
       for (int i = 0; i < si.numCompatibleInstantiations (); i++) {
         parms[si.next ()] = 0.0;
       }
     }
    }
    PotTable newTab = new PotTable (parms, this);
    
    // Now project the new table on each variable and return the projections.

    HashMap<Var,PotTable> ans =
      new HashMap<Var,PotTable> ();
    for (Var v : variables ()) {
      ans.put (
        v, ((PotTable)newTab.project(Collections.singleton(v))).
        makeBoolean ());
    }
    return ans;

  }

}
