package mark.reason.plg;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.Reader;
import java.util.*;
import mark.reason.core.*;

/**
 * A base class for a node in a plg.  We also view an object of this class
 * as a plg, since it is the root of exactly one plg.  This class implements
 * the following Theory operations:
 * <ul>
 *   <li> variables ()
 *   <li> modelTest ()
 *   <li> condition ()
 *   <li> conjoin ()
 *   <li> disjoin ()
 *   <li> negate ()
 *   <li> rename ()
 * </ul>
 * This class conforms to the immutable design pattern.
 *
 * @author Mark Chavira
 */

public abstract class Node extends mark.reason.inf.NodeTheory {

  //=========================================================================
  // Static Methods:
  //=========================================================================

  /**
   * Returns an array containing the same nodes as the given collection.
   *
   * @param nodes the collection.
   * @return the array.
   */
  
  public static Node[] toNodeArray (
   Collection<mark.core.rdag.Node> nodes) {
    Node[] ans = new Node[nodes.size ()];
    int index = 0;
    for (mark.core.rdag.Node n : nodes) {
      ans[index++] = (Node)n;
    }
    return ans;
  }

  /**
   * Constructs a cnf from the given reader using the given parser.  Does not
   * close the reader.
   * 
   * @param r the given reader.
   * @param p the given parser.
   */
  
  public static mark.core.util.Pair<Node,VariableTable> constructPlg (
   Reader r, Plgr p, VariableTable vt) throws Exception {
    return p.parse (r, vt);
  }

  /**
   * Constructs a plg from the file with the given name using the given
   * parser.
   * 
   * @param n the given name.
   * @param p the given parser.
   */
  
  public static mark.core.util.Pair<Node,VariableTable> constructPlg (
   String n, Plgr p, VariableTable vt) throws Exception {
    BufferedReader r = new BufferedReader (new FileReader (n));
    mark.core.util.Pair<Node,VariableTable> ans = constructPlg (r, p, vt);
    r.close ();
    return ans;
  }

  //=========================================================================
  // Node Methods:
  //=========================================================================

  /**
   * Returns the nth child as a plg node.
   *
   * @param n identifies the child.
   */
  
  public abstract Node nthCastedChild (int n);

  /**
   * Returns a new node of the same type with the given children.
   *
   * @param children the new children.
   * @return the new node.
   */
   
  protected abstract Node clone (Node[] children);

  /**
   * Uses the children as keys into the given memo and returns the nodes
   * referenced.
   *
   * @param memo the given memo
   * @return the nodes.
   */

  protected Node[] traversalChildren (Map memo) {
    Node[] ans = new Node[numOutgoingEdges ()];
    for (int i = 0; i < ans.length; i++) {
      ans[i] = (Node)memo.get (nthChild (i));
    }
    return ans;    
  }

  /**
   * Uses the children as keys into the given memo and returns the sets
   * referenced.
   *
   * @param memo the given memo.
   * @return the sets.
   */
  
  protected BitSet[] traversalSets (Map memo) {
    BitSet[] ans = new BitSet[numOutgoingEdges ()];
    for (int i = 0; i < ans.length; i++) {
      ans[i] = (BitSet)memo.get (nthChild (i));
    }
    return ans;    
  }

  //=========================================================================
  // Graph Hash Code:
  //
  // We use a caching scheme to avoid computing the hash code more than once
  // and to avoid computing it at all if it is not needed.  Originally, I
  // tried computing the hash code in the constructor, but this was a bit
  // complicated to implement.  For example, I wanted to write code to
  // compute a hash code in NodeAtom instead of computing it in all
  // subclasses of NodeAtom.  However, the hash code depends on the value
  // assigned to the variable, which is not available in the constructor
  // of NodeAtom.  The subclass could pass this information, but it was
  // just a bit more complicated.  So the scheme is to cache it.
  //=========================================================================

  /**
   * Whether the hash code has been computed.
   */

  private boolean fHashComputed = false;
  
  /**
   * The computed hash code.
   */
  
  private int fHashCode;
  
  /**
   * @see Object
   */
  
  public int hashCode () {
    if (!fHashComputed) {
      fHashCode = computeHash ();
      fHashComputed = true;
    }
    return fHashCode;
  }

  /**
   * Computes the hash code.
   */
  
  protected abstract int computeHash ();

  //=========================================================================
  // String Representations of Graph:
  //=========================================================================

  /**
   * @see java.lang.Object
   */

  public String toString () {
    return toInfixString ();
  }

  /**
   * Converts the plg to a string using prefix notation.
   * <p>
   * Warning : this method traverses the graph as if it were a tree.
   * 
   * @return the string.
   */

  public abstract String toPrefixString ();

  /** 
   * Converts the plg to a string using infix notation.
   * <p>
   * Warning : this method traverses the graph as if it were a tree.
   *
   * @return the string.
   */

  public abstract String toInfixString ();

  /**
   * Returns a label for the node.
   * 
   * @return the label.
   */
  
  public abstract String label ();
  
  //=========================================================================
  // Graph Transformations:
  //=========================================================================
  
  //-------------------------------------------------------------------------
  // Transform a plg into an nnf plg.
  //-------------------------------------------------------------------------

  /**
   * Assists nnf.
   */
  
  private Node nnfHelper (Node[] ch) throws Exception {
    if (this instanceof NodeLeaf) {
      return this;
    } else if (this instanceof NodeAnd || this instanceof NodeOr) {
      return clone (ch);
    } else if (this instanceof NodeNot) {
      if (ch[0] instanceof NodeTrue) {
        return NodeFalse.INSTANCE;
      } else if (ch[0] instanceof NodeFalse) {
        return NodeTrue.INSTANCE;
      } else if (ch[0] instanceof NodeAtom) {
        return new NodeNot (ch[0]);
      } else if (ch[0] instanceof NodeNot) {
        return ch[0].nthCastedChild (0);
      } else if (ch[0] instanceof NodeOr) {
        Node[] newCh = new Node[ch[0].numOutgoingEdges ()];
        for (int i = 0; i < newCh.length; i++) {
          newCh[i] = (new NodeNot (ch[0].nthCastedChild (i))).nnf ();
        }
        return new NodeAnd (newCh);
      } else if (ch[0] instanceof NodeAnd) {
        Node[] newCh = new Node[ch[0].numOutgoingEdges ()];
        for (int i = 0; i < newCh.length; i++) {
          newCh[i] = (new NodeNot (ch[0].nthCastedChild (i))).nnf ();
        }
        return new NodeOr (newCh);
      }
    }
    throw new Exception ("Must be an notAndOr plg to convert to nnf!");
  }

  /**
   * Converts a notAndOrPlg into an nnf.
   *
   * @return the resulting plg.
   */
  
  public Node nnf () throws Exception {
    ArrayList nodes = childrenBeforeParents ();
    HashMap memo = new HashMap ();
    for (int i = 0; i < nodes.size (); i++) {
      Node n = (Node)nodes.get (i);
      memo.put (n, n.nnfHelper (n.traversalChildren (memo)));
    }
    return (Node)memo.get (this);
  }

  //-------------------------------------------------------------------------
  // Transform a plg into a smooth plg.
  //-------------------------------------------------------------------------

  /**
   * Assists smooth ().
   */

  private void smoothHelper1 (
   Map memo, Map variables,
   mark.core.util.OrderedSet allVars, Node[] allSmoothPlgs) {
    BitSet vars = new BitSet (allVars.size ());
    if (this instanceof NodeAtom) {
      vars.set (allVars.indexOf (((NodeAtom)this).variable ()));
    }
    BitSet[] chSets = traversalSets (variables);
    for (int i = 0; i < chSets.length; i++) {
      vars.or (chSets[i]);
    }
    Node[] ch = traversalChildren (memo);
    if (this instanceof NodeOr) {
      for (int i = 0; i < ch.length; i++) {
        ch[i] =
          smoothHelper2 (vars, chSets[i], ch[i], allSmoothPlgs);
      }
    }
    variables.put (this, vars);
    memo.put (this, clone (ch));
  }
  
  /**
   * Assists smoothHelper1 () and smooth ().
   */
   
  private static Node smoothHelper2 (
   BitSet required, BitSet present, Node current,
   Node[] allSmoothPlgs) {
    BitSet diff = new BitSet (allSmoothPlgs.length);
    diff.or (required);
    diff.andNot (present);
    int card = diff.cardinality ();
    if (card == 0) {
      return current;
    }
    Node[] ch = new Node[card + 1];
    int chIndex = 0;
    for (int i = diff.nextSetBit (0); i >= 0; i = diff.nextSetBit (i + 1)) {
      ch[chIndex++] = allSmoothPlgs[i];
    }
    ch[chIndex] = current;
    return new NodeAnd (ch);
  }
  
  /**
   * Smooths the plg.  The plg must be in nnf.
   *
   * @param v if non-null, guarantees that variables in v are in the resut.
   * slashes: only works with binary variables
   * @return the new plg.
   */

  public Node smooth (Set<Var> v) throws Exception {
    ArrayList nodes = childrenBeforeParents ();
    HashMap variables = new HashMap ();
    HashMap memo = new HashMap ();
    mark.core.util.OrderedSet allVars = null;
    {
      HashSet allVarsSet =
        new HashSet (variables ());
      allVarsSet.addAll (v);
      allVars = new mark.core.util.OrderedSet (allVarsSet);
    }
    Node[] allSmoothPlgs = new Node[allVars.size ()];
    for (int i = 0; i < allSmoothPlgs.length; i++) {
      VarBinary bv = (VarBinary)allVars.get (i);
      NodeAtomBinary b = new NodeAtomBinary (bv);
      allSmoothPlgs[i] = new NodeOr (new Node[] {b, new NodeNot (b)});
    }
    for (int i = 0; i < nodes.size (); i++) {
      Node n = (Node)nodes.get (i);
      n.smoothHelper1 (memo, variables, allVars, allSmoothPlgs);
    }
    Node n = (Node)memo.get (this);
    if (v == null) {
      return n;
    }
    BitSet vBitSet = new BitSet (allVars.size ());
    for (Var vv : v) {
      vBitSet.set (allVars.indexOf (vv));
    }
    return
      smoothHelper2 (
        vBitSet, (BitSet)variables.get (this), n, allSmoothPlgs);
  }

  //-------------------------------------------------------------------------
  // Transform a plg into a not-and-or plg.
  //-------------------------------------------------------------------------

  /**
   * Transforms a plg into a notAndOr plg; that is, into a plg containing
   * only NOT nodes, AND nodes, OR nodes, and atoms.
   *
   * @return the resulting plg.
   */
  
  public Node notAndOrPlg () throws Exception {
    ArrayList nodes = childrenBeforeParents ();
    HashMap memo = new HashMap ();
    for (int i = 0; i < nodes.size (); i++) {
      Node n = (Node)nodes.get (i);
      memo.put (n, n.notAndOrPlgHelper (n.traversalChildren (memo)));
    }
    return (Node)memo.get (this);
  }
  
  /**
   * Assists notAndOrPlg ().
   * 
   * @param ch
   * @return
   * @throws Exception
   */
  
  protected abstract Node notAndOrPlgHelper (Node[] ch) throws Exception;

  //-------------------------------------------------------------------------
  // Transform a plg into a reduced plg.
  //-------------------------------------------------------------------------

  /**
   * Performs reductions on the plg.  For example, deletes some unnecessary
   * TrueNodes and FalseNodes, deletes AND and OR nodes with a single child,
   * and merges double negations.
   *
   * @return the resulting plg.
   */

  public Node reduce () throws Exception {
    ArrayList nodes = childrenBeforeParents ();
    HashMap memo = new HashMap ();
    for (int i = 0; i < nodes.size (); i++) {
      Node n = (Node)nodes.get (i);
      memo.put (n, n.reduceHelper (n.traversalChildren (memo)));
    }
    return (Node)memo.get (this);
  }
  
  /**
   * Assists reduce ().
   * 
   * @param ch
   * @return
   */
  
  protected abstract Node reduceHelper (Node[] ch);

  //-------------------------------------------------------------------------
  // Transforma a plg into a plg where each node has at most two children.
  //-------------------------------------------------------------------------

  /**
   * Returns a plg that has the same models where each node has at most two
   * children.
   *
   * @return the new plg.
   */
  
  public Node simplifyConnectives () throws Exception {
    ArrayList nodes = childrenBeforeParents ();
    HashMap memo = new HashMap ();
    for (int i = 0; i < nodes.size (); i++) {
      Node n = (Node)nodes.get (i);
      memo.put (n, n.simplifyConnectivesHelper (n.traversalChildren (memo)));
    }
    return (Node)memo.get (this);
  }
  
  /**
   * Assists simplifyConnectives ().
   * 
   * @param ch
   * @return
   */
  
  protected abstract Node simplifyConnectivesHelper (Node[] ch);

   //-------------------------------------------------------------------------
   // Transform a plg into a plg that replaces certain atoms with certain
   // plgs.
   //-------------------------------------------------------------------------
  
  /**
   * Returns a plg that replaces certain atoms with certain plgs.
   *
   * @param m a map from the variables of the atoms to replace to the plgs
   * with which to replace.
   * @return the new plg.
   */
  
  public Node replaceAtoms (Map m) throws Exception {
    ArrayList nodes = childrenBeforeParents ();
    HashMap memo = new HashMap ();
    for (int i = 0; i < nodes.size (); i++) {
      Node n = (Node)nodes.get (i);
      if (n instanceof NodeAtom) {
        Object v = ((NodeAtom)n).variable ();
        Object plg = m.get (v);
        memo.put (n, plg == null ? n : plg);
      } else if (n instanceof NodeLeaf) {
        memo.put (n, n);
      } else {
        memo.put (n, n.clone (n.traversalChildren (memo)));
      }
    }
    return (Node)memo.get (this);
  }

  //=========================================================================
  // Theory Implementations:
  //=========================================================================

  /**
   * @see Theory
   */
  
  public Set<Var> variables () throws Exception {
    ArrayList<mark.core.rdag.Node> nodes =
      childrenBeforeParents ();
    HashSet<Var> ans = new HashSet<Var> ();
    for (int i = 0; i < nodes.size (); i++) {
      Node n = (Node)nodes.get (i);
      if (n instanceof NodeAtom) {
        ans.add (((NodeAtom)n).variable ());
      }
    }
    return ans;
  }
  
  /**
   * @see Theory
   */
  
  public boolean modelTest (final Map w) throws Exception {
    ArrayList nodes = childrenBeforeParents ();
    HashMap<mark.core.rdag.Node,Boolean> memo =
      new HashMap<mark.core.rdag.Node,Boolean> ();
    for (int i = 0; i < nodes.size (); i++) {
      Node n = (Node)nodes.get (i);
      memo.put (
        n, n.modelTestHelper (n.traversalBooleans (memo), w) ? true : false);
    }
    return memo.get (this);
  }
  protected abstract boolean modelTestHelper (
   boolean[] memo, Map<Var,Double> w);
  
  /**
   * Theory.
   */

  public mark.reason.inf.Theory condition (Map pw) throws Exception {
    ArrayList nodes = childrenBeforeParents ();
    HashMap memo = new HashMap ();
    for (int i = 0; i < nodes.size (); i++) {
      Node n = (Node)nodes.get (i);
      memo.put (n, n.conditionHelper (n.traversalChildren (memo), pw));
    }
    return (Node)memo.get (this);
  }
  protected abstract Node conditionHelper (Node[] ch, Map pw);
  
  /**
   * @see Theory
   */

  public mark.reason.inf.Theory conjoin (mark.reason.inf.Theory rhs) {
    return new NodeAnd (new Node[] {this, (Node)rhs});
  }

  /**
   * @see Theory
   */

  public mark.reason.inf.Theory disjoin (mark.reason.inf.Theory rhs) {
    return new NodeOr (new Node[] {this, (Node)rhs});
  }

  /**
   * @see Theory
   */

  public mark.reason.inf.Theory negate () {
    return new NodeNot (this);
  }

  /**
   * @see Theory
   */
  
  public mark.reason.inf.Theory rename (final Map v) throws Exception {
    ArrayList nodes = new ArrayList ();
    HashMap memo = new HashMap ();
    for (int i = 0; i < nodes.size (); i++) {
      Node n = (Node)nodes.get (i);
      memo.put (n, n.renameHelper (n.traversalChildren (memo), v));
    }
    return (Node)memo.get (this);
  }
  
  /**
   * Assists rename ().
   * 
   * @param ch
   * @param v
   * @return
   */
  
  protected abstract Node renameHelper (Node[] ch, Map v);

  //=========================================================================
  // Other Graph Methods:
  //=========================================================================

  /**
   * Splits a plg into multiple plgs by introducing new binary variables.
   * For each non-root internal node, creates a new binary variable and new
   * binary atom.  Each non-root node N then has a corresponding atom
   * ATOM (N).  For each non-leaf node N representing
   * (<OP> ch1 ch2 ... chn), creates a plg:
   * <xmp>
   *  \equiv (X, (<op> (ATOM (ch1), ATOM (ch1), ..., ATOM (chn))))
   * </xmp>
   * where X is ATOM(N) if N is not a root and where X is the given plg if
   * N is a root.
   *
   * @param v the given binary variable.
   * @return the created plgs.
   */
  
  public ArrayList split (String name, Node v) throws Exception {
    int uniqueName = 0;
    ArrayList nodes = childrenBeforeParents ();
    HashMap memo = new HashMap ();
    ArrayList ans = new ArrayList ();
    for (int i = 0; i < nodes.size () - 1; i++) {
      Node n = (Node)nodes.get (i);
      if (n instanceof NodeAtom) {
        memo.put (n, n);
      } else {
        NodeAtom a =
          new NodeAtomBinary (
            new VarBinary (
              name + "_split_auxiliary_" + uniqueName++));
        memo.put (n, a);
        ans.add (
          new NodeEquivalent (a, n.clone (n.traversalChildren (memo))));
      }
    }
    ans.add (
      new NodeEquivalent (
        v,
        ((this instanceof NodeAtom) ?
        this : clone (this.traversalChildren (memo)))));
    return ans;
  }
  
}
