package mark.reason.struct;
import java.util.*;
import java.io.*;
import mark.reason.core.*;

/**
 * Utility functions for clause dtrees.  A clause dtree is a dtree over ints,
 * where each int represents the position of some clause.
 * 
 * @author Mark Chavira
 */

public class DtreeClauseUtils {

  /**
   * Writes the given dtree to the given writer and closes the given writer.
   *
   * @param pw the given writer.
   * @param dtree the given clause dtree.
   */
  
  public static void writeClauseDtree (
   mark.core.rdag.Dtree<Integer> dtree, String fn)
   throws Exception {
    PrintWriter dtreeWriter =
      new PrintWriter (
        new BufferedWriter (
          new FileWriter (fn)));
    ArrayList<mark.core.rdag.Node> nodes =
      dtree.childrenBeforeParents ();
    HashMap<mark.core.rdag.Node,Integer> nodeToIndex =
      mark.core.util.Util.indexMap (nodes);
    dtreeWriter.println ("dtree " + nodes.size ());
    for (mark.core.rdag.Node n : nodes) {
      if (n.numOutgoingEdges () == 0) {
        dtreeWriter.println ("L " + ((mark.core.rdag.Dtree)n).fObject);
      } else {
        dtreeWriter.println (
          "I " + nodeToIndex.get (n.nthChild (0)) + " " +
          nodeToIndex.get (n.nthChild (1)));
      }
    }
    dtreeWriter.close ();
  }

  /**
   * Creates the clause dtree for the given set of potentials and the results
   * of encoding and simplifying.
   *
   * @parm r a random generator.
   * @param pots the given potentials. 
   * @param rclauseToSrc the result from encoding that maps rclause to
   *   source.
   * @param eclauseToSrc the result from encoding that maps eclause to
   *   srouce.
   * @param posEvidence the result from simplifying that specifies positive
   *   evidence. 
   * @param negEvidenceVar the result from simplifying that specifies
   *   negative evidence variables.
   * @param negEvidenceVals the result from simplifying that specifies
   *   negative evidence values.
   * @param posEvidenceVarToSource a map from each variable in the positive
   *   evidence to the one potential that must necessarily contain that
   *   variable.
   * @param dtg the dtree generator to use to construct the bn dtree.
   * @param info an info object to pass to dtg or null if none desired.
   * @return the clause dtree or null if none requested.
   */
  
  public static mark.core.rdag.Dtree<Integer> clauseDtree (
   Random r, Set<mark.reason.net.Pot> pots,
   ArrayList<mark.reason.net.Pot> rclauseToSrc,
   ArrayList<mark.reason.net.Pot> eclauseToSrc,
   Map<Var,Double> posEvidence, ArrayList<Var> negEvidenceVars,
   ArrayList<Integer> negEvidenceVals,
   Map<Var,mark.reason.net.Pot> posEvidenceVarToSource,
   mark.reason.struct.Dtg dtg) throws Exception {
  
    // Create a new set of variables that removes values corresponding to the
    // evidence.
    
    HashMap<Var,Integer> oldVarToDomainSize = new HashMap<Var,Integer> ();
    for (Var oldVar : CoreUtil.variables (pots)) {
      if (posEvidence.containsKey (oldVar)) {
        oldVarToDomainSize.put (oldVar, 1);
      } else {
        oldVarToDomainSize.put (oldVar, oldVar.domainSize ());
      }
    }
    for (Var oldVar : negEvidenceVars) {
      mark.core.util.Util.cmIncrement (oldVarToDomainSize, oldVar, -1);
    }
    HashMap<Var,Var> oldVarToNewVar = new HashMap<Var,Var> ();
    for (Var oldVar : oldVarToDomainSize.keySet ()) {
      oldVarToNewVar.put (
        oldVar,
        new VarDiscrete (oldVar.name (), oldVarToDomainSize.get (oldVar)));
    }
    
    // For each potential, create a scope using the new variables.  We omit
    // each variable in the positive evidence, except in the one potential
    // that must contain it.
    
    HashMap<Scope,Scope> oldScopeToNewScope = new HashMap<Scope,Scope> ();
    for (Scope oldScope : pots) {
      ArrayList<Var> vars = new ArrayList<Var> ();
      for (Var oldVar : oldScope.variables ()) {
        if (
         !posEvidence.containsKey (oldVar) ||
         posEvidenceVarToSource.get (oldVar) == oldScope) {
          vars.add (oldVarToNewVar.get (oldVar));
        }
      }
      oldScopeToNewScope.put (oldScope, new Scope (vars));
    }
  
    // Create a 
    
    HashMap<Scope,HashSet<Integer>> newScopeToClauseIndexSet =
      new HashMap<Scope,HashSet<Integer>> ();
    for (Scope s : oldScopeToNewScope.values ()) {
      newScopeToClauseIndexSet.put (s, new HashSet<Integer> ());
    }
    for (int i = 0; i < rclauseToSrc.size (); i++) {
      newScopeToClauseIndexSet.get (
        oldScopeToNewScope.get (rclauseToSrc.get (i))).add (i);
    }
    for (int i = 0; i < eclauseToSrc.size (); i++) {
      newScopeToClauseIndexSet.get (
        oldScopeToNewScope.get (eclauseToSrc.get (i))).add (
          rclauseToSrc.size () + i);
    }
    
    // Create the bn dtree and get another method to convert the bn dtree to
    // a clause dtree.
    
    return
      mark.core.rdag.Dtree.targetDtreeExpand (
        dtg.generate (r, new HashSet<Scope> (oldScopeToNewScope.values ())),
        newScopeToClauseIndexSet);
    
  }

}
