package mark.reason.struct;
import java.util.*;
import java.io.*;
import mark.reason.core.*;

/**
 * A dtree having at each leaf a scope.
 * 
 * @author Mark Chavira
 */

public class DtreeScopeUtils {

  //=========================================================================
  // Private:
  //=========================================================================
  
  /**
   * A bucket eliminator for constructing a dtree from dtree leaves and an
   * elimination order.
   * <p>
   * This class conforms to the immutable design pattern.
   * <p>
   * This class conforms to the singleton design pattern.
   * 
   * @author Mark Chavira
   */
  
  private static mark.inf.core.Be<Var,mark.core.rdag.Dtree<Scope>> BE =
    new mark.inf.core.Be<Var,mark.core.rdag.Dtree<Scope>> () {
      protected mark.core.rdag.Dtree<Scope> combine (
       HashSet<mark.core.rdag.Dtree<Scope>> groups, Var elem,
       int currentBucket, int totalBuckets) throws Exception {
        return mark.core.rdag.Dtree.fullBinaryTree ((Collection)groups);
      }
      protected Set<Var> elements (mark.core.rdag.Dtree<Scope> g) {
        return g.fObject.variables ();
      }
    };

  /**
   * Assists the other version of computeClusters.
   * 
   * @param parentACutset if n is the root, then an empty set; otherwise, the
   *   a-cutset of the parent.  This set will be used as scratch during
   *   execution of the algorithm, but will be returned to its initial state
   *   upon return.
   * @param parentCutsetList if n is the root, then null; otherwise, the
   *   list of parent cutset variables.
   * @param m the map from node to cluster that will be filled.
   */
  
  private static void computeClusters (
   mark.core.rdag.Dtree<Scope> dtree,
   Map<mark.core.rdag.Dtree<Scope>,Set<Var>> nodeToVarSet,
   HashSet<Var> parentACutset, Var[] parentCutsetList,
   HashMap<mark.core.rdag.Dtree<Scope>,HashSet<Var>> m)
   throws Exception {

    // If the node is a leaf, then its cutset is just the set of nodes in the
    // corresponding potential.
     
    boolean hasParent = parentCutsetList != null;
    if (dtree.isLeaf ()) {
      m.put (dtree, new HashSet<Var> (nodeToVarSet.get (dtree)));
      return;
    }

    // Compute the cutset.
     
    HashSet<Var> currentCutset =
      new HashSet<Var> (nodeToVarSet.get (dtree.left ()));
    currentCutset.retainAll (nodeToVarSet.get (dtree.right ()));
    if (hasParent) {
      currentCutset.removeAll (parentACutset);
      for (Var v : parentCutsetList) {currentCutset.remove (v);}
    }
    Var[] currentCutsetList =
      currentCutset.toArray (new Var[currentCutset.size ()]);
     
    // Compute the context.
     
    HashSet<Var> currentContext = new HashSet<Var> ();
    if (hasParent) {
      currentContext.addAll (parentACutset);
      for (Var v : parentCutsetList) {currentContext.add (v);}
    }
    currentContext.retainAll (nodeToVarSet.get (dtree));

    // Now copute the cluster.
     
    HashSet<Var> currentCluster = new HashSet<Var> (currentCutset);
    currentCluster.addAll (currentContext);
    m.put (dtree, currentCluster);
    currentCluster = null;
    currentCutset = null;
    currentContext = null;
     
    // Now process children.
        
    if (hasParent) {for (Var v : parentCutsetList) {parentACutset.add (v);}}
    computeClusters (
      dtree.left (), nodeToVarSet, parentACutset, currentCutsetList, m);
    computeClusters (
      dtree.right (), nodeToVarSet, parentACutset, currentCutsetList, m);
    if (hasParent) {
      for (Var v : parentCutsetList) {parentACutset.remove (v);}
    }
        
  }

  //=========================================================================
  // Public:
  //=========================================================================

  /**
   * Builds a dtree from the given scopes using the given elimination order
   * and returns the result.
   * 
   * @param scopes the given scopes.
   * @param eo the given elimination order.
   * @return the resulting dtree.
   */
  
  public static <T extends Scope> mark.core.rdag.Dtree<Scope> construct (
   Set<T> scopes, List<Var> eo) throws Exception {
    HashSet<mark.core.rdag.Dtree<Scope>> leaves =
      new HashSet<mark.core.rdag.Dtree<Scope>> ();
    for (Scope s : scopes) {
      leaves.add (new mark.core.rdag.Dtree<Scope> (s));
    }
    mark.core.rdag.Dtree<Scope> ans =
      BE.eliminate (eo, leaves, true).iterator ().next ();
    return ans;
  }
  
  /**
   * Reads a bn dtree from the given reader.
   *
   * @param r the given reader.
   * @param potentials the dtree's potentials.
   * @return the dtree.
   */
  
  public static mark.core.rdag.Dtree<Scope> readBnDtree (
   BufferedReader b, Set<mark.reason.net.Pot> potentials)
   throws Exception {
    ArrayList<mark.reason.net.Pot> potList =
      new ArrayList<mark.reason.net.Pot> (potentials);
    Collections.sort (potList, mark.reason.net.Pot.TOTAL_COMPARATOR);
    int numNodes = Integer.parseInt (b.readLine ().split ("[\\s]+")[1]);
    ArrayList<mark.core.rdag.Dtree<Scope>> nodes =
      new ArrayList<mark.core.rdag.Dtree<Scope>> ();
    for (int i = 0; i < numNodes; i++) {
      String[] fields = b.readLine ().split ("[\\s]+");
      if (fields[0].equals ("I")) {
        mark.core.rdag.Dtree<Scope> l =
          nodes.get (Integer.parseInt (fields[1]));
        mark.core.rdag.Dtree<Scope> t =
          nodes.get (Integer.parseInt (fields[2]));
        nodes.add (new mark.core.rdag.Dtree<Scope> (l, t));
      } else {
        nodes.add (
          new mark.core.rdag.Dtree<Scope> (
            potList.get (Integer.parseInt (fields[1]))));
      }
    }
    return nodes.get (nodes.size () - 1);
  }
  
  /**
   * Writes the given bn dtree to the given writer.
   *
   * @param w the given writer.
   * @param potentials the dtree's potentials.
   * @param dtree the given dtree.
   */

  public static void writeBnDtree (
   PrintWriter pw, Set<mark.reason.net.Pot> potentials,
   mark.core.rdag.Dtree<Scope> dtree) throws Exception {
    ArrayList<mark.core.rdag.Dtree<Scope>> nodes =
      (ArrayList)dtree.childrenBeforeParents ();
    ArrayList<mark.reason.net.Pot> dists =
      new ArrayList<mark.reason.net.Pot> (potentials);
    Collections.sort (dists, mark.reason.net.Pot.TOTAL_COMPARATOR);
    HashMap<mark.core.rdag.Dtree<Scope>,Integer> nodeToIndex =
      mark.core.util.Util.indexMap (nodes);
    HashMap<mark.reason.net.Pot,Integer> distToIndex =
      mark.core.util.Util.indexMap (dists);
    pw.println ("dtree " + nodes.size ());
    for (int i = 0; i < nodes.size (); i++) {
      mark.core.rdag.Dtree<Scope> n = nodes.get (i);
      if (n.isInternal ()) {
        mark.core.rdag.Node l = n.nthChild (0);
        mark.core.rdag.Node r = n.nthChild (1);
        pw.println ("I " + nodeToIndex.get (l) + " " + nodeToIndex.get (r));
      } else {
        pw.println ("L " + distToIndex.get (n.fObject));
      }
    }
  }

  /**
   * Returns a map from dtree node to set of variables underneath the node.
   * The dtree must be over subclasses of class Scope.
   * 
   * @return the map.
   */
  
  public static HashMap<mark.core.rdag.Dtree<Scope>,HashSet<Var>>
   nodeToVarSetMap (mark.core.rdag.Dtree<Scope> dt) throws Exception {
    HashMap<mark.core.rdag.Dtree<Scope>,HashSet<Var>> ans =
      new HashMap<mark.core.rdag.Dtree<Scope>,HashSet<Var>> ();
    ArrayList<mark.core.rdag.Dtree<Scope>> nodes =
      (ArrayList)dt.childrenBeforeParents ();
    for (mark.core.rdag.Dtree<Scope> n : nodes) {
      if (n.isLeaf ()) {
        ans.put (n, new HashSet<Var> (n.fObject.variables ()));
      } else {
        HashSet<Var> s = new HashSet<Var> (ans.get (n.nthChild (0)));
        s.addAll (ans.get (n.nthChild (1)));
        ans.put (n, s);
      }
    }
    return ans;
  }
  
  /**
   * Returns a map from dtree node to cluster.  Phase 2 must have been run
   * prior to calling this method.
   * 
   * @return the map.
   */
  
  public static HashMap<mark.core.rdag.Dtree<Scope>,HashSet<Var>>
   nodeToClusterMap (
   mark.core.rdag.Dtree<Scope> dt,
   Map<mark.core.rdag.Dtree<Scope>,Set<Var>> nodeToVarSet) throws Exception {
    HashMap<mark.core.rdag.Dtree<Scope>,HashSet<Var>> ans =
      new HashMap<mark.core.rdag.Dtree<Scope>,HashSet<Var>> ();
    computeClusters (dt, nodeToVarSet, new HashSet<Var> (), null, ans);
    return ans;
  }
       
}
