package mark.reason.struct;
import java.util.*;
import mark.reason.core.*;

/**
 * A cloning strategy that makes use of a dtree and splits variables fully.
 * The strategy iterates in an outer loop and an inner loop.  In the outer
 * loop, it constructs a dtree and then enters the inner loop.  In the inner
 * loop, the strategy fully splits the variable, which when fully split will
 * reduce the separator and internal cluster entries most.  It continues in
 * this way until the max cluster limit is achieved.  Note that specifying
 * a sufficiently large value for the inner loop will result in only a single
 * dtree being produced. 
 * <p>
 * This class stores no modifiable state, so it may be reused.
 * <p>
 * This class is highly optimized.
 * <p>
 * This class conforms to the immutable design pattern.
 * <p>
 * This class conforms to the CliObject design patter.
 * 
 * @author Mark Chavira
 */

public class SpDtree
 implements SplittingStrategy, mark.core.shell.CliObject {
  
  /**
   * Each time we construct a dtree, generate elimination orders using these
   * engines.
   */
  
  private mark.inf.eo.Eoe[] fEoe;
  
  /**
   * Parallel to fEoe, the number of times to run each engine each time we
   * construct a dtree. 
   */
  
  private int[] fTimes;
  
  /**
   * The constructor, which accepts the list of elimination order engines and
   * the number of times to run each.  This constructor assumes control of
   * the given objects.
   * 
   * @param eoe the elimination order engines.
   * @param times parallel to eoe, the number of times to run each engine
   *   each time an order is produced.
   */
  
  public SpDtree (mark.inf.eo.Eoe[] eoe, int[] times) {
    fEoe = eoe;
    fTimes = times;
  }
    
  /**
   * @see SplittingStrategy
   */
  
  public <T extends Scope> Result<T> constructApproximateScopes (
   Random r, Set<T> originalScopes, Map<Var,Var[]> origToClones,
   double log2ClusterLimit) throws Exception {
    return
      constructApproximateScopes (
        r, originalScopes, origToClones, log2ClusterLimit, 1).elem0;
  }

  /**
   * A separator or internal cluster.  Variable indices into a Converter are
   * stored rather than the variables themselves.
   * 
   * @author Mark Chavira
   */
  
  private static class Cluster {
    
    /**
     * The variable indices.  The number of indices may be less than the
     * length.  This field may be modified after construction by removing
     * variable indices.
     */
    
    int[] fVarIndices;
    
    /**
     * The number of variable indices.  This field may be modified after
     * construction by removing variables.
     */
    
    int fNumVarIndices;
    
    /**
     * A map from variable index to domain size.  This field will not be
     * modified after construction and should not be modified externally
     * either.
     */
    
    int[] fVarIndexToDomainSize;
    
    /**
     * A map from variable index to (d-1)/d, where d is the domain size of
     * the variable.  This field will not be modified after construction and
     * should not be modified externally either.
     */
    double[] fVarIndexToReduceConstant;
    
    /**
     * The product of the cardinalities of the variables in the cluster.
     * This field will be recomputed each time a variable is removed from
     * the cluster.
     */
    
    double fSize;
    
    /**
     * Recaluates the fSize field based on the set of variable indices.
     */
    
    private void recalculateSize () {
      fSize = 1.0;
      int numVarIndices = fNumVarIndices;
      for (int i = 0; i < numVarIndices; i++) {
        int varIndex = fVarIndices[i];
        fSize *= fVarIndexToDomainSize[varIndex];
      }
    }
    
    /**
     * The constructor.
     * 
     * @param varIndices the variable indices.  Will assume ownership.
     * @param varIndexToDomainSize a map from variable index to domain size.
     *   This object is assumed to be shared with other objects, and no
     *   object may modify its value.
     * @param varIndexToReduceConstant a map from variable index to (d-1)/d,
     *   where d is the domain size of the corresponding variable.  This
     *   object is assumed to be shared with other objects, and no object may
     *   modify its value.
     */
    
    public Cluster (
     int[] varIndices, int[] varIndexToDomainSize,
     double[] varIndexToReduceConstant) {
      fVarIndices = varIndices;
      fNumVarIndices = fVarIndices.length;
      fVarIndexToDomainSize = varIndexToDomainSize;
      fVarIndexToReduceConstant = varIndexToReduceConstant;
      recalculateSize ();
    }
    
    /**
     * Removes the given variable index from the cluster.  The index must be
     * a part of the cluster.
     * 
     * @param varIndex the given index.
     */
    
    public void remove (int varIndex) {
      int numVarIndices = fNumVarIndices;
      for (int i = 0; i < numVarIndices; i++) {
        if (fVarIndices[i] == varIndex) {
          fVarIndices[i] = fVarIndices[--fNumVarIndices];
          break;
        }
      }
      recalculateSize ();
    }

    /**
     * In the given map from variable index to score, for each variable index
     * in the cluster, adds to its score the number of entries that will be
     * saved in this cluster if the variable were to be removed from this
     * cluster.
     * 
     * @param varIndexToScore the given map.
     */
    
    public void adjustScores (double[] varIndexToScore) {
      int numVarIndices = fNumVarIndices;
      for (int i = 0; i < numVarIndices; i++) {
        int varIndex = fVarIndices[i];
        varIndexToScore[varIndex] +=
          fSize * fVarIndexToReduceConstant[varIndex];
      }
    }
    
    /**
     * (1) Creates a cluster having the given variable indices, the first
     * given map from variable index to domain size, and the second given map
     * from variable index to (d-1)/d, where d is the domain size of the
     * given variable, (2) adds the new cluster to the given list of
     * clusters, and (3) adds the new cluster to each list L[v], where L is
     * the given map from variable index to list of clusters and v is the
     * variable indices in this cluster.
     * 
     * @param varIndices the given variable indices.
     * @param varIndexToDomainSize the first given map. 
     * @param varIndexToReduceConstant the second given map.
     * @param sepsAndInternalClusters the given list of clusters.
     * @param varIndexToSepAndInternalClusters the given map from variable
     *   index to list of clusters.
     */
    
    public static void create (
     int[] varIndices,
     int[] varIndexToDomainSize,
     double[] varIndexToReduceConstant,
     ArrayList<Cluster> sepsAndInternalClusters,
     ArrayList<Cluster>[] varIndexToSepAndInternalClusters) {
      Cluster cl =
        new Cluster (varIndices, varIndexToDomainSize, varIndexToReduceConstant);
      sepsAndInternalClusters.add (cl);
      for (int varIndex : varIndices) {
        ArrayList<Cluster> s = varIndexToSepAndInternalClusters[varIndex];
        if (s == null) {
          s = new ArrayList<Cluster> ();
          varIndexToSepAndInternalClusters[varIndex] = s;
        }
        s.add (cl);
      }
    }
    
  }
  
  /**
   * The implementation of the strategy described in the class header.  This
   * method is essentially the same as the other method of the same name, but
   * accepts more options and returns more information.  The other method
   * calls this one with default parameters.
   * 
   * @param r a random generator.
   * @param originalScopes the original network to approximate.
   * @param origToClones on entry, an empty map; on exit, the map from
   *   variable in the given network to set of corresponding variables in the
   *   approximate network.
   * @param logClusterLimit the given limit.
   * @param dtreeFrequency the number of times to run the inner loop.
   * @return the splitting result and also the dtree produced.
   */
  
  public <T extends Scope>
   mark.core.util.Pair<Result<T>,mark.core.rdag.Dtree<T>>
   constructApproximateScopes (
   Random r, Set<T> originalScopes, Map<Var,Var[]> origToClones,
   double log2ClusterLimit, int dtreeFrequency) throws Exception {

    // Convert the cluster limit to normal space from log space.  We want to
    // do so because we will be comparing numbers to this value that are in
    // normal space.  We convert this value to normal space rather than
    // converting each other value to log space, because it is more efficient
    // this way: we need to add and multiply such numbers, so we need to work
    // in normal space until the point at which we compare.
    
    double clusterLimit = Math.pow (2.0, log2ClusterLimit);
    
    // Define an object with the current set of scopes.  This object will
    // be updated at the end of each outer loop by fully splitting variables.
    
    HashSet<T> scopes = new HashSet<T> (originalScopes);
    
    // Enter the outer loop.  The output of this loop is a dtree and the
    // final set of separators and internal clusters in the dtree.

    Cluster[] sepsAndInternalClusters = null;
    mark.core.rdag.Dtree<T> dt = null;
    for (;;) {
      
      // Create a new dtree.

      dt =
        DtreeScopeUtils.construct ((Set)scopes,
        CoreUtil.eliminationOrder (
          fEoe, fTimes, r, scopes,
          Collections.singletonList ((Set<Var>)CoreUtil.variables (scopes)),
          null, Double.POSITIVE_INFINITY, false, null).fOrder);
      
      // Divide the dtree nodes into sinks and internal.
      
      ArrayList<mark.core.rdag.Dtree<T>> dtSinks =
        new ArrayList<mark.core.rdag.Dtree<T>> ();
      ArrayList<mark.core.rdag.Dtree<T>> dtInternal =
        new ArrayList<mark.core.rdag.Dtree<T>> ();
      for (
       mark.core.rdag.Dtree<T> n :
       (ArrayList<mark.core.rdag.Dtree<T>>)
       (ArrayList)dt.childrenBeforeParents ()) {
        (n.isLeaf () ? dtSinks : dtInternal).add (n);
      }
        
      // Create a converter.  For efficiency, create maps from variable index
      // to domain size and to (d-1)/d, where d is domain size.  We only
      // need consider variables that appear in an internal node cluster.
      
      mark.reason.core.Converter c = null;
      {
        HashSet<Var> varSet1 = new HashSet<Var> ();
        HashSet<Var> varSet2 = new HashSet<Var> ();
        for (mark.core.rdag.Dtree<T> n : dtSinks) {
          for (Var v : n.fObject.variables ()) {
            (varSet1.contains (v) ? varSet2 : varSet1).add (v);
          }
        }
        c = new mark.reason.core.Converter (varSet2);
      }
      int[] varIndexToDomainSize = c.cardinalities ();
      double[] varIndexToReduceConstant = new double[c.length ()];
      for (int varIndex = 0; varIndex < c.length (); varIndex++) {
        varIndexToReduceConstant[varIndex] =
          (varIndexToDomainSize[varIndex] - 1.0) /
          varIndexToDomainSize[varIndex];
      }
      
      // Create some objects that will make some further operations
      // convenient and efficient: the set of separators and internal
      // clusters and a map from variable index of variables that are in
      // internal clusters to set of separators and internal clusters
      // containing the variable.  Note that we convert to int[] for
      // efficiency (before did not and was much less efficient).
      
      Cluster[][] varIndexToSepAndInternalClusters = null;
      {
        ArrayList<Cluster> tempSepsAndInternalClusters =
          new ArrayList<Cluster> ();
        ArrayList<Cluster>[] tempVarIndexToSepAndInternalClusters =
          new ArrayList[c.length ()];
        HashMap<mark.core.rdag.Dtree<T>,HashSet<Var>> nodeToCluster =
          (HashMap)
          DtreeScopeUtils.nodeToClusterMap (
            (mark.core.rdag.Dtree)dt,
            (Map)DtreeScopeUtils.nodeToVarSetMap ((mark.core.rdag.Dtree)dt));
        for (mark.core.rdag.Dtree<T> n : dtInternal) {
          HashSet<Var> cl = nodeToCluster.get (n);
          Cluster.create (
            c.integerVariableArray (cl),
            varIndexToDomainSize, varIndexToReduceConstant,
            tempSepsAndInternalClusters,
            tempVarIndexToSepAndInternalClusters);
          HashSet<Var> sep = new HashSet<Var> (cl);
          sep.retainAll (nodeToCluster.get (n.left ()));
          Cluster.create (
            c.integerVariableArray (sep),
            varIndexToDomainSize, varIndexToReduceConstant,
            tempSepsAndInternalClusters,
            tempVarIndexToSepAndInternalClusters);
          sep = new HashSet<Var> (cl);
          sep.retainAll (nodeToCluster.get (n.right ()));
          Cluster.create (
            c.integerVariableArray (sep),
            varIndexToDomainSize, varIndexToReduceConstant,
            tempSepsAndInternalClusters,
            tempVarIndexToSepAndInternalClusters);
        }
        sepsAndInternalClusters =
          tempSepsAndInternalClusters.toArray (
            new Cluster[tempSepsAndInternalClusters.size ()]);
        varIndexToSepAndInternalClusters =
          new Cluster[tempVarIndexToSepAndInternalClusters.length][];
        for (
         int varIndex = 0;
         varIndex < varIndexToSepAndInternalClusters.length; varIndex++) {
          varIndexToSepAndInternalClusters[varIndex] =
            tempVarIndexToSepAndInternalClusters[varIndex].toArray (
              new Cluster
              [tempVarIndexToSepAndInternalClusters[varIndex].size ()]);
        }
      }
      
      // Enter the inner loop.

      double[] varIndexToScore = new double[c.length ()];
      int numLarger = -1;
      HashSet<Var> splitVars = new HashSet<Var> ();
      for (int iteration = 0; iteration < dtreeFrequency; iteration++) {

        // Compute the scores for this dtree.  Previously, we put the
        // following code outside of the inner loop and only updated scores
        // inside this loop.  However, updating ended up loosing precision.
        // Because computing from scratch does not appear to end up taking
        // much longer, I compute from scratch now.  To reinvoke the update
        // method, put this block (except last two lines) before the loop,
        // and uncomment code below.
        
        Arrays.fill (varIndexToScore, 0.0);
        numLarger = 0;
        for (Cluster cl : sepsAndInternalClusters) {
          cl.adjustScores (varIndexToScore);
          if (cl.fSize > clusterLimit) {numLarger++;}
        }
        if (numLarger == 0) {break;}
        //System.out.println (numLarger);

        // Select a variable to split.
        
        double max = Double.NEGATIVE_INFINITY;
        int maxVarIndex = -1;
        for (int varIndex = 0; varIndex < varIndexToScore.length; varIndex++) {
          double s = varIndexToScore[varIndex];
          if (s > max) {max = s; maxVarIndex = varIndex;}
        }
        splitVars.add (c.reasonVariable (maxVarIndex));

        // Split the variables in the clusters.  (the commented code here
        // is the updating code that I had in there previously).
        for (Cluster cl : varIndexToSepAndInternalClusters[maxVarIndex]) {
          //          double oldSize = cl.fSize;
          cl.remove (maxVarIndex);
          //double newSize = cl.fSize;
          //if (oldSize > clusterLimit && newSize <= clusterLimit) {
          //  --numLarger;
          //}
          //int numVarIndices = cl.fNumVarIndices;
          //for (int i = 0; i < numVarIndices; i++) {
          //  int varIndex = cl.fVarIndices[i];
          //  varIndexToScore[varIndex] +=
          //    (newSize - oldSize) * varIndexToReduceConstant[varIndex];
          //}
        }
        
        // Tidy up.
        
        varIndexToSepAndInternalClusters[maxVarIndex] = null;
        varIndexToScore[maxVarIndex] = Double.NEGATIVE_INFINITY;
        
      }
 
      // Count the number of occurrences of each split variable in the
      // potentials.
      
      HashMap<Var,Integer> varToNumOccurrences =
        new HashMap<Var,Integer> ();
      for (T s : scopes) {
        for (Var v : s.variables ()) {
          if (splitVars.contains (v)) {
            Integer num = varToNumOccurrences.get (v);
            if (num == null) {num = 0;}
            varToNumOccurrences.put (v, num + 1);
          }
        }
      }

      // Create clones.
      
      for (Var v : splitVars) {
        Var[] clones = new Var[varToNumOccurrences.get (v)];
        clones[0] = v;
        for (int i = 1; i < clones.length; i++) {
          clones[i] = v.clone (v.name () + "__clone__" + i);
        }
        origToClones.put (v, clones);
      }
      
      // Create the new scopes.
      
      HashMap<Var,Var> m = new HashMap<Var,Var> ();
      for (mark.core.rdag.Dtree<T> n : dtSinks) {
        Scope s = n.fObject;
        m.clear ();
        for (Var v : s.variables ()) {
          if (splitVars.contains (v)) {
            int index = varToNumOccurrences.get (v) - 1;
            varToNumOccurrences.put (v, index);
            Var clone = origToClones.get (v)[index];
            if (v != clone) {m.put (v, clone);}
          }
        }
        if (m.size () > 0) {
          T newScope = (T)s.replaceVariables (m);
          scopes.remove (s);
          scopes.add (newScope);
          n.fObject = newScope;
        }
      }

      // Get out if done.
      
      if (numLarger == 0) {break;}
    
    }
      
    // Package up the result and return.

    for (Var v : CoreUtil.variables (originalScopes)) {
      if (!origToClones.keySet ().contains (v)) {
        origToClones.put (v, new Var[] {v});
      }
    }
    Result ans = new Result ();
    ans.fScopes = scopes;
    ans.fLog2MaxClusterAchieved = Double.NEGATIVE_INFINITY; 
    for (Cluster cl : sepsAndInternalClusters) {
      ans.fLog2MaxClusterAchieved =
        Math.max (ans.fLog2MaxClusterAchieved, cl.fSize);
    }
    ans.fLog2MaxClusterAchieved =
      mark.core.math.Util.log2 (ans.fLog2MaxClusterAchieved);
    return
      new mark.core.util.Pair<Result<T>,mark.core.rdag.Dtree<T>> (ans, dt);

  }

}
